/*
 * SPDX-License-Identifier: BSD-3-Clause
 * Copyright(c) 2023 Napatech A/S
 */

#ifndef NT4GA_LINK_H_
#define NT4GA_LINK_H_

#include "common_adapter_defs.h"
#include "nthw_drv.h"
#include "i2c_nim.h"
#include "nthw_fpga_rst_nt200a0x.h"

/*
 * Link state.\n
 * Just after start of ntservice the link state might be unknown since the
 * monitoring routine is busy reading NIM state and NIM data. This might also
 * be the case after a NIM is plugged into an interface.
 * The error state indicates a HW reading error.
 */
enum NtLinkState_e {
  NT_LINK_STATE_UNKNOWN = 0, /* The link state has not been read yet */
  NT_LINK_STATE_DOWN    = 1, /* The link state is DOWN */
  NT_LINK_STATE_UP      = 2, /* The link state is UP */
  NT_LINK_STATE_ERROR   = 3  /* The link state could not be read */
};

typedef enum NtLinkState_e NtLinkState_t, *NtLinkState_p;

/*
 * Link duplex mode
 */
enum NtLinkDuplex_e {
  NT_LINK_DUPLEX_UNKNOWN=0,
  NT_LINK_DUPLEX_HALF = 0x01, /* Half duplex */
  NT_LINK_DUPLEX_FULL = 0x02, /* Full duplex */
};

typedef enum NtLinkDuplex_e NtLinkDuplex_t;

/*
 * Link loopback mode
 */
enum NtLinkLoopback_e {
  NT_LINK_LOOPBACK_OFF = 0,
  NT_LINK_LOOPBACK_HOST = 0x01, /* Host loopback mode */
  NT_LINK_LOOPBACK_LINE = 0x02, /* Line loopback mode */
};

/*
 * Link MDI mode
 */
enum NtLinkMDI_e {
  NT_LINK_MDI_NA = 0,
  NT_LINK_MDI_AUTO = 0x01,  /* MDI auto */
  NT_LINK_MDI_MDI  = 0x02,  /* MDI mode */
  NT_LINK_MDI_MDIX = 0x04,  /* MDIX mode */
};

typedef enum NtLinkMDI_e NtLinkMDI_t;

/*
 * Link Auto/Manual mode
 */
enum NtLinkAutoNeg_e {
  NT_LINK_AUTONEG_NA     = 0,
  NT_LINK_AUTONEG_MANUAL = 0x01,
  NT_LINK_AUTONEG_OFF    = NT_LINK_AUTONEG_MANUAL, /* Auto negotiation OFF */
  NT_LINK_AUTONEG_AUTO   = 0x02,
  NT_LINK_AUTONEG_ON     = NT_LINK_AUTONEG_AUTO,   /* Auto negotiation ON */
};

typedef enum NtLinkAutoNeg_e NtLinkAutoNeg_t;

/*
 * Callback functions to setup mac, pcs and phy
 */
typedef struct link_state_s {
  bool link_disabled;
  bool nim_present;
  bool lh_nim_absent;
  bool link_up;
  enum NtLinkState_e link_state;
  enum NtLinkState_e link_state_latched;
} link_state_t;

typedef struct link_info_s {
  enum NtLinkSpeed_e link_speed;
  enum NtLinkDuplex_e link_duplex;
  enum NtLinkAutoNeg_e link_auto_neg;
} link_info_t;

typedef struct port_action_s {
  bool port_disable;
  enum NtLinkSpeed_e port_speed;
  enum NtLinkDuplex_e port_duplex;
  uint32_t port_lpbk_mode;
} port_action_t;


typedef struct adapter_10g_s {
  nim_i2c_ctx_t nim_ctx[NUM_ADAPTER_PORTS_MAX]; /* Should be the first field */
  link_state_t link_state[NUM_ADAPTER_PORTS_MAX];
  NthwPhy3s10g_t phy_3s10g[NUM_ADAPTER_PORTS_MAX];
} adapter_10g_t;

typedef struct adapter_40g_s {
  nim_i2c_ctx_t nim_ctx[NUM_ADAPTER_PORTS_MAX]; /* Should be the first field */
  NthwPcs_t pcs40g[NUM_ADAPTER_PORTS_MAX];
  NthwMac_t mac40g[NUM_ADAPTER_PORTS_MAX];
  NthwFpga_rst_nt200a0x_t rst40g[NUM_ADAPTER_PORTS_MAX];
  NthwGpioPhy_t gpio_phy[NUM_ADAPTER_PORTS_MAX];
} adapter_40g_t;

typedef struct adapter_100g_s {
  nim_i2c_ctx_t nim_ctx[NUM_ADAPTER_PORTS_MAX]; /* Should be the first field */
  NthwMacPcs_t mac_pcs100g[NUM_ADAPTER_PORTS_MAX];
  NthwGfg_t gfg[NUM_ADAPTER_PORTS_MAX];
  NthwGpioPhy_t gpio_phy[NUM_ADAPTER_PORTS_MAX];
} adapter_100g_t;

typedef struct adapter_25g_s {
  nim_i2c_ctx_t nim_ctx[NUM_ADAPTER_PORTS_MAX]; /* Should be the first field */
  bool gpio_phy_present;
  bool gpio_sfpp_present;
  uint8_t mac_channel_index[NUM_ADAPTER_PORTS_MAX];
  NthwMacPcsXxv_t mac_pcs25g[NUM_ADAPTER_PORTS_MAX];
  NthwGfg_t gfg[NUM_ADAPTER_PORTS_MAX];
  NthwGpioPhy_t gpio_phy[NUM_ADAPTER_PORTS_MAX];
  NthwGpioSfpp_t gpio_sfpp[NUM_ADAPTER_PORTS_MAX];
} adapter_25g_t;

typedef struct adapterAgx_100g_s {
  nim_i2c_ctx_t nim_ctx[NUM_ADAPTER_PORTS_MAX]; /* should be the first field */
  NthwGfg_t gfg[NUM_ADAPTER_PORTS_MAX];
  int fecSupported[NUM_ADAPTER_PORTS_MAX];
} adapter_agx_100g_t;

typedef union Adapter_var_s {
  nim_i2c_ctx_t nim_ctx[NUM_ADAPTER_PORTS_MAX]; /* First field in all the adaptors type */
  adapter_10g_t var10g;
  adapter_25g_t var25g;
  adapter_40g_t var40g;
  adapter_100g_t var100g;
  adapter_agx_100g_t varA100g;
} adapter_var_u;

typedef struct Nt4gaLink_s
{
  link_state_t link_state[NUM_ADAPTER_PORTS_MAX];
  link_info_t link_info[NUM_ADAPTER_PORTS_MAX];
  port_action_t port_action[NUM_ADAPTER_PORTS_MAX];
  uint32_t speed_capa;
  /* */
  bool variables_initialized;
  adapter_var_u u;
} nt4ga_link_t;

bool nt4ga_port_get_nim_present(struct AdapterInfo_s* p, int port);

/*
 * port:s link mode
 */
void nt4ga_port_set_adm_state(struct AdapterInfo_s* p, int port, bool adm_state);
bool nt4ga_port_get_adm_state(struct AdapterInfo_s* p, int port);

/*
 * port:s link status
 */
void nt4ga_port_set_link_status(struct AdapterInfo_s* p, int port, bool status);
bool nt4ga_port_get_link_status(struct AdapterInfo_s* p, int port);

/*
 * port: link autoneg
 */
void nt4ga_port_set_link_autoneg(struct AdapterInfo_s* p, int port, bool autoneg);
bool nt4ga_port_get_link_autoneg(struct AdapterInfo_s* p, int port);

/*
 * port: link speed
 */
void nt4ga_port_set_link_speed(struct AdapterInfo_s* p, int port, NtLinkSpeed_t speed);
NtLinkSpeed_t nt4ga_port_get_link_speed(struct AdapterInfo_s* p, int port);

/*
 * port: link duplex
 */
void nt4ga_port_set_link_duplex(struct AdapterInfo_s* p, int port, NtLinkDuplex_t duplex);
NtLinkDuplex_t nt4ga_port_get_link_duplex(struct AdapterInfo_s* p, int port);

/*
 * port: loopback mode
 */
void nt4ga_port_set_loopback_mode(struct AdapterInfo_s* p, int port, uint32_t mode);
uint32_t nt4ga_port_get_loopback_mode(struct AdapterInfo_s* p, int port);

uint32_t nt4ga_port_get_link_speed_capabilities(struct AdapterInfo_s* p, int port);

/*
 * port: nim capabilities
 */
nim_i2c_ctx_t nt4ga_port_get_nim_capabilities(struct AdapterInfo_s* p, int port);

/*
 * port: tx power
 */
int nt4ga_port_tx_power(struct AdapterInfo_s* p, int port, bool disable);

#endif /* NT4GA_LINK_H_ */
