/*
 * SPDX-License-Identifier: BSD-3-Clause
 * Copyright(c) 2023 Napatech A/S
 */

#ifndef I2C_NIM_H_
#define I2C_NIM_H_

#include "nthw_drv.h"
#include "nim_defines.h"
#include "nthw_mac10g.h"
#include "nthw_phy3s10g.h"
#include "nt_link_speed.h"
#include "nthw_pcal6416a.h"

#include "sensors.h"

typedef enum i2c_type {
  I2C_HWIIC,
  I2C_HWAGX
} i2c_type_e;

typedef struct sfp_nim_state {
  uint8_t br;           /* bit rate, units of 100 MBits/sec */
} sfp_nim_state_t, *sfp_nim_state_p;

typedef struct nim_i2c_ctx {
  union {
    NthwIic_t hwiic; /* depends on *Fpga_t, instance number, and cycle time */
    struct {
      nthw_pcal6416a_t *p_io_nim;
      nthw_i2cm_t* p_nt_i2cm;
      nthw_pca9849_t *pCA9849;
      int mux_channel;
    } hwagx;
  };
  i2c_type_e type;           /* 0 = hwiic (xilinx) - 1 =  hwagx (agilex) */
  uint8_t instance;
  uint8_t devaddr;
  uint8_t regaddr;
  uint8_t nimId;
  NtPortType_t portType;

  char vendorName[17];
  char prodNo[17];
  char serialNo[17];
  char date[9];
  char rev[5];
  bool avgPwr;
  bool contentValid;
  uint8_t  pwrLevelReq;
  uint8_t  pwrLevelCur;
  uint16_t lenInfo[5];
  uint32_t speedMask; /* Speeds supported by the NIM */
  int8_t laneIdx; /* Is this associated with a single lane or all lanes (-1) */
  uint8_t laneCount;
  uint32_t options;
  bool txDisable;
  bool dmiSupp;

  union {
    struct {
      bool sfp28;
      bool sfpplus;
      bool dual_rate;
      bool hw_rate_sel;
      bool sw_rate_sel;
      bool cu_type;
      bool tri_speed;
      bool  extCal;
      bool  addrChg;
    } sfp;

    struct {
      bool rxOnly;
      bool qsfp28;
      union {
        struct {
            uint8_t revCompliance;
            bool mediaSideFecCtrl;
            bool hostSideFecCtrl;
            bool mediaSideFecEna;
            bool hostSideFecEna;
        } qsfp28;
      } specific_u;
    } qsfp;

  } specific_u;
} nim_i2c_ctx_t, *nim_i2c_ctx_p;

struct nim_sensor_group {
  struct nt_adapter_sensor* sensor;
  void (*read)(struct nim_sensor_group* sg, NthwSpis_t* t_spi);
  struct nim_i2c_ctx* ctx;
  struct nim_sensor_group* next;
};

struct nim_sensor_group* allocate_nim_sensor_group(uint8_t port, struct nim_i2c_ctx* ctx, enum NtSensorSource_e ssrc, struct nt_adapter_sensor_description* sd);

/*
 * Utility functions
 */

NtNimIdentifier_t translate_nimid(const nim_i2c_ctx_t *ctx);

/*
 * Builds an nim state for the port implied by `ctx`, returns zero
 * if successful, and non-zero otherwise. SFP and QSFP nims are supported
 */
int nim_state_build(nim_i2c_ctx_t *ctx, sfp_nim_state_t *state);

/*
 * Returns a type name such as "SFP/SFP+" for a given NIM type identifier,
 * or the string "ILLEGAL!".
 */
const char *nim_id_to_text(uint8_t nim_id);

int NIM_Sfp_Nim_setTxLaserDisable(nim_i2c_ctx_p ctx, bool disable);

int NIM_QsfpPlus_Nim_setTxLaserDisable(nim_i2c_ctx_t *ctx, bool disable, int laneIdx);

int nim_set_link_speed(nim_i2c_ctx_p ctx, NtLinkSpeed_t speed);

/*
 * This function tries to classify NIM based on it's ID and some register reads
 * and collects information into ctx structure. The @extra parameter could contain
 * the initialization argument for specific type of NIMS.
 */
int construct_and_preinit_nim(nim_i2c_ctx_p ctx, void *extra, uint8_t port, struct nim_sensor_group** nim_sensors_ptr, uint16_t* nim_sensors_cnt);

int free_nim_sensors(uint8_t port, struct nim_sensor_group** nim_sensors_ptr, uint16_t* nim_sensors_cnt);
int readDataLin(nim_i2c_ctx_p ctx, uint16_t lin_addr, uint16_t length, void *data);

void qsfp28_set_high_power(nim_i2c_ctx_p ctx);
bool qsfp28_set_fec_enable(nim_i2c_ctx_p ctx, bool mediaSideFec, bool hostSideFec);

/*
 * Agilex code from here
 */

void NIM_Agx_setup(struct nim_i2c_ctx* ctx, nthw_pcal6416a_t *p_io_nim, nthw_i2cm_t* p_nt_i2cm, nthw_pca9849_t *pCA9849);
void NthwNimAgxSetLowPower(struct nim_i2c_ctx* ctx, uint8_t nimIdx, bool lowPower);

#endif /* I2C_NIM_H_ */
