/*
 * SPDX-License-Identifier: BSD-3-Clause
 * Copyright(c) 2023 Napatech A/S
 */

#ifndef __NT4GA_SPI_V3__
#define __NT4GA_SPI_V3__

/* Must include v1.x series. The first v1.0a only had 248 bytes of storage. v2.0x have 255 */
#define MAX_AVR_CONTAINER_SIZE (248)

enum avr_opcodes {
  __AVR_OP_NOP               = 0, /* v2 NOP command */
  /* version handlers */
  AVR_OP_VERSION             = 1,
  AVR_OP_SPI_VERSION         = 2, /* v2.0+ command Get protocol version */
  AVR_OP_SYSINFO             = 3,
  /* Ping handlers */
  AVR_OP_PING                = 4,
  AVR_OP_PING_DELAY          = 5,
  /* i2c handlers */
  AVR_OP_I2C_READ            = 9,
  AVR_OP_I2C_WRITE           = 10,
  AVR_OP_I2C_RANDOM_READ     = 11,
  /* VPD handlers */
  AVR_OP_VPD_READ            = 19,
  AVR_OP_VPD_WRITE           = 20,
  /* SENSOR handlers */
  AVR_OP_SENSOR_FETCH        = 28,
  /* The following command are only relevant to V3 */
  AVR_OP_SENSOR_MON_CONTROL  = 42,
  AVR_OP_SENSOR_MON_SETUP    = 43,
  /* special version handler */
  AVR_OP_SYSINFO_2           = 62,
};

#define GEN2_AVR_IDENT_SIZE   (20)
#define GEN2_AVR_VERSION_SIZE (50)

#define GEN2_PN_SIZE          (13)
#define GEN2_PBA_SIZE         (16)
#define GEN2_SN_SIZE          (10)
#define GEN2_BNAME_SIZE       (14)
#define GEN2_PLATFORM_SIZE    (72)
#define GEN2_VPD_SIZE_TOTAL (1+GEN2_PN_SIZE+GEN2_PBA_SIZE+GEN2_SN_SIZE+GEN2_BNAME_SIZE+GEN2_PLATFORM_SIZE+2)

typedef struct vpdEeprom_s {
  uint8_t psu_hw_version;                           /* Hw revision - MUST NEVER ne overwritten. */
  /* Vital Product Data: P/N   (13bytes ascii 0-9) */
  uint8_t vpd_pn[GEN2_PN_SIZE];
  /* Vital Product Data: PBA   (16bytes ascii 0-9) */
  uint8_t vpd_pba[GEN2_PBA_SIZE];
  /* Vital Product Data: S/N   (10bytes ascii 0-9) */
  uint8_t vpd_sn[GEN2_SN_SIZE];
  /* Vital Product Data: Board Name (10bytes ascii) (e.g. "ntmainb1e2" or "ntfront20b1") */
  uint8_t vpd_board_name[GEN2_BNAME_SIZE];
  /*
   * Vital Product Data: Other (72bytes of MAC addresses or other stuff.. (gives up to 12 mac
   * addresses)
   */
  uint8_t vpd_platform_section[GEN2_PLATFORM_SIZE];
  /* CRC16 checksum of all of above. This field is not included in the checksum */
  uint16_t crc16;
} vpdEeprom_t;

typedef struct {
  uint8_t  psuHwRevision;
  char     boardType[GEN2_BNAME_SIZE+1];
  char     productId[GEN2_PN_SIZE+1];
  char     pbaId[GEN2_PBA_SIZE+1];
  char     serialNumber[GEN2_SN_SIZE+1];
  uint8_t  productFamily;
  uint32_t featureMask;
  uint32_t invfeatureMask;
  uint8_t  noOfMacs;
  uint8_t  macAddress[6];
  uint16_t customId;
  uint8_t  userId[8];
} BoardInfo_t;

struct TxRxBuf {
  uint16_t  size;
  void*  pBuf;
};

struct Nthw_Spi_V3
{
  int mTimeOut;
  int mnInstanceNo;
  NthwSpim_t* mpSpimMod;
  NthwSpis_t* mpSpisMod;
};

typedef struct Nthw_Spi_V3 Nthw_Spi_V3_t;
typedef struct Nthw_Spi_V3 Nthw_Spi_V3;

Nthw_Spi_V3_t* Nthw_Spi_V3_new(void);
int Nthw_Spi_V3_init(Nthw_Spi_V3_t* p, nthw_fpga_t* pFpga, int nInstanceNo);
void Nthw_Spi_V3_delete(Nthw_Spi_V3_t* p);

int Nthw_Spi_V3_setTimeout(Nthw_Spi_V3_t* p, int timeOut);
int Nthw_Spi_V3_getVersion(Nthw_Spi_V3_t* p);
int Nthw_Spi_V3_transfer(Nthw_Spi_V3_t* p, uint16_t Opcode, struct TxRxBuf* txBuf, struct TxRxBuf* rxBuf);

#endif /* __NT4GA_SPI_V3__ */
