/*
 * SPDX-License-Identifier: BSD-3-Clause
 * Copyright(c) 2023 Napatech A/S
 */

#include <stdio.h>
#include <stdarg.h>
#include <string.h>
#include <stdlib.h>
#include <assert.h>
#include <pthread.h>
#include <arpa/inet.h> /* htons, htonl, ntohs */

#include <inttypes.h>

#include "nt_util.h"
#include "ntlog.h"

#include "flow_api_nic_setup.h"
#include "stream_binary_flow_api.h"
#include "flow_api.h"
#include "flow_api_actions.h"
#include "flow_api_backend.h"
#include "flow_api_engine.h"
#include "flow_api_profile_inline.h"
#include "flow_api_profile_inline_config.h"
#include "flow_api_hw_db_inline.h"

static int rx_queue_idx_to_hw_id(struct flow_eth_dev *dev, int id)
{
  for (int i = 0; i < dev->num_queues; ++i) {
    if (dev->rx_queue[i].id == id) {
      return dev->rx_queue[i].hw_id;
    }
  }
  return -1;
}

/*
 * Flow Matcher functionality
 */

static int flm_sdram_calibrate(struct flow_nic_dev *ndev)
{
  int success = 0;

  hw_mod_flm_control_set(&ndev->be, HW_FLM_CONTROL_PRESET_ALL, 0x0);
  hw_mod_flm_control_set(&ndev->be, HW_FLM_CONTROL_SPLIT_SDRAM_USAGE, 0x10);
  hw_mod_flm_control_flush(&ndev->be);

  /* Wait for ddr4 calibration/init done */
  for (uint32_t i = 0; i < 1000000; ++i) {
    uint32_t value = 0;
    hw_mod_flm_status_update(&ndev->be);
    hw_mod_flm_status_get(&ndev->be, HW_FLM_STATUS_CALIBDONE, &value);
    if (value) {
      success = 1;
      break;
    }
    NtOs_WaituSec(1);
  }

  if (!success) {
    /* "Flow matcher initialization failed - SDRAM calibration failed"; */
    return -1;
  }

  /* Set the flow scan and timeout settings */
  if (ndev->be.flm.ver <= 20) {
    hw_mod_flm_timeout_set(&ndev->be, HW_FLM_TIMEOUT_T, 0);
    hw_mod_flm_timeout_flush(&ndev->be);
  }

  hw_mod_flm_scan_set(&ndev->be, HW_FLM_SCAN_I, 100);
  hw_mod_flm_scan_flush(&ndev->be);

  return 0;
}

static int flm_sdram_reset(struct flow_nic_dev *ndev, int enable)
{
  int success = 0;

#if 0
  /*
   * This sleep is to make sure no packets are being looked up in
   * the flow RAM while re-initializing the RAM
   */
  NtOs_WaituSec(1000);
#endif

  /*
   * Make sure no lookup is performed during init, i.e.
   * disable every category and disable FLM
   */
  hw_mod_flm_control_set(&ndev->be, HW_FLM_CONTROL_ENABLE, 0x0);
  hw_mod_flm_control_flush(&ndev->be);

  for (uint32_t i = 1; i < ndev->be.flm.nb_categories; ++i) {
    hw_mod_flm_rcp_set(&ndev->be, HW_FLM_RCP_PRESET_ALL, i, 0x0);
  }
  hw_mod_flm_rcp_flush(&ndev->be, 1, ndev->be.flm.nb_categories - 1);

  /* Wait for FLM to enter Idle state */
  for (uint32_t i = 0; i < 1000000; ++i) {
    uint32_t value = 0;
    hw_mod_flm_status_update(&ndev->be);
    hw_mod_flm_status_get(&ndev->be, HW_FLM_STATUS_IDLE, &value);
    if (value) {
      success = 1;
      break;
    }
    NtOs_WaituSec(1);
  }

  if (!success) {
    /* "Flow matcher initialization failed - never idle"; */
    return -1;
  }

  success = 0;

  /* Start SDRAM initialization */
  hw_mod_flm_control_set(&ndev->be, HW_FLM_CONTROL_INIT, 0x1);
  hw_mod_flm_control_flush(&ndev->be);

  for (uint32_t i = 0; i < 1000000; ++i) {
    uint32_t value = 0;
    hw_mod_flm_status_update(&ndev->be);
    hw_mod_flm_status_get(&ndev->be, HW_FLM_STATUS_INITDONE, &value);
    if (value) {
      success = 1;
      break;
    }
    NtOs_WaituSec(1);
  }

  if (!success) {
    /* "Flow matcher initialization failed - SDRAM initialization incomplete"; */
    return -1;
  }

  /* Set the INIT value back to zero to clear the bit in the SW register cache */
  hw_mod_flm_control_set(&ndev->be, HW_FLM_CONTROL_INIT, 0x0);
  hw_mod_flm_control_flush(&ndev->be);

  /* Enable FLM */
  hw_mod_flm_control_set(&ndev->be, HW_FLM_CONTROL_ENABLE, enable);
  hw_mod_flm_control_flush(&ndev->be);

  return 0;
}

struct flm_flow_key_def_s {
  union {
    struct {
      uint64_t qw0_dyn : 7;
      uint64_t qw0_ofs : 8;
      uint64_t qw4_dyn : 7;
      uint64_t qw4_ofs : 8;
      uint64_t sw8_dyn : 7;
      uint64_t sw8_ofs : 8;
      uint64_t sw9_dyn : 7;
      uint64_t sw9_ofs : 8;
      uint64_t outer_proto : 1;
      uint64_t inner_proto : 1;
      uint64_t pad : 2;
    };
    uint64_t data;
  };
  uint32_t mask[10];
};

static inline void set_key_def_qw(struct flm_flow_key_def_s *key_def, unsigned int qw, unsigned int dyn, unsigned int ofs)
{
  assert(qw < 2);
  if (qw == 0) {
    key_def->qw0_dyn = dyn & 0x7f;
    key_def->qw0_ofs = ofs & 0xff;
  }
  else {
    key_def->qw4_dyn = dyn & 0x7f;
    key_def->qw4_ofs = ofs & 0xff;
  }
}

static inline void set_key_def_sw(struct flm_flow_key_def_s *key_def, unsigned int sw, unsigned int dyn, unsigned int ofs)
{
  assert(sw < 2);
  if (sw == 0) {
    key_def->sw8_dyn = dyn & 0x7f;
    key_def->sw8_ofs = ofs & 0xff;
  }
  else {
    key_def->sw9_dyn = dyn & 0x7f;
    key_def->sw9_ofs = ofs & 0xff;
  }
}

#define FLM_MTR_PROFILE_SIZE  0x100000

struct flm_flow_mtr_handle_s {
  struct dual_buckets_s {
    uint16_t rate_a;
    uint16_t rate_b;
    uint16_t size_a;
    uint16_t size_b;
  } dual_buckets[FLM_MTR_PROFILE_SIZE];
};

int flow_mtr_supported(struct flow_eth_dev *dev)
{
  return hw_mod_flm_present(&dev->ndev->be) && dev->ndev->be.flm.nb_variant == 2;
}

uint64_t flow_mtr_meter_policy_n_max(void)
{
  return FLM_MTR_PROFILE_SIZE;
}

static inline uint64_t convert_to_bucket_size_units(uint64_t value)
{
  /* Assumes a 40-bit int as input */
  uint64_t lo_bits = (value & 0xfffff) * 1000000000;
  uint64_t hi_bits = ((value >> 20) & 0xfffff) * 1000000000;
  uint64_t round_up = (hi_bits & 0xfffff) || (lo_bits & 0xffffffffff) ? 1 : 0;
  return (hi_bits >> 20) + (lo_bits >> 40) + round_up;
}


int flow_mtr_set_profile(struct flow_eth_dev *dev,
    uint32_t profile_id,
    uint64_t bucket_rate_a, uint64_t bucket_size_a,
    uint64_t bucket_rate_b, uint64_t bucket_size_b)
{
  struct flow_nic_dev *ndev = dev->ndev;
  struct flm_flow_mtr_handle_s *handle = (struct flm_flow_mtr_handle_s*)ndev->flm_mtr_handle;
  struct dual_buckets_s *buckets = &handle->dual_buckets[profile_id];

  uint64_t bucket_rate_shift_a = 0;
  uint64_t bucket_rate_shift_b = 0;

  uint64_t bucket_size_shift_a = 0;
  uint64_t bucket_size_shift_b = 0;

  /* Round rates up to nearest 128 bytes/sec and shift to 128 bytes/sec units */
  bucket_rate_a = (bucket_rate_a & 0x7f) ? (bucket_rate_a >> 7) + 1 : (bucket_rate_a >> 7);
  bucket_rate_b = (bucket_rate_b & 0x7f) ? (bucket_rate_b >> 7) + 1 : (bucket_rate_b >> 7);

  /* Round rate down to max rate supported */
  if (bucket_rate_a > 0x7ff8000) bucket_rate_a = 0x7ff8000;
  if (bucket_rate_b > 0x7ff8000) bucket_rate_b = 0x7ff8000;

  /* Find shift to convert into 12-bit int */
  while ((bucket_rate_a >> bucket_rate_shift_a) > 0xfff) bucket_rate_shift_a += 1;
  while ((bucket_rate_b >> bucket_rate_shift_b) > 0xfff) bucket_rate_shift_b += 1;

  /* Store in format [11:0] shift-left [15:12] */
  buckets->rate_a = (bucket_rate_a >> bucket_rate_shift_a) | (bucket_rate_shift_a << 12);
  buckets->rate_b = (bucket_rate_b >> bucket_rate_shift_b) | (bucket_rate_shift_b << 12);

  /* Round size down to 38-bit int */
  if (bucket_size_a > 0x3fffffffff) bucket_size_a = 0x3fffffffff;
  if (bucket_size_b > 0x3fffffffff) bucket_size_b = 0x3fffffffff;

  /* Convert size to units of 2^40 / 10^9. Output is a 28-bit int. */
  bucket_size_a = convert_to_bucket_size_units(bucket_size_a);
  bucket_size_b = convert_to_bucket_size_units(bucket_size_b);

  /* Round rate down to max rate supported */
  if (bucket_size_a > 0x7ff8000) bucket_size_a = 0x7ff8000;
  if (bucket_size_b > 0x7ff8000) bucket_size_b = 0x7ff8000;

  /* Find shift to convert into 12-bit int */
  while ((bucket_size_a >> bucket_size_shift_a) > 0xfff) bucket_size_shift_a += 1;
  while ((bucket_size_b >> bucket_size_shift_b) > 0xfff) bucket_size_shift_b += 1;

  /* Store in format [11:0] shift-left [15:12] */
  buckets->size_a = (bucket_size_a >> bucket_size_shift_a) | (bucket_size_shift_a << 12);
  buckets->size_b = (bucket_size_b >> bucket_size_shift_b) | (bucket_size_shift_b << 12);

  return 0;
}

int flow_mtr_set_policy(UNUSED struct flow_eth_dev *dev,
    UNUSED uint32_t policy_id, UNUSED int drop)
{
  return 0;
}

#define FLM_MTR_STAT_SIZE  0x1000000
#define WORDS_PER_INF_DATA (sizeof(struct flm_v17_inf_data_s) / sizeof(uint32_t))
#define MAX_INF_DATA_RECORDS_PER_READ 20
#define UINT64_MSB ((uint64_t)1 << 63)

/* 2^23 bytes ~ 8MB */
#define FLM_PERIODIC_STATS_BYTE_LIMIT 8
/* 2^16 pkt ~ 64K pkt */
#define FLM_PERIODIC_STATS_PKT_LIMIT 5
/* 2^38 ns ~ 275 sec */
#define FLM_PERIODIC_STATS_BYTE_TIMEOUT 23


uint32_t flow_mtr_meters_supported(void) {
  return FLM_MTR_STAT_SIZE;
}

#include <stdatomic.h>

struct mtr_stat_s {
  struct dual_buckets_s *buckets;
  _Atomic volatile uint64_t n_pkt;
  _Atomic volatile uint64_t n_bytes;
  uint64_t n_pkt_base;
  uint64_t n_bytes_base;
  _Atomic volatile uint64_t stats_mask;
};

#define WORDS_PER_LEARN_DATA sizeof(struct flm_v17_lrn_data_s)
#define FLM_PROG_MAX_RETRY 100

static uint32_t flm_read_inf_rec_locked(struct flow_eth_dev *dev, uint32_t *data);

static int flow_flm_apply(struct flow_eth_dev *dev, struct flm_v17_lrn_data_s *learn_record) {
  uint32_t lrn_ready;
  uint32_t retry = 0;
  uint32_t data[WORDS_PER_INF_DATA * MAX_INF_DATA_RECORDS_PER_READ];

  hw_mod_flm_buf_ctrl_get(&dev->ndev->be, HW_FLM_BUF_CTRL_LRN_FREE, &lrn_ready);
  if (lrn_ready < WORDS_PER_LEARN_DATA) {
    hw_mod_flm_buf_ctrl_update(&dev->ndev->be);
    hw_mod_flm_buf_ctrl_get(&dev->ndev->be, HW_FLM_BUF_CTRL_LRN_FREE, &lrn_ready);
    while (lrn_ready < WORDS_PER_LEARN_DATA) {
      ++retry;
      if (retry > FLM_PROG_MAX_RETRY) {
        return 1;
      }

      flm_read_inf_rec_locked(dev, data);

      hw_mod_flm_buf_ctrl_update(&dev->ndev->be);
      hw_mod_flm_buf_ctrl_get(&dev->ndev->be, HW_FLM_BUF_CTRL_LRN_FREE, &lrn_ready);
    }
  }

  int res = hw_mod_flm_lrn_data_set_flush(&dev->ndev->be, HW_FLM_FLOW_LRN_DATA_V17, (uint32_t *)learn_record);
  return res;
}

int flow_mtr_create_meter(struct flow_eth_dev *dev,
                          uint32_t mtr_id,
                          uint32_t profile_id,
                          UNUSED uint32_t policy_id,
                          uint64_t stats_mask)
{
  pthread_mutex_lock(&dev->ndev->mtx);

  struct flm_flow_mtr_handle_s *handle = (struct flm_flow_mtr_handle_s*)dev->ndev->flm_mtr_handle;
  struct dual_buckets_s *buckets = &handle->dual_buckets[profile_id];

  struct flm_v17_lrn_data_s learn_record;
  memset(&learn_record, 0x0, sizeof(struct flm_v17_lrn_data_s));

  learn_record.sw9 = mtr_id + 1;
  learn_record.kid = 1;

  learn_record.rate = buckets->rate_a;
  learn_record.size = buckets->size_a;
  learn_record.fill = buckets->size_a & 0x0fff;

  learn_record.ft_mbr = 15;  /* FT to assign if MBR has been exceeded */

  learn_record.ent = 1;
  learn_record.op = 1;
  learn_record.eor = 1;

  learn_record.id[0] = mtr_id & 0xff;
  learn_record.id[1] = (mtr_id >> 8)  & 0xff;
  learn_record.id[2] = (mtr_id >> 16)  & 0xff;
  learn_record.id[3] = (mtr_id >> 24)  & 0xff;
  learn_record.id[8] = 1U << 7;

  if (stats_mask) {
    learn_record.vol_idx = 1;
  }

  int res = flow_flm_apply(dev, &learn_record);

  if (res == 0) {
    struct mtr_stat_s *mtr_stat = dev->ndev->mtr_stat_handle;
    mtr_stat[mtr_id].buckets = buckets;
    atomic_store(&mtr_stat[mtr_id].stats_mask, stats_mask);
  }

  pthread_mutex_unlock(&dev->ndev->mtx);

  return res;
}

int flow_mtr_destroy_meter(struct flow_eth_dev *dev,
    uint32_t mtr_id)
{
  pthread_mutex_lock(&dev->ndev->mtx);

  struct flm_v17_lrn_data_s learn_record;
  memset(&learn_record, 0x0, sizeof(struct flm_v17_lrn_data_s));

  learn_record.sw9 = mtr_id + 1;
  learn_record.kid = 1;

  learn_record.ent = 1;
  learn_record.op = 0;
  learn_record.eor = 1;

  learn_record.id[0] = mtr_id & 0xff;
  learn_record.id[1] = (mtr_id >> 8)  & 0xff;
  learn_record.id[2] = (mtr_id >> 16)  & 0xff;
  learn_record.id[3] = (mtr_id >> 24)  & 0xff;
  learn_record.id[8] = 1U << 7;

  /* Clear statistics so stats_mask prevents updates of counters on deleted meters */
  struct mtr_stat_s *mtr_stat = dev->ndev->mtr_stat_handle;
  atomic_store(&mtr_stat[mtr_id].stats_mask, 0);
  atomic_store(&mtr_stat[mtr_id].n_bytes, 0);
  atomic_store(&mtr_stat[mtr_id].n_pkt, 0);
  mtr_stat[mtr_id].n_bytes_base = 0;
  mtr_stat[mtr_id].n_pkt_base = 0;
  mtr_stat[mtr_id].buckets = NULL;

  int res = flow_flm_apply(dev, &learn_record);
  pthread_mutex_unlock(&dev->ndev->mtx);

  return res;
}

int flm_mtr_adjust_stats(struct flow_eth_dev *dev,
  uint32_t mtr_id, uint32_t adjust_value)
{
  pthread_mutex_lock(&dev->ndev->mtx);

  struct mtr_stat_s *mtr_stat = &((struct mtr_stat_s*)dev->ndev->mtr_stat_handle)[mtr_id];

  struct flm_v17_lrn_data_s learn_record;
  memset(&learn_record, 0x0, sizeof(struct flm_v17_lrn_data_s));

  learn_record.sw9 = mtr_id + 1;
  learn_record.kid = 1;

  learn_record.rate = mtr_stat->buckets->rate_a;
  learn_record.size = mtr_stat->buckets->size_a;
  learn_record.adj  = adjust_value;

  learn_record.ft_mbr = 15;

  learn_record.ent = 1;
  learn_record.op = 2;
  learn_record.eor = 1;

  if (atomic_load(&mtr_stat->stats_mask)) {
    learn_record.vol_idx = 1;
  }

  int res = flow_flm_apply(dev, &learn_record);

  pthread_mutex_unlock(&dev->ndev->mtx);

  return res;
}

static uint32_t flm_read_inf_rec_locked(struct flow_eth_dev *dev, uint32_t *data) {
  uint32_t inf_cnt = 0;

  hw_mod_flm_buf_ctrl_get(&dev->ndev->be, HW_FLM_BUF_CTRL_INF_AVAIL, &inf_cnt);
  if (inf_cnt < WORDS_PER_INF_DATA) {
    hw_mod_flm_buf_ctrl_update(&dev->ndev->be);
    hw_mod_flm_buf_ctrl_get(&dev->ndev->be, HW_FLM_BUF_CTRL_INF_AVAIL, &inf_cnt);
  }

  uint32_t records_to_read = inf_cnt / WORDS_PER_INF_DATA;
  if (records_to_read == 0) {
    return 0;
  }
  if (records_to_read > MAX_INF_DATA_RECORDS_PER_READ) {
    records_to_read = MAX_INF_DATA_RECORDS_PER_READ;
  }

  hw_mod_flm_inf_data_update_get(&dev->ndev->be, HW_FLM_FLOW_INF_DATA_V17, data, records_to_read * WORDS_PER_INF_DATA);

  return records_to_read;
}

uint32_t flm_mtr_update_stats(struct flow_eth_dev *dev) {
  uint32_t data[WORDS_PER_INF_DATA * MAX_INF_DATA_RECORDS_PER_READ];

  pthread_mutex_lock(&dev->ndev->mtx);
  uint32_t records = flm_read_inf_rec_locked(dev, data);
  pthread_mutex_unlock(&dev->ndev->mtx);

  struct mtr_stat_s *mtr_stat = dev->ndev->mtr_stat_handle;

  for (uint32_t i = 0; i < records; ++i) {
    uint32_t* p_record = &data[i * WORDS_PER_INF_DATA];

    /* Check that received record hold valid meter statistics */
    if ((p_record[6] < flow_mtr_meters_supported() && p_record[7] == 0 && (p_record[8] >> 31) == 1)) {
      uint32_t id = p_record[6];

      /* Don't update a deleted meter */
      uint64_t stats_mask = atomic_load(&mtr_stat[id].stats_mask);
      if (stats_mask) {
        uint64_t nb = ((uint64_t)p_record[1] << 32) | p_record[0];
        uint64_t np = ((uint64_t)p_record[3] << 32) | p_record[2];

        atomic_store(&mtr_stat[id].n_pkt, np | UINT64_MSB);
        atomic_store(&mtr_stat[id].n_bytes, nb);
        atomic_store(&mtr_stat[id].n_pkt, np);
#if 0
        printf("mtr_stat[%u].n_bytes %lu (%lx)\n", id, nb, nb);
        printf("mtr_stat[%u].n_pkt %lu (%lx)\n", np, np);
#endif
      }
    }
  }

  return records;
}

void flm_mtr_read_stats(struct flow_eth_dev *dev,
                        uint32_t id,
                        uint64_t *stats_mask,
                        uint64_t *green_pkt,
                        uint64_t *green_bytes,
                        int clear) {
  struct mtr_stat_s *mtr_stat = dev->ndev->mtr_stat_handle;
  *stats_mask = atomic_load(&mtr_stat[id].stats_mask);
  if (*stats_mask) {
    uint64_t pkt_1;
    uint64_t pkt_2;
    uint64_t nb;
    do {
      do {
        pkt_1 = atomic_load(&mtr_stat[id].n_pkt);
      } while(pkt_1 & UINT64_MSB);
      nb = atomic_load(&mtr_stat[id].n_bytes);
      pkt_2 = atomic_load(&mtr_stat[id].n_pkt);
    } while(pkt_1 != pkt_2);

    *green_pkt = pkt_1 - mtr_stat[id].n_pkt_base;
    *green_bytes = nb - mtr_stat[id].n_bytes_base;
    if (clear) {
      mtr_stat[id].n_pkt_base = pkt_1;
      mtr_stat[id].n_bytes_base = nb;
    }
  }
}


static inline uint8_t convert_port_to_ifr_mtu_recipe(uint32_t port) {
  return port + 1;
}

static uint8_t get_port_from_port_id(struct flow_nic_dev *ndev, uint32_t port_id)
{
  struct flow_eth_dev *dev = ndev->eth_base;

  while (dev) {
    if (dev->port_id == port_id) {
      return dev->port;
    }
    dev = dev->next;
  }

  return UINT8_MAX;
}

static void nic_insert_flow(struct flow_nic_dev *ndev, struct flow_handle *fh)
{
  if (ndev->flow_base) ndev->flow_base->prev = fh;
  fh->next = ndev->flow_base;
  fh->prev = NULL;
  ndev->flow_base = fh;
}

static void nic_remove_flow(struct flow_nic_dev *ndev, struct flow_handle *fh)
{
  struct flow_handle *next = fh->next;
  struct flow_handle *prev = fh->prev;

  if (next && prev) {
    prev->next = next;
    next->prev = prev;
  }
  else if (next) {
    ndev->flow_base = next;
    next->prev = NULL;
  }
  else if (prev) {
    prev->next = NULL;
  }
  else if (ndev->flow_base == fh) {
    ndev->flow_base = NULL;
  }
}

static void nic_insert_flow_flm(struct flow_nic_dev *ndev, struct flow_handle *fh)
{
  if (ndev->flow_base_flm) ndev->flow_base_flm->prev = fh;
  fh->next = ndev->flow_base_flm;
  fh->prev = NULL;
  ndev->flow_base_flm = fh;
}

static void nic_remove_flow_flm(struct flow_nic_dev *ndev, struct flow_handle *fh_flm)
{
  struct flow_handle *next = fh_flm->next;
  struct flow_handle *prev = fh_flm->prev;

  if (next && prev) {
    prev->next = next;
    next->prev = prev;
  }
  else if (next) {
    ndev->flow_base_flm = next;
    next->prev = NULL;
  }
  else if (prev) {
    prev->next = NULL;
  }
  else if (ndev->flow_base_flm == fh_flm) {
    ndev->flow_base_flm = NULL;
  }
}

static struct nic_flow_def* interpret_flow_elements(
    struct flow_eth_dev *dev,
    const struct flow_elem elem[],
    const struct flow_action action[],
    struct flow_error *error,
    uint16_t implicit_vlan_vid,
    uint32_t *in_port_id,
    uint32_t *num_dest_port,
    uint32_t *num_queues,
    uint32_t *packet_data,
    uint32_t *packet_mask,
    struct flm_flow_key_def_s *key_def)
{
  uint32_t any_count = 0;
  int mtr_count = 0;

  unsigned int encap_decap_order = 0;

  unsigned int qw_counter = 0;
  unsigned int sw_counter = 0;

  uint64_t modify_field_use_flags = 0x0;

  *in_port_id = UINT32_MAX;
  *num_dest_port = 0;
  *num_queues = 0;

  memset(packet_data, 0x0, sizeof(uint32_t) * 10);
  memset(packet_mask, 0x0, sizeof(uint32_t) * 10);
  key_def->data = 0;

  if (action == NULL || elem == NULL) {
    flow_nic_set_error(ERR_FAILED, error);
    NT_LOG(ERR, FILTER, "Flow items / actions missing\n");
    return NULL;
  }

  struct nic_flow_def *fd = calloc(1, sizeof(struct nic_flow_def));
  if (!fd) {
    flow_nic_set_error(ERR_MEMORY, error);
    NT_LOG(ERR, FILTER, "ERR Memory\n");
    return NULL;
  }

  /* Set default values for fd */
  fd->full_offload     = -1;
  fd->in_port_override = -1;
  fd->mark             = UINT32_MAX;
  fd->jump_to_group    = UINT32_MAX;

  fd->l2_prot = -1;
  fd->l3_prot = -1;
  fd->l4_prot = -1;
  fd->vlans = 0;
  fd->tunnel_prot = -1;
  fd->tunnel_l3_prot = -1;
  fd->tunnel_l4_prot = -1;
  fd->fragmentation = -1;

  NT_LOG(DBG, FILTER, ">>>>> [Dev %p] Nic %i, Port %i: fd %p - FLOW Interpretation <<<<<\n", dev, dev->ndev->adapter_no, dev->port, fd);

  /*
   * Gather flow match + actions and convert into internal flow definition structure (struct nic_flow_def_s)
   * This is the 1st step in the flow creation - validate, convert and prepare
   */
  for (int aidx = 0; action[aidx].type != FLOW_ACTION_TYPE_END; ++aidx) {

    switch (action[aidx].type) {
    case FLOW_ACTION_TYPE_PORT_ID:
      NT_LOG(DBG, FILTER, "Dev:%p: FLOW_ACTION_TYPE_PORT_ID\n", dev);
      if (action[aidx].conf)
      {
        if (*num_dest_port > 0) {
          NT_LOG(ERR, FILTER, "Multiple port_id actions for one flow is not supported\n");
          flow_nic_set_error(ERR_ACTION_MULTIPLE_PORT_ID_UNSUPPORTED, error);
          free(fd);
          return NULL;
        }

        uint32_t port_id = ((const struct flow_action_port_id *)action[aidx].conf)->id;
        uint8_t port = get_port_from_port_id(dev->ndev, port_id);

        if (fd->dst_num_avail == MAX_OUTPUT_DEST) {
          /* ERROR too many output destinations */
          NT_LOG(ERR, FILTER, "Too many output destinations\n");
          flow_nic_set_error(ERR_OUTPUT_TOO_MANY, error);
          free(fd);
          return NULL;
        }

        if (port >= dev->ndev->be.num_phy_ports) {
          /* ERROR phy port out of range */
          NT_LOG(ERR, FILTER, "Phy port out of range\n");
          flow_nic_set_error(ERR_OUTPUT_INVALID, error);
          free(fd);
          return NULL;
        }

        /* New destination port to add */
        fd->dst_id[fd->dst_num_avail].owning_port_id = port_id;
        fd->dst_id[fd->dst_num_avail].type = PORT_PHY;
        fd->dst_id[fd->dst_num_avail].id = (int)port;
        fd->dst_id[fd->dst_num_avail].active = 1;
        fd->dst_num_avail++;

        fd->flm_mtu_fragmentation_recipe = convert_port_to_ifr_mtu_recipe(port);

        if (fd->full_offload < 0) {
          fd->full_offload = 1;
        }

        *num_dest_port += 1;

        NT_LOG(DBG, FILTER, "Phy port ID: %i\n", (int)port);
      }
      break;

    case FLOW_ACTION_TYPE_QUEUE:
      NT_LOG(DBG, FILTER, "Dev:%p: FLOW_ACTION_TYPE_QUEUE\n", dev);
      if (action[aidx].conf)
      {
        const struct flow_action_queue *queue = (const struct flow_action_queue *)action[aidx].conf;

        int hw_id = rx_queue_idx_to_hw_id(dev, queue->index);

        fd->dst_id[fd->dst_num_avail].owning_port_id = dev->port;
        fd->dst_id[fd->dst_num_avail].id = hw_id;
        fd->dst_id[fd->dst_num_avail].type = PORT_VIRT;
        fd->dst_id[fd->dst_num_avail].active = 1;
        fd->dst_num_avail++;

        NT_LOG(DBG, FILTER, "Dev:%p: FLOW_ACTION_TYPE_QUEUE port %u, queue index: %u, hw id %u\n", dev, dev->port, queue->index, hw_id);

        fd->full_offload = 0;
        *num_queues += 1;
      }
      break;

    case FLOW_ACTION_TYPE_RSS:
      NT_LOG(DBG, FILTER, "Dev:%p: FLOW_ACTION_TYPE_RSS\n", dev);
      if (action[aidx].conf)
      {
        const struct flow_action_rss *rss = (const struct flow_action_rss *)action[aidx].conf;

        for (uint32_t i = 0; i < rss->queue_num; ++i) {
          int hw_id = rx_queue_idx_to_hw_id(dev, rss->queue[i]);

          fd->dst_id[fd->dst_num_avail].owning_port_id = dev->port;
          fd->dst_id[fd->dst_num_avail].id = hw_id;
          fd->dst_id[fd->dst_num_avail].type = PORT_VIRT;
          fd->dst_id[fd->dst_num_avail].active = 1;
          fd->dst_num_avail++;
        }

        fd->full_offload = 0;
        *num_queues += rss->queue_num;
      }
      break;

    case FLOW_ACTION_TYPE_MARK:
      NT_LOG(DBG, FILTER, "Dev:%p: FLOW_ACTION_TYPE_MARK\n", dev);
      if (action[aidx].conf)
      {
        fd->mark = ((const struct flow_action_mark *)action[aidx].conf)->id;
        NT_LOG(DBG, FILTER, "Mark: %i\n", fd->mark);
      }
      break;

    case FLOW_ACTION_TYPE_JUMP:
      NT_LOG(DBG, FILTER, "Dev:%p: FLOW_ACTION_TYPE_JUMP\n", dev);
      if (action[aidx].conf)
      {
        const struct flow_action_jump *jump = (const struct flow_action_jump *)action[aidx].conf;
        fd->jump_to_group = jump->group;
        NT_LOG(DBG, FILTER, "Dev:%p: FLOW_ACTION_TYPE_JUMP: group %u\n", dev, jump->group);
      }
      break;

    case FLOW_ACTION_TYPE_DROP:
      NT_LOG(DBG, FILTER, "Dev:%p: FLOW_ACTION_TYPE_DROP\n", dev);
      if (action[aidx].conf)
      {
        fd->dst_id[fd->dst_num_avail].owning_port_id = 0;
        fd->dst_id[fd->dst_num_avail].id = 0;
        fd->dst_id[fd->dst_num_avail].type = PORT_NONE;
        fd->dst_num_avail++;
      }
      break;

    case FLOW_ACTION_TYPE_METER:
      NT_LOG(DBG, FILTER, "Dev:%p: FLOW_ACTION_TYPE_METER\n", dev);
      if (action[aidx].conf)
      {
        const struct flow_action_meter *meter = (const struct flow_action_meter *)action[aidx].conf;
        if (mtr_count >= MAX_FLM_MTRS_SUPPORTED) {
          NT_LOG(ERR, FILTER, "ERROR: - Number of METER actions exceeds %d.\n", MAX_FLM_MTRS_SUPPORTED);
          flow_nic_set_error(ERR_ACTION_UNSUPPORTED, error);
          free(fd);
          return NULL;
        }
        fd->mtr_ids[mtr_count++] = meter->mtr_id + 1;
      }
      break;

    case FLOW_ACTION_TYPE_RAW_ENCAP:
      NT_LOG(DBG, FILTER, "Dev:%p: FLOW_ACTION_TYPE_RAW_ENCAP\n", dev);
      if (action[aidx].conf)
      {
        const struct flow_action_raw_encap *encap = (const struct flow_action_raw_encap *)action[aidx].conf;
        const struct flow_elem *items = encap->items;

        if (encap_decap_order != 1) {
          NT_LOG(ERR, FILTER, "ERROR: - RAW_ENCAP must follow RAW_DECAP.\n");
          flow_nic_set_error(ERR_ACTION_UNSUPPORTED, error);
          free(fd);
          return NULL;
        }

        if (encap->size == 0 || encap->size > 255 || encap->item_count < 2) {
          NT_LOG(ERR, FILTER, "ERROR: - RAW_ENCAP data/size invalid.\n");
          flow_nic_set_error(ERR_ACTION_UNSUPPORTED, error);
          free(fd);
          return NULL;
        }

        encap_decap_order = 2;

        fd->tun_hdr.len = (uint8_t)encap->size;
        memcpy(fd->tun_hdr.d.hdr8, encap->data, fd->tun_hdr.len);

        while (items->type != FLOW_ELEM_TYPE_END) {
          switch (items->type) {
          case FLOW_ELEM_TYPE_ETH:
            fd->tun_hdr.l2_len = 14;
            break;
          case FLOW_ELEM_TYPE_VLAN:
            fd->tun_hdr.nb_vlans += 1;
            fd->tun_hdr.l2_len += 4;
            break;
          case FLOW_ELEM_TYPE_IPV4:
            fd->tun_hdr.ip_version = 4;
            fd->tun_hdr.l3_len = sizeof(struct ipv4_hdr_s);
            fd->tun_hdr.new_outer = 1;

            /* Patch length */
            fd->tun_hdr.d.hdr8[fd->tun_hdr.l2_len + 2] = 0x07;
            fd->tun_hdr.d.hdr8[fd->tun_hdr.l2_len + 3] = 0xfd;
            break;
          case FLOW_ELEM_TYPE_IPV6:
            fd->tun_hdr.ip_version = 6;
            fd->tun_hdr.l3_len = sizeof(struct ipv6_hdr_s);
            fd->tun_hdr.new_outer = 1;

            /* Patch length */
            fd->tun_hdr.d.hdr8[fd->tun_hdr.l2_len + 4] = 0x07;
            fd->tun_hdr.d.hdr8[fd->tun_hdr.l2_len + 5] = 0xfd;
            break;
          case FLOW_ELEM_TYPE_SCTP:
            fd->tun_hdr.l4_len = sizeof(struct sctp_hdr_s);
            break;
          case FLOW_ELEM_TYPE_TCP:
            fd->tun_hdr.l4_len = sizeof(struct tcp_hdr_s);
            break;
          case FLOW_ELEM_TYPE_UDP:
            fd->tun_hdr.l4_len = sizeof(struct udp_hdr_s);

            /* Patch length */
            fd->tun_hdr.d.hdr8[fd->tun_hdr.l2_len + fd->tun_hdr.l3_len + 4] = 0x07;
            fd->tun_hdr.d.hdr8[fd->tun_hdr.l2_len + fd->tun_hdr.l3_len + 5] = 0xfd;
            break;
          case FLOW_ELEM_TYPE_ICMP:
            fd->tun_hdr.l4_len = sizeof(struct icmp_hdr_s);
            break;
          case FLOW_ELEM_TYPE_GTP:
            /* Patch length */
            fd->tun_hdr.d.hdr8[fd->tun_hdr.l2_len + fd->tun_hdr.l3_len + fd->tun_hdr.l4_len + 2] = 0x07;
            fd->tun_hdr.d.hdr8[fd->tun_hdr.l2_len + fd->tun_hdr.l3_len + fd->tun_hdr.l4_len + 3] = 0xfd;
            break;
          default:
            break;
          }
          items++;
        }

        if (fd->tun_hdr.nb_vlans > 3) {
          NT_LOG(ERR, FILTER, "ERROR: - Encapsulation with %d vlans not supported.\n", (int)fd->tun_hdr.nb_vlans);
          flow_nic_set_error(ERR_ACTION_UNSUPPORTED, error);
          free(fd);
          return NULL;
        }

        /* Convert encap data to 128-bit little endian */
        for (size_t i = 0; i < (encap->size + 15) / 16; ++i) {
          uint8_t *data = fd->tun_hdr.d.hdr8 + i * 16;
          for (unsigned int j = 0; j < 8; ++j) {
            uint8_t t = data[j];
            data[j] = data[15 - j];
            data[15 - j] = t;
          }
        }
      }
      break;

    case FLOW_ACTION_TYPE_RAW_DECAP:
      NT_LOG(DBG, FILTER, "Dev:%p: FLOW_ACTION_TYPE_RAW_DECAP\n", dev);
      if (action[aidx].conf)
      {
        const struct flow_action_raw_decap *decap = (const struct flow_action_raw_decap *)action[aidx].conf;

        if (encap_decap_order != 0) {
          NT_LOG(ERR, FILTER, "ERROR: - RAW_ENCAP must follow RAW_DECAP.\n");
          flow_nic_set_error(ERR_ACTION_UNSUPPORTED, error);
          free(fd);
          return NULL;
        }

        if (decap->item_count < 2) {
          NT_LOG(ERR, FILTER, "ERROR: - RAW_DECAP must decap something.\n");
          flow_nic_set_error(ERR_ACTION_UNSUPPORTED, error);
          free(fd);
          return NULL;
        }

        encap_decap_order = 1;

        switch (decap->items[decap->item_count - 2].type) {
        case FLOW_ELEM_TYPE_ETH:
        case FLOW_ELEM_TYPE_VLAN:
          fd->header_strip_end_dyn = 4;
          fd->header_strip_end_ofs = 0;
          break;
        case FLOW_ELEM_TYPE_IPV4:
        case FLOW_ELEM_TYPE_IPV6:
          fd->header_strip_end_dyn = 7;
          fd->header_strip_end_ofs = 0;
          break;
        case FLOW_ELEM_TYPE_SCTP:
        case FLOW_ELEM_TYPE_TCP:
        case FLOW_ELEM_TYPE_UDP:
        case FLOW_ELEM_TYPE_ICMP:
          fd->header_strip_end_dyn = 8;
          fd->header_strip_end_ofs = 0;
          break;
        case FLOW_ELEM_TYPE_GTP:
          fd->header_strip_end_dyn = 13;
          fd->header_strip_end_ofs = 0;
          break;
        default:
          fd->header_strip_end_dyn = 1;
          fd->header_strip_end_ofs = 0;
          break;
        }
      }
      break;

    case FLOW_ACTION_TYPE_MODIFY_FIELD:
      NT_LOG(DBG, FILTER, "Dev:%p: FLOW_ACTION_TYPE_MODIFY_FIELD\n", dev);
      {
        const struct flow_action_modify_field *modify_field = (const struct flow_action_modify_field *)action[aidx].conf;
        uint64_t modify_field_use_flag = 0;

        if (modify_field->src.field != FLOW_FIELD_VALUE) {
          NT_LOG(ERR, FILTER, "MODIFY_FIELD only src type VALUE is supported.\n");
          flow_nic_set_error(ERR_ACTION_UNSUPPORTED, error);
          free(fd);
          return NULL;
        }

        if (modify_field->dst.level > 2) {
          NT_LOG(ERR, FILTER, "MODIFY_FIELD only dst level 0, 1, and 2 is supported.\n");
          flow_nic_set_error(ERR_ACTION_UNSUPPORTED, error);
          free(fd);
          return NULL;
        }

        if (modify_field->dst.field == FLOW_FIELD_IPV4_TTL || modify_field->dst.field == FLOW_FIELD_IPV6_HOPLIMIT) {
          if (modify_field->operation != FLOW_MODIFY_SUB) {
            NT_LOG(ERR, FILTER, "MODIFY_FIELD only operation SUB is supported for TTL/HOPLIMIT.\n");
            flow_nic_set_error(ERR_ACTION_UNSUPPORTED, error);
            free(fd);
            return NULL;
          }

          if (fd->ttl_sub_enable) {
            NT_LOG(ERR, FILTER, "MODIFY_FIELD TTL/HOPLIMIT resource already in use.\n");
            flow_nic_set_error(ERR_ACTION_UNSUPPORTED, error);
            free(fd);
            return NULL;
          }

          fd->ttl_sub_enable = 1;
          fd->ttl_sub_ipv4   = (modify_field->dst.field == FLOW_FIELD_IPV4_TTL) ? 1 : 0;
          fd->ttl_sub_outer  = (modify_field->dst.level <= 1) ? 1 : 0;
        }
        else {
          if (modify_field->operation != FLOW_MODIFY_SET) {
            NT_LOG(ERR, FILTER, "MODIFY_FIELD only operation SET is supported in general.\n");
            flow_nic_set_error(ERR_ACTION_UNSUPPORTED, error);
            free(fd);
            return NULL;
          }

          if (fd->modify_field_count >= dev->ndev->be.tpe.nb_cpy_writers) {
            NT_LOG(ERR, FILTER, "MODIFY_FIELD exceeded maximum of %"PRIu32" MODIFY_FIELD actions.\n", dev->ndev->be.tpe.nb_cpy_writers);
            flow_nic_set_error(ERR_ACTION_UNSUPPORTED, error);
            free(fd);
            return NULL;
          }

          switch (modify_field->dst.field) {
          case FLOW_FIELD_IPV4_DSCP:
            fd->modify_field[fd->modify_field_count].select = CPY_SELECT_DSCP_IPV4;
            fd->modify_field[fd->modify_field_count].dyn = DYN_L3;
            fd->modify_field[fd->modify_field_count].ofs = 1;
            fd->modify_field[fd->modify_field_count].len = 1;
            break;
          case FLOW_FIELD_IPV6_DSCP:
            fd->modify_field[fd->modify_field_count].select = CPY_SELECT_DSCP_IPV6;
            fd->modify_field[fd->modify_field_count].dyn = DYN_L3;
            fd->modify_field[fd->modify_field_count].ofs = 0;
            fd->modify_field[fd->modify_field_count].len = 2;  /* len=2 is needed because IPv6 DSCP overlaps 2 bytes. */
            break;
          case FLOW_FIELD_GTP_PSC_QFI:
            fd->modify_field[fd->modify_field_count].select = CPY_SELECT_RQI_QFI;
            fd->modify_field[fd->modify_field_count].dyn = DYN_L4_PAYLOAD;
            fd->modify_field[fd->modify_field_count].ofs = 14;
            fd->modify_field[fd->modify_field_count].len = 1;
            break;
          case FLOW_FIELD_IPV4_SRC:
            fd->modify_field[fd->modify_field_count].select = CPY_SELECT_IPV4;
            fd->modify_field[fd->modify_field_count].dyn = DYN_L3;
            fd->modify_field[fd->modify_field_count].ofs = 12;
            fd->modify_field[fd->modify_field_count].len = 4;
            break;
          case FLOW_FIELD_IPV4_DST:
            fd->modify_field[fd->modify_field_count].select = CPY_SELECT_IPV4;
            fd->modify_field[fd->modify_field_count].dyn = DYN_L3;
            fd->modify_field[fd->modify_field_count].ofs = 16;
            fd->modify_field[fd->modify_field_count].len = 4;
            break;
          case FLOW_FIELD_TCP_PORT_SRC:
            /* fallthrough */
          case FLOW_FIELD_UDP_PORT_SRC:
            fd->modify_field[fd->modify_field_count].select = CPY_SELECT_PORT;
            fd->modify_field[fd->modify_field_count].dyn = DYN_L4;
            fd->modify_field[fd->modify_field_count].ofs = 0;
            fd->modify_field[fd->modify_field_count].len = 2;
            break;
          case FLOW_FIELD_TCP_PORT_DST:
            /* fallthrough */
          case FLOW_FIELD_UDP_PORT_DST:
            fd->modify_field[fd->modify_field_count].select = CPY_SELECT_PORT;
            fd->modify_field[fd->modify_field_count].dyn = DYN_L4;
            fd->modify_field[fd->modify_field_count].ofs = 2;
            fd->modify_field[fd->modify_field_count].len = 2;
            break;
          case FLOW_FIELD_GTP_TEID:
            fd->modify_field[fd->modify_field_count].select = CPY_SELECT_TEID;
            fd->modify_field[fd->modify_field_count].dyn = DYN_L4_PAYLOAD;
            fd->modify_field[fd->modify_field_count].ofs = 4;
            fd->modify_field[fd->modify_field_count].len = 4;
            break;
          default:
            NT_LOG(ERR, FILTER, "MODIFY_FIELD dst type is not supported.\n");
            flow_nic_set_error(ERR_ACTION_UNSUPPORTED, error);
            free(fd);
            return NULL;
          }

          modify_field_use_flag = 1 << fd->modify_field[fd->modify_field_count].select;
          if (modify_field_use_flag & modify_field_use_flags) {
            NT_LOG(ERR, FILTER, "MODIFY_FIELD dst type hardware resource already used.\n");
            flow_nic_set_error(ERR_ACTION_UNSUPPORTED, error);
            free(fd);
            return NULL;
          }

          memcpy(fd->modify_field[fd->modify_field_count].value8, modify_field->src.value, 16);

          fd->modify_field[fd->modify_field_count].level = modify_field->dst.level;

          modify_field_use_flags |= modify_field_use_flag;
          fd->modify_field_count += 1;
        }
      }
      break;

    default:
      NT_LOG(ERR, FILTER, "Invalid or unsupported flow action received - %i\n", action[aidx].type);
      flow_nic_set_error(ERR_ACTION_UNSUPPORTED, error);
      free(fd);
      return NULL;
    }
  }

  if (!(encap_decap_order == 0 || encap_decap_order == 2)) {
    NT_LOG(ERR, FILTER, "Invalid encap/decap actions\n");
    free(fd);
    return NULL;
  }

  if (implicit_vlan_vid > 0) {
    uint32_t *sw_data = &packet_data[1 - sw_counter];
    uint32_t *sw_mask = &packet_mask[1 - sw_counter];

    sw_mask[0] = 0x0fff;
    sw_data[0] = implicit_vlan_vid & sw_mask[0];

    km_add_match_elem(&fd->km, &sw_data[0], &sw_mask[0], 1, DYN_FIRST_VLAN, 0);
    set_key_def_sw(key_def, sw_counter, DYN_FIRST_VLAN, 0);
    sw_counter += 1;

    fd->vlans += 1;
  }

  /*
   * All Actions interpreted
   */

  int qw_reserved_mac = 0;
  int qw_reserved_ipv6 = 0;
  for (int eidx = 0; elem[eidx].type != FLOW_ELEM_TYPE_END; ++eidx) {

    switch (elem[eidx].type) {
    case FLOW_ELEM_TYPE_ETH:
      {
        const struct flow_elem_eth *eth_spec = (const struct flow_elem_eth *)elem[eidx].spec;
        const struct flow_elem_eth *eth_mask = (const struct flow_elem_eth *)elem[eidx].mask;

        if (eth_spec != NULL && eth_mask != NULL) {
          if (is_non_zero(eth_mask->d_addr.addr_b, 6) || is_non_zero(eth_mask->s_addr.addr_b, 6)) {
            qw_reserved_mac += 1;
          }
        }
      }
      break;

    case FLOW_ELEM_TYPE_IPV6:
      {
        const struct flow_elem_ipv6 *ipv6_spec = (const struct flow_elem_ipv6 *)elem[eidx].spec;
        const struct flow_elem_ipv6 *ipv6_mask = (const struct flow_elem_ipv6 *)elem[eidx].mask;

        if (ipv6_spec != NULL && ipv6_mask != NULL) {
          if (is_non_zero(ipv6_spec->hdr.src_addr, 16)) {
            qw_reserved_ipv6 += 1;
          }

          if (is_non_zero(ipv6_spec->hdr.dst_addr, 16)) {
            qw_reserved_ipv6 += 1;
          }
        }
      }
      break;

    default:
      break;
    }
  }

  int qw_free = 2 - qw_reserved_mac - qw_reserved_ipv6;
  if (qw_free < 0) {
    NT_LOG(ERR, FILTER, "Key size too big. Out of QW resources.\n");
    flow_nic_set_error(ERR_FAILED, error);
    free(fd);
    return NULL;
  }

  for (int eidx = 0; elem[eidx].type != FLOW_ELEM_TYPE_END; ++eidx) {

    switch (elem[eidx].type) {
    case FLOW_ELEM_TYPE_ANY:
      NT_LOG(DBG, FILTER, "Adap %i, Port %i: FLOW_ELEM_TYPE_ANY\n", dev->ndev->adapter_no, dev->port);
      {
        const struct flow_elem_any *any_spec = (const struct flow_elem_any *)elem[eidx].spec;
        const struct flow_elem_any *any_mask = (const struct flow_elem_any *)elem[eidx].mask;

        if (any_spec && any_mask) {
          any_count += any_spec->num & any_mask->num;
        }
      }
      break;

    case FLOW_ELEM_TYPE_ETH:
      NT_LOG(DBG, FILTER, "Adap %i, Port %i: FLOW_ELEM_TYPE_ETH\n", dev->ndev->adapter_no, dev->port);
      {
        const struct flow_elem_eth *eth_spec = (const struct flow_elem_eth *)elem[eidx].spec;
        const struct flow_elem_eth *eth_mask = (const struct flow_elem_eth *)elem[eidx].mask;

        if (any_count > 0) {
          NT_LOG(ERR, FILTER, "Tunneled L2 ethernet not supported\n");
          flow_nic_set_error(ERR_FAILED, error);
          free(fd);
          return NULL;
        }

        if (eth_spec != NULL && eth_mask != NULL) {
          int non_zero = is_non_zero(eth_mask->d_addr.addr_b, 6) || is_non_zero(eth_mask->s_addr.addr_b, 6);
          if (non_zero || (eth_mask->ether_type != 0 && sw_counter >= 2)) {
            if (qw_counter >= 2) {
              NT_LOG(ERR, FILTER, "Key size too big. Out of QW resources.\n");
              flow_nic_set_error(ERR_FAILED, error);
              free(fd);
              return NULL;
            }

            uint32_t *qw_data = &packet_data[2 + 4 - qw_counter * 4];
            uint32_t *qw_mask = &packet_mask[2 + 4 - qw_counter * 4];

            qw_data[0] =
              ((eth_spec->d_addr.addr_b[0] & eth_mask->d_addr.addr_b[0]) << 24) +
              ((eth_spec->d_addr.addr_b[1] & eth_mask->d_addr.addr_b[1]) << 16) +
              ((eth_spec->d_addr.addr_b[2] & eth_mask->d_addr.addr_b[2]) << 8) +
               (eth_spec->d_addr.addr_b[3] & eth_mask->d_addr.addr_b[3]);

            qw_data[1] =
              ((eth_spec->d_addr.addr_b[4] & eth_mask->d_addr.addr_b[4]) << 24) +
              ((eth_spec->d_addr.addr_b[5] & eth_mask->d_addr.addr_b[5]) << 16) +
              ((eth_spec->s_addr.addr_b[0] & eth_mask->s_addr.addr_b[0]) << 8) +
               (eth_spec->s_addr.addr_b[1] & eth_mask->s_addr.addr_b[1]);

            qw_data[2] =
              ((eth_spec->s_addr.addr_b[2] & eth_mask->s_addr.addr_b[2]) << 24) +
              ((eth_spec->s_addr.addr_b[3] & eth_mask->s_addr.addr_b[3]) << 16) +
              ((eth_spec->s_addr.addr_b[4] & eth_mask->s_addr.addr_b[4]) << 8) +
               (eth_spec->s_addr.addr_b[5] & eth_mask->s_addr.addr_b[5]);

            qw_data[3] = ntohs(eth_spec->ether_type & eth_mask->ether_type) << 16;

            qw_mask[0] =
              (eth_mask->d_addr.addr_b[0] << 24) +
              (eth_mask->d_addr.addr_b[1] << 16) +
              (eth_mask->d_addr.addr_b[2] << 8) +
               eth_mask->d_addr.addr_b[3];

            qw_mask[1] =
              (eth_mask->d_addr.addr_b[4] << 24) +
              (eth_mask->d_addr.addr_b[5] << 16) +
              (eth_mask->s_addr.addr_b[0] << 8) +
               eth_mask->s_addr.addr_b[1];

            qw_mask[2] =
              (eth_mask->s_addr.addr_b[2] << 24) +
              (eth_mask->s_addr.addr_b[3] << 16) +
              (eth_mask->s_addr.addr_b[4] << 8) +
               eth_mask->s_addr.addr_b[5];

            qw_mask[3] = ntohs(eth_mask->ether_type) << 16;

            km_add_match_elem(&fd->km, &qw_data[(size_t)(qw_counter * 4)], &qw_mask[(size_t)(qw_counter * 4)], 4, DYN_L2, 0);
            set_key_def_qw(key_def, qw_counter, DYN_L2, 0);
            qw_counter += 1;
            if (!non_zero) qw_free -= 1;
          } else if (eth_mask->ether_type != 0) {
            if (sw_counter >= 2) {
              NT_LOG(ERR, FILTER, "Key size too big. Out of SW-QW resources.\n");
              flow_nic_set_error(ERR_FAILED, error);
              free(fd);
              return NULL;
            }

            uint32_t *sw_data = &packet_data[1 - sw_counter];
            uint32_t *sw_mask = &packet_mask[1 - sw_counter];

            sw_mask[0] = ntohs(eth_mask->ether_type) << 16;
            sw_data[0] = ntohs(eth_spec->ether_type) << 16 & sw_mask[0];

            km_add_match_elem(&fd->km, &sw_data[0], &sw_mask[0], 1, DYN_L2, 12);
            set_key_def_sw(key_def, sw_counter, DYN_L2, 12);
            sw_counter += 1;
          }
        }

        fd->l2_prot = PROT_L2_ETH2;
      }
      break;

    case FLOW_ELEM_TYPE_VLAN:
      NT_LOG(DBG, FILTER, "Adap %i, Port %i: FLOW_ELEM_TYPE_VLAN\n", dev->ndev->adapter_no, dev->port);
      {
        const struct flow_elem_vlan *vlan_spec = (const struct flow_elem_vlan *)elem[eidx].spec;
        const struct flow_elem_vlan *vlan_mask = (const struct flow_elem_vlan *)elem[eidx].mask;

        if (vlan_spec != NULL && vlan_mask != NULL) {
          if (vlan_mask->tci || vlan_mask->inner_type) {
            if (implicit_vlan_vid > 0) {
              NT_LOG(ERR, FILTER, "Multiple VLANs not supported for implicit VLAN patterns.\n");
              flow_nic_set_error(ERR_MATCH_INVALID_OR_UNSUPPORTED_ELEM, error);
              free(fd);
              return NULL;
            }

            if (sw_counter < 2) {
              uint32_t *sw_data = &packet_data[1 - sw_counter];
              uint32_t *sw_mask = &packet_mask[1 - sw_counter];

              sw_mask[0] = ntohs(vlan_mask->tci) << 16 | ntohs(vlan_mask->inner_type);
              sw_data[0] = ntohs(vlan_spec->tci) << 16 | ntohs(vlan_spec->inner_type);
              sw_data[0] &= sw_mask[0];

              km_add_match_elem(&fd->km, &sw_data[0], &sw_mask[0], 1, DYN_FIRST_VLAN, 2 + 4 * fd->vlans);
              set_key_def_sw(key_def, sw_counter, DYN_FIRST_VLAN, 2 + 4 * fd->vlans);
              sw_counter += 1;
            } else if (qw_counter < 2 && qw_free > 0) {
              uint32_t *qw_data = &packet_data[2 + 4 - qw_counter * 4];
              uint32_t *qw_mask = &packet_mask[2 + 4 - qw_counter * 4];

              qw_data[0] = ntohs(vlan_spec->tci) << 16 | ntohs(vlan_spec->inner_type);
              qw_data[1] = 0;
              qw_data[2] = 0;
              qw_data[3] = 0;

              qw_mask[0] = ntohs(vlan_mask->tci) << 16 | ntohs(vlan_mask->inner_type);
              qw_mask[1] = 0;
              qw_mask[2] = 0;
              qw_mask[3] = 0;

              qw_data[0] &= qw_mask[0];
              qw_data[1] &= qw_mask[1];
              qw_data[2] &= qw_mask[2];
              qw_data[3] &= qw_mask[3];

              km_add_match_elem(&fd->km, &qw_data[0], &qw_mask[0], 4, DYN_FIRST_VLAN, 2 + 4 * fd->vlans);
              set_key_def_qw(key_def, qw_counter, DYN_FIRST_VLAN, 2 + 4 * fd->vlans);
              qw_counter += 1;
              qw_free -= 1;
            } else {
              NT_LOG(ERR, FILTER, "Key size too big. Out of SW-QW resources.\n");
              flow_nic_set_error(ERR_FAILED, error);
              free(fd);
              return NULL;
            }
          }
        }

        fd->vlans += 1;
      }
      break;

    case FLOW_ELEM_TYPE_IPV4:
      NT_LOG(DBG, FILTER, "Adap %i, Port %i: FLOW_ELEM_TYPE_IPV4\n", dev->ndev->adapter_no, dev->port);
      {
        const struct flow_elem_ipv4 *ipv4_spec = (const struct flow_elem_ipv4 *)elem[eidx].spec;
        const struct flow_elem_ipv4 *ipv4_mask = (const struct flow_elem_ipv4 *)elem[eidx].mask;

        if (ipv4_spec != NULL && ipv4_mask != NULL) {
          if (ipv4_mask->hdr.version_ihl != 0 ||
              ipv4_mask->hdr.tos != 0 ||
              ipv4_mask->hdr.length != 0 ||
              ipv4_mask->hdr.id != 0 ||
              (ipv4_mask->hdr.frag_offset != 0 && (ipv4_spec->hdr.frag_offset != 0xffff || ipv4_mask->hdr.frag_offset != 0xffff)) ||
              ipv4_mask->hdr.ttl != 0 ||
              ipv4_mask->hdr.next_proto_id != 0 ||
              ipv4_mask->hdr.hdr_csum != 0) {
            NT_LOG(ERR, FILTER, "Requested IPv4 field not support by running SW version.\n");
            flow_nic_set_error(ERR_FAILED, error);
            free(fd);
            return NULL;
          }

          if (ipv4_spec->hdr.frag_offset == 0xffff && ipv4_mask->hdr.frag_offset == 0xffff) {
            fd->fragmentation = 0xfe;
          }

          int match_cnt = (ipv4_mask->hdr.src_ip != 0) + (ipv4_mask->hdr.dst_ip != 0);
          if (match_cnt > 0) {
            if (qw_free > 0 && (match_cnt == 2 || (match_cnt == 1 && sw_counter >= 2))) {
              if (qw_counter >= 2) {
                NT_LOG(ERR, FILTER, "Key size too big. Out of QW resources.\n");
                flow_nic_set_error(ERR_FAILED, error);
                free(fd);
                return NULL;
              }

              uint32_t *qw_data = &packet_data[2 + 4 - qw_counter * 4];
              uint32_t *qw_mask = &packet_mask[2 + 4 - qw_counter * 4];

              qw_mask[0] = ntohl(ipv4_mask->hdr.src_ip);
              qw_mask[1] = ntohl(ipv4_mask->hdr.dst_ip);

              qw_data[0] = ntohl(ipv4_spec->hdr.src_ip) & qw_mask[0];
              qw_data[1] = ntohl(ipv4_spec->hdr.dst_ip) & qw_mask[1];

              km_add_match_elem(&fd->km, &qw_data[0], &qw_mask[0], 2, any_count > 0 ? DYN_TUN_L3 : DYN_L3, 12);
              set_key_def_qw(key_def, qw_counter, any_count > 0 ? DYN_TUN_L3 : DYN_L3, 12);
              qw_counter += 1;
              qw_free -= 1;
            } else {
              if (ipv4_mask->hdr.src_ip) {
                if (sw_counter >= 2) {
                  NT_LOG(ERR, FILTER, "Key size too big. Out of SW resources.\n");
                  flow_nic_set_error(ERR_FAILED, error);
                  free(fd);
                  return NULL;
                }

                uint32_t *sw_data = &packet_data[1 - sw_counter];
                uint32_t *sw_mask = &packet_mask[1 - sw_counter];

                sw_mask[0] = ntohl(ipv4_mask->hdr.src_ip);
                sw_data[0] = ntohl(ipv4_spec->hdr.src_ip) & sw_mask[0];

                km_add_match_elem(&fd->km, &sw_data[0], &sw_mask[0], 1, any_count > 0 ? DYN_TUN_L3 : DYN_L3, 12);
                set_key_def_sw(key_def, sw_counter, any_count > 0 ? DYN_TUN_L3 : DYN_L3, 12);
                sw_counter += 1;
              }

              if (ipv4_mask->hdr.dst_ip) {
                if (sw_counter >= 2) {
                  NT_LOG(ERR, FILTER, "Key size too big. Out of SW resources.\n");
                  flow_nic_set_error(ERR_FAILED, error);
                  free(fd);
                  return NULL;
                }

                uint32_t *sw_data = &packet_data[1 - sw_counter];
                uint32_t *sw_mask = &packet_mask[1 - sw_counter];

                sw_mask[0] = ntohl(ipv4_mask->hdr.dst_ip);
                sw_data[0] = ntohl(ipv4_spec->hdr.dst_ip) & sw_mask[0];

                km_add_match_elem(&fd->km, &sw_data[0], &sw_mask[0], 1, any_count > 0 ? DYN_TUN_L3 : DYN_L3, 16);
                set_key_def_sw(key_def, sw_counter, any_count > 0 ? DYN_TUN_L3 : DYN_L3, 16);
                sw_counter += 1;
              }
            }
          }
        }

        if (any_count > 0 || fd->l3_prot != -1) {
          fd->tunnel_l3_prot = PROT_TUN_L3_IPV4;
        }
        else {
          fd->l3_prot = PROT_L3_IPV4;
        }
      }
      break;

    case FLOW_ELEM_TYPE_IPV6:
      NT_LOG(DBG, FILTER, "Adap %i, Port %i: FLOW_ELEM_TYPE_IPV6\n", dev->ndev->adapter_no, dev->port);
      {
        const struct flow_elem_ipv6 *ipv6_spec = (const struct flow_elem_ipv6 *)elem[eidx].spec;
        const struct flow_elem_ipv6 *ipv6_mask = (const struct flow_elem_ipv6 *)elem[eidx].mask;

        if (ipv6_spec != NULL && ipv6_mask != NULL) {
          if (ipv6_mask->hdr.vtc_flow != 0 ||
              ipv6_mask->hdr.payload_len != 0 ||
              ipv6_mask->hdr.proto != 0 ||
              ipv6_mask->hdr.hop_limits != 0) {
            NT_LOG(ERR, FILTER, "Requested IPv6 field not support by running SW version\n");
            flow_nic_set_error(ERR_FAILED, error);
            free(fd);
            return NULL;
          }

          if (is_non_zero(ipv6_spec->hdr.src_addr, 16)) {
            if (qw_counter >= 2) {
              NT_LOG(ERR, FILTER, "Key size too big. Out of QW resources.\n");
              flow_nic_set_error(ERR_FAILED, error);
              free(fd);
              return NULL;
            }

            uint32_t *qw_data = &packet_data[2 + 4 - qw_counter * 4];
            uint32_t *qw_mask = &packet_mask[2 + 4 - qw_counter * 4];

            memcpy(&qw_data[0], ipv6_spec->hdr.src_addr, 16);
            memcpy(&qw_mask[0], ipv6_mask->hdr.src_addr, 16);

            qw_data[0] = ntohl(qw_data[0]);
            qw_data[1] = ntohl(qw_data[1]);
            qw_data[2] = ntohl(qw_data[2]);
            qw_data[3] = ntohl(qw_data[3]);

            qw_mask[0] = ntohl(qw_mask[0]);
            qw_mask[1] = ntohl(qw_mask[1]);
            qw_mask[2] = ntohl(qw_mask[2]);
            qw_mask[3] = ntohl(qw_mask[3]);

            qw_data[0] &= qw_mask[0];
            qw_data[1] &= qw_mask[1];
            qw_data[2] &= qw_mask[2];
            qw_data[3] &= qw_mask[3];

            km_add_match_elem(&fd->km, &qw_data[0], &qw_mask[0], 4, any_count > 0 ? DYN_TUN_L3 : DYN_L3, 8);
            set_key_def_qw(key_def, qw_counter, any_count > 0 ? DYN_TUN_L3 : DYN_L3, 8);
            qw_counter += 1;
          }

          if (is_non_zero(ipv6_spec->hdr.dst_addr, 16)) {
            if (qw_counter >= 2) {
              NT_LOG(ERR, FILTER, "Key size too big. Out of QW resources.\n");
              flow_nic_set_error(ERR_FAILED, error);
              free(fd);
              return NULL;
            }

            uint32_t *qw_data = &packet_data[2 + 4 - qw_counter * 4];
            uint32_t *qw_mask = &packet_mask[2 + 4 - qw_counter * 4];

            memcpy(&qw_data[0], ipv6_spec->hdr.dst_addr, 16);
            memcpy(&qw_mask[0], ipv6_mask->hdr.dst_addr, 16);

            qw_data[0] = ntohl(qw_data[0]);
            qw_data[1] = ntohl(qw_data[1]);
            qw_data[2] = ntohl(qw_data[2]);
            qw_data[3] = ntohl(qw_data[3]);

            qw_mask[0] = ntohl(qw_mask[0]);
            qw_mask[1] = ntohl(qw_mask[1]);
            qw_mask[2] = ntohl(qw_mask[2]);
            qw_mask[3] = ntohl(qw_mask[3]);

            qw_data[0] &= qw_mask[0];
            qw_data[1] &= qw_mask[1];
            qw_data[2] &= qw_mask[2];
            qw_data[3] &= qw_mask[3];

            km_add_match_elem(&fd->km, &qw_data[0], &qw_mask[0], 4, any_count > 0 ? DYN_TUN_L3 : DYN_L3, 24);
            set_key_def_qw(key_def, qw_counter, any_count > 0 ? DYN_TUN_L3 : DYN_L3, 24);
            qw_counter += 1;
          }
        }

        if (any_count > 0 || fd->l3_prot != -1) {
          fd->tunnel_l3_prot = PROT_TUN_L3_IPV6;
        }
        else {
          fd->l3_prot = PROT_L3_IPV6;
        }
      }
      break;

    case FLOW_ELEM_TYPE_UDP:
      NT_LOG(DBG, FILTER, "Adap %i, Port %i: FLOW_ELEM_TYPE_UDP\n", dev->ndev->adapter_no, dev->port);
      {
        const struct flow_elem_udp *udp_spec = (const struct flow_elem_udp *)elem[eidx].spec;
        const struct flow_elem_udp *udp_mask = (const struct flow_elem_udp *)elem[eidx].mask;

        if (udp_spec != NULL && udp_mask != NULL) {
          if (udp_mask->hdr.len != 0 ||
              udp_mask->hdr.cksum != 0) {
            NT_LOG(ERR, FILTER, "Requested UDP field not support by running SW version\n");
            flow_nic_set_error(ERR_FAILED, error);
            free(fd);
            return NULL;
          }

          if (udp_mask->hdr.src_port || udp_mask->hdr.dst_port) {
            if (sw_counter < 2) {
              uint32_t *sw_data = &packet_data[1 - sw_counter];
              uint32_t *sw_mask = &packet_mask[1 - sw_counter];

              sw_mask[0] =  (ntohs(udp_mask->hdr.src_port) << 16) | ntohs(udp_mask->hdr.dst_port);
              sw_data[0] = ((ntohs(udp_spec->hdr.src_port) << 16) | ntohs(udp_spec->hdr.dst_port)) & sw_mask[0];

              km_add_match_elem(&fd->km, &sw_data[0], &sw_mask[0], 1, any_count > 0 ? DYN_TUN_L4 : DYN_L4, 0);
              set_key_def_sw(key_def, sw_counter, any_count > 0 ? DYN_TUN_L4 : DYN_L4, 0);
              sw_counter += 1;
            } else if (qw_counter < 2 && qw_free > 0) {
              uint32_t *qw_data = &packet_data[2 + 4 - qw_counter * 4];
              uint32_t *qw_mask = &packet_mask[2 + 4 - qw_counter * 4];

              qw_data[0] = (ntohs(udp_spec->hdr.src_port) << 16) | ntohs(udp_spec->hdr.dst_port);
              qw_data[1] = 0;
              qw_data[2] = 0;
              qw_data[3] = 0;

              qw_mask[0] = (ntohs(udp_mask->hdr.src_port) << 16) | ntohs(udp_mask->hdr.dst_port);
              qw_mask[1] = 0;
              qw_mask[2] = 0;
              qw_mask[3] = 0;

              qw_data[0] &= qw_mask[0];
              qw_data[1] &= qw_mask[1];
              qw_data[2] &= qw_mask[2];
              qw_data[3] &= qw_mask[3];

              km_add_match_elem(&fd->km, &qw_data[0], &qw_mask[0], 4, any_count > 0 ? DYN_TUN_L4 : DYN_L4, 0);
              set_key_def_qw(key_def, qw_counter, any_count > 0 ? DYN_TUN_L4 : DYN_L4, 0);
              qw_counter += 1;
              qw_free -= 1;
            } else {
              NT_LOG(ERR, FILTER, "Key size too big. Out of SW-QW resources.\n");
              flow_nic_set_error(ERR_FAILED, error);
              free(fd);
              return NULL;
            }
          }
        }

        if (any_count > 0 || fd->l4_prot != -1) {
          fd->tunnel_l4_prot = PROT_TUN_L4_UDP;
          key_def->inner_proto = 1;
        }
        else {
          fd->l4_prot = PROT_L4_UDP;
          key_def->outer_proto = 1;
        }
      }
      break;

    case FLOW_ELEM_TYPE_SCTP:
      NT_LOG(DBG, FILTER, "Adap %i,Port %i:FLOW_ELEM_TYPE_SCTP\n", dev->ndev->adapter_no, dev->port);
      {
        const struct flow_elem_sctp *sctp_spec = (const struct flow_elem_sctp *)elem[eidx].spec;
        const struct flow_elem_sctp *sctp_mask = (const struct flow_elem_sctp *)elem[eidx].mask;

        if (sctp_spec != NULL && sctp_mask != NULL) {
          if (sctp_mask->hdr.tag != 0 ||
              sctp_mask->hdr.cksum != 0) {
            NT_LOG(ERR, FILTER, "Requested SCTP field not support by running SW version\n");
            flow_nic_set_error(ERR_FAILED, error);
            free(fd);
            return NULL;
          }

          if (sctp_mask->hdr.src_port || sctp_mask->hdr.dst_port) {
            if (sw_counter < 2) {
              uint32_t *sw_data = &packet_data[1 - sw_counter];
              uint32_t *sw_mask = &packet_mask[1 - sw_counter];

              sw_mask[0] =  (ntohs(sctp_mask->hdr.src_port) << 16) | ntohs(sctp_mask->hdr.dst_port);
              sw_data[0] = ((ntohs(sctp_spec->hdr.src_port) << 16) | ntohs(sctp_spec->hdr.dst_port)) & sw_mask[0];

              km_add_match_elem(&fd->km, &sw_data[0], &sw_mask[0], 1, any_count > 0 ? DYN_TUN_L4 : DYN_L4, 0);
              set_key_def_sw(key_def, sw_counter, any_count > 0 ? DYN_TUN_L4 : DYN_L4, 0);
              sw_counter += 1;
            } else if (qw_counter < 2 && qw_free > 0) {
              uint32_t *qw_data = &packet_data[2 + 4 - qw_counter * 4];
              uint32_t *qw_mask = &packet_mask[2 + 4 - qw_counter * 4];

              qw_data[0] = (ntohs(sctp_spec->hdr.src_port) << 16) | ntohs(sctp_spec->hdr.dst_port);
              qw_data[1] = 0;
              qw_data[2] = 0;
              qw_data[3] = 0;

              qw_mask[0] = (ntohs(sctp_mask->hdr.src_port) << 16) | ntohs(sctp_mask->hdr.dst_port);
              qw_mask[1] = 0;
              qw_mask[2] = 0;
              qw_mask[3] = 0;

              qw_data[0] &= qw_mask[0];
              qw_data[1] &= qw_mask[1];
              qw_data[2] &= qw_mask[2];
              qw_data[3] &= qw_mask[3];

              km_add_match_elem(&fd->km, &qw_data[0], &qw_mask[0], 4, any_count > 0 ? DYN_TUN_L4 : DYN_L4, 0);
              set_key_def_qw(key_def, qw_counter, any_count > 0 ? DYN_TUN_L4 : DYN_L4, 0);
              qw_counter += 1;
              qw_free -= 1;
            } else {
              NT_LOG(ERR, FILTER, "Key size too big. Out of SW-QW resources.\n");
              flow_nic_set_error(ERR_FAILED, error);
              free(fd);
              return NULL;
            }
          }
        }

        if (any_count > 0 || fd->l4_prot != -1) {
          fd->tunnel_l4_prot = PROT_TUN_L4_SCTP;
          key_def->inner_proto = 1;
        }
        else {
          fd->l4_prot = PROT_L4_SCTP;
          key_def->outer_proto = 1;
        }
      }
      break;

    case FLOW_ELEM_TYPE_TCP:
      NT_LOG(DBG, FILTER, "Adap %i, Port %i: FLOW_ELEM_TYPE_TCP\n", dev->ndev->adapter_no, dev->port);
      {
        const struct flow_elem_tcp *tcp_spec = (const struct flow_elem_tcp *)elem[eidx].spec;
        const struct flow_elem_tcp *tcp_mask = (const struct flow_elem_tcp *)elem[eidx].mask;

        if (tcp_spec != NULL && tcp_mask != NULL) {
          if (tcp_mask->hdr.sent_seq != 0 ||
              tcp_mask->hdr.recv_ack != 0 ||
              tcp_mask->hdr.data_off != 0 ||
              tcp_mask->hdr.tcp_flags != 0 ||
              tcp_mask->hdr.rx_win != 0 ||
              tcp_mask->hdr.cksum != 0 ||
              tcp_mask->hdr.tcp_urp != 0) {
            NT_LOG(ERR, FILTER, "Requested TCP field not support by running SW version\n");
            flow_nic_set_error(ERR_FAILED, error);
            free(fd);
            return NULL;
          }

          if (tcp_mask->hdr.src_port || tcp_mask->hdr.dst_port) {
            if (sw_counter < 2) {
              uint32_t *sw_data = &packet_data[1 - sw_counter];
              uint32_t *sw_mask = &packet_mask[1 - sw_counter];

              sw_mask[0] =  (ntohs(tcp_mask->hdr.src_port) << 16) | ntohs(tcp_mask->hdr.dst_port);
              sw_data[0] = ((ntohs(tcp_spec->hdr.src_port) << 16) | ntohs(tcp_spec->hdr.dst_port)) & sw_mask[0];

              km_add_match_elem(&fd->km, &sw_data[0], &sw_mask[0], 1, any_count > 0 ? DYN_TUN_L4 : DYN_L4, 0);
              set_key_def_sw(key_def, sw_counter, any_count > 0 ? DYN_TUN_L4 : DYN_L4, 0);
              sw_counter += 1;
            } else if (qw_counter < 2 && qw_free > 0) {
              uint32_t *qw_data = &packet_data[2 + 4 - qw_counter * 4];
              uint32_t *qw_mask = &packet_mask[2 + 4 - qw_counter * 4];

              qw_data[0] = (ntohs(tcp_spec->hdr.src_port) << 16) | ntohs(tcp_spec->hdr.dst_port);
              qw_data[1] = 0;
              qw_data[2] = 0;
              qw_data[3] = 0;

              qw_mask[0] = (ntohs(tcp_mask->hdr.src_port) << 16) | ntohs(tcp_mask->hdr.dst_port);
              qw_mask[1] = 0;
              qw_mask[2] = 0;
              qw_mask[3] = 0;

              qw_data[0] &= qw_mask[0];
              qw_data[1] &= qw_mask[1];
              qw_data[2] &= qw_mask[2];
              qw_data[3] &= qw_mask[3];

              km_add_match_elem(&fd->km, &qw_data[0], &qw_mask[0], 4, any_count > 0 ? DYN_TUN_L4 : DYN_L4, 0);
              set_key_def_qw(key_def, qw_counter, any_count > 0 ? DYN_TUN_L4 : DYN_L4, 0);
              qw_counter += 1;
              qw_free -= 1;
            } else {
              NT_LOG(ERR, FILTER, "Key size too big. Out of SW-QW resources.\n");
              flow_nic_set_error(ERR_FAILED, error);
              free(fd);
              return NULL;
            }
          }
        }

        if (any_count > 0 || fd->l4_prot != -1) {
          fd->tunnel_l4_prot = PROT_TUN_L4_TCP;
          key_def->inner_proto = 1;
        }
        else {
          fd->l4_prot = PROT_L4_TCP;
          key_def->outer_proto = 1;
        }
      }
      break;

    case FLOW_ELEM_TYPE_GTP:
      NT_LOG(DBG, FILTER, "Adap %i, Port %i: FLOW_ELEM_TYPE_GTP\n", dev->ndev->adapter_no, dev->port);
      {
        const struct flow_elem_gtp *gtp_spec = (const struct flow_elem_gtp *)elem[eidx].spec;
        const struct flow_elem_gtp *gtp_mask = (const struct flow_elem_gtp *)elem[eidx].mask;

        if (gtp_spec != NULL && gtp_mask != NULL) {
          if (gtp_mask->v_pt_rsv_flags != 0 ||
              gtp_mask->msg_type != 0 ||
              gtp_mask->msg_len != 0) {
            NT_LOG(ERR, FILTER, "Requested GTP field not support by running SW version\n");
            flow_nic_set_error(ERR_FAILED, error);
            free(fd);
            return NULL;
          }

          if (gtp_mask->teid) {
            if (sw_counter < 2) {
              uint32_t *sw_data = &packet_data[1 - sw_counter];
              uint32_t *sw_mask = &packet_mask[1 - sw_counter];

              sw_mask[0] = ntohl(gtp_mask->teid);
              sw_data[0] = ntohl(gtp_spec->teid) & sw_mask[0];

              km_add_match_elem(&fd->km, &sw_data[0], &sw_mask[0], 1, DYN_L4_PAYLOAD, 4);
              set_key_def_sw(key_def, sw_counter, DYN_L4_PAYLOAD, 4);
              sw_counter += 1;
            } else if (qw_counter < 2 && qw_free > 0) {
              uint32_t *qw_data = &packet_data[2 + 4 - qw_counter * 4];
              uint32_t *qw_mask = &packet_mask[2 + 4 - qw_counter * 4];

              qw_data[0] = ntohl(gtp_spec->teid);
              qw_data[1] = 0;
              qw_data[2] = 0;
              qw_data[3] = 0;

              qw_mask[0] = ntohl(gtp_mask->teid);
              qw_mask[1] = 0;
              qw_mask[2] = 0;
              qw_mask[3] = 0;

              qw_data[0] &= qw_mask[0];
              qw_data[1] &= qw_mask[1];
              qw_data[2] &= qw_mask[2];
              qw_data[3] &= qw_mask[3];

              km_add_match_elem(&fd->km, &qw_data[0], &qw_mask[0], 4, DYN_L4_PAYLOAD, 4);
              set_key_def_qw(key_def, qw_counter, DYN_L4_PAYLOAD, 4);
              qw_counter += 1;
              qw_free -= 1;
            } else {
              NT_LOG(ERR, FILTER, "Key size too big. Out of SW-QW resources.\n");
              flow_nic_set_error(ERR_FAILED, error);
              free(fd);
              return NULL;
            }
          }
        }

        fd->tunnel_prot = PROT_TUN_GTPV1U;
      }
      break;

    case FLOW_ELEM_TYPE_GTP_PSC:
      NT_LOG(DBG, FILTER, "Adap %i, Port %i: FLOW_ELEM_TYPE_GTP_PSC\n", dev->ndev->adapter_no, dev->port);
      {
        const struct flow_elem_gtp_psc *gtp_psc_spec = (const struct flow_elem_gtp_psc *)elem[eidx].spec;
        const struct flow_elem_gtp_psc *gtp_psc_mask = (const struct flow_elem_gtp_psc *)elem[eidx].mask;

        if (gtp_psc_spec != NULL && gtp_psc_mask != NULL) {
          if (gtp_psc_mask->pdu_type != 0 ||
              gtp_psc_mask->hdr_len != 0) {
            NT_LOG(ERR, FILTER, "Requested GTP PSC field is not supported by running SW version\n");
            flow_nic_set_error(ERR_FAILED, error);
            free(fd);
            return NULL;
          }

          if (gtp_psc_mask->qfi) {
            if (sw_counter < 2) {
              uint32_t *sw_data = &packet_data[1 - sw_counter];
              uint32_t *sw_mask = &packet_mask[1 - sw_counter];

              sw_mask[0] = ntohl(gtp_psc_mask->qfi);
              sw_data[0] = ntohl(gtp_psc_spec->qfi) & sw_mask[0];

              km_add_match_elem(&fd->km, &sw_data[0], &sw_mask[0], 1, DYN_L4_PAYLOAD, 14);
              set_key_def_sw(key_def, sw_counter, DYN_L4_PAYLOAD, 14);
              sw_counter += 1;
            } else if (qw_counter < 2 && qw_free > 0) {
              uint32_t *qw_data = &packet_data[2 + 4 - qw_counter * 4];
              uint32_t *qw_mask = &packet_mask[2 + 4 - qw_counter * 4];

              qw_data[0] = ntohl(gtp_psc_spec->qfi);
              qw_data[1] = 0;
              qw_data[2] = 0;
              qw_data[3] = 0;

              qw_mask[0] = ntohl(gtp_psc_mask->qfi);
              qw_mask[1] = 0;
              qw_mask[2] = 0;
              qw_mask[3] = 0;

              qw_data[0] &= qw_mask[0];
              qw_data[1] &= qw_mask[1];
              qw_data[2] &= qw_mask[2];
              qw_data[3] &= qw_mask[3];

              km_add_match_elem(&fd->km, &qw_data[0], &qw_mask[0], 4, DYN_L4_PAYLOAD, 14);
              set_key_def_qw(key_def, qw_counter, DYN_L4_PAYLOAD, 14);
              qw_counter += 1;
              qw_free -= 1;
            } else {
              NT_LOG(ERR, FILTER, "Key size too big. Out of SW-QW resources.\n");
              flow_nic_set_error(ERR_FAILED, error);
              free(fd);
              return NULL;
            }
          }
        }

        fd->tunnel_prot = PROT_TUN_GTPV1U;
      }
      break;

    case FLOW_ELEM_TYPE_PORT_ID:
      NT_LOG(DBG, FILTER, "Adap %i, Port %i: FLOW_ELEM_TYPE_PORT_ID\n", dev->ndev->adapter_no, dev->port);
      if (elem[eidx].spec) {
        *in_port_id = ((const struct flow_elem_port_id *)elem[eidx].spec)->id;
      }
      break;

    case FLOW_ELEM_TYPE_VOID:
      NT_LOG(DBG, FILTER, "Adap %i, Port %i: FLOW_ELEM_TYPE_VOID\n", dev->ndev->adapter_no, dev->port);
      break;

    default:
      NT_LOG(ERR, FILTER, "Invalid or unsupported flow request: %d\n", (int)elem[eidx].type);
      flow_nic_set_error(ERR_MATCH_INVALID_OR_UNSUPPORTED_ELEM, error);
      free(fd);
      return NULL;
      break;
    }
  }

  return fd;
}

static int convert_fh_to_fh_flm(struct flow_handle *fh,
  const uint32_t *packet_data, uint32_t flm_key_id, uint16_t rpl_ext_ptr,
  uint32_t priority)
{
  struct nic_flow_def *fd;
  struct flow_handle fh_copy;

  if (fh->type != FLOW_HANDLE_TYPE_FLOW) {
    return -1;
  }

  memcpy(&fh_copy, fh, sizeof(struct flow_handle));
  memset(fh, 0x0, sizeof(struct flow_handle));
  fd = fh_copy.fd;

  fh->type = FLOW_HANDLE_TYPE_FLM;
  fh->caller_id = fh_copy.caller_id;
  fh->dev = fh_copy.dev;
  fh->next = fh_copy.next;
  fh->prev = fh_copy.prev;

  fh->flm_db_idx_counter = fh_copy.db_idx_counter;
  for (int i = 0; i < RES_COUNT; ++i) {
    fh->flm_db_idxs[i] = fh_copy.db_idxs[i];
  }

  for (int i = 0; i < MAX_FLM_MTRS_SUPPORTED; ++i) {
    fh->flm_mtr_ids[i] = fd->mtr_ids[i];
  }

  switch (fd->l4_prot) {
    case PROT_L4_TCP:  fh->flm_prot = 6;   break;
    case PROT_L4_UDP:  fh->flm_prot = 17;  break;
    case PROT_L4_SCTP: fh->flm_prot = 132; break;
    case PROT_L4_ICMP: fh->flm_prot = 1;   break;
    default:
      switch (fd->tunnel_l4_prot) {
        case PROT_TUN_L4_TCP:  fh->flm_prot = 6;   break;
        case PROT_TUN_L4_UDP:  fh->flm_prot = 17;  break;
        case PROT_TUN_L4_SCTP: fh->flm_prot = 132; break;
        case PROT_TUN_L4_ICMP: fh->flm_prot = 1;   break;
        default: fh->flm_prot = 0; break;
      }
      break;
  }

  memcpy(fh->flm_data, packet_data, sizeof(uint32_t) * 10);

  fh->flm_kid = flm_key_id;
  fh->flm_rpl_ext_ptr = rpl_ext_ptr;
  fh->flm_prio = (uint8_t)priority;

  for (unsigned int i = 0; i < fd->modify_field_count; ++i) {
    switch (fd->modify_field[i].select) {
    case CPY_SELECT_DSCP_IPV4:
      /* fallthrough */
    case CPY_SELECT_DSCP_IPV6:
      fh->flm_dscp = fd->modify_field[i].value8[0];
      break;
    case CPY_SELECT_RQI_QFI:
      fh->flm_rqi = (fd->modify_field[i].value8[0] >> 6) & 0x1;
      fh->flm_qfi = fd->modify_field[i].value8[0] & 0x3f;
      break;
    case CPY_SELECT_IPV4:
      fh->flm_nat_ipv4 = ntohl(fd->modify_field[i].value32[0]);
      break;
    case CPY_SELECT_PORT:
      fh->flm_nat_port = ntohs(fd->modify_field[i].value16[0]);
      break;
    case CPY_SELECT_TEID:
      fh->flm_teid = ntohl(fd->modify_field[i].value32[0]);
      break;
    }
  }

  fh->flm_mtu_fragmentation_recipe = fd->flm_mtu_fragmentation_recipe;

  free(fd);

  return 0;
}

static int flm_flow_programming(struct flow_eth_dev *dev, struct flow_handle *fh,
  uint32_t flm_ft, uint32_t flm_op)
{
  struct flm_v17_lrn_data_s learn_record;

  if (fh->type != FLOW_HANDLE_TYPE_FLM) {
    return -1;
  }

  memset(&learn_record, 0x0, sizeof(struct flm_v17_lrn_data_s));

  learn_record.qw0[0] = fh->flm_data[9];
  learn_record.qw0[1] = fh->flm_data[8];
  learn_record.qw0[2] = fh->flm_data[7];
  learn_record.qw0[3] = fh->flm_data[6];
  learn_record.qw4[0] = fh->flm_data[5];
  learn_record.qw4[1] = fh->flm_data[4];
  learn_record.qw4[2] = fh->flm_data[3];
  learn_record.qw4[3] = fh->flm_data[2];
  learn_record.sw8    = fh->flm_data[1];
  learn_record.sw9    = fh->flm_data[0];
  learn_record.prot   = fh->flm_prot;

  if (fh->flm_mtr_ids) {
    FLM_V17_MBR_ID1(learn_record.mbr_idx) = fh->flm_mtr_ids[0];
    FLM_V17_MBR_ID2(learn_record.mbr_idx) = fh->flm_mtr_ids[1];
    FLM_V17_MBR_ID3(learn_record.mbr_idx) = fh->flm_mtr_ids[2];
    FLM_V17_MBR_ID4(learn_record.mbr_idx) = fh->flm_mtr_ids[3];

    /* Last non-zero mtr is used for statistics */
    uint8_t mbrs = 0;
    while (mbrs < MAX_FLM_MTRS_SUPPORTED && fh->flm_mtr_ids[mbrs] != 0) {
      ++mbrs;
    }
    learn_record.vol_idx = mbrs;
  }

  learn_record.nat_ip = fh->flm_nat_ipv4;
  learn_record.nat_port = fh->flm_nat_port;
  learn_record.nat_en = fh->flm_nat_ipv4 || fh->flm_nat_port ? 1 : 0;

  learn_record.dscp = fh->flm_dscp;
  learn_record.teid = fh->flm_teid;
  learn_record.qfi = fh->flm_qfi;
  learn_record.rqi = fh->flm_rqi;
  learn_record.color = fh->flm_rpl_ext_ptr & 0x3ff; /* Lower 10 bits used for RPL EXT PTR */
  learn_record.color |= (fh->flm_mtu_fragmentation_recipe & 0xf) << 10; /* Bit [13:10] used for MTU recipe */

  learn_record.ent = 0;
  learn_record.op = flm_op & 0xf;
  learn_record.prio = fh->flm_prio & 0x3;
  learn_record.ft = flm_ft;
  learn_record.kid = fh->flm_kid;
  learn_record.eor = 1;

  int res = flow_flm_apply(dev, &learn_record);
  return res;
}

static struct flow_handle* create_flow_filter(struct flow_eth_dev *dev, struct nic_flow_def *fd,
  const struct flow_attr *attr, struct flow_error *error, uint32_t port_id,
  uint32_t num_dest_port, uint32_t num_queues,
  uint32_t *packet_data, uint32_t *packet_mask, struct flm_flow_key_def_s *key_def)
{
  struct flow_handle *fh = calloc(1, sizeof(struct flow_handle));

  fh->type = FLOW_HANDLE_TYPE_FLOW;
  fh->port_id = port_id;
  fh->dev = dev;
  fh->fd = fd;

  struct hw_db_inline_qsl_data qsl_data = { 0 };

  /* Setup QSL */
  if (fd->dst_num_avail <= 0) {
    qsl_data.drop = 1;
  }
  else {
    assert(fd->dst_num_avail < HW_DB_INLINE_MAX_QST_PER_QSL);

    uint32_t ports[fd->dst_num_avail];
    uint32_t queues[fd->dst_num_avail];

    uint32_t port_index = 0;
    uint32_t queue_index = 0;
    uint32_t max = num_dest_port > num_queues ? num_dest_port : num_queues;

    memset(ports, 0, fd->dst_num_avail);
    memset(queues, 0, fd->dst_num_avail);

    qsl_data.table_size = max;
    qsl_data.retransmit = num_dest_port > 0 ? 1 : 0;

    for (int i = 0; i < fd->dst_num_avail; ++i) {
      if (fd->dst_id[i].type == PORT_PHY) {
        ports[port_index++] = fd->dst_id[i].id;
      }
      else if (fd->dst_id[i].type == PORT_VIRT) {
        queues[queue_index++] = fd->dst_id[i].id;
      }
    }

    for (uint32_t i = 0; i < max; ++i) {
      if (num_dest_port > 0) {
        qsl_data.table[i].tx_port = ports[i % num_dest_port];
        qsl_data.table[i].tx_port_en = 1;
      }
      if (num_queues > 0) {
        qsl_data.table[i].queue = queues[i % num_queues];
        qsl_data.table[i].queue_en = 1;
      }
    }
  }

  const int empty_pattern = fd->l2_prot < 0 && fd->l3_prot < 0 && fd->l4_prot < 0 && fd->vlans == 0 &&
    fd->tunnel_prot < 0 && fd->tunnel_l3_prot < 0 && fd->tunnel_l4_prot < 0;

  if (attr->group > 0 && empty_pattern) {
    /*
     * Default flow for group 1..32
     */

    /* Setup COT */
    struct hw_db_inline_cot_data cot_data = {
      .matcher_color_contrib = 0,
      .frag_rcp = fd->flm_mtu_fragmentation_recipe,
    };
    struct hw_db_cot_idx cot_idx = hw_db_inline_cot_add(dev->ndev, dev->ndev->hw_db_handle, &cot_data);
    fh->db_idxs[fh->db_idx_counter++] = cot_idx.raw;

    if (cot_idx.error) {
      NT_LOG(ERR, FILTER, "Could not reference COT resource\n");
      flow_nic_set_error(ERR_MATCH_RESOURCE_EXHAUSTION, error);
      goto error_out;
    }

    /* Finalize QSL */
    struct hw_db_qsl_idx qsl_idx = hw_db_inline_qsl_add(dev->ndev, dev->ndev->hw_db_handle, &qsl_data);
    fh->db_idxs[fh->db_idx_counter++] = qsl_idx.raw;

    if (qsl_idx.error) {
      NT_LOG(ERR, FILTER, "Could not reference QSL resource\n");
      flow_nic_set_error(ERR_MATCH_RESOURCE_EXHAUSTION, error);
      goto error_out;
    }

    /* Setup Action set */
    struct hw_db_inline_action_set_data action_set_data = {
      .contains_jump = 0,
      .cot = cot_idx,
      .qsl = qsl_idx,
    };

    /* Setup TPE */
    if (fd->ttl_sub_enable) {
      struct hw_db_inline_tpe_data tpe_data = {
        .insert_len = fd->tun_hdr.len,
        .new_outer = fd->tun_hdr.new_outer,
        .ttl_en = fd->ttl_sub_enable,
        .ttl_dyn = fd->ttl_sub_outer ? DYN_L3 : DYN_TUN_L3,
        .ttl_ofs = fd->ttl_sub_ipv4 ? 8 : 7,
      };
      struct hw_db_tpe_idx tpe_idx = hw_db_inline_tpe_add(dev->ndev, dev->ndev->hw_db_handle, &tpe_data);
      fh->db_idxs[fh->db_idx_counter++] = tpe_idx.raw;
      action_set_data.tpe = tpe_idx;

      if (tpe_idx.error) {
        NT_LOG(ERR, FILTER, "Could not reference TPE resource\n");
        flow_nic_set_error(ERR_MATCH_RESOURCE_EXHAUSTION, error);
        goto error_out;
      }
    }

    /* Finalize Action Set */
    struct hw_db_action_set_idx action_set_idx = hw_db_inline_action_set_add(dev->ndev, dev->ndev->hw_db_handle, &action_set_data);
    fh->db_idxs[fh->db_idx_counter++] = action_set_idx.raw;

    if (action_set_idx.error) {
      NT_LOG(ERR, FILTER, "Could not reference Action Set resource\n");
      flow_nic_set_error(ERR_MATCH_RESOURCE_EXHAUSTION, error);
      goto error_out;
    }

    /* Setup default FLM FT */
    struct hw_db_inline_flm_ft_data flm_ft_data = {
      .is_group_zero = 0,
      .group = attr->group,
      .action_set = action_set_idx,
    };
    struct hw_db_flm_ft flm_ft_idx = hw_db_inline_flm_ft_default(dev->ndev, dev->ndev->hw_db_handle, &flm_ft_data);
    fh->db_idxs[fh->db_idx_counter++] = flm_ft_idx.raw;

    if (flm_ft_idx.error) {
      NT_LOG(ERR, FILTER, "Could not reference FLM FT resource\n");
      flow_nic_set_error(ERR_MATCH_RESOURCE_EXHAUSTION, error);
      goto error_out;
    }

    nic_insert_flow(dev->ndev, fh);
  }
  else if (attr->group > 0) {
    /*
     * Flow for group 1..32
     */

    /* Setup FLM RCP */
    struct hw_db_inline_flm_rcp_data flm_data = {
      .qw0_dyn = key_def->qw0_dyn,
      .qw0_ofs = key_def->qw0_ofs,
      .qw4_dyn = key_def->qw4_dyn,
      .qw4_ofs = key_def->qw4_ofs,
      .sw8_dyn = key_def->sw8_dyn,
      .sw8_ofs = key_def->sw8_ofs,
      .sw9_dyn = key_def->sw9_dyn,
      .sw9_ofs = key_def->sw9_ofs,
      .outer_prot = key_def->outer_proto,
      .inner_prot = key_def->inner_proto,
    };
    memcpy(flm_data.mask, packet_mask, sizeof(uint32_t) * 10);
    struct hw_db_flm_idx flm_idx = hw_db_inline_flm_add(dev->ndev, dev->ndev->hw_db_handle, &flm_data, attr->group);
    fh->db_idxs[fh->db_idx_counter++] = flm_idx.raw;

    if (flm_idx.error) {
      NT_LOG(ERR, FILTER, "Could not reference FLM RPC resource\n");
      flow_nic_set_error(ERR_MATCH_RESOURCE_EXHAUSTION, error);
      goto error_out;
    }

    /* Setup COT */
    struct hw_db_inline_cot_data cot_data = {
      .matcher_color_contrib = 0x4,  /* FT key C */
      .frag_rcp = 0,
    };
    struct hw_db_cot_idx cot_idx = hw_db_inline_cot_add(dev->ndev, dev->ndev->hw_db_handle, &cot_data);
    fh->db_idxs[fh->db_idx_counter++] = cot_idx.raw;

    if (cot_idx.error) {
      NT_LOG(ERR, FILTER, "Could not reference COT resource\n");
      flow_nic_set_error(ERR_MATCH_RESOURCE_EXHAUSTION, error);
      goto error_out;
    }

    /* Finalize QSL */
    struct hw_db_qsl_idx qsl_idx = hw_db_inline_qsl_add(dev->ndev, dev->ndev->hw_db_handle, &qsl_data);
    fh->db_idxs[fh->db_idx_counter++] = qsl_idx.raw;

    if (qsl_idx.error) {
      NT_LOG(ERR, FILTER, "Could not reference QSL resource\n");
      flow_nic_set_error(ERR_MATCH_RESOURCE_EXHAUSTION, error);
      goto error_out;
    }

    /* Setup SLC LR */
    struct hw_db_inline_slc_lr_data slc_lr_data = {
      .head_slice_en = fd->header_strip_end_dyn != 0 || fd->header_strip_end_ofs != 0,
      .head_slice_dyn = fd->header_strip_end_dyn,
      .head_slice_ofs = fd->header_strip_end_ofs,
    };
    struct hw_db_slc_lr_idx slc_lr_idx = hw_db_inline_slc_lr_add(dev->ndev, dev->ndev->hw_db_handle, &slc_lr_data);
    fh->db_idxs[fh->db_idx_counter++] = slc_lr_idx.raw;

    if (slc_lr_idx.error) {
      NT_LOG(ERR, FILTER, "Could not reference SLC LR resource\n");
      flow_nic_set_error(ERR_MATCH_RESOURCE_EXHAUSTION, error);
      goto error_out;
    }

    /* Setup TPE EXT */
    assert(fd->tun_hdr.len <= HW_DB_INLINE_MAX_ENCAP_SIZE);

    struct hw_db_inline_tpe_ext_data tpe_ext_data = {
      .size = fd->tun_hdr.len,
    };

    memset(tpe_ext_data.hdr8, 0x0, HW_DB_INLINE_MAX_ENCAP_SIZE);
    memcpy(tpe_ext_data.hdr8, fd->tun_hdr.d.hdr8, (fd->tun_hdr.len + 3) & ~0x3);

    struct hw_db_tpe_ext_idx tpe_ext_idx = hw_db_inline_tpe_ext_add(dev->ndev, dev->ndev->hw_db_handle, &tpe_ext_data);
    fh->db_idxs[fh->db_idx_counter++] = tpe_ext_idx.raw;

    if (tpe_ext_idx.error) {
      NT_LOG(ERR, FILTER, "Could not reference TPE EXT resource\n");
      flow_nic_set_error(ERR_MATCH_RESOURCE_EXHAUSTION, error);
      goto error_out;
    }

    /* Setup TPE */
    assert(fd->modify_field_count <= 6);

    struct hw_db_inline_tpe_data tpe_data = {
      .insert_len = fd->tun_hdr.len,
      .new_outer = fd->tun_hdr.new_outer,
      .ttl_en = fd->ttl_sub_enable,
      .ttl_dyn = fd->ttl_sub_outer ? DYN_L3 : DYN_TUN_L3,
      .ttl_ofs = fd->ttl_sub_ipv4 ? 8 : 7,
    };

    for (unsigned int i = 0; i < fd->modify_field_count; ++i) {
      tpe_data.writer[i].en = 1;
      tpe_data.writer[i].reader_select = fd->modify_field[i].select;
      tpe_data.writer[i].dyn = fd->modify_field[i].dyn;
      tpe_data.writer[i].ofs = fd->modify_field[i].ofs;
      tpe_data.writer[i].len = fd->modify_field[i].len;
    }

    if (fd->tun_hdr.new_outer) {
      const int fcs_length = 4;

      /* L4 length */
      tpe_data.len_a_en = 1;
      tpe_data.len_a_pos_dyn = DYN_L4;
      tpe_data.len_a_pos_ofs = 4;
      tpe_data.len_a_add_dyn = 18;
      tpe_data.len_a_add_ofs = (uint32_t)(-fcs_length) & 0xff;
      tpe_data.len_a_sub_dyn = DYN_L4;

      /* L3 length */
      tpe_data.len_b_en = 1;
      tpe_data.len_b_pos_dyn = DYN_L3;
      tpe_data.len_b_pos_ofs = fd->tun_hdr.ip_version == 4 ? 2 : 4;
      tpe_data.len_b_add_dyn = 18;
      tpe_data.len_b_add_ofs = (uint32_t)(-fcs_length) & 0xff;
      tpe_data.len_b_sub_dyn = DYN_L3;

      /* GTP length */
      tpe_data.len_c_en = 1;
      tpe_data.len_c_pos_dyn = DYN_L4_PAYLOAD;
      tpe_data.len_c_pos_ofs = 2;
      tpe_data.len_c_add_dyn = 18;
      tpe_data.len_c_add_ofs = (uint32_t)(-8 - fcs_length) & 0xff;
      tpe_data.len_c_sub_dyn = DYN_L4_PAYLOAD;
    }

    struct hw_db_tpe_idx tpe_idx = hw_db_inline_tpe_add(dev->ndev, dev->ndev->hw_db_handle, &tpe_data);
    fh->db_idxs[fh->db_idx_counter++] = tpe_idx.raw;

    if (tpe_idx.error) {
      NT_LOG(ERR, FILTER, "Could not reference TPE resource\n");
      flow_nic_set_error(ERR_MATCH_RESOURCE_EXHAUSTION, error);
      goto error_out;
    }

    /* Setup Action Set */
    struct hw_db_inline_action_set_data action_set_data = {
      .contains_jump = 0,
      .cot = cot_idx,
      .qsl = qsl_idx,
      .slc_lr = slc_lr_idx,
      .tpe = tpe_idx,
    };
    struct hw_db_action_set_idx action_set_idx = hw_db_inline_action_set_add(dev->ndev, dev->ndev->hw_db_handle, &action_set_data);
    fh->db_idxs[fh->db_idx_counter++] = action_set_idx.raw;

    if (action_set_idx.error) {
      NT_LOG(ERR, FILTER, "Could not reference Action Set resource\n");
      flow_nic_set_error(ERR_MATCH_RESOURCE_EXHAUSTION, error);
      goto error_out;
    }

    /* Setup FLM FT */
    struct hw_db_inline_flm_ft_data flm_ft_data = {
      .is_group_zero = 0,
      .group = attr->group,
      .action_set = action_set_idx,
    };
    struct hw_db_flm_ft flm_ft_idx = hw_db_inline_flm_ft_add(dev->ndev, dev->ndev->hw_db_handle, &flm_ft_data);
    fh->db_idxs[fh->db_idx_counter++] = flm_ft_idx.raw;

    if (flm_ft_idx.error) {
      NT_LOG(ERR, FILTER, "Could not reference FLM FT resource\n");
      flow_nic_set_error(ERR_MATCH_RESOURCE_EXHAUSTION, error);
      goto error_out;
    }

    /* Program flow */
    convert_fh_to_fh_flm(fh, packet_data, flm_idx.id1 + 2, tpe_ext_idx.ids, attr->priority & 0x3);
    flm_flow_programming(dev, fh, flm_ft_idx.id1, 1);

    nic_insert_flow_flm(dev->ndev, fh);
  }
  else {
    /*
     * Flow for group 0
     */
    int identical_km_entry_ft = -1;

    /* Setup Action Set */
    struct hw_db_inline_action_set_data action_set_data = { 0 };
    if (fd->jump_to_group != UINT32_MAX) {
      /* Action Set only contains jump */
      action_set_data.contains_jump = 1;
      action_set_data.jump = fd->jump_to_group;
    }
    else {
      /* Action Set doesn't contain jump */
      action_set_data.contains_jump = 0;

      /* Setup COT */
      struct hw_db_inline_cot_data cot_data = {
        .matcher_color_contrib = 0,
        .frag_rcp = fd->flm_mtu_fragmentation_recipe,
      };
      struct hw_db_cot_idx cot_idx = hw_db_inline_cot_add(dev->ndev, dev->ndev->hw_db_handle, &cot_data);
      fh->db_idxs[fh->db_idx_counter++] = cot_idx.raw;
      action_set_data.cot = cot_idx;

      if (cot_idx.error) {
        NT_LOG(ERR, FILTER, "Could not reference COT resource\n");
        flow_nic_set_error(ERR_MATCH_RESOURCE_EXHAUSTION, error);
        goto error_out;
      }

      /* Finalize QSL */
      struct hw_db_qsl_idx qsl_idx = hw_db_inline_qsl_add(dev->ndev, dev->ndev->hw_db_handle, &qsl_data);
      fh->db_idxs[fh->db_idx_counter++] = qsl_idx.raw;
      action_set_data.qsl = qsl_idx;

      if (qsl_idx.error) {
        NT_LOG(ERR, FILTER, "Could not reference QSL resource\n");
        flow_nic_set_error(ERR_MATCH_RESOURCE_EXHAUSTION, error);
        goto error_out;
      }

      /* Setup TPE */
      if (fd->ttl_sub_enable) {
        struct hw_db_inline_tpe_data tpe_data = {
          .insert_len = fd->tun_hdr.len,
          .new_outer = fd->tun_hdr.new_outer,
          .ttl_en = fd->ttl_sub_enable,
          .ttl_dyn = fd->ttl_sub_outer ? DYN_L3 : DYN_TUN_L3,
          .ttl_ofs = fd->ttl_sub_ipv4 ? 8 : 7,
        };
        struct hw_db_tpe_idx tpe_idx = hw_db_inline_tpe_add(dev->ndev, dev->ndev->hw_db_handle, &tpe_data);
        fh->db_idxs[fh->db_idx_counter++] = tpe_idx.raw;
        action_set_data.tpe = tpe_idx;

        if (tpe_idx.error) {
          NT_LOG(ERR, FILTER, "Could not reference TPE resource\n");
          flow_nic_set_error(ERR_MATCH_RESOURCE_EXHAUSTION, error);
          goto error_out;
        }
      }
    }

    struct hw_db_action_set_idx action_set_idx = hw_db_inline_action_set_add(dev->ndev, dev->ndev->hw_db_handle, &action_set_data);
    fh->db_idxs[fh->db_idx_counter++] = action_set_idx.raw;

    if (action_set_idx.error) {
      NT_LOG(ERR, FILTER, "Could not reference Action Set resource\n");
      flow_nic_set_error(ERR_MATCH_RESOURCE_EXHAUSTION, error);
      goto error_out;
    }

    /* Setup CAT */
    struct hw_db_inline_cat_data cat_data = {
      .vlan_mask = (0xf << fd->vlans) & 0xf,
      .mac_port_mask = 1 << fh->port_id,
      .ptc_mask_frag = fd->fragmentation,
      .ptc_mask_l2 = fd->l2_prot != -1 ? (1 << fd->l2_prot) : -1,
      .ptc_mask_l3 = fd->l3_prot != -1 ? (1 << fd->l3_prot) : -1,
      .ptc_mask_l4 = fd->l4_prot != -1 ? (1 << fd->l4_prot) : -1,
      .err_mask_ttl = (fd->ttl_sub_enable && fd->ttl_sub_outer) ? -1 : 0x1,
      .ptc_mask_tunnel = fd->tunnel_prot != -1 ? (1 << fd->tunnel_prot) : -1,
      .ptc_mask_l3_tunnel = fd->tunnel_l3_prot != -1 ? (1 << fd->tunnel_l3_prot) : -1,
      .ptc_mask_l4_tunnel = fd->tunnel_l4_prot != -1 ? (1 << fd->tunnel_l4_prot) : -1,
      .err_mask_ttl_tunnel = (fd->ttl_sub_enable && !fd->ttl_sub_outer) ? -1 : 0x1,
    };
    struct hw_db_cat_idx cat_idx = hw_db_inline_cat_add(dev->ndev, dev->ndev->hw_db_handle, &cat_data);
    fh->db_idxs[fh->db_idx_counter++] = cat_idx.raw;

    if (cat_idx.error) {
      NT_LOG(ERR, FILTER, "Could not reference CAT resource\n");
      flow_nic_set_error(ERR_MATCH_RESOURCE_EXHAUSTION, error);
      goto error_out;
    }

    /* Setup KM RCP */
    struct hw_db_inline_km_rcp_data km_rcp_data = { .rcp = 0 };
    if (fd->km.num_ftype_elem) {
      struct flow_handle *flow = dev->ndev->flow_base, *found_flow = NULL;

      if (km_key_create(&fd->km, fh->port_id)) {
        NT_LOG(ERR, FILTER, "KM creation failed\n");
        flow_nic_set_error(ERR_MATCH_FAILED_BY_HW_LIMITS, error);
        goto error_out;
      }

      fd->km.be = &dev->ndev->be;

      /* Look for existing KM RCPs */
      while (flow) {
        if (flow->type == FLOW_HANDLE_TYPE_FLOW && flow->fd->km.flow_type) {
          int res = km_key_compare(&fd->km, &flow->fd->km);
          if (res < 0) {
            /* Flow rcp and match data is identical */
            identical_km_entry_ft = flow->fd->km.flow_type;
            found_flow = flow;
            break;
          }
          else if (res > 0) {
            /* Flow rcp found and match data is different */
            found_flow = flow;
          }
        }
        flow = flow->next;
      }

      km_attach_ndev_resource_management(&fd->km, &dev->ndev->km_res_handle);

      if (found_flow != NULL) {
        /* Reuse existing KM RCP */
        const struct hw_db_inline_km_rcp_data *other_km_rcp_data = hw_db_inline_find_data(dev->ndev,
          dev->ndev->hw_db_handle, HW_DB_IDX_TYPE_KM_RCP, (struct hw_db_idx*)found_flow->flm_db_idxs, found_flow->flm_db_idx_counter);

        if (other_km_rcp_data == NULL || flow_nic_ref_resource(dev->ndev, RES_KM_CATEGORY, other_km_rcp_data->rcp)) {
          NT_LOG(ERR, FILTER, "Could not reference existing KM RCP resource\n");
          flow_nic_set_error(ERR_MATCH_RESOURCE_EXHAUSTION, error);
          goto error_out;
        }

        km_rcp_data.rcp = other_km_rcp_data->rcp;
      }
      else {
        /* Alloc new KM RCP */
        int rcp = flow_nic_alloc_resource(dev->ndev, RES_KM_CATEGORY, 1);
        if (rcp < 0) {
          NT_LOG(ERR, FILTER, "Could not reference KM RCP resource (flow_nic_alloc)\n");
          flow_nic_set_error(ERR_MATCH_RESOURCE_EXHAUSTION, error);
          goto error_out;
        }

        km_rcp_set(&fd->km, rcp);
        km_rcp_data.rcp = (uint32_t)rcp;
      }
    }

    struct hw_db_km_idx km_idx = hw_db_inline_km_add(dev->ndev, dev->ndev->hw_db_handle, &km_rcp_data);
    fh->db_idxs[fh->db_idx_counter++] = km_idx.raw;

    if (km_idx.error) {
      NT_LOG(ERR, FILTER, "Could not reference KM RCP resource (db_inline)\n");
      flow_nic_set_error(ERR_MATCH_RESOURCE_EXHAUSTION, error);
      goto error_out;
    }

    /* Setup KM FT */
    struct hw_db_inline_km_ft_data km_ft_data = {
      .cat = cat_idx,
      .km = km_idx,
      .action_set = action_set_idx,
    };
    struct hw_db_km_ft km_ft_idx = hw_db_inline_km_ft_add(dev->ndev, dev->ndev->hw_db_handle, &km_ft_data);
    fh->db_idxs[fh->db_idx_counter++] = km_ft_idx.raw;

    if (km_ft_idx.error) {
      NT_LOG(ERR, FILTER, "Could not reference KM FT resource\n");
      flow_nic_set_error(ERR_MATCH_RESOURCE_EXHAUSTION, error);
      goto error_out;
    }

    /* Finalize KM RCP */
    if (fd->km.num_ftype_elem) {
      if (identical_km_entry_ft >= 0 && identical_km_entry_ft != km_ft_idx.id1) {
        NT_LOG(ERR, FILTER, "Identical KM matches cannot have different KM FTs\n");
        flow_nic_set_error(ERR_MATCH_FAILED_BY_HW_LIMITS, error);
        goto error_out;
      }

      fd->km.flow_type = km_ft_idx.id1;

      if (fd->km.target == KM_CAM) {
        uint32_t ft_a_mask = 0;
        hw_mod_km_rcp_get(&dev->ndev->be, HW_KM_RCP_FTM_A, (int)km_rcp_data.rcp, 0, &ft_a_mask);
        hw_mod_km_rcp_set(&dev->ndev->be, HW_KM_RCP_FTM_A, (int)km_rcp_data.rcp, 0, ft_a_mask | (1 << fd->km.flow_type));
      }

      hw_mod_km_rcp_flush(&dev->ndev->be, (int)km_rcp_data.rcp, 1);

      km_write_data_match_entry(&fd->km, 0);
    }

    /* Setup Match Set */
    struct hw_db_inline_match_set_data match_set_data = {
      .cat = cat_idx,
      .km = km_idx,
      .km_ft = km_ft_idx,
      .action_set = action_set_idx,
      .jump = fd->jump_to_group != UINT32_MAX ? fd->jump_to_group : 0,
      .priority = attr->priority & 0xff,
    };
    struct hw_db_match_set_idx match_set_idx = hw_db_inline_match_set_add(dev->ndev, dev->ndev->hw_db_handle, &match_set_data);
    fh->db_idxs[fh->db_idx_counter++] = match_set_idx.raw;

    if (match_set_idx.error) {
      NT_LOG(ERR, FILTER, "Could not reference Match Set resource\n");
      flow_nic_set_error(ERR_MATCH_RESOURCE_EXHAUSTION, error);
      goto error_out;
    }

    /* Setup FLM FT */
    struct hw_db_inline_flm_ft_data flm_ft_data = {
      .is_group_zero = 1,
      .jump = fd->jump_to_group != UINT32_MAX ? fd->jump_to_group : 0,
      .action_set = action_set_idx,
    };
    struct hw_db_flm_ft flm_ft_idx = hw_db_inline_flm_ft_add(dev->ndev, dev->ndev->hw_db_handle, &flm_ft_data);
    fh->db_idxs[fh->db_idx_counter++] = flm_ft_idx.raw;

    if (flm_ft_idx.error) {
      NT_LOG(ERR, FILTER, "Could not reference FLM FT resource\n");
      flow_nic_set_error(ERR_MATCH_RESOURCE_EXHAUSTION, error);
      goto error_out;
    }

    nic_insert_flow(dev->ndev, fh);
  }

  return fh;

error_out:
  if (fh->type == FLOW_HANDLE_TYPE_FLM) {
    hw_db_inline_deref_idxs(dev->ndev, dev->ndev->hw_db_handle, (struct hw_db_idx*)fh->flm_db_idxs, fh->flm_db_idx_counter);
  }
  else {
    hw_db_inline_deref_idxs(dev->ndev, dev->ndev->hw_db_handle, (struct hw_db_idx*)fh->db_idxs, fh->db_idx_counter);
    free(fh->fd);
  }

  free(fh);

  return NULL;
}

/*
 * Public functions
 */

int initialize_flow_management_of_ndev_profile_inline(struct flow_nic_dev *ndev)
{
  if (!ndev->flow_mgnt_prepared) {
    /* Check static arrays are big enough */
    assert(ndev->be.tpe.nb_cpy_writers <= MAX_CPY_WRITERS_SUPPORTED);

    /* KM Flow Type 0 is reserved */
    flow_nic_mark_resource_used(ndev, RES_KM_FLOW_TYPE, 0);
    flow_nic_mark_resource_used(ndev, RES_KM_CATEGORY,  0);

    /* FLM Flow Type 0 and 1 is reserved */
    flow_nic_mark_resource_used(ndev, RES_FLM_FLOW_TYPE, 0);
    flow_nic_mark_resource_used(ndev, RES_FLM_FLOW_TYPE, 1);
    flow_nic_mark_resource_used(ndev, RES_FLM_RCP,       0);

    /* CAT CFN 0 is reserved as a low priority catch all filter */
    hw_mod_cat_cfn_set(&ndev->be, HW_CAT_CFN_SET_ALL_DEFAULTS, 0, 0, 0);
    hw_mod_cat_cfn_flush(&ndev->be, 0, 1);

    /* COT is locked to CFN. Don't set color for CFN 0 */
    hw_mod_cat_cot_set(&ndev->be, HW_CAT_COT_PRESET_ALL, 0, 0);
    if (hw_mod_cat_cot_flush(&ndev->be, 0, 1) < 0)
      goto err_exit0;

    flow_nic_mark_resource_used(ndev, RES_CAT_CFN, 0);

    /* Initialize QSL with unmatched recipe index 0 - discard */
    if (hw_mod_qsl_rcp_set(&ndev->be, HW_QSL_RCP_DISCARD, 0, 0x1) < 0)
      goto err_exit0;
    if (hw_mod_qsl_rcp_flush(&ndev->be, 0, 1) < 0)
      goto err_exit0;

    flow_nic_mark_resource_used(ndev, RES_QSL_RCP, 0);

    /* Initialize QST with default index 0 */
    if (hw_mod_qsl_qst_set(&ndev->be, HW_QSL_QST_PRESET_ALL, 0, 0x0) < 0)
      goto err_exit0;
    if (hw_mod_qsl_qst_flush(&ndev->be, 0, 1) < 0)
      goto err_exit0;

    flow_nic_mark_resource_used(ndev, RES_QSL_QST, 0);

    /* SLC LR & TPE index 0 is reserved */
    flow_nic_mark_resource_used(ndev, RES_SLC_LR_RCP, 0);
    flow_nic_mark_resource_used(ndev, RES_TPE_RCP, 0);
    flow_nic_mark_resource_used(ndev, RES_TPE_EXT, 0);
    flow_nic_mark_resource_used(ndev, RES_TPE_RPL, 0);

    /* PDB setup Direct Virtio Scatter-Gather descriptor of 12 bytes for its recipe 0 */
    if (hw_mod_pdb_rcp_set(&ndev->be, HW_PDB_RCP_DESCRIPTOR, 0, 7) < 0)
      goto err_exit0;
    if (hw_mod_pdb_rcp_set(&ndev->be, HW_PDB_RCP_DESC_LEN, 0, 6) < 0)
      goto err_exit0;

    if (hw_mod_pdb_rcp_flush(&ndev->be, 0, 1) < 0)
      goto err_exit0;

    flow_nic_mark_resource_used(ndev, RES_PDB_RCP, 0);

    /* Set default hasher recipe to 5-tuple */
    flow_nic_set_hasher(ndev, 0, HASH_ALGO_5TUPLE);
    hw_mod_hsh_rcp_flush(&ndev->be, 0, 1);

    flow_nic_mark_resource_used(ndev, RES_HSH_RCP, 0);

    /* Handle RMC if available */
    if (ndev->be.rmc.base != NULL) {
      /* Unblock MAC and MAC statistics on this NIC */
      if (hw_mod_rmc_ctrl_set(&ndev->be, HW_RMC_BLOCK_STATT, 0) < 0)
        goto err_exit0;
      /* Block keep alive - not needed */
      if (hw_mod_rmc_ctrl_set(&ndev->be, HW_RMC_BLOCK_KEEPA, 1) < 0)
        goto err_exit0;
      /* Unblock all MAC ports */
      if (hw_mod_rmc_ctrl_set(&ndev->be, HW_RMC_BLOCK_MAC_PORT, 0) < 0)
        goto err_exit0;
      /* Unblock RPP slices */
      hw_mod_rmc_ctrl_set(&ndev->be, HW_RMC_BLOCK_RPP_SLICE, 0);

      if (hw_mod_rmc_ctrl_flush(&ndev->be) < 0)
        goto err_exit0;
    }

    /* FLM */
    if (flm_sdram_calibrate(ndev) < 0)
      goto err_exit0;
    if (flm_sdram_reset(ndev, 1) < 0)
      goto err_exit0;

    hw_mod_flm_control_set(&ndev->be, HW_FLM_CONTROL_LDS, 0);  /* Learn done status */
    hw_mod_flm_control_set(&ndev->be, HW_FLM_CONTROL_LFS, 0);  /* Learn fail status */
    hw_mod_flm_control_set(&ndev->be, HW_FLM_CONTROL_LIS, 0);  /* Learn ignore status */
    hw_mod_flm_control_set(&ndev->be, HW_FLM_CONTROL_UDS, 0);  /* Unlearn done status */
    hw_mod_flm_control_set(&ndev->be, HW_FLM_CONTROL_UIS, 0);  /* Unlearn ignore status */
    hw_mod_flm_control_set(&ndev->be, HW_FLM_CONTROL_RDS, 0);  /* Relearn done status */
    hw_mod_flm_control_set(&ndev->be, HW_FLM_CONTROL_RIS, 0);  /* Relearn ignore status */
    hw_mod_flm_control_set(&ndev->be, HW_FLM_CONTROL_RBL, 4);
    hw_mod_flm_control_flush(&ndev->be);

    hw_mod_flm_prio_set(&ndev->be, HW_FLM_PRIO_LIMIT0, 0);  /* Drop at 100% FIFO fill level */
    hw_mod_flm_prio_set(&ndev->be, HW_FLM_PRIO_FT0, 1);
    hw_mod_flm_prio_set(&ndev->be, HW_FLM_PRIO_LIMIT1, 6);  /* Drop at 37,5% FIFO fill level */
    hw_mod_flm_prio_set(&ndev->be, HW_FLM_PRIO_FT1, 1);
    hw_mod_flm_prio_set(&ndev->be, HW_FLM_PRIO_LIMIT2, 4);  /* Drop at 25% FIFO fill level */
    hw_mod_flm_prio_set(&ndev->be, HW_FLM_PRIO_FT2, 1);
    hw_mod_flm_prio_set(&ndev->be, HW_FLM_PRIO_LIMIT3, 2);  /* Drop at 12,5% FIFO fill level */
    hw_mod_flm_prio_set(&ndev->be, HW_FLM_PRIO_FT3, 1);
    hw_mod_flm_prio_flush(&ndev->be);

    /* TODO How to set and use these limits */
    for (uint32_t i = 0; i < ndev->be.flm.nb_pst_profiles; ++i) {
      hw_mod_flm_pst_set(&ndev->be, HW_FLM_PST_BP, i, FLM_PERIODIC_STATS_BYTE_LIMIT);
      hw_mod_flm_pst_set(&ndev->be, HW_FLM_PST_PP, i, FLM_PERIODIC_STATS_PKT_LIMIT);
      hw_mod_flm_pst_set(&ndev->be, HW_FLM_PST_TP, i, FLM_PERIODIC_STATS_BYTE_TIMEOUT);
    }
    hw_mod_flm_pst_flush(&ndev->be, 0, ALL_ENTRIES);

    hw_mod_flm_stat_update(&ndev->be);

    ndev->flm_mtr_handle = calloc(1, sizeof(struct flm_flow_mtr_handle_s));
    ndev->mtr_stat_handle = calloc(FLM_MTR_STAT_SIZE, sizeof(struct mtr_stat_s));

    if (ndev->flm_mtr_handle == NULL || ndev->mtr_stat_handle == NULL) {
      goto err_exit0;
    }

    struct mtr_stat_s* mtr_stat = ndev->mtr_stat_handle;
    for (uint32_t i = 0; i < FLM_MTR_STAT_SIZE; ++i) {
      atomic_init(&mtr_stat[i].n_pkt, 0);
      atomic_init(&mtr_stat[i].n_bytes, 0);
      atomic_init(&mtr_stat[i].stats_mask, 0);
    }

    if (flow_group_handle_create(&ndev->group_handle, ndev->be.flm.nb_categories)) {
      goto err_exit0;
    }

    if (hw_db_inline_create(ndev, &ndev->hw_db_handle)) {
      goto err_exit0;
    }

    ndev->flow_mgnt_prepared = 1;
  }
  return 0;

err_exit0:
  done_flow_management_of_ndev_profile_inline(ndev);
  return -1;
}


int done_flow_management_of_ndev_profile_inline(struct flow_nic_dev *ndev)
{
  #ifdef FLOW_DEBUG
   ndev->be.iface->set_debug_mode(ndev->be.be_dev, FLOW_BACKEND_DEBUG_MODE_WRITE);
  #endif

  if (ndev->flow_mgnt_prepared) {
    flm_sdram_reset(ndev, 0);

    flow_nic_free_resource(ndev, RES_KM_FLOW_TYPE, 0);
    flow_nic_free_resource(ndev, RES_KM_CATEGORY, 0);

    hw_mod_flm_rcp_set(&ndev->be, HW_FLM_RCP_PRESET_ALL, 0, 0);
    hw_mod_flm_rcp_flush(&ndev->be, 0, 1);
    flow_nic_free_resource(ndev, RES_FLM_FLOW_TYPE, 0);
    flow_nic_free_resource(ndev, RES_FLM_FLOW_TYPE, 1);
    flow_nic_free_resource(ndev, RES_FLM_RCP, 0);

    free(ndev->flm_mtr_handle);
    free(ndev->mtr_stat_handle);
    flow_group_handle_destroy(&ndev->group_handle);

    hw_mod_cat_cfn_set(&ndev->be, HW_CAT_CFN_PRESET_ALL, 0, 0, 0);
    hw_mod_cat_cfn_flush(&ndev->be, 0, 1);
    hw_mod_cat_cot_set(&ndev->be, HW_CAT_COT_PRESET_ALL, 0, 0);
    hw_mod_cat_cot_flush(&ndev->be, 0, 1);
    flow_nic_free_resource(ndev, RES_CAT_CFN, 0);

    hw_mod_qsl_rcp_set(&ndev->be, HW_QSL_RCP_PRESET_ALL, 0, 0);
    hw_mod_qsl_rcp_flush(&ndev->be, 0, 1);
    flow_nic_free_resource(ndev, RES_QSL_RCP, 0);

    hw_mod_slc_lr_rcp_set(&ndev->be, HW_SLC_LR_RCP_PRESET_ALL, 0, 0);
    hw_mod_slc_lr_rcp_flush(&ndev->be, 0, 1);
    flow_nic_free_resource(ndev, RES_SLC_LR_RCP, 0);

    hw_mod_tpe_reset(&ndev->be);
    flow_nic_free_resource(ndev, RES_TPE_RCP, 0);
    flow_nic_free_resource(ndev, RES_TPE_EXT, 0);
    flow_nic_free_resource(ndev, RES_TPE_RPL, 0);

    hw_mod_pdb_rcp_set(&ndev->be, HW_PDB_RCP_PRESET_ALL, 0, 0);
    hw_mod_pdb_rcp_flush(&ndev->be, 0, 1);
    flow_nic_free_resource(ndev, RES_PDB_RCP, 0);

    hw_mod_hsh_rcp_set(&ndev->be, HW_HSH_RCP_PRESET_ALL, 0, 0, 0);
    hw_mod_hsh_rcp_flush(&ndev->be, 0, 1);
    flow_nic_free_resource(ndev, RES_HSH_RCP, 0);

    hw_db_inline_destroy(ndev->hw_db_handle);

    #ifdef FLOW_DEBUG
    ndev->be.iface->set_debug_mode(ndev->be.be_dev, FLOW_BACKEND_DEBUG_MODE_NONE);
    #endif

    ndev->flow_mgnt_prepared = 0;
  }

  return 0;
}

int flow_validate_profile_inline(
    struct flow_eth_dev *dev,
    const struct flow_elem elem[],
    const struct flow_action action[],
    struct flow_error *error)
{
  uint32_t port_id = 0;
  uint32_t num_dest_port = 0;
  uint32_t num_queues = 0;

  uint32_t packet_data[10];
  uint32_t packet_mask[10];
  struct flm_flow_key_def_s key_def;

  flow_nic_set_error(ERR_SUCCESS, error);

  pthread_mutex_lock(&dev->ndev->mtx);
  struct nic_flow_def *fd = interpret_flow_elements(dev, elem, action, error, 0,
    &port_id, &num_dest_port, &num_queues, packet_data, packet_mask, &key_def);
  pthread_mutex_unlock(&dev->ndev->mtx);

  if (!fd) {
    return -1;
  }

  free(fd);
  return 0;
}

struct flow_handle *flow_create_profile_inline(struct flow_eth_dev *dev,
   const struct flow_attr *attr,
   const struct flow_elem elem[],
   const struct flow_action action[],
   struct flow_error *error)
{
  struct flow_handle *fh = NULL;

  uint32_t port_id = UINT32_MAX;
  uint32_t num_dest_port;
  uint32_t num_queues;

  uint32_t packet_data[10];
  uint32_t packet_mask[10];
  struct flm_flow_key_def_s key_def;

  struct flow_attr attr_local;
  memcpy(&attr_local, attr, sizeof(struct flow_attr));
  if (attr_local.group > 0) {
    attr_local.forced_vlan_vid = 0;
  }

  flow_nic_set_error(ERR_SUCCESS, error);

  pthread_mutex_lock(&dev->ndev->mtx);

  struct nic_flow_def *fd = interpret_flow_elements(dev, elem, action, error, attr_local.forced_vlan_vid,
    &port_id, &num_dest_port, &num_queues, packet_data, packet_mask, &key_def);
  if (!fd) {
    goto err_exit;
  }

  /* Translate group IDs */
  if (fd->jump_to_group != UINT32_MAX && flow_group_translate_get(dev->ndev->group_handle, attr_local.caller_id, dev->port, fd->jump_to_group, &fd->jump_to_group)) {
    NT_LOG(ERR, FILTER, "ERROR: Could not get group resource\n");
    flow_nic_set_error(ERR_MATCH_RESOURCE_EXHAUSTION, error);
    goto err_exit;
  }
  if (attr_local.group > 0 && flow_group_translate_get(dev->ndev->group_handle, attr_local.caller_id, dev->port, attr_local.group, &attr_local.group)) {
    NT_LOG(ERR, FILTER, "ERROR: Could not get group resource\n");
    flow_nic_set_error(ERR_MATCH_RESOURCE_EXHAUSTION, error);
    goto err_exit;
  }

  if (port_id == UINT32_MAX) {
    port_id = dev->port_id;
  }

  /* Create and flush filter to NIC */
  fh = create_flow_filter(dev, fd, &attr_local, error,
    port_id, num_dest_port, num_queues, packet_data, packet_mask, &key_def);
  if (!fh) {
    goto err_exit;
  }

  NT_LOG(DBG, FILTER, "New FlOW: fh (flow handle) %p, fd (flow definition) %p\n", fh, fd);
  NT_LOG(DBG, FILTER, ">>>>> [Dev %p] Nic %i, Port %i: fh %p fd %p - implementation <<<<<\n",
      dev, dev->ndev->adapter_no, dev->port, fh, fd);

  fh->caller_id = attr->caller_id;
  pthread_mutex_unlock(&dev->ndev->mtx);

  return fh;

err_exit:
  if (fh) {
    flow_destroy_locked_profile_inline(dev, fh, NULL);
  }

  pthread_mutex_unlock(&dev->ndev->mtx);

  NT_LOG(ERR, FILTER, "ERR: %s\n", __func__);
  return NULL;
}

int flow_destroy_locked_profile_inline(
    struct flow_eth_dev *dev,
    struct flow_handle *fh,
    struct flow_error *error)
{
  assert(dev);
  assert(fh);

  int err = 0;

  flow_nic_set_error(ERR_SUCCESS, error);

  /* take flow out of ndev list - may not have been put there yet */
  if (fh->type == FLOW_HANDLE_TYPE_FLM) {
    nic_remove_flow_flm(dev->ndev, fh);
  }
  else {
    nic_remove_flow(dev->ndev, fh);
  }

#ifdef FLOW_DEBUG
  dev->ndev->be.iface->set_debug_mode(dev->ndev->be.be_dev, FLOW_BACKEND_DEBUG_MODE_WRITE);
#endif

  if (fh->type == FLOW_HANDLE_TYPE_FLM) {
    hw_db_inline_deref_idxs(dev->ndev, dev->ndev->hw_db_handle, (struct hw_db_idx*)fh->flm_db_idxs, fh->flm_db_idx_counter);
    flm_flow_programming(dev, fh, 0, 0);
  }
  else {
    NT_LOG(DBG, FILTER, "removing flow :%p\n", fh);
    if (fh->fd->km.num_ftype_elem) {
      km_clear_data_match_entry(&fh->fd->km);

      const struct hw_db_inline_km_rcp_data *other_km_rcp_data = hw_db_inline_find_data(dev->ndev,
        dev->ndev->hw_db_handle, HW_DB_IDX_TYPE_KM_RCP, (struct hw_db_idx*)fh->flm_db_idxs, fh->flm_db_idx_counter);

      if (other_km_rcp_data != NULL &&
        flow_nic_deref_resource(dev->ndev, RES_KM_CATEGORY, (int)other_km_rcp_data->rcp) == 0) {
        hw_mod_km_rcp_set(&dev->ndev->be, HW_KM_RCP_PRESET_ALL, (int)other_km_rcp_data->rcp, 0, 0);
        hw_mod_km_rcp_flush(&dev->ndev->be, (int)other_km_rcp_data->rcp, 1);
      }
    }
    hw_db_inline_deref_idxs(dev->ndev, dev->ndev->hw_db_handle, (struct hw_db_idx*)fh->db_idxs, fh->db_idx_counter);
    free(fh->fd);
  }

  if (err) {
    NT_LOG(ERR, FILTER, "FAILED removing flow: %p\n", fh);
    flow_nic_set_error(ERR_REMOVE_FLOW_FAILED, error);
  }

  free(fh);

#ifdef FLOW_DEBUG
  dev->ndev->be.iface->set_debug_mode(dev->ndev->be.be_dev, FLOW_BACKEND_DEBUG_MODE_NONE);
#endif

  return err;
}


int flow_destroy_profile_inline(struct flow_eth_dev *dev, struct flow_handle *flow, struct flow_error *error)
{
  int err = 0;

  flow_nic_set_error(ERR_SUCCESS, error);

  if (flow) {
    /* Delete this flow */
    pthread_mutex_lock(&dev->ndev->mtx);
    err = flow_destroy_locked_profile_inline(dev, flow, error);
    pthread_mutex_unlock(&dev->ndev->mtx);
  }

  return err;
}

int flow_flush_profile_inline(
    struct flow_eth_dev *dev,
    uint16_t caller_id,
    struct flow_error *error)
{
  int err = 0;

  flow_nic_set_error(ERR_SUCCESS, error);

  /*
   * Delete all created FLM flows from this eth device.
   * FLM flows must be deleted first because normal flows are their parents.
   */
  struct flow_handle *flow = dev->ndev->flow_base_flm;

  pthread_mutex_lock(&dev->ndev->mtx);

  while (flow && !err) {
    if (flow->dev == dev && flow->caller_id == caller_id) {
      struct flow_handle *flow_next = flow->next;
      err = flow_destroy_locked_profile_inline(dev, flow, error);
      flow = flow_next;
    } else {
      flow = flow->next;
    }
  }

  /* Delete all created flows from this eth device */
  flow = dev->ndev->flow_base;

  while (flow && !err) {
    if (flow->dev == dev && flow->caller_id == caller_id) {
      struct flow_handle *flow_next = flow->next;
      err = flow_destroy_locked_profile_inline(dev, flow, error);
      flow = flow_next;
    } else {
      flow = flow->next;
    }
  }

  pthread_mutex_unlock(&dev->ndev->mtx);

  return err;
}

int flow_query_profile_inline(
    UNUSED struct flow_eth_dev *dev,
    UNUSED struct flow_handle *flow,
    UNUSED const struct flow_action *action,
    void **data,
    uint32_t *length,
    struct flow_error *error)
{
  NT_LOG(ERR, FILTER, "ERROR: Not implemented yet\n");

  *length = 0;
  *data = NULL;
  error->type = FLOW_ERROR_GENERAL;
  error->message = "rte_flow_query is not supported";
  return -1;
}


int flow_dev_dump_profile_inline(UNUSED struct flow_eth_dev *dev,
  UNUSED struct flow_handle *flow,
  UNUSED FILE *file,
  UNUSED struct flow_error *error)
{
  return 0;
}


int flow_get_flm_stats_profile_inline(struct flow_nic_dev *ndev,
    uint64_t *data, uint64_t size)
{
  const enum hw_flm_e fields[] = {
    HW_FLM_STAT_FLOWS,
    HW_FLM_STAT_LRN_DONE,
    HW_FLM_STAT_LRN_IGNORE,
    HW_FLM_STAT_LRN_FAIL,
    HW_FLM_STAT_UNL_DONE,
    HW_FLM_STAT_UNL_IGNORE,
    HW_FLM_STAT_AUL_DONE,
    HW_FLM_STAT_AUL_IGNORE,
    HW_FLM_STAT_AUL_FAIL,
    HW_FLM_STAT_TUL_DONE,
    HW_FLM_STAT_REL_DONE,
    HW_FLM_STAT_REL_IGNORE,
    HW_FLM_STAT_PRB_DONE,
    HW_FLM_STAT_PRB_IGNORE,

    HW_FLM_STAT_STA_DONE,
    HW_FLM_STAT_INF_DONE,
    HW_FLM_STAT_INF_SKIP,
    HW_FLM_STAT_PCK_HIT,
    HW_FLM_STAT_PCK_MISS,
    HW_FLM_STAT_PCK_UNH,
    HW_FLM_STAT_PCK_DIS,
    HW_FLM_STAT_CSH_HIT,
    HW_FLM_STAT_CSH_MISS,
    HW_FLM_STAT_CSH_UNH,
    HW_FLM_STAT_CUC_START,
    HW_FLM_STAT_CUC_MOVE,
  };

  const uint64_t fields_cnt = sizeof(fields) / sizeof(enum hw_flm_e);

  if (size < fields_cnt) {
    return -1;
  }

  hw_mod_flm_stat_update(&ndev->be);

  for (uint64_t i = 0; i < fields_cnt; ++i) {
    uint32_t value = 0;
    hw_mod_flm_stat_get(&ndev->be, fields[i], &value);
    data[i] = (fields[i] == HW_FLM_STAT_FLOWS) ? value : data[i] + value;
    if (ndev->be.flm.ver < 18 && fields[i] == HW_FLM_STAT_PRB_IGNORE) {
      break;
    }
  }

  return 0;
}

int flow_set_mtu_inline(struct flow_eth_dev *dev, uint32_t port, uint16_t mtu)
{
  if (port >= 255) {
    return -1;
  }

  uint32_t ipv4_en_frag;
  uint32_t ipv4_action;
  uint32_t ipv6_en_frag;
  uint32_t ipv6_action;
  if (port == 0) {
    ipv4_en_frag = PORT_0_IPV4_FRAGMENTATION;
    ipv4_action  = PORT_0_IPV4_DF_ACTION;
    ipv6_en_frag = PORT_0_IPV6_FRAGMENTATION;
    ipv6_action  = PORT_0_IPV6_ACTION;
  } else if (port == 1) {
    ipv4_en_frag = PORT_1_IPV4_FRAGMENTATION;
    ipv4_action  = PORT_1_IPV4_DF_ACTION;
    ipv6_en_frag = PORT_1_IPV6_FRAGMENTATION;
    ipv6_action  = PORT_1_IPV6_ACTION;
  } else {
    ipv4_en_frag = DISABLE_FRAGMENTATION;
    ipv4_action  = IPV4_DF_DROP;
    ipv6_en_frag = DISABLE_FRAGMENTATION;
    ipv6_action  = IPV6_DROP;
  }


  int err = 0;
  uint8_t ifr_mtu_recipe = convert_port_to_ifr_mtu_recipe(port);
  struct flow_nic_dev *ndev = dev->ndev;

  err |= hw_mod_tpe_rpp_ifr_rcp_set(&ndev->be, HW_TPE_IFR_RCP_IPV4_EN, ifr_mtu_recipe, ipv4_en_frag);
  err |= hw_mod_tpe_rpp_ifr_rcp_set(&ndev->be, HW_TPE_IFR_RCP_IPV6_EN, ifr_mtu_recipe, ipv6_en_frag);
  err |= hw_mod_tpe_rpp_ifr_rcp_set(&ndev->be, HW_TPE_IFR_RCP_MTU, ifr_mtu_recipe, mtu);
  err |= hw_mod_tpe_rpp_ifr_rcp_set(&ndev->be, HW_TPE_IFR_RCP_IPV4_DF_DROP, ifr_mtu_recipe, ipv4_action);
  err |= hw_mod_tpe_rpp_ifr_rcp_set(&ndev->be, HW_TPE_IFR_RCP_IPV6_DROP, ifr_mtu_recipe, ipv6_action);

  err |= hw_mod_tpe_ifr_rcp_set(&ndev->be, HW_TPE_IFR_RCP_IPV4_EN, ifr_mtu_recipe, ipv4_en_frag);
  err |= hw_mod_tpe_ifr_rcp_set(&ndev->be, HW_TPE_IFR_RCP_IPV6_EN, ifr_mtu_recipe, ipv6_en_frag);
  err |= hw_mod_tpe_ifr_rcp_set(&ndev->be, HW_TPE_IFR_RCP_MTU, ifr_mtu_recipe, mtu);
  err |= hw_mod_tpe_ifr_rcp_set(&ndev->be, HW_TPE_IFR_RCP_IPV4_DF_DROP, ifr_mtu_recipe, ipv4_action);
  err |= hw_mod_tpe_ifr_rcp_set(&ndev->be, HW_TPE_IFR_RCP_IPV6_DROP, ifr_mtu_recipe, ipv6_action);

  if (err == 0) {
    err |= hw_mod_tpe_rpp_ifr_rcp_flush(&ndev->be, ifr_mtu_recipe, 1);
    err |= hw_mod_tpe_ifr_rcp_flush(&ndev->be, ifr_mtu_recipe, 1);
  }

  return err;
}
