/*
 * SPDX-License-Identifier: BSD-3-Clause
 * Copyright(c) 2023 Napatech A/S
 */

#include <stdio.h>
#include <stdarg.h>
#include <string.h>
#include <stdlib.h>
#include <assert.h>
#include <pthread.h>
#include <arpa/inet.h> /* htons, htonl, ntohs */

#include "nt_util.h"
#include "ntlog.h"

#include "flow_api_nic_setup.h"
#include "stream_binary_flow_api.h"
#include "flow_api.h"
#include "flow_api_actions.h"
#include "flow_api_backend.h"
#include "flow_api_engine.h"
#include "flow_api_profile_vswitch.h"

#define SCATTER_GATHER


/* Convert to v15+ CTS layout from prior versions of CAT_CTS */
struct cts_field_map_s {
  enum hw_cat_e cts_field_id;
  int offset;
};

#define CTS_CFN_OFFSET(num, cfn) (cfn * ((num + 1) / 2))

static struct cts_field_map_s cts_offset[CT_CNT] = {
/* 0  CT_COL */    { HW_CAT_CTS_CAT_A, 0 },
/* 1  CT_COR */    { HW_CAT_CTS_CAT_B, 0 },
/* 2  CT_HSH */    { HW_CAT_CTS_CAT_A, 1 },
/* 3  CT_QSL */    { HW_CAT_CTS_CAT_B, 1 },
/* 4  CT_IPF */    { HW_CAT_CTS_CAT_A, 2 },
/* 5  CT_SLC */    { HW_CAT_CTS_CAT_B, 2 },
/* 6  CT_PDB */    { HW_CAT_CTS_CAT_A, 3 },
/* 7  CT_MSK */    { HW_CAT_CTS_CAT_B, 3 },
/* 8  CT_HST */    { HW_CAT_CTS_CAT_A, 4 },
/* 9  CT_EPP */    { HW_CAT_CTS_CAT_B, 4 },
/* 10 CT_TPE */    { HW_CAT_CTS_CAT_A, 5 },
/* 11 CT_RRB */    { HW_CAT_CTS_CAT_B, 5 },
};

static const uint8_t _in6addr_any[16];
#define ipv6_addr_is_set(a) memcmp(a, &_in6addr_any, sizeof(*a))

/*
 * Flow NIC and Eth port device management
 */

enum PROTO_TYPES {
  PROTO_ICMP = 1,    /* FLOW_ELEM_TYPE_ICMP */
  PROTO_TCP  = 6,    /* FLOW_ELEM_TYPE_TCP */
  PROTO_UDP  = 17,   /* FLOW_ELEM_TYPE_UDP */
  PROTO_SCTP = 132   /* FLOW_ELEM_TYPE_SCTP */
};

static int port_to_rx_queue_idx(struct flow_nic_dev *ndev, uint32_t portid, int queue_n, int bypass)
{
  struct flow_eth_dev *dev = ndev->eth_base;
  while (dev) {
    if (portid == (uint32_t)dev->port) {
      /*
       * bypass must be set to 0 to get exception queue for a virtual port
       */
      if ((bypass) && ndev->flow_profile == FLOW_ETH_DEV_PROFILE_VSWITCH) {
        /* reserved 0 for exception */
        queue_n++;
      }
      if (dev->num_queues <= queue_n) return -1;

      return dev->rx_queue[queue_n].hw_id;
    }
    dev = dev->next;
  }
  return -1;
}

enum queue_type_e {
  SWITCH_INPUT_QUEUES,
  SWITCH_OUTPUT_QUEUES
};

static int get_port_nb_queues(struct flow_nic_dev *ndev, uint32_t portid, enum queue_type_e queue_type)
{
  struct flow_eth_dev *dev = ndev->eth_base;
  while (dev) {
    if (portid == (uint32_t)dev->port) {

      if (ndev->flow_profile == FLOW_ETH_DEV_PROFILE_VSWITCH) {
        /*
         * on vSwitch index 0 is exception Rx queue - aka switch input queue
         */
        if (queue_type == SWITCH_OUTPUT_QUEUES) {
          return dev->num_queues ? dev->num_queues - 1 : 0; /* have reserved queue 0 for input exception queue */
        } else {
          return dev->num_queues ? dev->num_queues : 0; /* only 1 exception queue - for now */
        }

      } else {
        /* all considered input queues */
        return dev->num_queues;
      }
    }
    dev = dev->next;
  }
  return 0; /* None - may be a virtual port */
}


static int get_port_from_port_id(struct flow_nic_dev *ndev, uint32_t port_id)
{
  struct flow_eth_dev *dev = ndev->eth_base;

  while (dev) {
    if (dev->port_id == port_id) {
      return dev->port;
    }
    dev = dev->next;
  }

  return -1;
}

static struct flow_eth_dev *get_dev_from_port(struct flow_nic_dev *ndev, uint8_t port)
{
  struct flow_eth_dev *dev = ndev->eth_base;

  while (dev) {
    if (dev->port == port) {
      return dev;
    }
    dev = dev->next;
  }

  return NULL;
}

/*
 * Map Flow MARK to flow stat id
 */
static uint32_t create_flow_stat_id(struct flow_nic_dev *ndev, uint32_t mark)
{
  uint32_t flow_stat_id = mark & (MAX_COLOR_FLOW_STATS - 1);

#if 0 /* Use when running on Open Source */
  uint32_t old_flow_stat_id = flow_stat_id;
  do {
    if (!ndev->flow_stat_id_map[flow_stat_id]) {
      ndev->flow_stat_id_map[flow_stat_id] = mark + 1;
      return flow_stat_id;
    }
    flow_stat_id = (flow_stat_id + 1) & (MAX_COLOR_FLOW_STATS - 1);
  } while (flow_stat_id != old_flow_stat_id);
#else
  /* for NTAPI version to easily replicate algorithm for DPDK PMD */
#if 0
  if (!ndev->flow_stat_id_map[flow_stat_id]) {
#endif
    NT_LOG(DBG, FILTER, "Set flow stat id %i -> mark %i\n", flow_stat_id, mark);
  ndev->flow_stat_id_map[flow_stat_id] = mark + 1;
  return flow_stat_id;
#if 0
  }
#endif

#if 0
  NT_LOG(DBG, FILTER, "ERROR: MARKS mapped:\n");
  for (int i = 0; i < MAX_COLOR_FLOW_STATS; i++) {
    if (ndev->flow_stat_id_map[i]) {
      NT_LOG(DBG, FILTER, "  idx: %i -> mark %i\n", i, ndev->flow_stat_id_map[i] - 1);
    }
  }
#endif
#endif

  return INVALID_FLOW_STAT_ID;
}


static void delete_flow_stat_id(struct flow_nic_dev *ndev, uint32_t flow_stat_id)
{
  if (flow_stat_id < MAX_COLOR_FLOW_STATS) {
    NT_LOG(DBG, FILTER, "Clear flow stat id %i -> mark %i\n", flow_stat_id, ndev->flow_stat_id_map[flow_stat_id] - 1);

    ndev->flow_stat_id_map[flow_stat_id] = 0;
#if 0
    NT_LOG(DBG, FILTER, "MARKS mapped:\n");
    for (int i = 0; i < MAX_COLOR_FLOW_STATS; i++) {
      if (ndev->flow_stat_id_map[i]) {
        NT_LOG(DBG, FILTER, "  idx: %i -> mark %08x\n", i, ndev->flow_stat_id_map[i]);
      }
    }
#endif
  }
}


#if 0
/* RES_QSL_QST */             "RES_QSL_QST",
#endif

int initialize_flow_management_of_ndev_profile_vswitch(struct flow_nic_dev *ndev)
{
  if (!ndev->flow_mgnt_prepared) {
    /* Flow Type 0 is reserved */
    flow_nic_mark_resource_used(ndev, RES_KM_FLOW_TYPE, 0);

    /* preserve category 0 for KCC */
    flow_nic_mark_resource_used(ndev, RES_KM_CATEGORY, 0);

    if (ndev->flow_profile == FLOW_ETH_DEV_PROFILE_VSWITCH) {

      /* initialize QSL with unmatched recipe index 0 - discard */
      if (hw_mod_qsl_rcp_set(&ndev->be, HW_QSL_RCP_DISCARD, 0, 1) < 0)
        goto err_exit0;
      if (hw_mod_qsl_rcp_flush(&ndev->be, 0, 1) < 0)
        goto err_exit0;

      flow_nic_mark_resource_used(ndev, RES_QSL_RCP, 0);

      /*
       * Create default QSL Drop recipe to use for CAT
       * Allocate it so it can be referenced by flows
       */
      ndev->default_qsl_drop_index = flow_nic_alloc_resource(ndev, RES_QSL_RCP, 1);
      hw_mod_qsl_rcp_set(&ndev->be, HW_QSL_RCP_PRESET_ALL, ndev->default_qsl_drop_index, 0);
      if (hw_mod_qsl_rcp_set(&ndev->be, HW_QSL_RCP_DROP, ndev->default_qsl_drop_index, 0xff) < 0)
        goto err_exit0;
      if (hw_mod_qsl_rcp_flush(&ndev->be, ndev->default_qsl_drop_index, 1) < 0)
        goto err_exit0;

      /* Set default hasher recipe to 5-tuple */
      flow_nic_set_hasher(ndev, ndev->default_qsl_drop_index, HASH_ALGO_5TUPLE);
      hw_mod_hsh_rcp_flush(&ndev->be, ndev->default_qsl_drop_index, 1);

      /*
       * Create default QSL Discard recipe to use for CAT
       * Allocate it so it can be referenced by flows
       */
      ndev->default_qsl_discard_index = flow_nic_alloc_resource(ndev, RES_QSL_RCP, 1);
      hw_mod_qsl_rcp_set(&ndev->be, HW_QSL_RCP_PRESET_ALL, ndev->default_qsl_discard_index, 0);
      if (hw_mod_qsl_rcp_set(&ndev->be, HW_QSL_RCP_DISCARD, ndev->default_qsl_discard_index, 1) < 0)
        goto err_exit0;
      if (hw_mod_qsl_rcp_flush(&ndev->be, ndev->default_qsl_discard_index, 1) < 0)
        goto err_exit0;

      /* Set default hasher recipe to 5-tuple */
      flow_nic_set_hasher(ndev, ndev->default_qsl_discard_index, HASH_ALGO_5TUPLE);
      hw_mod_hsh_rcp_flush(&ndev->be, ndev->default_qsl_discard_index, 1);

#ifdef SCATTER_GATHER
      /* PDB setup Direct Virtio Scatter-Gather descriptor of 12 bytes for its recipe 0 */

      if (hw_mod_pdb_rcp_set(&ndev->be, HW_PDB_RCP_DESCRIPTOR, 0, 7) < 0)
        goto err_exit0;
      if (hw_mod_pdb_rcp_set(&ndev->be, HW_PDB_RCP_DESC_LEN, 0, 6) < 0)
        goto err_exit0;
#else
      /* PDB setup dyndescr 5 len 16 for its recipe 0 */
      if (hw_mod_pdb_rcp_set(&ndev->be, HW_PDB_RCP_DESCRIPTOR, 0, 7) < 0)
        goto err_exit0;
      if (hw_mod_pdb_rcp_set(&ndev->be, HW_PDB_RCP_DESC_LEN, 0, 8) < 0)
        goto err_exit0;
#endif
      if (hw_mod_pdb_rcp_flush(&ndev->be, 0, 1) < 0)
        goto err_exit0;

      flow_nic_mark_resource_used(ndev, RES_PDB_RCP, 0);

      /* Set default hasher recipe to 5-tuple */
      flow_nic_set_hasher(ndev, 0, HASH_ALGO_5TUPLE);
      hw_mod_hsh_rcp_flush(&ndev->be, 0, 1);

      flow_nic_mark_resource_used(ndev, RES_HSH_RCP, 0);


#ifdef SCATTER_GATHER
      /* SLC -no slicing */
      hw_mod_slc_rcp_set(&ndev->be, HW_SLC_RCP_PRESET_ALL, 0, 0);
#else
      /* SLC - recipe 0 exception. must slice FCS also : EOF - 4 */
      if (hw_mod_slc_rcp_set(&ndev->be, HW_SLC_RCP_SLC_EN, 0, 1) < 0)
        goto err_exit0;
      if (hw_mod_slc_rcp_set(&ndev->be, HW_SLC_RCP_DYN, 0, 18) < 0)
        goto err_exit0;
      if (hw_mod_slc_rcp_set(&ndev->be, HW_SLC_RCP_OFS, 0, -4) < 0)
        goto err_exit0;
#endif
      if (hw_mod_slc_rcp_flush(&ndev->be, 0, 1) < 0)
        goto err_exit0;

      flow_nic_mark_resource_used(ndev, RES_SLC_RCP, 0);

      /* IOA - set recipe 0 to pass through -  is all zero OK */
      hw_mod_ioa_rcp_set(&ndev->be, HW_IOA_RCP_PRESET_ALL, 0, 0);
      if (hw_mod_ioa_rcp_flush(&ndev->be, 0, 1) < 0)
        goto err_exit0;

      flow_nic_mark_resource_used(ndev, RES_IOA_RCP, 0);

      /* ROA - set recipe 0 - set index 0 to drop. Must be activated by CAO in QSL */
      if (hw_mod_roa_tuncfg_set(&ndev->be, HW_ROA_TUNCFG_RECIRC_BYPASS, 0, 255) < 0)
        goto err_exit0;
      if (hw_mod_roa_tuncfg_flush(&ndev->be, 0, 1) < 0)
        goto err_exit0;

      flow_nic_mark_resource_used(ndev, RES_ROA_RCP, 0); /* RES_IOA_ROA_EPP follows ROA RCP indexes */

#ifdef FLOW_DEBUG
  ndev->be.iface->set_debug_mode(ndev->be.be_dev, FLOW_BACKEND_DEBUG_MODE_NONE);
#endif

      for (unsigned int i = 0; i < ndev->be.roa.nb_lag_entries; i += 4) {
        hw_mod_roa_lagcfg_set(&ndev->be, HW_ROA_LAGCFG_TXPHY_PORT, i, 0);
        hw_mod_roa_lagcfg_set(&ndev->be, HW_ROA_LAGCFG_TXPHY_PORT, i + 1, 1);
        hw_mod_roa_lagcfg_set(&ndev->be, HW_ROA_LAGCFG_TXPHY_PORT, i + 2, 2);
        hw_mod_roa_lagcfg_set(&ndev->be, HW_ROA_LAGCFG_TXPHY_PORT, i + 3, 0);
        hw_mod_roa_lagcfg_flush(&ndev->be, i, 4);
      }
      hw_mod_roa_config_set(&ndev->be, HW_ROA_CONFIG_FWD_RECIRCULATE, 1);
      hw_mod_roa_config_set(&ndev->be, HW_ROA_CONFIG_FWD_NORMAL_PCKS, 1);
      hw_mod_roa_config_set(&ndev->be, HW_ROA_CONFIG_FWD_TXPORT0, 1);
      hw_mod_roa_config_set(&ndev->be, HW_ROA_CONFIG_FWD_TXPORT1, 1);
      hw_mod_roa_config_flush(&ndev->be);

#ifdef FLOW_DEBUG
  ndev->be.iface->set_debug_mode(ndev->be.be_dev, FLOW_BACKEND_DEBUG_MODE_WRITE);
#endif

      /* COT - set color to 0 for unmatched - color encoding must not have CAO enabled for this entry */
      hw_mod_cat_cot_set(&ndev->be, HW_CAT_COT_PRESET_ALL, 0, 0);
      if (hw_mod_cat_cot_flush(&ndev->be, 0, 1) < 0)
        goto err_exit0;

      flow_nic_mark_resource_used(ndev, RES_CAT_COT, 0);


      /*
       * OK, all recipe 0's are setup for no-match
       *
       * Unblock MAC and MAC statistics on this NIC
       */
      if (hw_mod_rmc_ctrl_set(&ndev->be, HW_RMC_BLOCK_STATT, 0) < 0)
        goto err_exit0;
      /* block keep alive - not needed */
      if (hw_mod_rmc_ctrl_set(&ndev->be, HW_RMC_BLOCK_KEEPA, 1) < 0)
        goto err_exit0;
      /*
       * Unblock all MAC ports
       */
      if (hw_mod_rmc_ctrl_set(&ndev->be, HW_RMC_BLOCK_MAC_PORT, 0) < 0)
        goto err_exit0;

      /*
       * unblock RPP slices
       */
      hw_mod_rmc_ctrl_set(&ndev->be, HW_RMC_BLOCK_RPP_SLICE, 0);

      if (hw_mod_rmc_ctrl_flush(&ndev->be) < 0)
        goto err_exit0;
    } /* else for other profiles! */


    for (int i = 0; i < MAX_COLOR_FLOW_STATS; i++) {
      ndev->flow_stat_id_map[i] = 0;
    }

    ndev->flow_mgnt_prepared = 1;
  }
  return 0;

err_exit0:
  done_flow_management_of_ndev_profile_vswitch(ndev);
  return -1;
}


int done_flow_management_of_ndev_profile_vswitch(struct flow_nic_dev *ndev)
{
#ifdef FLOW_DEBUG
   ndev->be.iface->set_debug_mode(ndev->be.be_dev, FLOW_BACKEND_DEBUG_MODE_WRITE);
#endif

  if (ndev->flow_mgnt_prepared) {
    if (ndev->flow_profile == FLOW_ETH_DEV_PROFILE_VSWITCH) {

      /* reserved by HW */
      flow_nic_free_resource(ndev, RES_KM_FLOW_TYPE, 0);

      /* default recipe for KCC disable */
      flow_nic_free_resource(ndev, RES_KM_CATEGORY, 0);

      hw_mod_qsl_rcp_set(&ndev->be, HW_QSL_RCP_PRESET_ALL, 0, 0);
      hw_mod_qsl_rcp_flush(&ndev->be, 0, 1);
      flow_nic_free_resource(ndev, RES_QSL_RCP, 0);

      hw_mod_qsl_rcp_set(&ndev->be, HW_QSL_RCP_PRESET_ALL, ndev->default_qsl_drop_index, 0);
      hw_mod_qsl_rcp_flush(&ndev->be, ndev->default_qsl_drop_index, 1);

      flow_nic_deref_resource(ndev, RES_QSL_RCP, ndev->default_qsl_drop_index);
      ndev->default_qsl_drop_index = 0;

      hw_mod_qsl_rcp_set(&ndev->be, HW_QSL_RCP_PRESET_ALL, ndev->default_qsl_discard_index, 0);
      hw_mod_qsl_rcp_flush(&ndev->be, ndev->default_qsl_discard_index, 1);

      flow_nic_deref_resource(ndev, RES_QSL_RCP, ndev->default_qsl_discard_index);
      ndev->default_qsl_discard_index = 0;

      hw_mod_pdb_rcp_set(&ndev->be, HW_PDB_RCP_PRESET_ALL, 0, 0);
      hw_mod_pdb_rcp_flush(&ndev->be, 0, 1);
      flow_nic_free_resource(ndev, RES_PDB_RCP, 0);

      hw_mod_hsh_rcp_set(&ndev->be, HW_HSH_RCP_PRESET_ALL, 0, 0, 0);
      hw_mod_hsh_rcp_flush(&ndev->be, 0, 1);
      flow_nic_free_resource(ndev, RES_HSH_RCP, 0);

      hw_mod_slc_rcp_set(&ndev->be, HW_SLC_RCP_PRESET_ALL, 0, 0);
      hw_mod_slc_rcp_flush(&ndev->be, 0, 1);
      flow_nic_free_resource(ndev, RES_SLC_RCP, 0);

      hw_mod_cat_cot_set(&ndev->be, HW_CAT_COT_PRESET_ALL, 0, 0);
      hw_mod_cat_cot_flush(&ndev->be, 0, 1);
      flow_nic_free_resource(ndev, RES_CAT_COT, 0);

      hw_mod_ioa_rcp_set(&ndev->be, HW_IOA_RCP_PRESET_ALL, 0, 0);
      hw_mod_ioa_rcp_flush(&ndev->be, 0, 1);
      flow_nic_free_resource(ndev, RES_IOA_RCP, 0);

      hw_mod_roa_tuncfg_set(&ndev->be, HW_ROA_TUNCFG_PRESET_ALL, 0, 0);
      hw_mod_roa_tuncfg_flush(&ndev->be, 0, 1);
      flow_nic_free_resource(ndev, RES_ROA_RCP, 0);

    }
  #ifdef FLOW_DEBUG
    ndev->be.iface->set_debug_mode(ndev->be.be_dev, FLOW_BACKEND_DEBUG_MODE_NONE);
  #endif

    ndev->flow_mgnt_prepared = 0;
  }
  return 0;
}

static struct nic_flow_def* new_flow_def(int kcc_avail)
{
  struct nic_flow_def *fd = calloc(1, sizeof(struct nic_flow_def));
  if (fd) {
    fd->in_port_override = -1;
    fd->mark = (uint32_t) -1;
    if (kcc_avail) {
      fd->kcc = calloc(1, sizeof(struct kcc_flow_def_s));
      fd->kcc->id = KCC_ID_INVALID;
      fd->kcc->bank_used = -1;
    }
  }
  return fd;
}



static void delete_flow_def(struct nic_flow_def *fd)
{
  if (fd) {
    if (fd->kcc) {
      int new_refcnt = kcc_key_ref_count_dec(fd->kcc);
      /* on last ref */
      if (new_refcnt == 0)
        kcc_free_unique_id(fd->kcc);

      /* not yet created or last ref */
      if (new_refcnt <= 0)
        free(fd->kcc);
    }

    if (fd->tnl) {
      tunnel_release(fd->tnl);
    }

    free(fd);
  }
}

static struct flow_handle *new_flow_handle(struct flow_eth_dev *dev, struct nic_flow_def *fd)
{
  struct flow_handle *hdl = calloc(1, sizeof(struct flow_handle));
  if (hdl) {
    hdl->type = FLOW_HANDLE_TYPE_FLOW;
    hdl->dev = dev;
    hdl->fd = fd;
  }
  return hdl;
}


static int reuse_fh_action_resource(struct flow_nic_dev *ndev, enum res_type_e res_type, struct flow_handle *fh)
{
  int next_res = 0;
  int res;
  uint32_t requested_iptype = 0;

  if (res_type == RES_ROA_RCP) {
    /*
     * For resource type "RES_ROA_RCP" get the IP type of the resource we are searching for
     * 0:IPv4, 1:IPv6
     */
    hw_mod_roa_tuncfg_get(&ndev->be, HW_ROA_TUNCFG_IP_TYPE, fh->resource[RES_ROA_RCP].index, &requested_iptype);
  }

  /*
   * The idea is to find a matching resource in a lower layer compare but that
   * has been allocated in the resource manager
   */
  do {
    next_res = flow_nic_find_next_used_resource(ndev, res_type, next_res);
    if (next_res < 0) return -1; /* no more found */
    uint32_t result = (uint32_t)next_res;
    uint32_t found_iptype = 0;

    switch (res_type) {
    case RES_IOA_RCP:
      res = hw_mod_ioa_rcp_get(&ndev->be, HW_IOA_RCP_COMPARE, fh->resource[RES_IOA_RCP].index, &result);
      break;
    case RES_ROA_RCP:
      res = hw_mod_roa_tuncfg_get(&ndev->be, HW_ROA_TUNCFG_COMPARE, fh->resource[RES_ROA_RCP].index, &result);

      if (res == 1 && fh->fd->tun_hdr.len) {
        res = hw_mod_roa_tunhdr_get(&ndev->be, HW_ROA_TUNHDR_COMPARE, fh->resource[RES_ROA_RCP].index, next_res, &result);
        /* Get the IP type of the found resource. 0:IPv4, 1:IPv6 */
        hw_mod_roa_tuncfg_get(&ndev->be, HW_ROA_TUNCFG_IP_TYPE, next_res, &found_iptype);
        if (res == 1 && (found_iptype == 1) && (requested_iptype == 1)) {
          /* IPv6 TUNHDR takes up 2 indexes - i.e. we must compare the next index also */
          res = hw_mod_roa_tunhdr_get(&ndev->be, HW_ROA_TUNHDR_COMPARE, fh->resource[RES_ROA_RCP].index + 1, next_res + 1, &result);
        }
      }
      break;
    default:
      return -1;
    }
    if (res < 0) break;

    if (res == 1 && flow_nic_ref_resource(ndev, res_type, next_res) == 0) {
      NT_LOG(DBG, FILTER, "Resource reuse found %s, idx %i\n", dbg_res_descr[res_type], next_res);

      /* Referenced, reset cache */
      switch (res_type) {
      case RES_IOA_RCP:
        hw_mod_ioa_rcp_set(&ndev->be, HW_IOA_RCP_PRESET_ALL, fh->resource[res_type].index, 0);
        break;
      case RES_ROA_RCP:
        hw_mod_roa_tuncfg_set(&ndev->be, HW_ROA_TUNCFG_PRESET_ALL, fh->resource[res_type].index, 0);
        /* this IOA memory follows ROA index */
        hw_mod_ioa_roa_epp_set(&ndev->be, HW_IOA_ROA_EPP_PRESET_ALL, fh->resource[res_type].index, 0);
        break;
      default:
        return -1;
      }

      flow_nic_deref_resource(ndev, res_type, fh->resource[res_type].index);

      /* change index and create reference */
      fh->resource[res_type].index = next_res;
      fh->resource[res_type].referenced = 1;
      return next_res;
    }

    next_res++;
    if ((res_type == RES_ROA_RCP) && (found_iptype == 1)) {
      /*
       * A resource of type RES_ROA_RCP and IPv6 was found
       * Such a resource takes up 2 indexes. Hence increment next_res twice
       */
      next_res++;
    }
  } while (1);

  return -1;
}


static void nic_insert_flow(struct flow_nic_dev *ndev, struct flow_handle *fh)
{
  fh->next = ndev->flow_base;
  ndev->flow_base = fh;
}

static int nic_remove_flow(struct flow_nic_dev *ndev, struct flow_handle *fh)
{
  assert(fh);
  assert(ndev);
  struct flow_handle *flow = ndev->flow_base, *prev_flow = NULL;

  while (flow) {
    if (flow == fh) {
      if (prev_flow) prev_flow->next = fh->next;
      else ndev->flow_base = fh->next;
      return 0;  /* found and removed */
    }
    prev_flow = flow;
    flow = flow->next;
  }
  return -1;  /* not found */
}

/*
 * Check if the given IP proto field is separately specified in an FLOW ELEM
 * If found return 1
 */
static int protoSpecified(const uint8_t proto, const struct flow_elem elem[])
{
  enum flow_elem_type type;
  if (proto == PROTO_ICMP) type = FLOW_ELEM_TYPE_ICMP;
  else
    if (proto == PROTO_TCP) type = FLOW_ELEM_TYPE_TCP;
    else
      if (proto == PROTO_UDP) type = FLOW_ELEM_TYPE_UDP;
      else
        if (proto == PROTO_SCTP) type = FLOW_ELEM_TYPE_SCTP;
        else
          return 0;

  int idx = 0;
  while (elem[idx].type != FLOW_ELEM_TYPE_END) {
    if ((enum flow_elem_type)elem[idx].type == type) return 1;
    idx++;
  }
  return 0;
}

static void add_layer4_port_match(struct nic_flow_def *fd, uint16_t src_p,
             uint16_t dst_p, uint16_t mask_src_p, uint16_t mask_dst_p)
{
  uint32_t e_word[4];
  uint32_t e_mask[4];

  if (!mask_src_p && !mask_dst_p) return;

  if (mask_src_p && mask_dst_p) {
    e_word[0] = ((uint32_t)(ntohs(src_p) & ntohs(mask_src_p)) << 16) |
                 (ntohs(dst_p) & ntohs(mask_dst_p));
    e_mask[0] = (ntohs(mask_src_p) << 16) | ntohs(mask_dst_p);

    km_add_match_elem(&fd->km, e_word, e_mask, 1, DYN_L4, 0);
  } else {
    if (mask_src_p) {
      e_word[0] = ((uint32_t)(ntohs(src_p) & ntohs(mask_src_p)) << 16);
      e_mask[0] = (ntohs(mask_src_p) << 16);
    }
    if (mask_dst_p) {
      e_word[0] = (uint32_t)(ntohs(dst_p) & ntohs(mask_dst_p));
      e_mask[0] = (uint32_t)ntohs(mask_dst_p);
    }

    if (mask_dst_p == 0xffff) {
      fd->l4_dst_port = ntohs(dst_p);
    }

    km_add_match_elem(&fd->km, e_word, e_mask, 1, DYN_L4, 0);
  }
}


/*
 * Pre-matching CAT FN for use by KCC CAT dependencies
 * is later verified and flow dropped if mismatch
 */
static int prematch_cat_fn(struct flow_eth_dev *dev, struct nic_flow_def *fd) {
  unsigned int i;
  NT_LOG(DBG, FILTER, "CAT CFN Prematch checking:\n");
  NT_LOG(DBG, FILTER, "  L2     : %s\n", (fd->l2_prot == -1) ? "Any":
                          (fd->l2_prot == 1) ? "None(other)" :
                          (fd->l2_prot == 2) ? "Eth2" : "unknown");
  NT_LOG(DBG, FILTER, "  L3     : %s\n", (fd->l3_prot == -1) ? "Any":
                          (fd->l3_prot == 1) ? "None(other)" :
                          (fd->l3_prot == 2) ? "IPv4" :
                          (fd->l3_prot == 4) ? "IPv6" : "unknown");
  NT_LOG(DBG, FILTER, "  L4     : %s\n", (fd->l4_prot == -1) ? "Any":
                          (fd->l4_prot == 1) ? "None(other)" :
                          (fd->l4_prot == 2) ? "TCP" :
                          (fd->l4_prot == 4) ? "UDP" :
                          (fd->l4_prot == 8) ? "SCTP" :
                          (fd->l4_prot == 16) ? "ICMP" : "unknown");
  NT_LOG(DBG, FILTER, "  VLANS  : %i\n", fd->vlans);
  NT_LOG(DBG, FILTER, "  Tunnel : %s\n", (fd->tunnel_prot == -1) ? "Any":
                          (fd->tunnel_prot == 1) ? "None(other)" :
                          (fd->tunnel_prot == 0x200) ? "VxLan" :
                          (fd->tunnel_prot == 0) ? "None" : "unknown");

  for (i = 0; i < dev->ndev->be.cat.nb_cat_funcs; i++) {
    if (flow_nic_is_resource_used(dev->ndev, RES_CAT_CFN, i)) {
      uint32_t val;
      hw_mod_cat_cfn_get(&dev->ndev->be, HW_CAT_CFN_PTC_L2, i, 0, &val);
      if (fd->l2_prot != (int) val)
        continue;

      hw_mod_cat_cfn_get(&dev->ndev->be, HW_CAT_CFN_PTC_L3, i, 0, &val);
      if (fd->l3_prot != (int) val)
        continue;

      hw_mod_cat_cfn_get(&dev->ndev->be, HW_CAT_CFN_PTC_L4, i, 0, &val);
      if (fd->l4_prot != (int) val)
        continue;

      hw_mod_cat_cfn_get(&dev->ndev->be, HW_CAT_CFN_PTC_VLAN, i, 0, &val);
      val &= 0xf;
      if (((1U << fd->vlans) & 0xf) != val)
        continue;

      hw_mod_cat_cfn_get(&dev->ndev->be, HW_CAT_CFN_PTC_TUNNEL, i, 0, &val);
      if (fd->tunnel_prot != (int) val)
        continue;

      NT_LOG(DBG, FILTER, "YES, Found CAT FN match %i\n", i);
      return i; /* found match */
    }
  }
  NT_LOG(DBG, FILTER, "NO matching CAT FN Found\n");
  return -1; /* no match found */
}

static void init_prematch_cat_fn(struct nic_flow_def *fd) {
  fd->l2_prot = -1;
  fd->l3_prot = -1;
  fd->l4_prot = -1;
  fd->vlans = 0;
  fd->tunnel_prot = 1;
}


static struct nic_flow_def* interpret_flow_elements(
    struct flow_eth_dev *dev,
    const struct flow_elem elem[],
    const struct flow_action action[],
    struct flow_error *error,
    uint32_t *in_port_id,
    int *km_rcp_index)
{
  int valid = 1;
  int eidx = 0;
  int aidx = 0;
  struct rss_queues_s {
    enum flow_hash_function func;
    uint16_t queue[FLOW_MAX_QUEUES + 1];
    int queue_num;
  } rss;

  memset(&rss, 0, sizeof(rss));

  if (dev->ndev->flow_profile == FLOW_ETH_DEV_PROFILE_VSWITCH) {
    /*
     * Default setup RSS for HW-RSS - OVS uses always only 1 queue
     */
    rss.queue_num = FLOW_MAX_QUEUES;

    for (int i = 0; i < rss.queue_num; i++) {
      rss.queue[i] = (uint16_t)i;
    }
  }

  struct nic_flow_def *fd = new_flow_def(dev->ndev->be.cat.kcc_size);

  if (!fd) {
    flow_nic_set_error(ERR_MEMORY, error);
    NT_LOG(DBG, FILTER, "ERR Memory\n");
    return NULL;
  }

  init_prematch_cat_fn(fd);

  NT_LOG(DBG, FILTER, ">>>>> [Dev %p] Nic %i, Port %i: fd %p - FLOW Interpretation <<<<<\n", dev, dev->ndev->adapter_no, dev->port, fd);

  /*
   * Gather flow match + actions and convert into internal flow definition structure (struct nic_flow_def_s)
   * This is the 1st step in the flow creation - validate, convert and prepare
   */
  if (action) {
    while (action[aidx].type != FLOW_ACTION_TYPE_END && valid) {

      switch (action[aidx].type) {
      case FLOW_ACTION_TYPE_PORT_ID:
        NT_LOG(DBG, FILTER, "Dev:%p: FLOW_ACTION_TYPE_PORT_ID\n", dev);
        if (action[aidx].conf)
        {
          uint32_t port_id = ((const struct flow_action_port_id *)action[aidx].conf)->id;

          /*
           * Check if output destination is reachable for this nic-dev
           */
          int port = get_port_from_port_id(dev->ndev, port_id);
          if (port < 0 || port > 255) {
            /* ERROR too many output destinations */
            NT_LOG(DBG, FILTER, "port not reachable from this nic device. Port ID:  %i\n", port_id);
            flow_nic_set_error(ERR_OUTPUT_INVALID, error);
            delete_flow_def(fd);
            return NULL;
          }

          uint32_t portid = port;

          if (fd->dst_num_avail == MAX_OUTPUT_DEST) {
            /* ERROR too many output destinations */
            flow_nic_set_error(ERR_OUTPUT_TOO_MANY, error);
            NT_LOG(DBG, FILTER, "ERR ERR_OUTPUT_TOO_MANY\n");
            delete_flow_def(fd);
            return NULL;
          }

          /* if drop also specified, ignore it */
          if (fd->dst_num_avail == 1 && fd->dst_id[0].type == PORT_NONE) {
            fd->dst_num_avail = 0;
          } else {
            uint32_t last_dest_port = -1;
            int num_dest_ports = 0;

            /* check how many ports requested and if this port already set */
            int found = 0;
            for (int i = 0; i < fd->dst_num_avail; i++) {

              if (fd->dst_id[i].owning_port_id == portid) {
                found = 1;
              }

              if (last_dest_port != fd->dst_id[i].owning_port_id) {
                if (num_dest_ports == 2) {
                  NT_LOG(ERR, FILTER, "ERROR: Too many destination ports. Max 2 is supported\n");
                  flow_nic_set_error(ERR_OUTPUT_TOO_MANY, error);
                  delete_flow_def(fd);
                  return NULL;

                }
                last_dest_port = fd->dst_id[i].owning_port_id;
                num_dest_ports++;
              }
            }

            if (found)
              break;
          }

          /*
           * New destination port to add
           */

          /* get queue_idx's or phy portid for port */

          if (portid < dev->ndev->be.num_phy_ports) {
            fd->dst_id[fd->dst_num_avail].owning_port_id = portid;
            fd->dst_id[fd->dst_num_avail].type = PORT_PHY;
            fd->dst_id[fd->dst_num_avail].id = portid;
            fd->dst_id[fd->dst_num_avail].active = 1;
            fd->dst_num_avail++;

            NT_LOG(DBG, FILTER, "Phy port ID: %i\n", portid);
          } else {
            int nb_queues = get_port_nb_queues(dev->ndev, portid, SWITCH_OUTPUT_QUEUES);

            if (nb_queues == 0) {
              /* No physical queues attached to this rx-port */
              NT_LOG(DBG, FILTER, "No Queues attached to port %i: Internal port\n", portid);
              fd->dst_id[fd->dst_num_avail].owning_port_id = portid;
              fd->dst_id[fd->dst_num_avail].type = PORT_INTERNAL;
              fd->dst_id[fd->dst_num_avail].active = 1;
              fd->dst_num_avail++;
            } else {

              NT_LOG(DBG, FILTER, "Number of Queues to port %i: %i\n", portid, nb_queues);

              /* activate the first only */
              fd->dst_id[fd->dst_num_avail].active = 1;

              for (int i = 0; i < nb_queues; i++) {
                fd->dst_id[fd->dst_num_avail].owning_port_id = portid;
                fd->dst_id[fd->dst_num_avail].id = port_to_rx_queue_idx(dev->ndev, portid, i, 1);
                fd->dst_id[fd->dst_num_avail].type = PORT_VIRT;
                NT_LOG(DBG, FILTER, "Output queue index #%i: ID (for virtual dest port %i): %i\n", i, portid, fd->dst_id[fd->dst_num_avail].id);
                fd->dst_num_avail++;
              }
            }
          }

          fd->full_offload = 1;
        }
        break;
      case FLOW_ACTION_TYPE_QUEUE:
        NT_LOG(DBG, FILTER, "Dev:%p: FLOW_ACTION_TYPE_QUEUE\n", dev);
        /*
         * This queue index always refers to this dev port rx queue
         * furthermore, it is always considered switch input queues in vswitch mode
         */
        if (action[aidx].conf)
        {
          const struct flow_action_queue *queue = (const struct flow_action_queue *)action[aidx].conf;
          unsigned int nb_queues = get_port_nb_queues(dev->ndev, dev->port, SWITCH_INPUT_QUEUES);

          if (nb_queues > queue->index) {
            int port_idx = -1;

            for (int i = 0; i < fd->dst_num_avail; i++) {
              if (fd->dst_id[i].owning_port_id == dev->port) {
                port_idx = i;
                break;
              }
            }

            int idx = (port_idx < 0) ? fd->dst_num_avail : port_idx;

            for (unsigned int i = 0; i < nb_queues; i++) {
              int hw_id = port_to_rx_queue_idx(dev->ndev, dev->port, i, 1);

              if (i == queue->index) {
                fd->dst_id[idx + i].active = 1;
                NT_LOG(DBG, FILTER, "Enable this port queue index %i (hw queue id %i)\n", i, hw_id);
              }
              /* if not created yet, add */
              if (port_idx < 0) {
                fd->dst_id[fd->dst_num_avail].owning_port_id = dev->port;
                fd->dst_id[fd->dst_num_avail].id = hw_id;
                fd->dst_id[fd->dst_num_avail].type = PORT_VIRT;
                fd->dst_num_avail++;
              }
            }

          }
        }
        break;
      case FLOW_ACTION_TYPE_RSS:
        NT_LOG(DBG, FILTER, "Dev:%p: FLOW_ACTION_TYPE_RSS\n", dev);
        if (action[aidx].conf)
        {
          const struct flow_action_rss *rss_in = (const struct flow_action_rss *)action[aidx].conf;

          if (rss_in->queue_num > FLOW_MAX_QUEUES) {
            NT_LOG(DBG, FILTER, "ERROR: too many queues. Max supported %i\n", FLOW_MAX_QUEUES);
            flow_nic_set_error(ERR_RSS_TOO_MANY_QUEUES, error);
            delete_flow_def(fd);
            return NULL;
          }
          /*
           * record it for later merge into output destinations
           * need to make sure all output destinations has been
           * received
           */


          /*
           * Override rss setup from OVS - we want to do RSS HW offload, but still only have a single exception queue
           * on the representer port in OVS. Thus OVS does not know about our RSS intentions from HW to VM
           * Right now, we override to always have full RSS enabled
           */
          if (dev->ndev->flow_profile == FLOW_ETH_DEV_PROFILE_VSWITCH && rss_in->queue_num == 1) {
            /* if only 1 queue, use default setup */
          } else {
            rss.queue_num = rss_in->queue_num;

            for (int i = 0; i < rss.queue_num; i++) {
              rss.queue[i] = (uint16_t)((rss_in->queue) ? rss_in->queue[i]: i);
            }
          }

          rss.func = rss_in->func;
        }
        break;
      case FLOW_ACTION_TYPE_MARK:
        NT_LOG(DBG, FILTER, "Dev:%p: FLOW_ACTION_TYPE_MARK\n", dev);
        if (action[aidx].conf)
        {
          fd->mark = ((const struct flow_action_mark *)action[aidx].conf)->id;
          NT_LOG(DBG, FILTER, "Mark: %i\n", fd->mark);
        }
        break;
      case FLOW_ACTION_TYPE_COUNT:
        NT_LOG(DBG, FILTER, "Dev:%p: FLOW_ACTION_TYPE_COUNT\n", dev);
        /* supported */
        break;

      case FLOW_ACTION_TYPE_DROP:
        NT_LOG(DBG, FILTER, "Dev:%p: FLOW_ACTION_TYPE_DROP\n", dev);
        /* if only output action available, register it, otherwise ignore */
        if (fd->dst_num_avail == 0) {
          fd->dst_id[0].type = PORT_NONE;
          fd->dst_id[0].owning_port_id = PORT_ID_NONE;
          fd->dst_id[0].active = 1;
          fd->dst_num_avail = 1;
          NT_LOG(DBG, FILTER, "DROP set\n");

          fd->full_offload = 1;
        } else {
          NT_LOG(DBG, FILTER, "DROP ignored\n");
        }
        break;

      case FLOW_ACTION_TYPE_POP_VLAN:
        NT_LOG(DBG, FILTER, "Dev:%p: FLOW_ACTION_TYPE_POP_VLAN\n", dev);

        fd->ioa_actions = ioa_set_vlan_pop(fd->ioa_actions);
        fd->full_offload = 1;
        break;
      case FLOW_ACTION_TYPE_PUSH_VLAN:
        NT_LOG(DBG, FILTER, "Dev:%p: FLOW_ACTION_TYPE_PUSH_VLAN\n", dev);
        if (action[aidx].conf)
        {
          const struct flow_action_push_vlan *vlan = (const struct flow_action_push_vlan *)action[aidx].conf;
          if (vlan->ethertype == 0x0081)
          {
            /* OK default - nothing to do */
            NT_LOG(DBG, FILTER, "Vlan - 0x8100\n");
          }
          else if (vlan->ethertype == 0xa888)
          {
            NT_LOG(DBG, FILTER, "QinQ - 0x88a8\n");
            fd->ioa_actions = ioa_set_vlan_push_qinq(fd->ioa_actions);
            fd->full_offload = 1;
          }
          else
          {
            NT_LOG(DBG, FILTER, "ERROR: - VLAN type 0x%04x not supported\n", vlan->ethertype);
            flow_nic_set_error(ERR_VLAN_TYPE_NOT_SUPPORTED, error);
            delete_flow_def(fd);
            return NULL;
          }

        }
        break;

      case FLOW_ACTION_TYPE_SET_VLAN_VID:
        NT_LOG(DBG, FILTER, "Dev:%p: FLOW_ACTION_TYPE_SET_VLAN_VID\n", dev);
        if (action[aidx].conf)
        {
          const struct flow_action_set_vlan_vid *vid = (const struct flow_action_set_vlan_vid *)action[aidx].conf;
          NT_LOG(DBG, FILTER, "Vlan VID: %04x\n", htons(vid->vlan_vid) & 0x0fff);
          fd->ioa_actions = ioa_set_vlan_push(fd->ioa_actions, htons(vid->vlan_vid) & 0x0fff);
          fd->full_offload = 1;
        }
        break;

      case FLOW_ACTION_TYPE_SET_VLAN_PCP:
        NT_LOG(DBG, FILTER, "Dev:%p: FLOW_ACTION_TYPE_SET_VLAN_PCP\n", dev);
        if (action[aidx].conf)
        {
          const struct flow_action_set_vlan_pcp *pcp = (const struct flow_action_set_vlan_pcp *)action[aidx].conf;
          NT_LOG(DBG, FILTER, "Vlan PCP: %01x\n", pcp->vlan_pcp);
          fd->ioa_actions = ioa_set_vlan_pcp(fd->ioa_actions, pcp->vlan_pcp);
          NT_LOG(DBG, FILTER, "IOA actions %016lx\n", fd->ioa_actions);
          fd->full_offload = 1;
        }
        break;

      case FLOW_ACTION_TYPE_VXLAN_ENCAP:
        NT_LOG(DBG, FILTER, "Dev:%p: FLOW_ACTION_TYPE_VXLAN_ENCAP\n", dev);
        if (action[aidx].conf)
        {
          const struct flow_action_vxlan_encap *tun = (const struct flow_action_vxlan_encap *)action[aidx].conf;

          if (flow_tunnel_create_vxlan_hdr(&dev->ndev->be, fd, tun->vxlan_tunnel) == 0) {
            NT_LOG(DBG, FILTER, "set ROA tun hdr len %i\n", fd->tun_hdr.len);
            fd->roa_actions = set_roa_tun_type(fd->roa_actions, TUN_TYPE_VXLAN);
            fd->roa_actions = set_roa_tunhdr_len(fd->roa_actions, fd->tun_hdr.len);
            fd->roa_actions = set_roa_tun_ip_type(fd->roa_actions, (fd->tun_hdr.ip_version == 4) ? TUN_IPV4 : TUN_IPV6);
            fd->roa_actions = set_roa_tun_ip_csum(fd->roa_actions, fd->tun_hdr.ip_csum_precalc);
          } else {
            NT_LOG(DBG, FILTER, "ERROR: VXLAN Header in PUSH not supported or invalid\n");
            flow_nic_set_error(ERR_VXLAN_HEADER_NOT_ACCEPTED, error);
            delete_flow_def(fd);
            return NULL;
          }

          fd->full_offload = 1;
        }
        break;

      case FLOW_ACTION_TYPE_TUNNEL_SET:
        NT_LOG(DBG, FILTER, "Dev:%p: FLOW_ACTION_TYPE_TUNNEL_SET\n", dev);
        {
          int idx = 0; /* don't eat the tunnel definition (by incrementing on eidx) - we need to match on it */
          fd->tnl = tunnel_parse(elem, &idx, NULL);

          if (!fd->tnl) {
            NT_LOG(DBG, FILTER, "ERROR: Extraction of RTE Tunnel Set failed\n");
#if 0
            flow_nic_set_error(ERR_VXLAN_HEADER_NOT_ACCEPTED, error);
#endif
            delete_flow_def(fd);
            return NULL;
          }

        }
        break;

      case FLOW_ACTION_TYPE_VXLAN_DECAP:
        NT_LOG(DBG, FILTER, "Dev:%p: FLOW_ACTION_TYPE_VXLAN_DECAP\n", dev);
        {
          uint8_t virt_port = 0xff;

          if (fd->tnl) {
            /*
             * DPDK RTE tunnel helper API used
             */
            virt_port = get_tunnel_vport(fd->tnl);

            NT_LOG(DBG, FILTER, "VXLAN POP - recirc port: %i\n", virt_port);
          } else {
            NT_LOG(DBG, FILTER, "ERROR: Failed getting/creating new internal-virtual port\n");
            flow_nic_set_error(ERR_VXLAN_POP_FAILED_CREATING_VTEP, error);
            delete_flow_def(fd);
            return NULL;
          }

          /* mark newly created. needs default queue assignment - done later */
          fd->roa_actions = set_roa_new_recirc_port(fd->roa_actions);

          fd->ioa_actions = ioa_set_vxlan_pop(fd->ioa_actions);
          fd->roa_actions = set_roa_recirculate(fd->roa_actions, virt_port);

          fd->full_offload = 1;

          /* if VxLAN POP action, we then match only on VxLan tunneled packets */
          fd->tunnel_prot = (1 << PROT_TUN_VXLAN);
        }
        break;
      case FLOW_ACTION_TYPE_JUMP:
        {
          const struct flow_action_jump *jmp _unused = (const struct flow_action_jump *)action[aidx].conf;
          NT_LOG(DBG, FILTER, "Dev:%p: FLOW_ACTION_TYPE_JUMP\n", dev);

          NT_LOG(DBG, FILTER, "Jump to : %i\n", jmp->group);
        }
        break;
      default:
        NT_LOG(DBG, FILTER, "ERROR: Invalid or unsupported flow action received - %i\n", action[aidx].type);
        flow_nic_set_error(ERR_ACTION_UNSUPPORTED, error);
        delete_flow_def(fd);
        return NULL;
      }

      aidx++;
    }
  }



  be16_t eth_type = 0;
  be16_t eth_type_mask = 0xffff;
  uint32_t e_word[4];
  uint32_t e_mask[4];
  int svx_used = 0;

  /*
   * All Actions interpreted
   */

  if (valid) {
    /*
     * If no output destination redirect on this flow - setup this dev queue
     * to steer the packets to its input queue
     */
    if (fd->dst_num_avail == 0) {
      int nb_queues = 0;
      int recirc_port = roa_get_recirc_port(fd->roa_actions);

      /*
       * An action may have instructed recirculate to a new port, otherwise
       * it is considered partial offload in vswitch mode or just use this dev's defined input queues
       */
      if (recirc_port < 0) {
        nb_queues = get_port_nb_queues(dev->ndev, dev->port, SWITCH_INPUT_QUEUES);

        if (nb_queues) {
          /* if no rss specified, activate the first only */
          if (!rss.queue_num)
            fd->dst_id[fd->dst_num_avail].active = 1;

          for (int i = 0; i < nb_queues; i++) {
            fd->dst_id[fd->dst_num_avail].owning_port_id = dev->port;
            fd->dst_id[fd->dst_num_avail].id = port_to_rx_queue_idx(dev->ndev, dev->port, i, 0);
            fd->dst_id[fd->dst_num_avail].type = PORT_VIRT;

            NT_LOG(DBG, FILTER, "Dev queue(exception) index %i: ID (for Rx on this dev: port %i): %i\n", i, dev->port, fd->dst_id[fd->dst_num_avail].id);
            fd->dst_num_avail++;
          }
        } else {
            NT_LOG(DBG, FILTER, "ERROR: No destination queue(s) specified\n");
          flow_nic_set_error(ERR_ACTION_NO_OUTPUT_QUEUE_FOUND, error);
          delete_flow_def(fd);
          return NULL;
        }
      }
    }

    /*
     * Check if RSS has been specified, then collect it into final
     * destination setup. This is needed because we need to know all
     * output destinations before compiling RSS in
     * All destination queues must be activated up to max rss.queue_num.
     * There might be less queues to a specific destination port,
     * but that will not generate an error, but rather scale the output
     * accordingly.
     * If more output queues available only rss.queue_num is activated
     */

    if (rss.queue_num) {
      uint32_t last_port_id = (uint32_t)-1;
      uint16_t num = 0;

      for (int i = 0; i < fd->dst_num_avail; i++) {
        if (last_port_id != fd->dst_id[i].owning_port_id) {
          num = 0;
          last_port_id = fd->dst_id[i].owning_port_id;
        }

        /* is this activated in list */
        int found = 0;
        for (int idx = 0; idx < rss.queue_num; idx++) {
          if (num == rss.queue[idx]) {
            found = 1;
            NT_LOG(DBG, FILTER, "Enable port %i queue index %i (hw queue id %i)\n", last_port_id, num, fd->dst_id[i].id);
            break;
          }
        }
        /* enable if found otherwise disable */
        fd->dst_id[i].active = found;
        num++;
      }
    }

    if (fd->kcc) {
      fd->kcc->be = &dev->ndev->be;
      kcc_attach_ndev_resource_management(fd->kcc, &dev->ndev->kcc_res_handle);

      /*
       * printf("Attach KCC Flow management\n");
       * start out with no sideband as default in KCC
       */
      kcc_key_add_no_sideband(fd->kcc);
    }

    /*
     * Set received action mark (set in our intermediate flow conversion layer) into tunnel here
     * It is actually a NT stat_id. Set it before tunnel_parse on elements (FLOW_ELEM_TYPE_TUNNEL)
     */
    if (fd->tnl && fd->mark != (uint32_t)-1) {
      tunnel_set_flow_stat_id(fd->tnl, fd->mark);
    }

    if (elem) {
      while (elem[eidx].type != FLOW_ELEM_TYPE_END && valid) {

        switch (elem[eidx].type) {
        case FLOW_ELEM_TYPE_TUNNEL:
          NT_LOG(DBG, FILTER, "Adap %i, Port %i: FLOW_ELEM_TYPE_TUNNEL\n", dev->ndev->adapter_no, dev->port);
          /*
           * This is a tunnel encapsulated inner packet
           * digest tunnel definition and register/match to get vport
           */
          uint32_t vni;
          fd->tnl = tunnel_parse(elem, &eidx, &vni);

          if (!fd->tnl || vni == (uint32_t)-1) {
#if 0
            flow_nic_set_error(ERR_MATCH_TOO_MANY_TUNNEL_PORTS, error);
#endif
            delete_flow_def(fd);
            return NULL;
          }

          /* after break it will increment for next loop - but we are already at next element */
          if (eidx) eidx--;

          NT_LOG(DBG, FILTER, "IN PORT override from %i -> %i\n", dev->port, get_tunnel_vport(fd->tnl));
          fd->in_port_override = get_tunnel_vport(fd->tnl);

          /*
           * The VNI is carried through this tunnel encapsulation information, thus
           * use it to program the inner match part together with this VNI
           * NOTE: the VNI is not part of the tunnel configuration
           */
          NT_LOG(DBG, FILTER, "VNI (Network Identifier) %i\n", vni);

          e_word[0] = vni;
          e_mask[0] = 0x00ffffff;

          if (fd->kcc) {
            NT_LOG(DBG, FILTER, "Add KCC KEY VxLAN VNI: %08x\n", e_word[0]);
            kcc_key_add_vxlan(fd->kcc, e_word[0]);
          } else {
            /* Use as metadata carrying VNI from a recently pop'ed VXLAN header */
            e_word[0] |= 2 << 24;
            e_mask[0] |= 0xff << 24;
            km_add_match_elem(&fd->km, e_word, e_mask, 1, SB_VNI, 0);
          }
          svx_used = 1;
          break;

        case FLOW_ELEM_TYPE_ANY:
          /* Any accepted */
          break;
        case FLOW_ELEM_TYPE_ETH:
          NT_LOG(DBG, FILTER, "Adap %i, Port %i: FLOW_ELEM_TYPE_ETH\n", dev->ndev->adapter_no, dev->port);

          fd->l2_prot = (1 << PROT_L2_ETH2);

          const struct flow_elem_eth *eth = (const struct flow_elem_eth *)elem[eidx].spec;
          const struct flow_elem_eth *mask = (const struct flow_elem_eth *)elem[eidx].mask;
          int elx = 0;

          if (eth == NULL || mask == NULL)
          {
            /* Silently accepted */
            break;
          }

          if (is_non_zero(mask->d_addr.addr_b, 6) || is_non_zero(mask->s_addr.addr_b, 6)) {
            e_word[elx + 0] = ((eth->d_addr.addr_b[0] & mask->d_addr.addr_b[0]) << 24) +
                              ((eth->d_addr.addr_b[1] & mask->d_addr.addr_b[1]) << 16) +
                              ((eth->d_addr.addr_b[2] & mask->d_addr.addr_b[2]) << 8) +
                               (eth->d_addr.addr_b[3] & mask->d_addr.addr_b[3]);
            e_word[elx + 1] = ((eth->d_addr.addr_b[4] & mask->d_addr.addr_b[4]) << 24) +
                              ((eth->d_addr.addr_b[5] & mask->d_addr.addr_b[5]) << 16) +
                              ((eth->s_addr.addr_b[0] & mask->s_addr.addr_b[0]) << 8) +
                               (eth->s_addr.addr_b[1] & mask->s_addr.addr_b[1]);
            e_word[elx + 2] = ((eth->s_addr.addr_b[2] & mask->s_addr.addr_b[2]) << 24) +
                              ((eth->s_addr.addr_b[3] & mask->s_addr.addr_b[3]) << 16) +
                              ((eth->s_addr.addr_b[4] & mask->s_addr.addr_b[4]) << 8) +
                               (eth->s_addr.addr_b[5] & mask->s_addr.addr_b[5]);

            e_mask[elx + 0] = (mask->d_addr.addr_b[0] << 24) + (mask->d_addr.addr_b[1] << 16) +
                              (mask->d_addr.addr_b[2] << 8) + mask->d_addr.addr_b[3];
            e_mask[elx + 1] = (mask->d_addr.addr_b[4] << 24) + (mask->d_addr.addr_b[5] << 16) +
                              (mask->s_addr.addr_b[0] << 8) + mask->s_addr.addr_b[1];
            e_mask[elx + 2] = (mask->s_addr.addr_b[2] << 24) + (mask->s_addr.addr_b[3] << 16) +
                              (mask->s_addr.addr_b[4] << 8) + mask->s_addr.addr_b[5];
            elx += 3;

            km_add_match_elem(&fd->km, e_word, e_mask, elx, DYN_L2, 0);
          }

          eth_type = eth->ether_type;
          eth_type_mask = mask->ether_type;

          /*
           * Set the L3 protocol type up front for supported protocols.
           * This ensures that we match on the specific eth type, even if there are
           * no follow-up IPv4 or IPv6 elements in the rte flow.
           */
          if ((eth_type & eth_type_mask) == htons(0x0800)) {
            fd->l3_prot = (1 << PROT_L3_IPV4);
          } else if ((eth_type & eth_type_mask) == htons(0x86dd)) {
            fd->l3_prot = (1 << PROT_L3_IPV6);
          }
          break;
        case FLOW_ELEM_TYPE_VLAN:
          NT_LOG(DBG, FILTER, "Adap %i, Port %i: FLOW_ELEM_TYPE_VLAN\n", dev->ndev->adapter_no, dev->port);
          if (fd->vlans > 3) {
            NT_LOG(DBG, FILTER, "ERROR: Too many VLANs in flow match\n");
            flow_nic_set_error(ERR_MATCH_VLAN_TOO_MANY, error);
            delete_flow_def(fd);
            return NULL;
          }

          fd->vlans++;
          if (elem[eidx].spec && elem[eidx].mask)
          {
            const struct flow_elem_vlan *vlan = (const struct flow_elem_vlan *)elem[eidx].spec;
            const struct flow_elem_vlan *vlan_mask = (const struct flow_elem_vlan *)elem[eidx].mask;

            if (fd->kcc && fd->vlans == 1) {
              NT_LOG(DBG, FILTER, "Add KCC KEY VLAN TPID,TCI: %04x,%04x\n",
                  ntohs(eth_type/* & eth_type_mask*/),
                  ntohs(vlan->tci & vlan_mask->tci));

              /* Cannot mask in KCC, thus a limitation here!!! */
              kcc_key_add_vlan(fd->kcc, ntohs(eth_type /* & eth_type_mask*/),
                  ntohs(vlan->tci /* & vlan_mask->tci*/));
            } else {
              e_word[0] = ((uint32_t)ntohs(eth_type & eth_type_mask) << 16) |
                          ((uint32_t)ntohs(vlan->tci & vlan_mask->tci));
              e_mask[0] = ((uint32_t) ntohs(eth_type_mask) << 16) |
                          ((uint32_t) ntohs(vlan_mask->tci));

              km_add_match_elem(&fd->km, e_word, e_mask, 1, DYN_FIRST_VLAN, (int8_t)((fd->vlans - 1) * 4));
            }

            eth_type = vlan->inner_type;
            eth_type_mask = vlan_mask->inner_type;

            /*
             * Set the L3 protocol type up front for supported protocols.
             * This ensures that we match on the specified vlan "inner_type" protocol type,
             * even if there are no follow-up IPv4 or IPv6 elements in the rte flow.
             */
            if ((eth_type & eth_type_mask) == htons(0x0800)) {
              fd->l3_prot = (1 << PROT_L3_IPV4);
            } else if ((eth_type & eth_type_mask) == htons(0x86dd)) {
              fd->l3_prot = (1 << PROT_L3_IPV6);
            }
          }
          break;
        case FLOW_ELEM_TYPE_IPV4:
          NT_LOG(DBG, FILTER, "Adap %i, Port %i: FLOW_ELEM_TYPE_IPV4\n", dev->ndev->adapter_no, dev->port);
          fd->l3_prot = (1 << PROT_L3_IPV4);
          eth_type = 0x0008;

          if (elem[eidx].spec && elem[eidx].mask)
          {
            const struct flow_elem_ipv4 *ipv4 = (const struct flow_elem_ipv4 *)elem[eidx].spec;
            const struct flow_elem_ipv4 *ipv4_mask = (const struct flow_elem_ipv4 *)elem[eidx].mask;

            if ((ipv4_mask->hdr.version_ihl) || ipv4_mask->hdr.tos) {
              e_word[0] = ((ipv4->hdr.version_ihl & ipv4_mask->hdr.version_ihl) << 24) |
                          ((ipv4->hdr.tos & ipv4_mask->hdr.tos) << 16);
              e_mask[0] = (((ipv4_mask->hdr.version_ihl) << 24) |
                          ((ipv4_mask->hdr.tos) << 16)) & 0xffff0000;

              km_add_match_elem(&fd->km, e_word, e_mask, 1, DYN_L3, 0);
            }

            /* if protocol is individually specified, ignore it here */
            uint8_t proto_mask = ipv4_mask->hdr.next_proto_id;
            if (protoSpecified(ipv4->hdr.next_proto_id & proto_mask, elem)) proto_mask = 0;

            if (ipv4_mask->hdr.ttl || ipv4_mask->hdr.frag_offset || proto_mask) {
              e_word[0] = (ntohs(ipv4->hdr.frag_offset) & ntohs(ipv4_mask->hdr.frag_offset)) << 16 |
                          (((ipv4->hdr.ttl & ipv4_mask->hdr.ttl) & 0xff) << 8) |
                          ((ipv4->hdr.next_proto_id & proto_mask) & 0xff);

              e_mask[0] = (ntohs(ipv4_mask->hdr.frag_offset) << 16) |
                          ((ipv4_mask->hdr.ttl & 0xff) << 8) |
                           (proto_mask & 0xff);

              km_add_match_elem(&fd->km, e_word, e_mask, 1, DYN_L3, 6);
            }

            /* Both IP addresses together */
            if (ipv4_mask->hdr.src_ip && ipv4_mask->hdr.dst_ip) {
              e_word[0] = ntohl(ipv4->hdr.src_ip) & ntohl(ipv4_mask->hdr.src_ip);
              e_word[1] = ntohl(ipv4->hdr.dst_ip) & ntohl(ipv4_mask->hdr.dst_ip);
              e_mask[0] = ntohl(ipv4_mask->hdr.src_ip);
              e_mask[1] = ntohl(ipv4_mask->hdr.dst_ip);

              km_add_match_elem(&fd->km, e_word, e_mask, 2, DYN_L3, 12);
            } else {

              if (ipv4_mask->hdr.src_ip) {
                e_word[0] = ntohl(ipv4->hdr.src_ip) & ntohl(ipv4_mask->hdr.src_ip);
                e_mask[0] = ntohl(ipv4_mask->hdr.src_ip);

                km_add_match_elem(&fd->km, e_word, e_mask, 1, DYN_L3, 12);
              }

              if (ipv4_mask->hdr.dst_ip) {
                e_word[0] = ntohl(ipv4->hdr.dst_ip) & ntohl(ipv4_mask->hdr.dst_ip);
                e_mask[0] = ntohl(ipv4_mask->hdr.dst_ip);

                km_add_match_elem(&fd->km, e_word, e_mask, 1, DYN_L3, 16);
              }
            }

          }
          break;
        case FLOW_ELEM_TYPE_IPV6:
          NT_LOG(DBG, FILTER, "Adap %i, Port %i: FLOW_ELEM_TYPE_IPV6\n", dev->ndev->adapter_no, dev->port);
          fd->l3_prot = (1 << PROT_L3_IPV6);
          eth_type = 0xdd86;

          if (elem[eidx].spec && elem[eidx].mask)
          {
            const struct flow_elem_ipv6 *ipv6 = (const struct flow_elem_ipv6 *)elem[eidx].spec;
            const struct flow_elem_ipv6 *ipv6_mask = (const struct flow_elem_ipv6 *)elem[eidx].mask;

            uint32_t vtc_flow = ntohl(ipv6->hdr.vtc_flow);
            uint32_t vtc_flow_mask = ntohl(ipv6_mask->hdr.vtc_flow);
            uint32_t vtc_flow_masked = vtc_flow & vtc_flow_mask;

            if ((vtc_flow_masked >> 28) != 6 && (vtc_flow_masked >> 28) != 0) {
              flow_nic_set_error(ERR_MATCH_INVALID_IPV6_HDR, error);
              delete_flow_def(fd);
              NT_LOG(DBG, FILTER, "ERR: ERR_MATCH_INVALID_IPV6_HDR\n");
              return NULL;
            }
            vtc_flow_mask &= ~(0xf << 28);

            if (vtc_flow_mask) {
              e_word[0] = vtc_flow & vtc_flow_mask;
              e_mask[0] = vtc_flow_mask;

              km_add_match_elem(&fd->km, e_word, e_mask, 1, DYN_L3, 0);

            }

            /* if protocol is individually specified, ignore it here */
            uint8_t proto_mask = ipv6_mask->hdr.proto;
            if (protoSpecified(ipv6->hdr.proto & proto_mask, elem)) proto_mask = 0;

            if (ipv6_mask->hdr.hop_limits || ipv6_mask->hdr.payload_len ||
                proto_mask) {

              e_word[0] = (((ntohs(ipv6->hdr.payload_len & ipv6_mask->hdr.payload_len)) & 0xffff) << 16) |
                          (((ipv6->hdr.proto & proto_mask) & 0xff) << 8) |
                          ((ipv6->hdr.hop_limits & ipv6_mask->hdr.hop_limits) & 0xff);

              e_mask[0] = (((ntohs(ipv6_mask->hdr.payload_len) & 0xffff) << 16) |
                           ((uint32_t)proto_mask << 8) |
                           ipv6_mask->hdr.hop_limits);

              km_add_match_elem(&fd->km, e_word, e_mask, 1, DYN_L3, 4);
            }

            if (ipv6_addr_is_set(ipv6_mask->hdr.src_addr) != 0) {
              e_word[0] = ((uint32_t)(ipv6->hdr.src_addr[0] & ipv6_mask->hdr.src_addr[0]) << 24)   + ((uint32_t)(ipv6->hdr.src_addr[1] & ipv6_mask->hdr.src_addr[1])<< 16) +
                          ((uint32_t)(ipv6->hdr.src_addr[2] & ipv6_mask->hdr.src_addr[2]) << 8)    + ((uint32_t)(ipv6->hdr.src_addr[3] & ipv6_mask->hdr.src_addr[3]));
              e_word[1] = ((uint32_t)(ipv6->hdr.src_addr[4] & ipv6_mask->hdr.src_addr[4]) << 24)   + ((uint32_t)(ipv6->hdr.src_addr[5] & ipv6_mask->hdr.src_addr[5]) << 16) +
                          ((uint32_t)(ipv6->hdr.src_addr[6] & ipv6_mask->hdr.src_addr[6]) << 8)    + ((uint32_t)(ipv6->hdr.src_addr[7] & ipv6_mask->hdr.src_addr[7]));
              e_word[2] = ((uint32_t)(ipv6->hdr.src_addr[8] & ipv6_mask->hdr.src_addr[8]) << 24)   + ((uint32_t)(ipv6->hdr.src_addr[9] & ipv6_mask->hdr.src_addr[9]) << 16) +
                          ((uint32_t)(ipv6->hdr.src_addr[10] & ipv6_mask->hdr.src_addr[10]) << 8)  + ((uint32_t)(ipv6->hdr.src_addr[11] & ipv6_mask->hdr.src_addr[11]));
              e_word[3] = ((uint32_t)(ipv6->hdr.src_addr[12] & ipv6_mask->hdr.src_addr[12]) << 24) + ((uint32_t)(ipv6->hdr.src_addr[13] & ipv6_mask->hdr.src_addr[13]) << 16) +
                          ((uint32_t)(ipv6->hdr.src_addr[14] & ipv6_mask->hdr.src_addr[14]) << 8)  + ((uint32_t)(ipv6->hdr.src_addr[15] & ipv6_mask->hdr.src_addr[15]));

              e_mask[0] = ((uint32_t)ipv6_mask->hdr.src_addr[0] << 24)  + ((uint32_t)ipv6_mask->hdr.src_addr[1] << 16) +
                          ((uint32_t)ipv6_mask->hdr.src_addr[2] << 8)   + ((uint32_t)ipv6_mask->hdr.src_addr[3]);
              e_mask[1] = ((uint32_t)ipv6_mask->hdr.src_addr[4] << 24)  + ((uint32_t)ipv6_mask->hdr.src_addr[5] << 16) +
                          ((uint32_t)ipv6_mask->hdr.src_addr[6] << 8)   + ((uint32_t)ipv6_mask->hdr.src_addr[7]);
              e_mask[2] = ((uint32_t)ipv6_mask->hdr.src_addr[8] << 24)  + ((uint32_t)ipv6_mask->hdr.src_addr[9] << 16) +
                          ((uint32_t)ipv6_mask->hdr.src_addr[10] << 8)  + ((uint32_t)ipv6_mask->hdr.src_addr[11]);
              e_mask[3] = ((uint32_t)ipv6_mask->hdr.src_addr[12] << 24) + ((uint32_t)ipv6_mask->hdr.src_addr[13] << 16) +
                          ((uint32_t)ipv6_mask->hdr.src_addr[14] << 8)  + ((uint32_t)ipv6_mask->hdr.src_addr[15]);

              km_add_match_elem(&fd->km, e_word, e_mask, 4, DYN_L3, 8);
            }

            if (ipv6_addr_is_set(ipv6_mask->hdr.dst_addr) != 0) {
              e_word[0] = ((uint32_t)(ipv6->hdr.dst_addr[0] & ipv6_mask->hdr.dst_addr[0]) << 24)   + ((uint32_t)(ipv6->hdr.dst_addr[1] & ipv6_mask->hdr.dst_addr[1])<< 16) +
                          ((uint32_t)(ipv6->hdr.dst_addr[2] & ipv6_mask->hdr.dst_addr[2]) << 8)    + ((uint32_t)(ipv6->hdr.dst_addr[3] & ipv6_mask->hdr.dst_addr[3]));
              e_word[1] = ((uint32_t)(ipv6->hdr.dst_addr[4] & ipv6_mask->hdr.dst_addr[4]) << 24)   + ((uint32_t)(ipv6->hdr.dst_addr[5] & ipv6_mask->hdr.dst_addr[5]) << 16) +
                          ((uint32_t)(ipv6->hdr.dst_addr[6] & ipv6_mask->hdr.dst_addr[6]) << 8)    + ((uint32_t)(ipv6->hdr.dst_addr[7] & ipv6_mask->hdr.dst_addr[7]));
              e_word[2] = ((uint32_t)(ipv6->hdr.dst_addr[8] & ipv6_mask->hdr.dst_addr[8]) << 24)   + ((uint32_t)(ipv6->hdr.dst_addr[9] & ipv6_mask->hdr.dst_addr[9]) << 16) +
                          ((uint32_t)(ipv6->hdr.dst_addr[10] & ipv6_mask->hdr.dst_addr[10]) << 8)  + ((uint32_t)(ipv6->hdr.dst_addr[11] & ipv6_mask->hdr.dst_addr[11]));
              e_word[3] = ((uint32_t)(ipv6->hdr.dst_addr[12] & ipv6_mask->hdr.dst_addr[12]) << 24) + ((uint32_t)(ipv6->hdr.dst_addr[13] & ipv6_mask->hdr.dst_addr[13]) << 16) +
                          ((uint32_t)(ipv6->hdr.dst_addr[14] & ipv6_mask->hdr.dst_addr[14]) << 8)  + ((uint32_t)(ipv6->hdr.dst_addr[15] & ipv6_mask->hdr.dst_addr[15]));

              e_mask[0] = ((uint32_t)ipv6_mask->hdr.dst_addr[0] << 24)  + ((uint32_t)ipv6_mask->hdr.dst_addr[1] << 16) +
                          ((uint32_t)ipv6_mask->hdr.dst_addr[2] << 8)   + ((uint32_t)ipv6_mask->hdr.dst_addr[3]);
              e_mask[1] = ((uint32_t)ipv6_mask->hdr.dst_addr[4] << 24)  + ((uint32_t)ipv6_mask->hdr.dst_addr[5] << 16) +
                          ((uint32_t)ipv6_mask->hdr.dst_addr[6] << 8)   + ((uint32_t)ipv6_mask->hdr.dst_addr[7]);
              e_mask[2] = ((uint32_t)ipv6_mask->hdr.dst_addr[8] << 24)  + ((uint32_t)ipv6_mask->hdr.dst_addr[9] << 16) +
                          ((uint32_t)ipv6_mask->hdr.dst_addr[10] << 8)  + ((uint32_t)ipv6_mask->hdr.dst_addr[11]);
              e_mask[3] = ((uint32_t)ipv6_mask->hdr.dst_addr[12] << 24) + ((uint32_t)ipv6_mask->hdr.dst_addr[13] << 16) +
                          ((uint32_t)ipv6_mask->hdr.dst_addr[14] << 8)  + ((uint32_t)ipv6_mask->hdr.dst_addr[15]);

              km_add_match_elem(&fd->km, e_word, e_mask, 4, DYN_L3, 24);
            }

          }
          break;

        case FLOW_ELEM_TYPE_UDP:
          NT_LOG(DBG, FILTER, "Adap %i, Port %i: FLOW_ELEM_TYPE_UDP\n", dev->ndev->adapter_no, dev->port);
          fd->l4_prot = (1 << PROT_L4_UDP);

          if (elem[eidx].spec && elem[eidx].mask)
          {
            const struct flow_elem_udp *udp = (const struct flow_elem_udp *)elem[eidx].spec;
            const struct flow_elem_udp *udp_mask = (const struct flow_elem_udp *)elem[eidx].mask;

            add_layer4_port_match(fd, udp->hdr.src_port,
                udp->hdr.dst_port, udp_mask->hdr.src_port, udp_mask->hdr.dst_port);
          }
          break;
        case FLOW_ELEM_TYPE_SCTP:
          NT_LOG(DBG, FILTER, "Adap %i,Port %i:FLOW_ELEM_TYPE_SCTP\n", dev->ndev->adapter_no, dev->port);
          fd->l4_prot = (1 << PROT_L4_SCTP);

          if (elem[eidx].spec && elem[eidx].mask)
          {
            const struct flow_elem_sctp *sctp = (const struct flow_elem_sctp *)elem[eidx].spec;
            const struct flow_elem_sctp *sctp_mask = (const struct flow_elem_sctp *)elem[eidx].mask;

            add_layer4_port_match(fd, sctp->hdr.src_port,
                sctp->hdr.dst_port, sctp_mask->hdr.src_port, sctp_mask->hdr.dst_port);
          }
          break;
        case FLOW_ELEM_TYPE_TCP:
          NT_LOG(DBG, FILTER, "Adap %i, Port %i: FLOW_ELEM_TYPE_TCP\n", dev->ndev->adapter_no, dev->port);
          fd->l4_prot = (1 << PROT_L4_TCP);

          if (elem[eidx].spec && elem[eidx].mask)
          {
            const struct flow_elem_tcp *tcp = (const struct flow_elem_tcp *)elem[eidx].spec;
            const struct flow_elem_tcp *tcp_mask = (const struct flow_elem_tcp *)elem[eidx].mask;

            add_layer4_port_match(fd, tcp->hdr.src_port,
                tcp->hdr.dst_port, tcp_mask->hdr.src_port, tcp_mask->hdr.dst_port);

            if (tcp_mask->hdr.data_off || tcp_mask->hdr.tcp_flags) {
              uint8_t mask_offs  = tcp_mask->hdr.data_off & 0xf0;
              uint8_t mask_flags = tcp_mask->hdr.tcp_flags & 0x8f;

              e_word[0] = (uint32_t)((tcp->hdr.data_off & mask_offs) << 24) +
                                    ((tcp->hdr.tcp_flags & mask_flags) << 16);
              e_mask[0] = (uint32_t)((mask_offs << 24) + (mask_flags << 16)) & 0xffff0000;

              km_add_match_elem(&fd->km, e_word, e_mask, 1, DYN_L4, 12);
            }
          }
          break;
        case FLOW_ELEM_TYPE_ICMP:
          NT_LOG(DBG, FILTER, "Adap %i, Port %i: FLOW_ELEM_TYPE_ICMP\n", dev->ndev->adapter_no, dev->port);
          fd->l4_prot = (1 << PROT_L4_ICMP);

          if (elem[eidx].spec && elem[eidx].mask)
          {
            const struct flow_elem_icmp *icmp = (const struct flow_elem_icmp *)elem[eidx].spec;
            const struct flow_elem_icmp *icmp_mask = (const struct flow_elem_icmp *)elem[eidx].mask;

            if (icmp_mask->hdr.type || icmp_mask->hdr.code) {
              e_word[0] = (uint32_t)((icmp->hdr.type & icmp_mask->hdr.type) << 24) +
                                    ((icmp->hdr.code & icmp_mask->hdr.code) << 16);
              e_mask[0] = (uint32_t)((icmp_mask->hdr.type << 24) + (icmp_mask->hdr.code << 16)) & 0xffff0000;

              km_add_match_elem(&fd->km, e_word, e_mask, 1, DYN_L4, 0);
            }
          }
          break;

        case FLOW_ELEM_TYPE_VXLAN:
          NT_LOG(DBG, FILTER, "Adap %i, Port %i: FLOW_ELEM_TYPE_VXLAN\n", dev->ndev->adapter_no, dev->port);

          if (elem[eidx].spec && elem[eidx].mask)
          {
            const struct flow_elem_vxlan *vxlan = (const struct flow_elem_vxlan *)elem[eidx].spec;
            const struct flow_elem_vxlan *vxlan_mask = (const struct flow_elem_vxlan *)elem[eidx].mask;

            uint32_t vni = (uint32_t)(((uint32_t)vxlan->vni[0] << 16) | ((uint32_t)vxlan->vni[1] << 8) |
                                      ((uint32_t)vxlan->vni[2]));
            uint32_t vni_mask = (uint32_t)(((uint32_t)vxlan_mask->vni[0] << 16) | ((uint32_t)vxlan_mask->vni[1] << 8) |
                                         ((uint32_t)vxlan_mask->vni[2]));

            e_word[0] = vni & vni_mask;
            e_mask[0] = vni_mask;

            if (fd->kcc) {
              NT_LOG(DBG, FILTER, "Add KCC KEY VxLAN VNI: %08x\n", e_word[0]);
              kcc_key_add_vxlan(fd->kcc, e_word[0]);
            } else {
              /* Use as metadata carrying VNI from a recently pop'ed VXLAN header */
              e_word[0] |= 2 << 24;
              e_mask[0] |= 0xff << 24;
              km_add_match_elem(&fd->km, e_word, e_mask, 1, SB_VNI, 0);
            }
            svx_used = 1;
        }
        break;

        default:
          valid = 0;
          break;
        }

        eidx++;
      }

    }
  } /* valid - from actions */

  if (valid == 0) {
    NT_LOG(ERR, FILTER, "INFO: Invalid or unsupported flow request\n");
    flow_nic_set_error(ERR_MATCH_INVALID_OR_UNSUPPORTED_ELEM, error);
    delete_flow_def(fd);
    return NULL;
  }

  if (eth_type & eth_type_mask) {
    NT_LOG(DBG, FILTER, "Match Packet EtherType: %04x\n", ntohs(eth_type & eth_type_mask));
  }

  /*
   * in vSwitch mode we don't want to offload everything
   */
  if (dev->ndev->flow_profile == FLOW_ETH_DEV_PROFILE_VSWITCH) {
    if (eth_type == 0x0988 && fd->mark == FLOW_MARK_LACP) {
      /*
       * register LACP match
       */
      fd->special_match = SPECIAL_MATCH_LACP;
      /*
       * This EtherType is not known to CAT frame decoder, therefore
       * we need to specifically match it here in KM also
       */
      e_word[0] = (uint32_t)(htons(eth_type)<<16);
      e_mask[0] = (uint32_t)0xffff0000;

      km_add_match_elem(&fd->km, e_word, e_mask, 1, DYN_L2, 12);

    } else
    if (eth_type != 0x0008 && eth_type != 0xdd86) {
      NT_LOG(DBG, FILTER, "INFO: Only hw-offload IP traffic IPv4 or IPv6 - (%04x) rejected by policy\n", eth_type);
      flow_nic_set_error(ERR_MATCH_UNSUPPORTED_ETHER_TYPE, error);
      delete_flow_def(fd);

      return NULL;
    }

#if 1
    /* For now, partial offload has been disabled */
    if (!fd->full_offload && fd->special_match == SPCIAL_MATCH_NONE) {
        NT_LOG(DBG, FILTER, "INFO: Partial offload not supported - rejected by policy\n");
      flow_nic_set_error(ERR_MATCH_PARTIAL_OFFLOAD_NOT_SUPPORTED, error);
      delete_flow_def(fd);
      return NULL;
    }
#endif
  }

  *in_port_id = (fd->in_port_override >= 0)? fd->in_port_override : dev->port;

  /* Find out if we need KM for this flow offload */

  if (fd->km.num_ftype_elem || fd->kcc) {
    int kcc_unique_id = -1;
    fd->kcc_referenced = 0;
    /*
     * if KCC is available, we need to use a unique ID through SWX word to bind KCC entry with KM entry
     */
    if (fd->kcc) {
      NT_LOG(DBG, FILTER, "Add in_port to KCC Key. Port: %i\n", *in_port_id);
      kcc_key_add_port(fd->kcc, (uint16_t) *in_port_id);

      int catfn = prematch_cat_fn(dev, fd);
      if (catfn >= 0) {
        NT_LOG(DBG, FILTER, "Found matching CAT FN - check for matching KCC key...\n");
        kcc_key_add_cat_cfn(fd->kcc, (uint8_t) catfn);

        /* search for a KCC match */
        struct flow_handle *flow = dev->ndev->flow_base;
        while (flow) {
          if (kcc_key_compare(fd->kcc, flow->fd->kcc)) {
            /* if match found, a match in KM-recipe must be found too, otherwise discard flow */
            NT_LOG(DBG, FILTER, "KCC key match found: Reuse KCC Key %08x\n",
                flow->fd->kcc->key32[1]);
            kcc_key_ref_count_add(flow->fd->kcc);
            fd->kcc_referenced = 1;
            free(fd->kcc);
            fd->kcc = flow->fd->kcc;
            break;
          }
          flow = flow->next;
        }

        if (!fd->kcc_referenced) {
          kcc_key_add_cat_cfn(fd->kcc, (uint8_t)catfn);
        }

      } else {
        /* No CAT_FN match found */
        NT_LOG(DBG, FILTER, "Found NO matching CAT FN - need to create a new one later\n");
        kcc_key_add_cat_cfn(fd->kcc, (uint8_t) 0xff); /* indicate no matches found - need new CAT_FN recipe */
      }

      if (!fd->kcc_referenced) {
        /* allocate new kcc unique ID */
        kcc_unique_id = kcc_alloc_unique_id(fd->kcc);

        if (kcc_unique_id < 0) {
          NT_LOG(DBG, FILTER, "INFO: CAT CAM exhausted\n");
          flow_nic_set_error(ERR_MATCH_CAT_CAM_EXHAUSTED, error);
          delete_flow_def(fd);
          return NULL;
        }
      } else {
        kcc_unique_id = fd->kcc->id;
      }

      NT_LOG(DBG, FILTER, "Add KCC ID to KM Key. Unique ID: %04x\n", kcc_unique_id);
      e_word[0] = (uint32_t) kcc_unique_id;
      e_mask[0] = 0xffffffff;
      km_add_match_elem(&fd->km, e_word, e_mask, 1, SB_KCC_ID, 0);
      svx_used = 1;

    } else {
      /*
       * Make sure we add MAC port information in CAM, but only when possible.
       * If VNI is matched we cannot match MAC port also
       */
      if (!svx_used) {
        NT_LOG(DBG, FILTER, "Adap %i, Port %i: Add Port as Sideband\n",
            dev->ndev->adapter_no, dev->port);
        e_word[0] = *in_port_id;
        e_mask[0] = 0x000000ff;
        km_add_match_elem(&fd->km, e_word, e_mask, 1, SB_MAC_PORT, 0);
        svx_used = 1;
      }
    }

    int res = km_key_create(&fd->km, *in_port_id);

    if (res != 0) {
      NT_LOG(DBG, FILTER, "KM creation failed\n");
      flow_nic_set_error(ERR_MATCH_FAILED_BY_HW_LIMITS, error);
      delete_flow_def(fd);
      return NULL;
    }

    /*
     * Color as info not finally set yet, but will come later
     * however, must already now be announced it will be there
     * before we compare against existing flows - VSWITCH
     */
    km_set_info(&fd->km, 1);

    NT_LOG(DBG, FILTER, "Search for identical key flow type...\n");
    /* Find existing, identical KM flow type if possible */
    struct flow_handle *flow = dev->ndev->flow_base, *found_flow = NULL;
    while (flow) {
      /*
       * Does this flow already exist or otherwise contain a valid KM RCP
       * and flow type that we may reuse.
       * Note, only non ref'ed flows. They contain the original recipe
       * information
       */
      if (flow->fd->km.flow_type &&
          flow->resource[RES_KM_CATEGORY].count &&
          !flow->resource[RES_KM_CATEGORY].referenced) {
        int type = km_key_compare(&fd->km, &flow->fd->km);

        if (type < 0) {
          /* Identical flow found already setup */
          NT_LOG(DBG, FILTER, "ERROR: Identical flows found. Rejecting this flow\n");
          delete_flow_def(fd);
          return NULL;
        }

        if (type > 0) {
          /* only set it once, but continue checking for flow match clash possibilities */
          if (fd->km.flow_type == 0) {
            fd->km.flow_type = type;
            found_flow = flow;
          }
        }

      }
      flow = flow->next;
    }

    if (found_flow && fd->km.flow_type) {
      NT_LOG(DBG, FILTER, "YES, found an identical flow type %i, used by KM recipe %i\n", fd->km.flow_type, found_flow->resource[RES_KM_CATEGORY].index);
      assert(found_flow->resource[RES_KM_CATEGORY].count == 1);

      *km_rcp_index = found_flow->resource[RES_KM_CATEGORY].index;
    } else {

      /* if no KM recipe reuse found and kcc and fd->kcc_referenced - discard */
      if (fd->kcc && fd->kcc_referenced) {
        NT_LOG(DBG, FILTER,
            "ERROR: Identical KCC Key found for reuse but no identical KM key type found - discard\n");
        delete_flow_def(fd);
        return NULL;
      }

      NT_LOG(DBG, FILTER, "None identical found. Will create a new one\n");
    }

  }

  return fd;
}






static int reset_cat_function_setup(struct flow_eth_dev *dev, int cfn, int km, int flushed)
{
  uint32_t val;

  hw_mod_cat_cfn_set(&dev->ndev->be, HW_CAT_CFN_PRESET_ALL, cfn, 0, 0);
  if (flushed) {
    hw_mod_cat_cfn_flush(&dev->ndev->be, cfn, 1);
  }
  /*
   * Remove and clear CFN refs in
   * KCE
   * KCS
   * FTE
   * CTE
   * CTS
   * (CCT)
   */
  /* KCE (KM Enable) - Disable KM if enabled - only if KCC not enabled */
  if (km && dev->ndev->be.cat.kcc_size == 0) {

    hw_mod_cat_kce_km_get(&dev->ndev->be, HW_CAT_KCE_ENABLE_BM, KM_FLM_IF_FIRST, cfn / 8, &val);
    hw_mod_cat_kce_km_set(&dev->ndev->be, HW_CAT_KCE_ENABLE_BM, KM_FLM_IF_FIRST, cfn / 8, val & ~(1 << (cfn % 8)));
    hw_mod_cat_kcs_km_set(&dev->ndev->be, HW_CAT_KCS_CATEGORY, KM_FLM_IF_FIRST, cfn, 0);

    if (flushed) {
      hw_mod_cat_kce_km_flush(&dev->ndev->be, KM_FLM_IF_FIRST, cfn / 8, 1);
      hw_mod_cat_kcs_km_flush(&dev->ndev->be, KM_FLM_IF_FIRST, cfn, 1);
    }
  }

  /* FTE (Flow Type Enable) - Disable all flow types for both lookups */
  for (unsigned int ft = 0; ft < dev->ndev->be.cat.nb_flow_types; ft++) {
    int index = (8 * ft + cfn / 8) * 2;
    hw_mod_cat_fte_km_get(&dev->ndev->be, HW_CAT_FTE_ENABLE_BM, KM_FLM_IF_FIRST, index, &val);                           /* lookup A */
    if (val & (1 << (cfn % 8))) {
      val &= ~(1 << (cfn % 8));
      hw_mod_cat_fte_km_set(&dev->ndev->be, HW_CAT_FTE_ENABLE_BM, KM_FLM_IF_FIRST, index, val);      /* lookup A */
      if (flushed) {
        hw_mod_cat_fte_km_flush(&dev->ndev->be, KM_FLM_IF_FIRST, index, 1);
      } else {
        NT_LOG(DBG, FILTER, "RESET: SET FTE VAL %02x  addr %i\n", val, index);
      }
    }
    hw_mod_cat_fte_km_get(&dev->ndev->be, HW_CAT_FTE_ENABLE_BM, KM_FLM_IF_FIRST, index + 1, &val);                       /* lookup B */
    if (val & (1 << (cfn % 8))) {
      hw_mod_cat_fte_km_set(&dev->ndev->be, HW_CAT_FTE_ENABLE_BM, KM_FLM_IF_FIRST, index + 1, val & ~(1 << (cfn % 8)));  /* lookup B */
      if (flushed) {
        hw_mod_cat_fte_km_flush(&dev->ndev->be, KM_FLM_IF_FIRST, index + 1, 1);
      }
    }
  }

  uint32_t cte;
  hw_mod_cat_cte_get(&dev->ndev->be, HW_CAT_CTE_ENABLE_BM, cfn, &cte);

  if (cte) {
    /* CTE (Category Table Enable) - Disable all categories */
    hw_mod_cat_cte_set(&dev->ndev->be, HW_CAT_CTE_ENABLE_BM, cfn, 0);
    if (flushed) {
      hw_mod_cat_cte_flush(&dev->ndev->be, cfn, 1);
    }

  for (uint32_t cte_type = 0; cte_type < dev->ndev->be.cat.cts_num; cte_type++) {
    if (cts_offset[cte_type].offset >= 0) {
      hw_mod_cat_cts_set(&dev->ndev->be, HW_CAT_CTS_CAT_A, CTS_CFN_OFFSET(dev->ndev->be.cat.cts_num, cfn) + cts_offset[cte_type].offset, 0);
      hw_mod_cat_cts_set(&dev->ndev->be, HW_CAT_CTS_CAT_B, CTS_CFN_OFFSET(dev->ndev->be.cat.cts_num, cfn) + cts_offset[cte_type].offset, 0);
      }
    }

    if (flushed) {
    hw_mod_cat_cts_flush(&dev->ndev->be, CTS_CFN_OFFSET(dev->ndev->be.cat.cts_num, cfn), (dev->ndev->be.cat.cts_num + 1) / 2);
    }
  }

  return 0;
}


int flow_validate_profile_vswitch(
    struct flow_eth_dev *dev,
    const struct flow_elem elem[],
    const struct flow_action action[],
    struct flow_error *error)
{
  uint32_t port_id;
  int km_rcp_index;

  flow_nic_set_error(ERR_SUCCESS, error);

  pthread_mutex_lock(&dev->ndev->mtx);
  struct nic_flow_def *fd = interpret_flow_elements(dev, elem, action, error, &port_id, &km_rcp_index);
  pthread_mutex_unlock(&dev->ndev->mtx);

  if (!fd) {
    return -1;
  }

  delete_flow_def(fd);
  return 0;
}


static int flush_flow_handle(struct flow_eth_dev *dev, struct flow_handle *fh, struct flow_error *error)
{

  for (unsigned int res = 0; res < RES_COUNT; res++) {

    /* if resource needed and not already referenced(reusing another) */
    if (fh->resource[res].count && !fh->resource[res].referenced) {
      switch (res) {

      case RES_KM_FLOW_TYPE:
        /*
         * nothing to do. A resource needed by KM
         * used in address in FTE
         */
        break;

      case RES_KM_CATEGORY:
        assert(fh->resource[res].count == 1);
        hw_mod_km_rcp_flush(&dev->ndev->be, fh->resource[res].index, fh->resource[res].count);
        break;

      /*
       * QSL QST: Queue Select Table
       */
      case RES_QSL_QST:
        hw_mod_qsl_qst_flush(&dev->ndev->be, fh->resource[res].index, fh->resource[res].count);
        break;

      /*
       * CAT COT: CAT Color Table Memory
       */
      case RES_CAT_COT:
        assert(fh->resource[res].count == 1);
        hw_mod_cat_cot_flush(&dev->ndev->be, fh->resource[res].index, fh->resource[res].count);
        break;
#if 0
      /*
       * COR: Correlation Key Module
       *
       * COR - TODO
       */
#endif

      /*
       * HSH RCP: Hasher Recipe
       */
      case RES_HSH_RCP:
        assert(fh->resource[res].count == 1);
        hw_mod_hsh_rcp_flush(&dev->ndev->be, fh->resource[res].index, fh->resource[res].count);
        break;

      /*
       * QSL RCP: Queue Select Recipe
       */
      case RES_QSL_RCP:
        assert(fh->resource[res].count == 1);
        hw_mod_qsl_rcp_flush(&dev->ndev->be, fh->resource[res].index, fh->resource[res].count);
        break;

      /*
       * SLC RCP: Slicer Recipe
       */
      case RES_SLC_RCP:
        assert(fh->resource[res].count == 1);
        hw_mod_slc_rcp_flush(&dev->ndev->be, fh->resource[res].index, fh->resource[res].count);
        break;

      /*
       * PDB RCP: Packet Descriptor Builder Recipe
       */
      case RES_PDB_RCP:
        assert(fh->resource[res].count == 1);
        hw_mod_pdb_rcp_flush(&dev->ndev->be, fh->resource[res].index, fh->resource[res].count);
        break;

      /*
       * ROA RCP: Retransmit Offload Actions Recipe
       */
      case RES_ROA_RCP:
        assert((fh->resource[res].count == 1) || (fh->resource[res].count == 2)); /* 1 for IPv4, 2 for IPv6 */
        uint32_t hdrlen = 0;
        uint32_t iptype = 0;
        hw_mod_roa_tuncfg_get(&dev->ndev->be, HW_ROA_TUNCFG_TUN_LEN, fh->resource[res].index, &hdrlen);
        if (hdrlen) {
          hw_mod_roa_tuncfg_get(&dev->ndev->be, HW_ROA_TUNCFG_IP_TYPE, fh->resource[res].index, &iptype);
          if (iptype == 0) {
            hw_mod_roa_tunhdr_flush(&dev->ndev->be, fh->resource[res].index * 4, 1); /* Count is 1 for IPv4 */
          } else {
            hw_mod_roa_tunhdr_flush(&dev->ndev->be, fh->resource[res].index * 4, 2); /* Count is 2 for IPv6 */
          }
        }
        hw_mod_roa_tuncfg_flush(&dev->ndev->be, fh->resource[res].index, fh->resource[res].count);

        /* this IOA memory follows ROA RCP */
        hw_mod_ioa_roa_epp_flush(&dev->ndev->be, fh->resource[res].index, fh->resource[res].count);

        /*
         * if new created internal virtual port - with no queue configured (VTEP),
         * we need to redirect non-matched/exception packets to the port exception
         * queue, which uses the initial match carrying this recirculate ->
         * internal virtual port, Additional information is carried in color bits
         * and in-port will be changed in Packet Descriptor
         */
        if (get_roa_new_recirc_port(fh->fd->roa_actions)) {
          /* create QSL UNM - unmatched - redirects for default/exception queue */
          int32_t virt_port = roa_get_recirc_port(fh->fd->roa_actions);

          if (virt_port < 0 || virt_port > 255) {
#if 0
            /* TODO error */
#endif
            NT_LOG(DBG, FILTER, "ERROR: New virtual recirculation port creation indicated, but none found!\n");
            flow_nic_set_error(ERR_INTERNAL_VIRTUAL_PORT_CREATION_FAILED, error);
            break;
          }
          uint32_t val, hw_id;
          hw_mod_qsl_unmq_get(&dev->ndev->be, HW_QSL_UNMQ_EN, virt_port, &val);
          hw_mod_qsl_unmq_get(&dev->ndev->be, HW_QSL_UNMQ_DEST_QUEUE, virt_port, &hw_id);
          NT_LOG(DBG, FILTER, "Is default queue setup on index %i : %s, hw id (%i, %i)\n",virt_port, val?"yes":"no",
              hw_id, dev->rx_queue[0].hw_id);
          if (!val || (dev->rx_queue[0].hw_id != (int)hw_id)) {
            hw_mod_qsl_unmq_set(&dev->ndev->be, HW_QSL_UNMQ_DEST_QUEUE, virt_port, dev->rx_queue[0].hw_id);
            hw_mod_qsl_unmq_set(&dev->ndev->be, HW_QSL_UNMQ_EN, virt_port, 1);
            hw_mod_qsl_unmq_flush(&dev->ndev->be, virt_port, 1);

            NT_LOG(DBG, FILTER, "Write new default destination for virtual port %i, exception queue %i\n",
                virt_port, dev->rx_queue[0].hw_id);
          }
        }
        break;

      /*
       * IOA RCP: Inline Offload Actions Recipe
       */
      case RES_IOA_RCP:
        assert(fh->resource[res].count == 1);
        hw_mod_ioa_rcp_flush(&dev->ndev->be, fh->resource[res].index, fh->resource[res].count);
        break;

      /*
       * CAT CFN: CAT Function Memory
       */
      case RES_CAT_CFN:
        assert(fh->resource[RES_CAT_CFN].count == 1);
        /* Key Matcher enabler */
        if (fh->resource[RES_KM_CATEGORY].count) {

          /* if KCC enabled don't write to KCE/KCS */
          if (dev->ndev->be.cat.kcc_size == 0) {
            hw_mod_cat_kce_km_flush(&dev->ndev->be, KM_FLM_IF_FIRST, fh->resource[RES_CAT_CFN].index / 8, 1);
            hw_mod_cat_kcs_km_flush(&dev->ndev->be, KM_FLM_IF_FIRST, fh->resource[RES_CAT_CFN].index, 1);
          }

          int index = (8 * fh->resource[RES_KM_FLOW_TYPE].index
              + fh->resource[RES_CAT_CFN].index / 8) * 2;
          hw_mod_cat_fte_km_flush(&dev->ndev->be, KM_FLM_IF_FIRST, index, 1);

        } else {
          uint32_t val;
          /* No KM recipe setup - enable all flow types */
          int cfn_adr = fh->resource[RES_CAT_CFN].index / 8;
          int cfn_bit = fh->resource[RES_CAT_CFN].index % 8;

          for (unsigned int ft = 0; ft < dev->ndev->be.cat.nb_flow_types; ft++) {
            int index = (8 * ft + cfn_adr) * 2;

            hw_mod_cat_fte_km_get(&dev->ndev->be, HW_CAT_FTE_ENABLE_BM, KM_FLM_IF_FIRST, index, &val);     /* lookup A */
            val |= (1 << cfn_bit);
            hw_mod_cat_fte_km_set(&dev->ndev->be, HW_CAT_FTE_ENABLE_BM, KM_FLM_IF_FIRST, index, val);      /* lookup A */
#if 0  /* not using B for now */
            hw_mod_cat_fte_km_get(&dev->ndev->be, HW_CAT_FTE_ENABLE_BM, KM_FLM_IF_FIRST, index + 1, &val); /* lookup B */
            val |= (1 << cfn_bit);
            hw_mod_cat_fte_km_set(&dev->ndev->be, HW_CAT_FTE_ENABLE_BM, KM_FLM_IF_FIRST, index + 1, val);  /* lookup B */
#endif
            hw_mod_cat_fte_km_flush(&dev->ndev->be, KM_FLM_IF_FIRST, index, 1);
          }
        }

        /* Category Table enabling */
        if (fh->cte) {
          int last = -1;
          for (unsigned int cte_type = 0; cte_type < CT_CNT; cte_type++) {
            if (fh->cte & (1 << cte_type)) {
              /* HW_CAT_CTS_CATEGORY */

              int offset = cts_offset[cte_type].offset;
              assert(offset >= 0);

              if (offset != last) {
                hw_mod_cat_cts_flush(&dev->ndev->be, CTS_CFN_OFFSET(dev->ndev->be.cat.cts_num, fh->resource[RES_CAT_CFN].index) + offset, 1);
                last = offset;
              }

            }
          }
          hw_mod_cat_cte_flush(&dev->ndev->be, fh->resource[RES_CAT_CFN].index, 1);
        }

        /* Flush CAT CFN later */
        break;
#if 0
        RES_CAT_EXO,
        RES_CAT_LEN,
#endif
      default:
        assert(0);
        flow_nic_set_error(ERR_FAILED, error);
        goto err_exit;
      }
    } /* count - any defined and !referenced */
  }

  /* If new to be created, flush CAT CFN as "first"-last thing to enable */
  if (fh->resource[RES_CAT_CFN].count && !fh->resource[RES_CAT_CFN].referenced) {
    hw_mod_cat_cfn_flush(&dev->ndev->be, fh->resource[RES_CAT_CFN].index, fh->resource[RES_CAT_CFN].count);
  }

  /* Do KM CAM/TCAM as last thing */
  if (fh->fd->km.flow_type) {
    int res;
    /*
     * if KCC-CAM available, we need to populate that first
     */
    if (fh->fd->kcc && !fh->fd->kcc_referenced) {
      res = kcc_write_data_match_entry(fh->fd->kcc);
      if (res < 0) {
        flow_nic_set_error(ERR_MATCH_CAT_CAM_FAILED, error);
        goto err_exit;
      }
    }
    /*
     * Now since KM is used we need to populate the CAM/TCAM entries + color information
     */
    res = km_write_data_match_entry(&fh->fd->km, fh->color);
    if (res < 0) {
      flow_nic_set_error(ERR_MATCH_CAM_EXHAUSTED, error);
      goto err_exit;
    }
  }

  fh->flushed = 1;
  return 0;

err_exit:
  return -1;

}


struct flow_handle *flow_create_profile_vswitch(struct flow_eth_dev *dev,
   const struct flow_elem elem[],
   const struct flow_action action[],
   UNUSED struct flow_error *error)
{
  uint32_t port_id;
  int km_rcp_index = -1;
  struct flow_handle *fh = NULL;

  flow_nic_set_error(ERR_SUCCESS, error);

  pthread_mutex_lock(&dev->ndev->mtx);
  struct nic_flow_def *fd = interpret_flow_elements(dev, elem, action, error, &port_id, &km_rcp_index);
  if (!fd) {
    goto err_exit;
  }

#if 0
  /* TODO - handle flow profiles - much will be shared between profiles, but some of it must be different */
#endif
  fh = new_flow_handle(dev, fd);
  if (!fh) {
    flow_nic_set_error(ERR_MEMORY, error);
    goto err_exit;
  }

  NT_LOG(DBG, FILTER, "New FlOW: fh (flow handle) %p, fd (flow definition) %p\n", fh, fd);
  NT_LOG(DBG, FILTER, ">>>>> [Dev %p] Nic %i, Port %i: fh %p fd %p - implementation <<<<<\n",
      dev, dev->ndev->adapter_no, dev->port, fh, fd);
  /*
   * Start implement into HW
   * First create KM part if needed
   */

  fh->port_id = port_id;

  if (fh->fd->km.num_ftype_elem) {
    /*
     * Need Key Matcher
     * Attach NIC scoped Key Matcher resource management
     * to this flow KM handler (create on first time)
     */
    fd->km.be = &dev->ndev->be;
    km_attach_ndev_resource_management(&fd->km, &dev->ndev->km_res_handle);

    /* if already shared by another KM RCP */
    if (fh->fd->km.flow_type) {
      assert(km_rcp_index >= 0);

      int err = flow_nic_ref_resource(dev->ndev, RES_KM_CATEGORY, km_rcp_index);
      if (err) {
        NT_LOG(DBG, FILTER, "ERROR: Could not reference KM RCP resource\n");
        flow_nic_set_error(ERR_MATCH_RESOURCE_EXHAUSTION, error);
        goto err_exit;
      }
      fh->resource[RES_KM_CATEGORY].count = 1;
      fh->resource[RES_KM_CATEGORY].index = km_rcp_index;
      fh->resource[RES_KM_CATEGORY].referenced = 1;

      err = flow_nic_ref_resource(dev->ndev, RES_KM_FLOW_TYPE, fh->fd->km.flow_type);
      if (err) {
        NT_LOG(DBG, FILTER, "ERROR: Could not reference Flow Type resource\n");
        flow_nic_set_error(ERR_MATCH_RESOURCE_EXHAUSTION, error);
        goto err_exit;
      }
      fh->resource[RES_KM_FLOW_TYPE].count = 1;
      fh->resource[RES_KM_FLOW_TYPE].index = fh->fd->km.flow_type;
      fh->resource[RES_KM_FLOW_TYPE].referenced = 1;
      /* KM RCP and Flow Type done - shared */

      NT_LOG(DBG, FILTER, "reuse existing flow type %i (fd:%p)\n", fh->fd->km.flow_type, fh->fd);

    } else {
      /* need to create key definition - KM RCP */
      if (flow_nic_allocate_fh_resource(dev->ndev, RES_KM_CATEGORY, fh, 1, 1)) {
        NT_LOG(ERR, FILTER, "ERROR: Could not get KM RCP resource\n");
        flow_nic_set_error(ERR_MATCH_RESOURCE_EXHAUSTION, error);
        goto err_exit;
      }

      /* if KCC */
      if (fh->fd->kcc) {
        /* use same flow type for all flows in KM */
        int err = 0;

        fh->resource[RES_KM_FLOW_TYPE].count = 1;
        fh->resource[RES_KM_FLOW_TYPE].index = 1;

        /* Fixing flow_type to always use 1 */
        if (!flow_nic_is_resource_used(dev->ndev, RES_KM_FLOW_TYPE, 1)) {
          flow_nic_alloc_resource_index(dev->ndev, 1, RES_KM_FLOW_TYPE);
          fh->resource[RES_KM_FLOW_TYPE].referenced = 0;
        } else {
          err = flow_nic_ref_resource(dev->ndev, RES_KM_FLOW_TYPE, 1);
          fh->resource[RES_KM_FLOW_TYPE].referenced = 1;
        }
        if (err) {
          NT_LOG(DBG, FILTER, "ERROR: Could not reference Flow Type resource index 1\n");
          flow_nic_set_error(ERR_MATCH_RESOURCE_EXHAUSTION, error);
          goto err_exit;
        }

        fh->fd->km.flow_type = 1;
      } else {
        if (flow_nic_allocate_fh_resource(dev->ndev, RES_KM_FLOW_TYPE, fh, 1, 1)) {
          NT_LOG(ERR, FILTER, "ERROR: Could not get KM FLOW TYPE resource\n");
          flow_nic_set_error(ERR_MATCH_RESOURCE_EXHAUSTION, error);
          goto err_exit;
        }
        fh->fd->km.flow_type = fh->resource[RES_KM_FLOW_TYPE].index;
      }

      NT_LOG(DBG, FILTER, "Got new flow type %i (fd:%p)\n", fh->fd->km.flow_type, fh->fd);
      /*
       * Create KM Recipe
       */
      if (km_rcp_set(&fh->fd->km, fh->resource[RES_KM_CATEGORY].index) < 0) {
        flow_nic_set_error(ERR_MATCH_FAILED_TOO_COMPLEX, error);
        goto err_exit;
      }
    }
    /*
     * add CAM/TCAM flow entry - done later down the path, when color(actions) has been
     * ironed out!
     */

    /*
     * KCC CAM entry needs KM category
     */
    if (fh->fd->kcc) {
      /* key and id is already set, so only category needs to be set */
      NT_LOG(DBG, FILTER, "Add KM Category to KCC CAM entry. KM index : %i\n",
          fh->resource[RES_KM_CATEGORY].index);
      kcc_add_km_category(fd->kcc,
          (uint8_t) fh->resource[RES_KM_CATEGORY].index);
    }

  }


  /*
   * Output check preparation/conversion and allocate NIC resources
   */


  /*
   * Create output/destination setup
   */
  int num_queues = 0;
  /* collecting information for replication */
  int port_idx[MAX_REPLICATION_PORTS];
  int phy_idx[MAX_REPLICATION_PORTS];
  int num_vport_idx = 0;
  int num_phy_idx = 0;
  int num_dest_ports = 0;
  uint32_t last_dest_port = -1;

  memset(port_idx, 0, sizeof(port_idx));
  memset(phy_idx, 0, sizeof(phy_idx));

  /* count and prepare destination queues */
  for (int i = 0; i < fh->fd->dst_num_avail; i++) {
    if (fh->fd->dst_id[i].active) {

      if (fh->fd->dst_id[i].type == PORT_VIRT) {
        num_queues++;
      }

      if (last_dest_port != fh->fd->dst_id[i].owning_port_id) {

        last_dest_port = fh->fd->dst_id[i].owning_port_id;

        /* collect for replication - mirroring */
        if (fh->fd->dst_id[i].type == PORT_VIRT) {
          assert(num_vport_idx < MAX_REPLICATION_PORTS);
          port_idx[num_vport_idx++] = i;  /* first queue of port */
        } else
        if (fh->fd->dst_id[i].type == PORT_PHY) {
          assert(num_phy_idx < MAX_REPLICATION_PORTS);
          phy_idx[num_phy_idx++] = i;
        }
        num_dest_ports++;

      }

    }
  }


  int setup_ioa_dest = 0;


  /* if replication requested - mirroring */
  if (num_dest_ports > 1) {
    /*
     * Cannot handle RSS or more than 2 destinations while replicating - yet!
     */
    if ((num_queues + num_phy_idx) > 2) {
      /* ERROR */
      NT_LOG(ERR, FILTER, "ERROR: Too many destination queues for mirroring\n");
      flow_nic_set_error(ERR_ACTION_REPLICATION_FAILED, error);
      goto err_exit;
    }

    if (num_vport_idx > 1) {
      /* recirculate - bypass to second port queue */
      fh->fd->dst_id[port_idx[1]].active = 0;
      num_queues--;
    }
  }

  if (num_queues > 1) {
    struct flow_eth_dev *out_dev = NULL;
    if (num_dest_ports == 1) {
      out_dev = get_dev_from_port(dev->ndev, last_dest_port);
    }

    if (out_dev && out_dev->rss_target_id >= 0) {
      fh->resource[RES_QSL_RCP].index = out_dev->rss_target_id;
      fh->resource[RES_QSL_RCP].count = 1;
      fh->resource[RES_QSL_RCP].referenced = 1;
      flow_nic_ref_resource(dev->ndev, RES_QSL_RCP, out_dev->rss_target_id);
    } else {

      NT_LOG(ERR, FILTER, "ERROR: No RSS output queues found for destination port\n");
      flow_nic_set_error(ERR_ACTION_NO_OUTPUT_QUEUE_FOUND, error);
      goto err_exit;
    }

  } else {
    if (fh->fd->full_offload) {
      /*
       * DROP in QSL
       * if not multi queue (RSS)
       * use default drop/discard QSL index to bind to CAT
       * if any targets defined use discard otherwise drop
       */
      int drop_discard = (num_phy_idx || num_queues) ? dev->ndev->default_qsl_discard_index : dev->ndev->default_qsl_drop_index;

      /* TODO Temp override to Drop */
      drop_discard = dev->ndev->default_qsl_drop_index;

      fh->resource[RES_QSL_RCP].index = drop_discard;
      fh->resource[RES_QSL_RCP].count = 1;
      fh->resource[RES_QSL_RCP].referenced = 1;
      flow_nic_ref_resource(dev->ndev, RES_QSL_RCP, drop_discard);

      /* single queue target destination or none */
      setup_ioa_dest = num_queues; /* only if any defined */
    }
    /* else Partial Offload - may also be special high priority LACP flows */
  }


  /*
   * setup any IOA queue targets
   */
  int next_queue_idx = 0;
  if (setup_ioa_dest) {
    /* create ioa and put in the dest queue id */
    assert(num_vport_idx >= 0 && fh->fd->dst_id[port_idx[0]].id >= 0 && fh->fd->dst_id[port_idx[0]].id < 256);

    NT_LOG(DBG, FILTER, "IOA set queue %i\n", fh->fd->dst_id[port_idx[0]].id);
    fh->fd->ioa_actions = ioa_set_queue(fh->fd->ioa_actions, (uint8_t)fh->fd->dst_id[port_idx[0]].id);
    next_queue_idx = 1;

    /*
     * error checking - if push header is active, thus only 1 destination queue possible
     */
    if (get_roa_tunhdr_len(fh->fd->roa_actions) && num_vport_idx > 1) {
      NT_LOG(ERR, FILTER, "ERROR: PUSH Tunnel Header action in ROA needs recirculate-bypass to output to queues. Recirculate-bypass cannot output to multiple destination queues\n");
      flow_nic_set_error(ERR_ACTION_TUNNEL_HEADER_PUSH_OUTPUT_LIMIT, error);
      goto err_exit;
    }
  }

  /*
   * if replication/mirroring to another vport recirculate and bypass to vport
   */
  if (num_vport_idx > 1) {
    NT_LOG(DBG, FILTER, "ROA recirc bypass to queue %i\n", fh->fd->dst_id[port_idx[next_queue_idx]].id);
    fh->fd->roa_actions = set_roa_recirc_bypass(fh->fd->roa_actions, (uint8_t)fh->fd->dst_id[port_idx[next_queue_idx]].id);
  }

  /*
   * if phy port is destination target
   */
  if (num_phy_idx) {
    NT_LOG(DBG, FILTER, "ROA Tx to %i\n", fh->fd->dst_id[phy_idx[0]].id);
    fh->fd->roa_actions = set_roa_tx(fh->fd->roa_actions, (uint8_t)fh->fd->dst_id[phy_idx[0]].id);

    /*
     * if replication/mirroring to another phy port recirculate and bypass to phy
     */
    if (num_phy_idx > 1) {
      uint8_t port = (uint8_t)(ROA_RECIRC_BYPASS_PHY_OFFSET + (fh->fd->dst_id[phy_idx[1]].id));
      NT_LOG(DBG, FILTER, "ROA recirc bypass to phy %i\n", port);
      fh->fd->roa_actions = set_roa_recirc_bypass(fh->fd->roa_actions, port);
    }
  }


  if (fh->fd->full_offload)
  {
    fh->cao_enabled = 1;
  }


  /*
   * Handle Color and Action settings
   */

  if (fh->fd->ioa_actions) {
    if (flow_nic_allocate_fh_resource(dev->ndev, RES_IOA_RCP, fh, 1, 1) < 0) {
      NT_LOG(ERR, FILTER, "ERROR: Could not get IOA resource\n");
      flow_nic_set_error(ERR_ACTION_INLINE_MOD_RESOURCE_EXHAUSTION, error);
      goto err_exit;
    }
  }
  hw_mod_ioa_rcp_set(&dev->ndev->be, HW_IOA_RCP_PRESET_ALL, fh->resource[RES_IOA_RCP].index, 0);

  if (fh->fd->roa_actions) {
    /* Get the number of resources required (2 for IPv6, 1 for IPv4) */
    uint32_t count = (get_roa_tun_ip_type(fh->fd->roa_actions) == 1) ? 2 : 1;

    if (flow_nic_allocate_fh_resource(dev->ndev, RES_ROA_RCP, fh, count, count) < 0) {
      NT_LOG(ERR, FILTER, "ERROR: Could not get ROA resource\n");
      flow_nic_set_error(ERR_ACTION_RETRANSMIT_RESOURCE_EXHAUSTION, error);
      goto err_exit;
    }
    hw_mod_roa_tuncfg_set(&dev->ndev->be, HW_ROA_TUNCFG_PRESET_ALL, fh->resource[RES_ROA_RCP].index, 0);
  }


  /*
   * Get a flow stat id
   */
  if (fh->fd->mark != (uint32_t)-1) {
    if (fh->cao_enabled) {
      fh->flow_stat_id = create_flow_stat_id(dev->ndev, fh->fd->mark);
      if (fh->flow_stat_id == INVALID_FLOW_STAT_ID) {
        NT_LOG(ERR, FILTER, "ERROR: no more flow stat counters available\n");
        flow_nic_set_error(ERR_ACTION_FLOW_COUNTER_EXHAUSTION, error);
        goto err_exit;
      }
    } else {
      /* Partial Offload Flow */
      if (fh->fd->mark > FLOW_MARK_MAX && fd->special_match == SPCIAL_MATCH_NONE) {
        NT_LOG(DBG, FILTER, "ERROR: flow mark too big for hw-offload\n");
        flow_nic_set_error(ERR_PARTIAL_FLOW_MARK_TOO_BIG, error);
        goto err_exit;
      }
      fh->flow_stat_id = fh->fd->mark;
    }
    NT_LOG(DBG, FILTER, "Flow Mark received %08x, Internal Flow stat ID %i\n", fh->fd->mark, fh->flow_stat_id);
  } else {
    fh->flow_stat_id = INVALID_FLOW_STAT_ID;
  }

  /*
   * Format Color information into Mark and Actions
   */
  int ioa_rcp = (fh->resource[RES_IOA_RCP].count) ? fh->resource[RES_IOA_RCP].index : 0;
  int roa_rcp = (fh->resource[RES_ROA_RCP].count) ? fh->resource[RES_ROA_RCP].index : 0;


  if (fh->resource[RES_IOA_RCP].count ||
      fh->resource[RES_ROA_RCP].count) {

    if (fh->resource[RES_IOA_RCP].count) {
      flow_actions_create_ioa_config(&dev->ndev->be, fh->resource[RES_IOA_RCP].index, fh->fd->ioa_actions);
      /*
       * Check possible reuse of IOA recipes
       */
      int new_res = reuse_fh_action_resource(dev->ndev, RES_IOA_RCP, fh);
      if (new_res >= 0) {
        /* referenced instead */
        ioa_rcp = new_res;
      }
    }

    if (fh->resource[RES_ROA_RCP].count) {
      if (fh->fd->tun_hdr.len) {
        flow_actions_create_roa_tunhdr(&dev->ndev->be, fh->resource[RES_ROA_RCP].index, &fh->fd->tun_hdr);
      }
      flow_actions_create_roa_tuncfg(&dev->ndev->be, fh->resource[RES_ROA_RCP].index, fh->fd->roa_actions);
      /*
       * Check possible reuse of ROA recipes
       */
      int new_res = reuse_fh_action_resource(dev->ndev, RES_ROA_RCP, fh);
      if (new_res >= 0) {
        /* referenced instead */
        roa_rcp = new_res;
      }
    }

  }


  if (fh->cao_enabled) {
    /* CAO is set - format color into action enabled color format */
    int qsl_hsh = 0;

    if (fh->resource[RES_QSL_RCP].count) {
      /*
       * Even though we have setup a default QSL for CAT, then in QSL v6+ when
       * color action is activated it overrides the QSL selection, thus
       * we needs to select the default drop qsl here in color also
       */
      qsl_hsh = fh->resource[RES_QSL_RCP].index;
      NT_LOG(DBG, FILTER, "Set QSL_HSH index %i drop/discard index = %i/%i\n", qsl_hsh,
          dev->ndev->default_qsl_drop_index, dev->ndev->default_qsl_discard_index);
      /*
       * OTOH if index is not equal to default QSL Drop/Discard recipe then allocate and configure hasher too
       */
      if (fh->resource[RES_QSL_RCP].index != dev->ndev->default_qsl_drop_index &&
          fh->resource[RES_QSL_RCP].index != dev->ndev->default_qsl_discard_index &&
          fh->resource[RES_QSL_RCP].referenced == 0) {
        int err = flow_nic_allocate_fh_resource_index(dev->ndev, RES_HSH_RCP, qsl_hsh, fh);
        if (err) {
          NT_LOG(ERR, FILTER, "ERROR: hasher recipe index %i - seems allocated, but should not be!\n", qsl_hsh);
          flow_nic_set_error(ERR_FAILED, error);
          goto err_exit;
        }
        flow_nic_set_hasher(dev->ndev, fh->resource[RES_QSL_RCP].index, HASH_ALGO_5TUPLE);  /* resource is following QSL index */
      }
    }

    /*
     * Create final color value combined from all actions defined
     */
    fh->color = set_color_action(fh->flow_stat_id, ioa_rcp, roa_rcp, qsl_hsh);

  } else {
    /* only Mark information or none */
    fh->color = fh->flow_stat_id;
    NT_LOG(ERR, FILTER, "Partial Offload: no actions defined. Set mark in color: %08x\n", fh->color);
  }


  /*
   * If we need a Color information, either for Mark or actions
   */
  if (fh->fd->mark < ((uint32_t)-1) ||
      fh->resource[RES_ROA_RCP].count ||
      fh->resource[RES_IOA_RCP].count) {
      if (flow_nic_allocate_fh_resource(dev->ndev, RES_CAT_COT, fh, 1, 1) < 0) {
        NT_LOG(ERR, FILTER, "ERROR: Could not get CAT COT resource\n");
        flow_nic_set_error(ERR_ACTION_INTERNAL_RESOURCE_EXHAUSTION, error);
        goto err_exit;
      }
  }

  /*
   * Prepare and create CAT CFN root flow match function
   */

  /* Allocate new CAT CFN */
  if (flow_nic_allocate_fh_resource(dev->ndev, RES_CAT_CFN, fh, 1, 1)) {
    NT_LOG(ERR, FILTER, "ERROR: Could not get CAT CFN resource\n");
    flow_nic_set_error(ERR_MATCH_RESOURCE_EXHAUSTION, error);
    goto err_exit;
  }

  hw_mod_cat_cfn_set(&dev->ndev->be, HW_CAT_CFN_SET_ALL_DEFAULTS, fh->resource[RES_CAT_CFN].index, 0, 0);

  /* Enable KM */
  uint32_t val = 0;
  int cfn_addr8 = fh->resource[RES_CAT_CFN].index / 8;
  int cfn_bit8 = fh->resource[RES_CAT_CFN].index % 8;

#if 0
/*  int */
/*  uint8_t kcc_key_get_cat_cfn(struct kcc_flow_def_s *kcc) */
#endif
  if (fh->fd->km.flow_type) {

    if (!fh->fd->kcc) {
      /* else enable KM in CAT KCE/KCS */
      hw_mod_cat_kce_km_get(&dev->ndev->be, HW_CAT_KCE_ENABLE_BM, KM_FLM_IF_FIRST, cfn_addr8, &val);
      val |= (1 << cfn_bit8);
      hw_mod_cat_kce_km_set(&dev->ndev->be, HW_CAT_KCE_ENABLE_BM, KM_FLM_IF_FIRST, cfn_addr8, val);

      hw_mod_cat_kcs_km_set(&dev->ndev->be, HW_CAT_KCS_CATEGORY, KM_FLM_IF_FIRST,
          fh->resource[RES_CAT_CFN].index, fh->resource[RES_KM_CATEGORY].index);
    }

    /* Setup flow type - flow_type is zero if no KM flow types allocated */
    int index = (8 * fh->fd->km.flow_type + cfn_addr8) * 2;

    hw_mod_cat_fte_km_get(&dev->ndev->be, HW_CAT_FTE_ENABLE_BM, KM_FLM_IF_FIRST, index, &val);
    val |= (1 << cfn_bit8);
    hw_mod_cat_fte_km_set(&dev->ndev->be, HW_CAT_FTE_ENABLE_BM, KM_FLM_IF_FIRST, index,  val);  /* lookup A */

    NT_LOG(DBG, FILTER, "SET FTE VAL %02x  addr %i\n", val, index);
  }

  /* Do we need CAT COL - Action or Mark if partial - even when KCC enabled - because it needs to inform about KM color importance! */
  if (fh->resource[RES_CAT_COT].count) {
    fh->cte |= (1 << CT_COL);

    if (fh->fd->km.flow_type) {
      /* Color is carried by KM entry */
      hw_mod_cat_cot_set(&dev->ndev->be, HW_CAT_COT_COLOR, fh->resource[RES_CAT_COT].index, 0);
      /* lookup A and B has each a bit enable - not used yet - only A is used */
      hw_mod_cat_cot_set(&dev->ndev->be, HW_CAT_COT_KM, fh->resource[RES_CAT_COT].index, 1);
    } else {
      hw_mod_cat_cot_set(&dev->ndev->be, HW_CAT_COT_COLOR, fh->resource[RES_CAT_COT].index, fh->color);
      hw_mod_cat_cot_set(&dev->ndev->be, HW_CAT_COT_KM, fh->resource[RES_CAT_COT].index,  0);
    }

    hw_mod_cat_cts_set(&dev->ndev->be, cts_offset[CT_COL].cts_field_id,
        CTS_CFN_OFFSET(dev->ndev->be.cat.cts_num,
            fh->resource[RES_CAT_CFN].index)
            + cts_offset[CT_COL].offset, fh->resource[RES_CAT_COT].index);

    /*
     * QSL destination if setup - ASL and HSH follows closely (same index)
     * QSL/HSH is selected in Color if CAO enabled
     */
    if (!fh->cao_enabled) {
      if (fh->resource[RES_QSL_RCP].count) {
        fh->cte |= (1 << CT_QSL);
        fh->cte |= (1 << CT_HSH);

        hw_mod_cat_cts_set(&dev->ndev->be, cts_offset[CT_QSL].cts_field_id,
            CTS_CFN_OFFSET(dev->ndev->be.cat.cts_num,
                fh->resource[RES_CAT_CFN].index)
                + cts_offset[CT_QSL].offset, fh->resource[RES_QSL_RCP].index);
        hw_mod_cat_cts_set(&dev->ndev->be, cts_offset[CT_HSH].cts_field_id,
            CTS_CFN_OFFSET(dev->ndev->be.cat.cts_num,
                fh->resource[RES_CAT_CFN].index)
                + cts_offset[CT_HSH].offset, fh->resource[RES_QSL_RCP].index);
      }
    }

  }



  /*
   * Others not used yet
   * CT_COR: not used in virt
   * CT_HSH: to be used - is set up for 5 tuple hashing on recipe 0
   * CT_IPF: not used in virt
   * CT_SLC: setup to slice FCS always in recipe 0
   * CT_PDB: setup to DVIO0 descriptor
   * CT_MSK: not used in virt
   * CT_HST: not used in virt
   * CT_EPP: not used in virt by binary flow code
   */

  if (fh->cte) {
    hw_mod_cat_cte_set(&dev->ndev->be, HW_CAT_CTE_ENABLE_BM, fh->resource[RES_CAT_CFN].index, fh->cte);
  }

  /* Frame decoder element match */
  hw_mod_cat_cfn_set(&dev->ndev->be, HW_CAT_CFN_PTC_L2,
      fh->resource[RES_CAT_CFN].index, 0, fh->fd->l2_prot);

  hw_mod_cat_cfn_set(&dev->ndev->be, HW_CAT_CFN_PTC_L3,
      fh->resource[RES_CAT_CFN].index, 0, fh->fd->l3_prot);

  hw_mod_cat_cfn_set(&dev->ndev->be, HW_CAT_CFN_PTC_L4,
      fh->resource[RES_CAT_CFN].index, 0, fh->fd->l4_prot);

  hw_mod_cat_cfn_set(&dev->ndev->be, HW_CAT_CFN_PTC_VLAN,
      fh->resource[RES_CAT_CFN].index, 0, (1U << fh->fd->vlans) & 0xf);

  hw_mod_cat_cfn_set(&dev->ndev->be, HW_CAT_CFN_PTC_TUNNEL,
      fh->resource[RES_CAT_CFN].index, 0, fh->fd->tunnel_prot);

  if (fh->fd->kcc) {
    NT_LOG(DBG, FILTER, "KCC active CAT RCP port set to 0xff\n");
    hw_mod_cat_cfn_set(&dev->ndev->be, HW_CAT_CFN_MAC_PORT,
        fh->resource[RES_CAT_CFN].index, 0, 0xff);
#if 0
    NT_LOG(DBG, FILTER, "KCC active CAT RCP port set to 0xff - AVOIDED - NEED PORT ID IN CAT TO SEPARATE DIFFERENT SRC_PORT\n");
    hw_mod_cat_cfn_set(&dev->ndev->be, HW_CAT_CFN_MAC_PORT, fh->resource[RES_CAT_CFN].index, 0, fh->port_id);
#endif
  } else {
    hw_mod_cat_cfn_set(&dev->ndev->be, HW_CAT_CFN_MAC_PORT,
        fh->resource[RES_CAT_CFN].index, 0, fh->port_id);
  }

  /* Enable CAT function */
  hw_mod_cat_cfn_set(&dev->ndev->be, HW_CAT_CFN_ENABLE,
      fh->resource[RES_CAT_CFN].index, 0, 1);

  /*
   * OK - finished converting Flow into NIC cached resource configuration
   */


  /*
   * check possible reuse of existing Match Function
   * reuse and free current and reference resources if possible
   *
   * Note: the KCC Key is already unique caused by a new CAT CFN taken
   * Therefore we only need to consider KCC Key clashes when reuse of CAT CFN
   * is found possible
   */
  int reuse = 0;
  if (fh->resource[RES_CAT_CFN].count) {
    /*
     * Structure to collect info for referencing later, if match found
     */
    struct {
      enum res_type_e res_type;
      int idx;
      int equ_idx;
      int count;
    } equal_cats[RES_COUNT];
    int num_equal_cats = 0;

    uint32_t val = 0;
    uint32_t val1 = 0;
    uint32_t equal_idx = 0;

    do {
      int res = hw_mod_cat_cfn_get(&dev->ndev->be, HW_CAT_CFN_FIND, fh->resource[RES_CAT_CFN].index, equal_idx, &equal_idx);

      NT_LOG(DBG, FILTER, "Search result for CAT CFN sharing: res %i, equal %08x\n", res, equal_idx);
      if (res == 0 && equal_idx != NOT_FOUND) {
        /* First step : equal CAT CFN */
        uint32_t cfn_addr = fh->resource[RES_CAT_CFN].index / 8;
        uint32_t cfn_bit = (1 << (fh->resource[RES_CAT_CFN].index % 8));
        uint32_t equ_addr = equal_idx / 8;
        uint32_t equ_bit = (1 << (equal_idx % 8));

        int pass = 0;

        /* KCC takes over KM selection */
        if (dev->ndev->be.cat.kcc_size) {
          pass = 1;
        } else {
          /* check if KM enabled/disabled same in both functions */
          hw_mod_cat_kce_km_get(&dev->ndev->be, HW_CAT_KCE_ENABLE_BM, KM_FLM_IF_FIRST, cfn_addr,
              &val);
          hw_mod_cat_kce_km_get(&dev->ndev->be, HW_CAT_KCE_ENABLE_BM, KM_FLM_IF_FIRST, equ_addr,
              &val1);
          pass = !!(val & cfn_bit) == !!(val1 & equ_bit);
        }
#if 0
        printf("CHECK KCE: val %02x, cfn_bit %02x, val1 %02x, equ_bit %02x\n", val, cfn_bit, val1, equ_bit);
#endif
        if (pass) {

          /* KM enabled on both */
          if (val1 & equ_bit) {
            /* if KCC, no KM category select in CAT */
            if (dev->ndev->be.cat.kcc_size) {
              val1 = val;
            } else {
              /* check same category (must be recognized in KM to get same category function for reuse done earlier) */
              res = hw_mod_cat_kcs_km_get(&dev->ndev->be, HW_CAT_KCS_CATEGORY, KM_FLM_IF_FIRST,
                  fh->resource[RES_CAT_CFN].index, &val);
              res |= hw_mod_cat_kcs_km_get(&dev->ndev->be, HW_CAT_KCS_CATEGORY, KM_FLM_IF_FIRST,
                  equal_idx, &val1);
            }

            /* Expecting KM to be reused earlier, thus same category of KM */

            if (!res && val1 == val) {
              /* Same KM category. Then also check all enabled Flow Types */
              reuse = 1; /* break when fail */

              for (unsigned int ft = 0; ft < dev->ndev->be.cat.nb_flow_types; ft++) {
                int cfn_ft_idx = (8 * ft + cfn_addr) * 2;
                int equ_ft_idx = (8 * ft + equ_addr) * 2;
                hw_mod_cat_fte_km_get(&dev->ndev->be, HW_CAT_FTE_ENABLE_BM, KM_FLM_IF_FIRST, cfn_ft_idx, &val);   /* lookup A */
                hw_mod_cat_fte_km_get(&dev->ndev->be, HW_CAT_FTE_ENABLE_BM, KM_FLM_IF_FIRST, equ_ft_idx, &val1);  /* lookup A */

#if 0
                printf("ft %02i: cfn %i, cfn_ft_idx %03i, cfn_bit %02x, val  %02x\n", ft, fh->resource[RES_CAT_CFN].index,
                    cfn_ft_idx, cfn_bit, val);
                printf("       equ %i, equ_ft_idx %03i, equ_bit %02x, val1 %02x\n", equal_idx,
                    equ_ft_idx, equ_bit, val1);
#endif
                if (!!(val & cfn_bit) != !!(val1 & equ_bit)) {
                  reuse = 0;
                  NT_LOG(DBG, FILTER, "Lookup A flow type %i failed idx %i  val %i, cfn_bit %02x, val1 %i, equ_bit %02x\n", ft, equal_idx,
                      val, cfn_bit, val1, equ_bit);
                  break;
                }
                hw_mod_cat_fte_km_get(&dev->ndev->be, HW_CAT_FTE_ENABLE_BM, KM_FLM_IF_FIRST, cfn_ft_idx + 1, &val);   /* lookup B */
                hw_mod_cat_fte_km_get(&dev->ndev->be, HW_CAT_FTE_ENABLE_BM, KM_FLM_IF_FIRST, equ_ft_idx + 1, &val1);  /* lookup B */

                if (!!(val & cfn_bit) != !!(val1 & equ_bit)) {
                  reuse = 0;
                  NT_LOG(DBG, FILTER, "Lookup B flow type failed idx %i  val %i, cfn_bit %02x, val1 %i, equ_bit %02x\n", equal_idx,
                      val, cfn_bit, val1, equ_bit);
                  break;
                }
              }
            } else {
              NT_LOG(DBG, FILTER, "Found not equal HW_CAT_KCS_CATEGORY setup %i:%i, %i:%i\n", fh->resource[RES_CAT_CFN].index, val, equal_idx, val1);
            }

          } else {
            /* KM not used */
            reuse = 1;
          }


          if (reuse) {
            /* same KM Category and KM flow types used - or no KM used */

            /* check for CTE identical (Category Enable Bitmap) */
            hw_mod_cat_cte_get(&dev->ndev->be, HW_CAT_CTE_ENABLE_BM, fh->resource[RES_CAT_CFN].index, &val);
            hw_mod_cat_cte_get(&dev->ndev->be, HW_CAT_CTE_ENABLE_BM, equal_idx, &val1);
            if (val == val1) {

              /* for later referencing */
              equal_cats[num_equal_cats].res_type = RES_CAT_CFN;
              equal_cats[num_equal_cats].idx      = fh->resource[RES_CAT_CFN].index;
              equal_cats[num_equal_cats].equ_idx  = equal_idx;
              equal_cats[num_equal_cats].count    = fh->resource[RES_CAT_CFN].count;
              num_equal_cats++;

              /* Check that each category configuration in both functions are identical */
              for (unsigned int cte_type = CT_COL; (cte_type < CT_CNT) && reuse; cte_type++) {
              /*
               * COL, COR, HSH, QSL, (LTX), IPF, SLC, PDB (MSK, HST)
               */
                if (val & (1 << cte_type)) {
                  /* if set, check identical configuration */
                  uint32_t category_idx, equ_category_idx;

                  assert(cts_offset[cte_type].offset >= 0);  /* LTX must not be selected here */
                  hw_mod_cat_cts_get(&dev->ndev->be, cts_offset[cte_type].cts_field_id, CTS_CFN_OFFSET(dev->ndev->be.cat.cts_num, fh->resource[RES_CAT_CFN].index) + cts_offset[cte_type].offset,
                      &category_idx);
                  hw_mod_cat_cts_get(&dev->ndev->be, cts_offset[cte_type].cts_field_id, CTS_CFN_OFFSET(dev->ndev->be.cat.cts_num, equal_idx) + cts_offset[cte_type].offset,
                      &equ_category_idx);

                  if (category_idx != equ_category_idx) {

                    uint32_t val = equ_category_idx;
                    int res = 0;

                    switch (cte_type) {
                    case CT_COL:
                      equal_cats[num_equal_cats].res_type = RES_CAT_COT;
                      res = hw_mod_cat_cot_get(&dev->ndev->be, HW_CAT_COT_COMPARE, category_idx, &val);
                      if (!res) {
                        NT_LOG(DBG, FILTER, "CTE mismatch HW_CAT_COT_COMPARE res %i, compare val %i\n", res, val);
                      }
                      break;
                    case CT_HSH:
                        /* ignore in QSL v6+ interpret as being free resource */
                        res = 2; /* ignore */
                        break;
                    case CT_QSL:
                      /* ignore in QSL v6+ interpret as being free resource - this includes QST */
                      res = 2; /* ignore */
                      break;
                    case CT_SLC:
                      equal_cats[num_equal_cats].res_type = RES_SLC_RCP;
                      res = hw_mod_slc_rcp_get(&dev->ndev->be, HW_SLC_RCP_COMPARE, category_idx, &val);
                      if (!res) {
                        NT_LOG(DBG, FILTER, "CTE minmatch HW_SLC_RCP_COMPARE res %i, compare val %i\n", res, val);
                      }
                      break;
                    case CT_PDB:
                      equal_cats[num_equal_cats].res_type = RES_PDB_RCP;
                      res = hw_mod_pdb_rcp_get(&dev->ndev->be, HW_PDB_RCP_COMPARE, category_idx, &val);
                      if (!res) {
                        NT_LOG(DBG, FILTER, "CTE msmatch HW_PDB_RCP_COMPARE res %i, compare val %i\n", res, val);
                      }
                      break;

                    /* error or not supported */
                    case CT_COR:
                    case CT_IPF:
                    default:
                      NT_LOG(DBG, FILTER, "Unsupported Category Type found in CFN!\n");
                      res = -1;
                      break;
                    }

                    if (res <= 0) {
                      NT_LOG(DBG, FILTER, "Error or different Category Types found in CFNs\n");
                      reuse = 0;
                    } else {
                      if (res == 1) {
                        /* collect info for reuse later */
                        equal_cats[num_equal_cats].idx     = category_idx;
                        equal_cats[num_equal_cats].equ_idx = equ_category_idx;
                        equal_cats[num_equal_cats].count   = fh->resource[equal_cats[num_equal_cats].res_type].count;
                        num_equal_cats++;
                      } /* else ignore */
                    }
                  } /* if pointing to same category, let it be, no ref changes needed but pass */

                }
              }

            } else {
              /* Not same categories enabled in both bitmaps */
              NT_LOG(DBG, FILTER, "Not same Categories found enabled in CAT CFN %i:%i, %i:%i\n", fh->resource[RES_CAT_CFN].index, val, equal_idx, val1);
              reuse = 0;
            }
          }


          if (reuse) {
            /*
             * Yes, we got an identical CAT CFN function and all its used referenced resources
             * reuse the already created, identical function and sub-resources
             */

            int reuse_error = 0;

            if (fh->fd->kcc && kcc_key_get_cat_cfn(fh->fd->kcc) == 0xff) {
              /* Reuse of CAT FN found, but no match found in pre-match for KCC Key - error */
              NT_LOG(DBG, FILTER, "ERROR: CAT FN reuse found, but no match found in PRE-MATCH of CAT FN!!\n");

              flow_nic_set_error(ERR_INTERNAL_CAT_FUNC_REUSE_FAILED, error);
              goto err_exit;
            }


            for (int i = 0; i < num_equal_cats && !reuse_error; i++) {
              /*
               * Reuse resource
               */
              for (int idx_offs = 0;
                  idx_offs < equal_cats[i].count && !reuse_error; idx_offs++) {
                /* reference equal resource index */
                reuse_error = flow_nic_ref_resource(dev->ndev, equal_cats[i].res_type, equal_cats[i].equ_idx + idx_offs);
                if (!reuse_error) {
                  /* free(deref) the taken resource */
                  flow_nic_deref_resource(dev->ndev, equal_cats[i].res_type, equal_cats[i].idx + idx_offs);
                }
              }
              if (reuse_error) break;

              /* reset settings in cache */
              switch (equal_cats[i].res_type) {
              case RES_CAT_CFN:

                /* If KCC, check that we found same CAT FN for reuse in PRE-MATCH and now */
                if (fh->fd->kcc &&
                    equal_cats[i].equ_idx != kcc_key_get_cat_cfn(fh->fd->kcc)) {
                  reuse_error = 1;
                  break;
                }

#ifdef FLOW_DEBUG
                dev->ndev->be.iface->set_debug_mode(dev->ndev->be.be_dev, FLOW_BACKEND_DEBUG_MODE_WRITE);
#endif
                reuse_error = reset_cat_function_setup(dev, fh->resource[RES_CAT_CFN].index, fh->resource[RES_KM_CATEGORY].count, 0);
#ifdef FLOW_DEBUG
                dev->ndev->be.iface->set_debug_mode(dev->ndev->be.be_dev, FLOW_BACKEND_DEBUG_MODE_NONE);
#endif
                break;
              case RES_CAT_COT: /* CT_COL */
                hw_mod_cat_cot_set(&dev->ndev->be, HW_CAT_COT_PRESET_ALL, equal_cats[i].idx, 0);
                break;
              case RES_HSH_RCP: /* CT_HSH */
                hw_mod_hsh_rcp_set(&dev->ndev->be, HW_HSH_RCP_PRESET_ALL, equal_cats[i].idx, 0, 0);
                break;
              case RES_QSL_QST: {
                uint32_t queue_id;
                for (int ii = 0; ii < equal_cats[i].count; ii++) {
                    hw_mod_qsl_qst_get(&dev->ndev->be, HW_QSL_QST_QUEUE, equal_cats[i].idx + ii, &queue_id); /* get queue_id */
                    hw_mod_qsl_qst_set(&dev->ndev->be, HW_QSL_QST_EN, equal_cats[i].idx + ii, 0);            /* disable */
                }
              }
                break;
              case RES_QSL_RCP: /* CT_QSL */
                hw_mod_qsl_rcp_set(&dev->ndev->be, HW_QSL_RCP_PRESET_ALL, equal_cats[i].idx, 0);
                break;
              case RES_SLC_RCP: /* CT_SLC */
                hw_mod_slc_rcp_set(&dev->ndev->be, HW_SLC_RCP_PRESET_ALL, equal_cats[i].idx, 0);
                break;
              case RES_PDB_RCP: /* CT_PDB */
                hw_mod_pdb_rcp_set(&dev->ndev->be, HW_PDB_RCP_PRESET_ALL, equal_cats[i].idx, 0);
                break;
              default: /* CT_IPF, CT_COR - not supported */
                /* ERROR */
                break;
              }

              /* change index and create reference */
              if (!reuse_error) {
                fh->resource[equal_cats[i].res_type].index = equal_cats[i].equ_idx;
                fh->resource[equal_cats[i].res_type].referenced = 1;
              }

            }

            if (reuse_error) {
              NT_LOG(ERR, FILTER, "ERROR: Internal CAT CFN reuse failed\n");
              flow_nic_set_error(ERR_INTERNAL_CAT_FUNC_REUSE_FAILED, error);
              goto err_exit;
            }

            /* already setup, differences are setup in KM */

            NT_LOG(DBG, FILTER, "YES, CAT CFN can be reused, with all references\n");
            break; /* Found a CFN for reuse :) */
          }

          if (dev->ndev->flow_profile == FLOW_ETH_DEV_PROFILE_VSWITCH) {
            /*
             * VSWITCH
             * Match in CAT match criteria, but not in output - entropy clash
             * means need to distinguish in CAT CFN before different output selections including KM recipe
             */
            flow_nic_set_error(ERR_MATCH_ENTROPHY_FAILED, error);
            goto err_exit;
          }

          /* otherwise try next CFN match */

        } else {
          NT_LOG(DBG, FILTER, "CAT Equal failed on KM Enable - HW_CAT_KCE_ENABLE_BM - CLASH - drop offload\n");

          if (dev->ndev->flow_profile == FLOW_ETH_DEV_PROFILE_VSWITCH) {
            /*
             * VSWITCH
             * Match in CAT match criteria, but NOT same KM recipe selected
             * Prevent an entropy clash. it means that same CAT CFN input match
             * but used for different KM recipes (Key Definitions)
             */
            flow_nic_set_error(ERR_MATCH_ENTROPHY_FAILED, error);
            goto err_exit;
          }

        }
      } else {
        /* CAT don't match - Not found or error - break out */
        NT_LOG(DBG, FILTER, "CAT Equal failed on CFN - no more to match\n");
        break;
      }

      equal_idx++; /* step over this CFN match, maybe more exists... */
    } while (equal_idx < dev->ndev->be.cat.nb_cat_funcs);


    if (fh->fd->kcc && !reuse) {
     /* if KCC and the pre-match found CAT FN, but this exact match failed, discard flow */
     if (kcc_key_get_cat_cfn(fh->fd->kcc) != 0xff) {
       NT_LOG(ERR, FILTER, "ERROR: Internal CAT CFN reuse failed - PRE-MATCH of CAT FN found, but exact match failed\n");
       flow_nic_set_error(ERR_INTERNAL_CAT_FUNC_REUSE_FAILED, error);
       goto err_exit;

     } else {
       /* last thing, set new created CAT FN index in KCC Key */
       NT_LOG(DBG, FILTER, "Add CAT CFN to KCC key. CAT CFN idx %i\n",
                 fh->resource[RES_CAT_CFN].index);
       kcc_key_add_cat_cfn(fh->fd->kcc, (uint8_t)fh->resource[RES_CAT_CFN].index);
     }
    }

  }

  /*
   * OK flush everything to NIC HW
   */
#ifdef FLOW_DEBUG
  dev->ndev->be.iface->set_debug_mode(dev->ndev->be.be_dev, FLOW_BACKEND_DEBUG_MODE_WRITE);
#endif

  int err = flush_flow_handle(dev, fh, error);
  if (err) goto err_exit;

#ifdef FLOW_DEBUG
  dev->ndev->be.iface->set_debug_mode(dev->ndev->be.be_dev, FLOW_BACKEND_DEBUG_MODE_NONE);
#endif

  /* link flow def into created flow list on NIC device */
  nic_insert_flow(dev->ndev, fh);

  pthread_mutex_unlock(&dev->ndev->mtx);

#ifdef FLOW_DEBUG
  {
    NT_LOG(DBG, FILTER, "**********************************************************************\n");
    char *s = ntlog_helper_str_alloc("");
    ntlog_helper_str_add(s, ">>>>> [%p] Flow %p, Nic %i, Port %i: FLOW Created Successfully : Mark", dev, fh, dev->ndev->adapter_no, dev->port);
    if (fh->fd->mark != (uint32_t)-1) {
      ntlog_helper_str_add(s, " %i <<<<<\n", fh->fd->mark);
    } else {
      ntlog_helper_str_add(s, " (none) <<<<<\n");
    }
    NT_LOG(DBG, FILTER, "%s", s);
    ntlog_helper_str_free(s);
    NT_LOG(DBG, FILTER, "**********************************************************************\n");
  }
#endif

  return fh;

err_exit:

#ifdef FLOW_DEBUG
  NT_LOG(DBG, FILTER, "**********************************************************************\n");
  char *s = ntlog_helper_str_alloc("");

  ntlog_helper_str_add(s, ">>>>> [%p] Nic %i, Port %i: FLOW Created FAILED : Mark", dev, dev->ndev->adapter_no, dev->port);
  if (fh &&
      fh->fd &&
      fh->fd->mark != (uint32_t)-1) {
    ntlog_helper_str_add(s, " %i <<<<<\n", fh->fd->mark);
  } else {
    ntlog_helper_str_add(s, " (none) <<<<<\n");
  }
  NT_LOG(DBG, FILTER, "%s", s);
  NT_LOG(DBG, FILTER, "**********************************************************************\n");
  ntlog_helper_str_free(s);
#else
  NT_LOG(DBG, FILTER, "NOT OFFLOADED (not an error): %s\n", __func__);
#endif

  if (fh) {
    flow_destroy_locked_profile_vswitch(dev, fh, NULL);
  }

  pthread_mutex_unlock(&dev->ndev->mtx);
  return NULL;
}



int flow_destroy_locked_profile_vswitch(
    struct flow_eth_dev *dev,
    struct flow_handle *fh,
    UNUSED struct flow_error *error)
{
  int err = 0;
  NT_LOG(DBG, FILTER, "removing flow :%p\n", fh);

  flow_nic_set_error(ERR_SUCCESS, error);
#if 0
  /* TODO - consider preventing eth_dev interfaces on same nic_dev interface can delete flows. ? */
#endif

  /* take flow out of ndev list - may not have been put there yet */
  nic_remove_flow(dev->ndev, fh);

#ifdef FLOW_DEBUG
  dev->ndev->be.iface->set_debug_mode(dev->ndev->be.be_dev, FLOW_BACKEND_DEBUG_MODE_WRITE);
#endif

  /* Clear first KM-CAM/TCAM entry to stop matching */
  if (fh->fd) {
    km_clear_data_match_entry(&fh->fd->km);
  }

  for (int res_type = 0; res_type < RES_COUNT; res_type++) {
    if (fh->resource[res_type].count > 0) {

      for (int ii = 0; ii < fh->resource[res_type].count; ii++) {
        /* If last ref count of this resource, free it */
        if (flow_nic_deref_resource(dev->ndev, res_type, fh->resource[res_type].index + ii) == 0) {
          /* Free resource up in NIC */
          switch (res_type) {
          case RES_CAT_CFN:
          {
            /* Disable CAT function in NIC */
            assert(ii == 0);
            err |= reset_cat_function_setup(dev, fh->resource[RES_CAT_CFN].index, fh->resource[RES_KM_CATEGORY].count, fh->flushed);
          }
          break;
          case RES_QSL_QST:
          {
            /* Disable QSL output Queue */
            uint32_t queue_id;
            hw_mod_qsl_qst_get(&dev->ndev->be, HW_QSL_QST_QUEUE, fh->resource[RES_QSL_QST].index + ii, &queue_id); /* get queue_id */
            hw_mod_qsl_qst_set(&dev->ndev->be, HW_QSL_QST_EN,    fh->resource[RES_QSL_QST].index + ii, 0);         /* disable */

            if (fh->flushed) {
              hw_mod_qsl_qst_flush(&dev->ndev->be, fh->resource[RES_QSL_QST].index + ii, 1);
            }
          }
          break;
          case RES_QSL_RCP:
            hw_mod_qsl_rcp_set(&dev->ndev->be, HW_QSL_RCP_PRESET_ALL, fh->resource[RES_QSL_RCP].index + ii, 0);
            if (fh->flushed) {
              hw_mod_qsl_rcp_flush(&dev->ndev->be, fh->resource[RES_QSL_RCP].index + ii, 1);
            }
            break;
          case RES_CAT_COT:
            hw_mod_cat_cot_set(&dev->ndev->be, HW_CAT_COT_PRESET_ALL, fh->resource[res_type].index + ii, 0);
            if (fh->flushed) {
              hw_mod_cat_cot_flush(&dev->ndev->be, fh->resource[res_type].index + ii, 1);
            }
            break;
          case RES_KM_CATEGORY:
            assert(ii == 0);
            hw_mod_km_rcp_set(&dev->ndev->be, HW_KM_RCP_PRESET_ALL, fh->resource[res_type].index + ii, 0, 0);
            if (fh->flushed) {
              hw_mod_km_rcp_flush(&dev->ndev->be, fh->resource[res_type].index + ii, 1);
            }
            break;
          case RES_KM_FLOW_TYPE:
            /* Nothing needed */
            break;
          case RES_HSH_RCP: /* CT_HSH */
            hw_mod_hsh_rcp_set(&dev->ndev->be, HW_HSH_RCP_PRESET_ALL, fh->resource[res_type].index + ii, 0, 0);
            if (fh->flushed) {
              hw_mod_hsh_rcp_flush(&dev->ndev->be, fh->resource[res_type].index + ii, 1);
            }
            break;
          case RES_SLC_RCP: /* CT_SLC */
            hw_mod_slc_rcp_set(&dev->ndev->be, HW_SLC_RCP_PRESET_ALL, fh->resource[res_type].index + ii, 0);
            if (fh->flushed) {
              hw_mod_slc_rcp_flush(&dev->ndev->be, fh->resource[res_type].index + ii, 1);
            }
            break;
          case RES_PDB_RCP: /* CT_PDB */
            hw_mod_pdb_rcp_set(&dev->ndev->be, HW_PDB_RCP_PRESET_ALL,fh->resource[res_type].index + ii, 0);
            if (fh->flushed) {
              hw_mod_pdb_rcp_flush(&dev->ndev->be, fh->resource[res_type].index + ii, 1);
            }
            break;

          /* Actions modules */
          case RES_IOA_RCP:
            hw_mod_ioa_rcp_set(&dev->ndev->be, HW_IOA_RCP_PRESET_ALL, fh->resource[res_type].index + ii, 0);
            if (fh->flushed) {
              hw_mod_ioa_rcp_flush(&dev->ndev->be, fh->resource[res_type].index + ii, 1);
            }
            break;
          case RES_ROA_RCP:
            hw_mod_roa_tuncfg_set(&dev->ndev->be, HW_ROA_TUNCFG_PRESET_ALL, fh->resource[res_type].index + ii, 0);

            /* this IOA memory follows ROA */
            hw_mod_ioa_roa_epp_set(&dev->ndev->be, HW_IOA_ROA_EPP_PRESET_ALL, fh->resource[res_type].index + ii, 0);

            assert(fh->fd);
            uint8_t virt_port = (uint8_t)roa_get_recirc_port(fh->fd->roa_actions);
            int virt_r = is_virtual_port(virt_port);

            if (fh->flushed) {
              hw_mod_roa_tuncfg_flush(&dev->ndev->be, fh->resource[res_type].index + ii, 1);
              if (virt_r) {
                hw_mod_qsl_unmq_set(&dev->ndev->be, HW_QSL_UNMQ_DEST_QUEUE, virt_port, 0);
                hw_mod_qsl_unmq_set(&dev->ndev->be, HW_QSL_UNMQ_EN, virt_port, 0);
                hw_mod_qsl_unmq_flush(&dev->ndev->be, virt_port, 1);
              }

              /* this IOA memory follows ROA */
              hw_mod_ioa_roa_epp_flush(&dev->ndev->be, fh->resource[res_type].index + ii, 1);
            }

            break;
          default:
            err |= 1;
            break;
          }
#if 0
RES_CAT_EXO,
RES_CAT_LEN,
#endif

        } else {

          /* This resource is still in use by another flow */
          if (fh->fd && res_type == RES_KM_CATEGORY &&   /* is this a KM recipe */
              fh->resource[res_type].referenced == 0) { /* if then, does it hold the original definition? */
              /*
               * A special case where KM-category/flow-type is reused and this is original
               * We would need to take over the KM-definition to a still living flow, referencing this KM def
               * Note: original not-ref'ed instance is used to match against on new flows. need to keep it on a not ref'ed instance
               */
            NT_LOG(DBG, FILTER,
                "removing flow with original KM definition %p, but still referenced - KM idx %i\n",
                fh, fh->resource[res_type].index);

            struct flow_handle *flow = dev->ndev->flow_base;
            while (flow) {
              if (fh->fd != flow->fd) { /* jumping over itself */
                if (flow->resource[RES_KM_CATEGORY].referenced &&
                    flow->resource[RES_KM_CATEGORY].index == fh->resource[res_type].index)
                  {
                    NT_LOG(DBG, FILTER, "taking over definition of KM idx %i by flow %p\n", fh->resource[res_type].index, flow);
                    flow->resource[RES_KM_CATEGORY].referenced = 0; /* make new original/definition holder */
                    flow->resource[RES_KM_FLOW_TYPE].referenced = 0;
                    break;
                  }
              }
              flow = flow->next;
            }
          }

        } /* end else referenced still */

      }  /* end for loop count - typical only 1 */
    }  /* if resource in use */
  }


  if (fh) {
    if (fh->flow_stat_id < MAX_COLOR_FLOW_STATS) {
      delete_flow_stat_id(dev->ndev, fh->flow_stat_id);
    }

    if (fh->fd)
      delete_flow_def(fh->fd);

    if (!err)
      NT_LOG(DBG, FILTER, "Successfully removed flow :%p\n\n", fh);
    else {
      NT_LOG(DBG, FILTER, "FAILED removing flow :%p\n", fh);
      flow_nic_set_error(ERR_REMOVE_FLOW_FAILED, error);
    }

    free(fh);
  }


#ifdef FLOW_DEBUG
  dev->ndev->be.iface->set_debug_mode(dev->ndev->be.be_dev, FLOW_BACKEND_DEBUG_MODE_NONE);
#endif

  return err;
}



int flow_destroy_profile_vswitch(struct flow_eth_dev *dev, struct flow_handle *flow, struct flow_error *error)
{
  int err = 0;

  flow_nic_set_error(ERR_SUCCESS, error);

  pthread_mutex_lock(&dev->ndev->mtx);
  if (flow) {
    /* delete this flow */
    err = flow_destroy_locked_profile_vswitch(dev, flow, error);
  } else {
    /* delete all created flows from this eth device */
    flow = dev->ndev->flow_base;

    while (flow && !err) {
      if (flow->dev == dev) {
        struct flow_handle *flow_next = flow->next;
        err = flow_destroy_locked_profile_vswitch(dev, flow, NULL);
        flow = flow_next;
      } else {
        flow = flow->next;
      }
    }
  }

  pthread_mutex_unlock(&dev->ndev->mtx);
  return err;
}


int flow_flush_profile_vswitch(
    UNUSED struct flow_eth_dev *dev,
    struct flow_error *error)
{
  NT_LOG(ERR, FILTER, "ERROR: Not implemented yet\n");
  error->type = FLOW_ERROR_GENERAL;
  error->message = "rte_flow_flush is not supported";
  return -1;
}

int flow_query_profile_vswitch(
    UNUSED struct flow_eth_dev *dev,
    UNUSED struct flow_handle *flow,
    UNUSED const struct flow_action *action,
    UNUSED void **data,
    UNUSED uint32_t *length,
    UNUSED struct flow_error *error)
{
  NT_LOG(ERR, FILTER, "ERROR: Not implemented yet\n");

  *data = NULL;
  *length = 0;
#if 0
  /* TODO TODO */
#endif
  return 0;
}


#if 0 /* only for devel-debugging */
void flow_dump_stat(struct flow_eth_dev *dev)
{
  uint32_t cam_elem[6];
  uint32_t cuckoo_moves;

  int err = km_get_cam_population_level(dev->ndev->km_res_handle, cam_elem, &cuckoo_moves);
  if (!err) {
    printf("CAM [0] : %i\n", cam_elem[0]);
    printf("CAM [1] : %i\n", cam_elem[1]);
    printf("CAM [2] : %i\n", cam_elem[2]);
    printf("Cuckoo moves: %i\n", cuckoo_moves);

    printf("CAM utilization %i %%\n", ((cam_elem[0]+cam_elem[1]+cam_elem[2])*100)/(3*2048));

  }
}
#endif
