/*
 * SPDX-License-Identifier: BSD-3-Clause
 * Copyright(c) 2023 Napatech A/S
 */

#include "ntlog.h"

#include "nthw_drv.h"
#include "nthw_register.h"
#include "nthw_rac.h"

#include "flow_nthw_flm.h"

#include <stdlib.h> /* malloc */
#include <string.h> /* memset */

struct FlmNthw *FlmNthw_new(void)
{
  struct FlmNthw *p = malloc(sizeof(struct FlmNthw));
  if (p)
    (void)memset(p, 0, sizeof(*p));

  return p;
}

void FlmNthw_delete(struct FlmNthw *p)
{
  if (p) {
    (void)memset(p, 0, sizeof(*p));
    free(p);
  }
}

void flmNthw_setDebugMode(struct FlmNthw *p, unsigned int nDebugMode)
{
  nthw_module_set_debug_mode(p->mFlm, nDebugMode);
}

int FlmNthw_init(struct FlmNthw *p, nthw_fpga_t *pFpga, int nInstance)
{
  const char *const pAdapterIdStr = pFpga->p_fpga_info->mpAdapterIdStr;
  nthw_module_t* pMod = nthw_fpga_query_module(pFpga, MOD_FLM, nInstance);

  assert(nInstance >= 0 && nInstance < 256);

  if (p == NULL) {
    return pMod == NULL ? -1 : 0;
  }

  if (pMod == NULL) {
    NT_LOG(ERR, NTHW, "%s: Flm %d: no such instance\n", pAdapterIdStr, nInstance);
    return -1;
  }

  p->mpRac = pFpga->p_fpga_info->mp_nthw_rac;

  p->mpFpga = pFpga;
  p->mPhysicalAdapterNo = (uint8_t)nInstance;
  p->mFlm = pMod;

#if defined(DEBUG) && (0)
  nthw_module_set_debug_mode(p->mFlm, 0xFF); /* DEBUG: */
#endif

  p->mpControl = nthw_module_get_register(p->mFlm, FLM_CONTROL);
  p->mpControlEnable = nthw_register_get_field(p->mpControl, FLM_CONTROL_ENABLE);
  p->mpControlInit = nthw_register_get_field(p->mpControl, FLM_CONTROL_INIT);
  p->mpControlLds = nthw_register_get_field(p->mpControl, FLM_CONTROL_LDS);
  p->mpControlLfs = nthw_register_get_field(p->mpControl, FLM_CONTROL_LFS);
  p->mpControlLis = nthw_register_get_field(p->mpControl, FLM_CONTROL_LIS);
  p->mpControlUds = nthw_register_get_field(p->mpControl, FLM_CONTROL_UDS);
  p->mpControlUis = nthw_register_get_field(p->mpControl, FLM_CONTROL_UIS);
  p->mpControlRds = nthw_register_get_field(p->mpControl, FLM_CONTROL_RDS);
  p->mpControlRis = nthw_register_get_field(p->mpControl, FLM_CONTROL_RIS);
  p->mpControlPds = nthw_register_query_field(p->mpControl, FLM_CONTROL_PDS);
  p->mpControlPis = nthw_register_query_field(p->mpControl, FLM_CONTROL_PIS);
  p->mpControlCrcwr = nthw_register_get_field(p->mpControl, FLM_CONTROL_CRCWR);
  p->mpControlCrcrd = nthw_register_get_field(p->mpControl, FLM_CONTROL_CRCRD);
  p->mpControlRbl = nthw_register_get_field(p->mpControl, FLM_CONTROL_RBL);
  p->mpControlEab = nthw_register_get_field(p->mpControl, FLM_CONTROL_EAB);
  p->mpControlSplitSdramUsage = nthw_register_get_field(p->mpControl, FLM_CONTROL_SPLIT_SDRAM_USAGE);

  p->mpStatus = nthw_module_get_register(p->mFlm, FLM_STATUS);
  p->mpStatusCalibdone = nthw_register_get_field(p->mpStatus, FLM_STATUS_CALIBDONE);
  p->mpStatusInitdone = nthw_register_get_field(p->mpStatus, FLM_STATUS_INITDONE);
  p->mpStatusIdle = nthw_register_get_field(p->mpStatus, FLM_STATUS_IDLE);
  p->mpStatusCritical = nthw_register_get_field(p->mpStatus, FLM_STATUS_CRITICAL);
  p->mpStatusPanic = nthw_register_get_field(p->mpStatus, FLM_STATUS_PANIC);
  p->mpStatusCrcerr = nthw_register_get_field(p->mpStatus, FLM_STATUS_CRCERR);
  p->mpStatusEftBp = nthw_register_get_field(p->mpStatus, FLM_STATUS_EFT_BP);

  /* FLM_TIMEOUT was removed in the 55-30 */
  p->mpTimeout = nthw_module_query_register(p->mFlm, FLM_TIMEOUT);
  p->mpTimeoutT = nthw_register_query_field(p->mpTimeout, FLM_TIMEOUT_T);

  /* FLM_SCRUB was renamed to FLM_SCAN in the 55-30 */
  p->mpScan = nthw_module_query_register(p->mFlm, FLM_SCAN);
  if (p->mpScan == NULL) {
    p->mpScan = nthw_module_get_register(p->mFlm, FLM_SCRUB);
    p->mpScanI = nthw_register_get_field(p->mpScan, FLM_SCRUB_I);
  } else {
    p->mpScanI = nthw_register_get_field(p->mpScan, FLM_SCAN_I);
  }

  p->mpLoadBin = nthw_module_get_register(p->mFlm, FLM_LOAD_BIN);
  p->mpLoadBinBin = nthw_register_get_field(p->mpLoadBin, FLM_LOAD_BIN_BIN);

  p->mpLoadPps = nthw_module_get_register(p->mFlm, FLM_LOAD_PPS);
  p->mpLoadPpsPps = nthw_register_get_field(p->mpLoadPps, FLM_LOAD_PPS_PPS);

  p->mpLoadLps = nthw_module_get_register(p->mFlm, FLM_LOAD_LPS);
  p->mpLoadLpsLps = nthw_register_get_field(p->mpLoadLps, FLM_LOAD_LPS_LPS);

  p->mpLoadAps = nthw_module_get_register(p->mFlm, FLM_LOAD_APS);
  p->mpLoadApsAps = nthw_register_get_field(p->mpLoadAps, FLM_LOAD_APS_APS);

  p->mpPrio = nthw_module_get_register(p->mFlm, FLM_PRIO);
  p->mpPrioLimit0 = nthw_register_get_field(p->mpPrio, FLM_PRIO_LIMIT0);
  p->mpPrioFt0 = nthw_register_get_field(p->mpPrio, FLM_PRIO_FT0);
  p->mpPrioLimit1 = nthw_register_get_field(p->mpPrio, FLM_PRIO_LIMIT1);
  p->mpPrioFt1 = nthw_register_get_field(p->mpPrio, FLM_PRIO_FT1);
  p->mpPrioLimit2 = nthw_register_get_field(p->mpPrio, FLM_PRIO_LIMIT2);
  p->mpPrioFt2 = nthw_register_get_field(p->mpPrio, FLM_PRIO_FT2);
  p->mpPrioLimit3 = nthw_register_get_field(p->mpPrio, FLM_PRIO_LIMIT3);
  p->mpPrioFt3 = nthw_register_get_field(p->mpPrio, FLM_PRIO_FT3);

  p->mpPstCtrl = nthw_module_get_register(p->mFlm, FLM_PST_CTRL);
  p->mpPstCtrlAdr = nthw_register_get_field(p->mpPstCtrl, FLM_PST_CTRL_ADR);
  p->mpPstCtrlCnt = nthw_register_get_field(p->mpPstCtrl, FLM_PST_CTRL_CNT);
  p->mpPstData = nthw_module_get_register(p->mFlm, FLM_PST_DATA);
  p->mpPstDataBp = nthw_register_get_field(p->mpPstData, FLM_PST_DATA_BP);
  p->mpPstDataPp = nthw_register_get_field(p->mpPstData, FLM_PST_DATA_PP);
  p->mpPstDataTp = nthw_register_get_field(p->mpPstData, FLM_PST_DATA_TP);

  p->mpRcpCtrl = nthw_module_get_register(p->mFlm, FLM_RCP_CTRL);
  p->mpRcpCtrlAdr = nthw_register_get_field(p->mpRcpCtrl, FLM_RCP_CTRL_ADR);
  p->mpRcpCtrlCnt = nthw_register_get_field(p->mpRcpCtrl, FLM_RCP_CTRL_CNT);
  p->mpRcpData = nthw_module_get_register(p->mFlm, FLM_RCP_DATA);
  p->mpRcpDataLookup = nthw_register_get_field(p->mpRcpData, FLM_RCP_DATA_LOOKUP);
  p->mpRcpDataQw0Dyn = nthw_register_get_field(p->mpRcpData, FLM_RCP_DATA_QW0_DYN);
  p->mpRcpDataQw0Ofs = nthw_register_get_field(p->mpRcpData, FLM_RCP_DATA_QW0_OFS);
  p->mpRcpDataQw0Sel = nthw_register_get_field(p->mpRcpData, FLM_RCP_DATA_QW0_SEL);
  p->mpRcpDataQw4Dyn = nthw_register_get_field(p->mpRcpData, FLM_RCP_DATA_QW4_DYN);
  p->mpRcpDataQw4Ofs = nthw_register_get_field(p->mpRcpData, FLM_RCP_DATA_QW4_OFS);
  p->mpRcpDataSw8Dyn = nthw_register_get_field(p->mpRcpData, FLM_RCP_DATA_SW8_DYN);
  p->mpRcpDataSw8Ofs = nthw_register_get_field(p->mpRcpData, FLM_RCP_DATA_SW8_OFS);
  p->mpRcpDataSw8Sel = nthw_register_get_field(p->mpRcpData, FLM_RCP_DATA_SW8_SEL);
  p->mpRcpDataSw9Dyn = nthw_register_get_field(p->mpRcpData, FLM_RCP_DATA_SW9_DYN);
  p->mpRcpDataSw9Ofs = nthw_register_get_field(p->mpRcpData, FLM_RCP_DATA_SW9_OFS);
  p->mpRcpDataMask = nthw_register_get_field(p->mpRcpData, FLM_RCP_DATA_MASK);
  p->mpRcpDataKid = nthw_register_get_field(p->mpRcpData, FLM_RCP_DATA_KID);
  p->mpRcpDataOpn = nthw_register_get_field(p->mpRcpData, FLM_RCP_DATA_OPN);
  p->mpRcpDataIpn = nthw_register_get_field(p->mpRcpData, FLM_RCP_DATA_IPN);
  p->mpRcpDataBytDyn = nthw_register_get_field(p->mpRcpData, FLM_RCP_DATA_BYT_DYN);
  p->mpRcpDataBytOfs = nthw_register_get_field(p->mpRcpData, FLM_RCP_DATA_BYT_OFS);
  p->mpRcpDataTxplm = nthw_register_get_field(p->mpRcpData, FLM_RCP_DATA_TXPLM);
  p->mpRcpDataAutoIpv4Mask = nthw_register_get_field(p->mpRcpData, FLM_RCP_DATA_AUTO_IPV4_MASK);

  p->mpBufCtrl = nthw_module_get_register(p->mFlm, FLM_BUF_CTRL);

  p->mpLrnData = nthw_module_get_register(p->mFlm, FLM_LRN_DATA);
  p->mpInfData = nthw_module_get_register(p->mFlm, FLM_INF_DATA);
  p->mpStaData = nthw_module_get_register(p->mFlm, FLM_STA_DATA);

  p->mpStatLrnDone = nthw_module_get_register(p->mFlm, FLM_STAT_LRN_DONE);
  p->mpStatLrnDoneCnt = nthw_register_get_field(p->mpStatLrnDone, FLM_STAT_LRN_DONE_CNT);

  p->mpStatLrnIgnore = nthw_module_get_register(p->mFlm, FLM_STAT_LRN_IGNORE);
  p->mpStatLrnIgnoreCnt = nthw_register_get_field(p->mpStatLrnIgnore, FLM_STAT_LRN_IGNORE_CNT);

  p->mpStatLrnFail = nthw_module_get_register(p->mFlm, FLM_STAT_LRN_FAIL);
  p->mpStatLrnFailCnt = nthw_register_get_field(p->mpStatLrnFail, FLM_STAT_LRN_FAIL_CNT);

  p->mpStatUnlDone = nthw_module_get_register(p->mFlm, FLM_STAT_UNL_DONE);
  p->mpStatUnlDoneCnt = nthw_register_get_field(p->mpStatUnlDone, FLM_STAT_UNL_DONE_CNT);

  p->mpStatUnlIgnore = nthw_module_get_register(p->mFlm, FLM_STAT_UNL_IGNORE);
  p->mpStatUnlIgnoreCnt = nthw_register_get_field(p->mpStatUnlIgnore, FLM_STAT_UNL_IGNORE_CNT);

  p->mpStatPrbDone = nthw_module_query_register(p->mFlm, FLM_STAT_PRB_DONE);
  p->mpStatPrbDoneCnt = nthw_register_query_field(p->mpStatPrbDone, FLM_STAT_PRB_DONE_CNT);

  p->mpStatPrbIgnore = nthw_module_query_register(p->mFlm, FLM_STAT_PRB_IGNORE);
  p->mpStatPrbIgnoreCnt = nthw_register_query_field(p->mpStatPrbIgnore, FLM_STAT_PRB_IGNORE_CNT);

  p->mpStatRelDone = nthw_module_get_register(p->mFlm, FLM_STAT_REL_DONE);
  p->mpStatRelDoneCnt = nthw_register_get_field(p->mpStatRelDone, FLM_STAT_REL_DONE_CNT);

  p->mpStatRelIgnore = nthw_module_get_register(p->mFlm, FLM_STAT_REL_IGNORE);
  p->mpStatRelIgnoreCnt = nthw_register_get_field(p->mpStatRelIgnore, FLM_STAT_REL_IGNORE_CNT);

  p->mpStatAulDone = nthw_module_get_register(p->mFlm, FLM_STAT_AUL_DONE);
  p->mpStatAulDoneCnt = nthw_register_get_field(p->mpStatAulDone, FLM_STAT_AUL_DONE_CNT);

  p->mpStatAulIgnore = nthw_module_get_register(p->mFlm, FLM_STAT_AUL_IGNORE);
  p->mpStatAulIgnoreCnt = nthw_register_get_field(p->mpStatAulIgnore, FLM_STAT_AUL_IGNORE_CNT);

  p->mpStatAulFail = nthw_module_get_register(p->mFlm, FLM_STAT_AUL_FAIL);
  p->mpStatAulFailCnt = nthw_register_get_field(p->mpStatAulFail, FLM_STAT_AUL_FAIL_CNT);

  p->mpStatTulDone = nthw_module_get_register(p->mFlm, FLM_STAT_TUL_DONE);
  p->mpStatTulDoneCnt = nthw_register_get_field(p->mpStatTulDone, FLM_STAT_TUL_DONE_CNT);

  p->mpStatFlows = nthw_module_get_register(p->mFlm, FLM_STAT_FLOWS);
  p->mpStatFlowsCnt = nthw_register_get_field(p->mpStatFlows, FLM_STAT_FLOWS_CNT);

  p->mpStatStaDone = nthw_module_query_register(p->mFlm, FLM_STAT_STA_DONE);
  p->mpStatStaDoneCnt = nthw_register_query_field(p->mpStatStaDone, FLM_STAT_STA_DONE_CNT);

  p->mpStatInfDone = nthw_module_query_register(p->mFlm, FLM_STAT_INF_DONE);
  p->mpStatInfDoneCnt = nthw_register_query_field(p->mpStatInfDone, FLM_STAT_INF_DONE_CNT);

  p->mpStatInfSkip = nthw_module_query_register(p->mFlm, FLM_STAT_INF_SKIP);
  p->mpStatInfSkipCnt = nthw_register_query_field(p->mpStatInfSkip, FLM_STAT_INF_SKIP_CNT);

  p->mpStatPckHit = nthw_module_query_register(p->mFlm, FLM_STAT_PCK_HIT);
  p->mpStatPckHitCnt = nthw_register_query_field(p->mpStatPckHit, FLM_STAT_PCK_HIT_CNT);

  p->mpStatPckMiss = nthw_module_query_register(p->mFlm, FLM_STAT_PCK_MISS);
  p->mpStatPckMissCnt = nthw_register_query_field(p->mpStatPckMiss, FLM_STAT_PCK_MISS_CNT);

  p->mpStatPckUnh = nthw_module_query_register(p->mFlm, FLM_STAT_PCK_UNH);
  p->mpStatPckUnhCnt = nthw_register_query_field(p->mpStatPckUnh, FLM_STAT_PCK_UNH_CNT);

  p->mpStatPckDis = nthw_module_query_register(p->mFlm, FLM_STAT_PCK_DIS);
  p->mpStatPckDisCnt = nthw_register_query_field(p->mpStatPckDis, FLM_STAT_PCK_DIS_CNT);

  p->mpStatCshHit = nthw_module_query_register(p->mFlm, FLM_STAT_CSH_HIT);
  p->mpStatCshHitCnt = nthw_register_query_field(p->mpStatCshHit, FLM_STAT_CSH_HIT_CNT);

  p->mpStatCshMiss = nthw_module_query_register(p->mFlm, FLM_STAT_CSH_MISS);
  p->mpStatCshMissCnt = nthw_register_query_field(p->mpStatCshMiss, FLM_STAT_CSH_MISS_CNT);

  p->mpStatCshUnh = nthw_module_query_register(p->mFlm, FLM_STAT_CSH_UNH);
  p->mpStatCshUnhCnt = nthw_register_query_field(p->mpStatCshUnh, FLM_STAT_CSH_UNH_CNT);

  p->mpStatCucStart = nthw_module_query_register(p->mFlm, FLM_STAT_CUC_START);
  p->mpStatCucStartCnt = nthw_register_query_field(p->mpStatCucStart, FLM_STAT_CUC_START_CNT);

  p->mpStatCucMove = nthw_module_query_register(p->mFlm, FLM_STAT_CUC_MOVE);
  p->mpStatCucMoveCnt = nthw_register_query_field(p->mpStatCucMove, FLM_STAT_CUC_MOVE_CNT);

  p->mpScrubCtrl = nthw_module_query_register(p->mFlm, FLM_SCRUB_CTRL);
  p->mpScrubCtrlAdr = nthw_register_query_field(p->mpScrubCtrl, FLM_SCRUB_CTRL_ADR);
  p->mpScrubCtrlCnt = nthw_register_query_field(p->mpScrubCtrl, FLM_SCRUB_CTRL_CNT);

  p->mpScrubData = nthw_module_query_register(p->mFlm, FLM_SCRUB_DATA);
  p->mpScrubDataT = nthw_register_query_field(p->mpScrubData, FLM_SCRUB_DATA_T);

  return 0;
}

void FlmNthw_ControlEnable(const struct FlmNthw *p, uint32_t val)
{
  nthw_field_set_val32(p->mpControlEnable, val);
}

void FlmNthw_ControlInit(const struct FlmNthw *p, uint32_t val)
{
  nthw_field_set_val32(p->mpControlInit, val);
}

void FlmNthw_ControlLds(const struct FlmNthw *p, uint32_t val)
{
  nthw_field_set_val32(p->mpControlLds, val);
}

void FlmNthw_ControlLfs(const struct FlmNthw *p, uint32_t val)
{
  nthw_field_set_val32(p->mpControlLfs, val);
}

void FlmNthw_ControlLis(const struct FlmNthw *p, uint32_t val)
{
  nthw_field_set_val32(p->mpControlLis, val);
}

void FlmNthw_ControlUds(const struct FlmNthw *p, uint32_t val)
{
  nthw_field_set_val32(p->mpControlUds, val);
}

void FlmNthw_ControlUis(const struct FlmNthw *p, uint32_t val)
{
  nthw_field_set_val32(p->mpControlUis, val);
}

void FlmNthw_ControlRds(const struct FlmNthw *p, uint32_t val)
{
  nthw_field_set_val32(p->mpControlRds, val);
}

void FlmNthw_ControlRis(const struct FlmNthw *p, uint32_t val)
{
  nthw_field_set_val32(p->mpControlRis, val);
}

void FlmNthw_ControlPds(const struct FlmNthw *p, uint32_t val)
{
  assert(p->mpControlPds);
  nthw_field_set_val32(p->mpControlPds, val);
}

void FlmNthw_ControlPis(const struct FlmNthw *p, uint32_t val)
{
  assert(p->mpControlPis);
  nthw_field_set_val32(p->mpControlPis, val);
}

void FlmNthw_ControlCrcwr(const struct FlmNthw *p, uint32_t val)
{
  nthw_field_set_val32(p->mpControlCrcwr, val);
}

void FlmNthw_ControlCrcrd(const struct FlmNthw *p, uint32_t val)
{
  nthw_field_set_val32(p->mpControlCrcrd, val);
}

void FlmNthw_ControlRbl(const struct FlmNthw *p, uint32_t val)
{
  nthw_field_set_val32(p->mpControlRbl, val);
}

void FlmNthw_ControlEab(const struct FlmNthw *p, uint32_t val)
{
  nthw_field_set_val32(p->mpControlEab, val);
}

void FlmNthw_ControlSplitSdramUsage(const struct FlmNthw *p, uint32_t val)
{
  nthw_field_set_val32(p->mpControlSplitSdramUsage, val);
}

void FlmNthw_ControlFlush(const struct FlmNthw *p)
{
  nthw_register_flush(p->mpControl, 1);
}

void FlmNthw_StatusCalibdone(const struct FlmNthw *p, uint32_t *val, int get)
{
  if (get) {
    *val = nthw_field_get_val32(p->mpStatusCalibdone);
  }
}

void FlmNthw_StatusInitdone(const struct FlmNthw *p, uint32_t *val, int get)
{
  if (get) {
    *val = nthw_field_get_val32(p->mpStatusInitdone);
  }
}

void FlmNthw_StatusIdle(const struct FlmNthw *p, uint32_t *val, int get)
{
  if (get) {
    *val = nthw_field_get_val32(p->mpStatusIdle);
  }
}

void FlmNthw_StatusCritical(const struct FlmNthw *p, uint32_t *val, int get)
{
  if (get) {
    *val = nthw_field_get_val32(p->mpStatusCritical);
  }
  else {
    nthw_field_set_val32(p->mpStatusCritical, *val);
  }
}

void FlmNthw_StatusPanic(const struct FlmNthw *p, uint32_t *val, int get)
{
  if (get) {
    *val = nthw_field_get_val32(p->mpStatusPanic);
  }
  else {
    nthw_field_set_val32(p->mpStatusPanic, *val);
  }
}

void FlmNthw_StatusCrcerr(const struct FlmNthw *p, uint32_t *val, int get)
{
  if (get) {
    *val = nthw_field_get_val32(p->mpStatusCrcerr);
  }
  else {
    nthw_field_set_val32(p->mpStatusCrcerr, *val);
  }
}

void FlmNthw_StatusEftBp(const struct FlmNthw *p, uint32_t *val, int get)
{
  if (get) {
    *val = nthw_field_get_val32(p->mpStatusEftBp);
  }
}

void FlmNthw_StatusFlush(const struct FlmNthw *p)
{
  nthw_register_flush(p->mpStatus, 1);
}

void FlmNthw_StatusUpdate(const struct FlmNthw *p)
{
  nthw_register_update(p->mpStatus);
}

void FlmNthw_TimeoutT(const struct FlmNthw *p, uint32_t val)
{
  nthw_field_set_val32(p->mpTimeoutT, val);
}

void FlmNthw_TimeoutFlush(const struct FlmNthw *p)
{
  nthw_register_flush(p->mpTimeout, 1);
}

void FlmNthw_ScanI(const struct FlmNthw *p, uint32_t val)
{
  nthw_field_set_val32(p->mpScanI, val);
}

void FlmNthw_ScanFlush(const struct FlmNthw *p)
{
  nthw_register_flush(p->mpScan, 1);
}

void FlmNthw_LoadBin(const struct FlmNthw *p, uint32_t val)
{
  nthw_field_set_val32(p->mpLoadBinBin, val);
}

void FlmNthw_LoadBinFlush(const struct FlmNthw *p)
{
  nthw_register_flush(p->mpLoadBin, 1);
}

void FlmNthw_LoadPps(const struct FlmNthw *p, uint32_t val)
{
  nthw_field_set_val32(p->mpLoadPpsPps, val);
}

void FlmNthw_LoadPpsFlush(const struct FlmNthw *p)
{
  nthw_register_flush(p->mpLoadPps, 1);
}

void FlmNthw_LoadLps(const struct FlmNthw *p, uint32_t val)
{
  nthw_field_set_val32(p->mpLoadLpsLps, val);
}

void FlmNthw_LoadLpsFlush(const struct FlmNthw *p)
{
  nthw_register_flush(p->mpLoadLps, 1);
}

void FlmNthw_LoadAps(const struct FlmNthw *p, uint32_t val)
{
  nthw_field_set_val32(p->mpLoadApsAps, val);
}

void FlmNthw_LoadApsFlush(const struct FlmNthw *p)
{
  nthw_register_flush(p->mpLoadAps, 1);
}

void FlmNthw_PrioLimit0(const struct FlmNthw *p, uint32_t val)
{
  nthw_field_set_val32(p->mpPrioLimit0, val);
}

void FlmNthw_PrioFt0(const struct FlmNthw *p, uint32_t val)
{
  nthw_field_set_val32(p->mpPrioFt0, val);
}

void FlmNthw_PrioLimit1(const struct FlmNthw *p, uint32_t val)
{
  nthw_field_set_val32(p->mpPrioLimit1, val);
}

void FlmNthw_PrioFt1(const struct FlmNthw *p, uint32_t val)
{
  nthw_field_set_val32(p->mpPrioFt1, val);
}

void FlmNthw_PrioLimit2(const struct FlmNthw *p, uint32_t val)
{
  nthw_field_set_val32(p->mpPrioLimit2, val);
}

void FlmNthw_PrioFt2(const struct FlmNthw *p, uint32_t val)
{
  nthw_field_set_val32(p->mpPrioFt2, val);
}

void FlmNthw_PrioLimit3(const struct FlmNthw *p, uint32_t val)
{
  nthw_field_set_val32(p->mpPrioLimit3, val);
}

void FlmNthw_PrioFt3(const struct FlmNthw *p, uint32_t val)
{
  nthw_field_set_val32(p->mpPrioFt3, val);
}

void FlmNthw_PrioFlush(const struct FlmNthw *p)
{
  nthw_register_flush(p->mpPrio, 1);
}

void FlmNthw_PstSelect(const struct FlmNthw *p, uint32_t val)
{
  nthw_field_set_val32(p->mpPstCtrlAdr, val);
}

void FlmNthw_PstCnt(const struct FlmNthw *p, uint32_t val)
{
  nthw_field_set_val32(p->mpPstCtrlCnt, val);
}

void FlmNthw_PstBp(const struct FlmNthw *p, uint32_t val)
{
  nthw_field_set_val32(p->mpPstDataBp, val);
}

void FlmNthw_PstPp(const struct FlmNthw *p, uint32_t val)
{
  nthw_field_set_val32(p->mpPstDataPp, val);
}

void FlmNthw_PstTp(const struct FlmNthw *p, uint32_t val)
{
  nthw_field_set_val32(p->mpPstDataTp, val);
}

void FlmNthw_PstFlush(const struct FlmNthw *p)
{
  nthw_register_flush(p->mpPstCtrl, 1);
  nthw_register_flush(p->mpPstData, 1);
}

void FlmNthw_RcpSelect(const struct FlmNthw *p, uint32_t val)
{
  nthw_field_set_val32(p->mpRcpCtrlAdr, val);
}

void FlmNthw_RcpCnt(const struct FlmNthw *p, uint32_t val)
{
  nthw_field_set_val32(p->mpRcpCtrlCnt, val);
}

void FlmNthw_RcpLookup(const struct FlmNthw *p, uint32_t val)
{
  nthw_field_set_val32(p->mpRcpDataLookup, val);
}

void FlmNthw_RcpQw0Dyn(const struct FlmNthw *p, uint32_t val)
{
  nthw_field_set_val32(p->mpRcpDataQw0Dyn, val);
}

void FlmNthw_RcpQw0Ofs(const struct FlmNthw *p, uint32_t val)
{
  nthw_field_set_val32(p->mpRcpDataQw0Ofs, val);
}

void FlmNthw_RcpQw0Sel(const struct FlmNthw *p, uint32_t val)
{
  nthw_field_set_val32(p->mpRcpDataQw0Sel, val);
}

void FlmNthw_RcpQw4Dyn(const struct FlmNthw *p, uint32_t val)
{
  nthw_field_set_val32(p->mpRcpDataQw4Dyn, val);
}

void FlmNthw_RcpQw4Ofs(const struct FlmNthw *p, uint32_t val)
{
  nthw_field_set_val32(p->mpRcpDataQw4Ofs, val);
}

void FlmNthw_RcpSw8Dyn(const struct FlmNthw *p, uint32_t val)
{
  nthw_field_set_val32(p->mpRcpDataSw8Dyn, val);
}

void FlmNthw_RcpSw8Ofs(const struct FlmNthw *p, uint32_t val)
{
  nthw_field_set_val32(p->mpRcpDataSw8Ofs, val);
}

void FlmNthw_RcpSw8Sel(const struct FlmNthw *p, uint32_t val)
{
  nthw_field_set_val32(p->mpRcpDataSw8Sel, val);
}

void FlmNthw_RcpSw9Dyn(const struct FlmNthw *p, uint32_t val)
{
  nthw_field_set_val32(p->mpRcpDataSw9Dyn, val);
}

void FlmNthw_RcpSw9Ofs(const struct FlmNthw *p, uint32_t val)
{
  nthw_field_set_val32(p->mpRcpDataSw9Ofs, val);
}

void FlmNthw_RcpMask(const struct FlmNthw *p, const uint32_t *val)
{
  nthw_field_set_val(p->mpRcpDataMask, val, 10);
}

void FlmNthw_RcpKid(const struct FlmNthw *p, uint32_t val)
{
  nthw_field_set_val32(p->mpRcpDataKid, val);
}

void FlmNthw_RcpOpn(const struct FlmNthw *p, uint32_t val)
{
  nthw_field_set_val32(p->mpRcpDataOpn, val);
}

void FlmNthw_RcpIpn(const struct FlmNthw *p, uint32_t val)
{
  nthw_field_set_val32(p->mpRcpDataIpn, val);
}

void FlmNthw_RcpBytDyn(const struct FlmNthw *p, uint32_t val)
{
  nthw_field_set_val32(p->mpRcpDataBytDyn, val);
}

void FlmNthw_RcpBytOfs(const struct FlmNthw *p, uint32_t val)
{
  nthw_field_set_val32(p->mpRcpDataBytOfs, val);
}

void FlmNthw_RcpTxplm(const struct FlmNthw *p, uint32_t val)
{
  nthw_field_set_val32(p->mpRcpDataTxplm, val);
}

void FlmNthw_RcpAutoIpv4Mask(const struct FlmNthw *p, uint32_t val)
{
  nthw_field_set_val32(p->mpRcpDataAutoIpv4Mask, val);
}

void FlmNthw_RcpFlush(const struct FlmNthw *p)
{
  nthw_register_flush(p->mpRcpCtrl, 1);
  nthw_register_flush(p->mpRcpData, 1);
}

int FlmNthw_BufCtrlUpdate(const struct FlmNthw *p, uint32_t* lrn_free, uint32_t* inf_avail, uint32_t* sta_avail)
{
  int ret = -1;

  struct nthw_rac* rac = (struct nthw_rac*)p->mpRac;
  uint32_t address_bufctrl = nthw_register_get_address(p->mpBufCtrl);
  RabBusId_t busId = 1;
  struct DmaBufPtr bc_buf;
  ret = nthw_rac_rab_dma_begin(rac);
  if (ret == 0) {
    nthw_rac_rab_read32_dma(rac, busId, address_bufctrl, 2, &bc_buf);
    ret = nthw_rac_rab_dma_commit(rac);
    if (ret != 0) {
      return ret;
    }

    uint32_t bc_mask = bc_buf.size - 1;
    uint32_t bc_index = bc_buf.index;
    *lrn_free = bc_buf.base[bc_index & bc_mask] & 0xffff;
    *inf_avail = (bc_buf.base[bc_index & bc_mask] >> 16) & 0xffff;
    *sta_avail = bc_buf.base[(bc_index + 1) & bc_mask] & 0xffff;
  }

  return ret;
}

int FlmNthw_LrnDataFlush(const struct FlmNthw *p, const uint32_t *data, uint32_t wordCount, uint32_t* lrn_free, uint32_t* inf_avail, uint32_t* sta_avail)
{
  int ret = -1;

  struct nthw_rac* rac = (struct nthw_rac*)p->mpRac;
  uint32_t address = nthw_register_get_address(p->mpLrnData);
  uint32_t address_bufctrl = nthw_register_get_address(p->mpBufCtrl);
  RabBusId_t busId = 1;
  struct DmaBufPtr bc_buf;

  if (nthw_rac_rab_dma_begin(rac) == 0) {
    /* Announce the number of words to write to LRN_DATA */
    uint32_t bufctrl_data[2];
    bufctrl_data[0] = wordCount;
    bufctrl_data[1]= 0;
    nthw_rac_rab_write32_dma(rac, busId, address_bufctrl, 2, bufctrl_data);
    nthw_rac_rab_write32_dma(rac, busId, address, wordCount, data);
    nthw_rac_rab_read32_dma(rac, busId, address_bufctrl, 2, &bc_buf);
    ret = nthw_rac_rab_dma_commit(rac);
    if (ret != 0) {
      return ret;
    }

    uint32_t bc_mask = bc_buf.size - 1;
    uint32_t bc_index = bc_buf.index;
    *lrn_free = bc_buf.base[bc_index & bc_mask] & 0xffff;
    *inf_avail = (bc_buf.base[bc_index & bc_mask] >> 16) & 0xffff;
    *sta_avail = bc_buf.base[(bc_index + 1) & bc_mask] & 0xffff;
  }

  return ret;
}

int FlmNthw_InfDataUpdate(const struct FlmNthw *p, uint32_t *data, uint32_t wordCount, uint32_t* lrn_free, uint32_t* inf_avail, uint32_t* sta_avail)
{
  int ret = -1;

  struct nthw_rac* rac = (struct nthw_rac*)p->mpRac;
  uint32_t address_infdata = nthw_register_get_address(p->mpInfData);
  uint32_t address_bufctrl = nthw_register_get_address(p->mpBufCtrl);
  RabBusId_t busId = 1;
  struct DmaBufPtr buf;
  struct DmaBufPtr bc_buf;
  ret = nthw_rac_rab_dma_begin(rac);
  if (ret == 0) {
    /* Announce the number of words to read from INF_DATA */
    uint32_t bufctrl_data[2];
    bufctrl_data[0] = wordCount << 16;
    bufctrl_data[1]= 0;
    nthw_rac_rab_write32_dma(rac, busId, address_bufctrl, 2, bufctrl_data);
    nthw_rac_rab_read32_dma(rac, busId, address_infdata, wordCount, &buf);
    nthw_rac_rab_read32_dma(rac, busId, address_bufctrl, 2, &bc_buf);
    ret = nthw_rac_rab_dma_commit(rac);
    if (ret != 0) {
      return ret;
    }

    uint32_t mask = buf.size - 1;
    uint32_t index = buf.index;
    for (uint32_t i = 0; i < wordCount; ++index, ++i) {
      data[i] = buf.base[index & mask];
    }

    uint32_t bc_mask = bc_buf.size - 1;
    uint32_t bc_index = bc_buf.index;
    *lrn_free = bc_buf.base[bc_index & bc_mask] & 0xffff;
    *inf_avail = (bc_buf.base[bc_index & bc_mask] >> 16) & 0xffff;
    *sta_avail = bc_buf.base[(bc_index + 1) & bc_mask] & 0xffff;
  }

  return ret;
}

int FlmNthw_StaDataUpdate(const struct FlmNthw *p, uint32_t *data, uint32_t wordCount, uint32_t* lrn_free, uint32_t* inf_avail, uint32_t* sta_avail)
{
  int ret = -1;

  struct nthw_rac* rac = (struct nthw_rac*)p->mpRac;
  uint32_t address_stadata = nthw_register_get_address(p->mpStaData);
  uint32_t address_bufctrl = nthw_register_get_address(p->mpBufCtrl);
  RabBusId_t busId = 1;
  struct DmaBufPtr buf;
  struct DmaBufPtr bc_buf;
  ret = nthw_rac_rab_dma_begin(rac);
  if (ret == 0) {
    /* Announce the number of words to read from STA_DATA */
    uint32_t bufctrl_data[2];
    bufctrl_data[0] = 0;
    bufctrl_data[1]= wordCount;
    nthw_rac_rab_write32_dma(rac, busId, address_bufctrl, 2, bufctrl_data);
    nthw_rac_rab_read32_dma(rac, busId, address_stadata, wordCount, &buf);
    nthw_rac_rab_read32_dma(rac, busId, address_bufctrl, 2, &bc_buf);
    ret = nthw_rac_rab_dma_commit(rac);
    if (ret != 0) {
      return ret;
    }

    uint32_t mask = buf.size - 1;
    uint32_t index = buf.index;
    for (uint32_t i = 0; i < wordCount; ++index, ++i) {
      data[i] = buf.base[index & mask];
    }

    uint32_t bc_mask = bc_buf.size - 1;
    uint32_t bc_index = bc_buf.index;
    *lrn_free = bc_buf.base[bc_index & bc_mask] & 0xffff;
    *inf_avail = (bc_buf.base[bc_index & bc_mask] >> 16) & 0xffff;
    *sta_avail = bc_buf.base[(bc_index + 1) & bc_mask] & 0xffff;
  }

  return ret;
}

void FlmNthw_StatLrnDoneCnt(const struct FlmNthw *p, uint32_t *val, int get)
{
  if (get) {
    *val = nthw_field_get_val32(p->mpStatLrnDoneCnt);
  }
}

void FlmNthw_StatLrnDoneUpdate(const struct FlmNthw *p)
{
  nthw_register_update(p->mpStatLrnDone);
}

void FlmNthw_StatLrnIgnoreCnt(const struct FlmNthw *p, uint32_t *val, int get)
{
  if (get) {
    *val = nthw_field_get_val32(p->mpStatLrnIgnoreCnt);
  }
}

void FlmNthw_StatLrnIgnoreUpdate(const struct FlmNthw *p)
{
  nthw_register_update(p->mpStatLrnIgnore);
}

void FlmNthw_StatLrnFailCnt(const struct FlmNthw *p, uint32_t *val, int get)
{
  if (get) {
    *val = nthw_field_get_val32(p->mpStatLrnFailCnt);
  }
}

void FlmNthw_StatLrnFailUpdate(const struct FlmNthw *p)
{
  nthw_register_update(p->mpStatLrnFail);
}

void FlmNthw_StatUnlDoneCnt(const struct FlmNthw *p, uint32_t *val, int get)
{
  if (get) {
    *val = nthw_field_get_val32(p->mpStatUnlDoneCnt);
  }
}

void FlmNthw_StatUnlDoneUpdate(const struct FlmNthw *p)
{
  nthw_register_update(p->mpStatUnlDone);
}

void FlmNthw_StatUnlIgnoreCnt(const struct FlmNthw *p, uint32_t *val, int get)
{
  if (get) {
    *val = nthw_field_get_val32(p->mpStatUnlIgnoreCnt);
  }
}

void FlmNthw_StatUnlIgnoreUpdate(const struct FlmNthw *p)
{
  nthw_register_update(p->mpStatUnlIgnore);
}

void FlmNthw_StatPrbDoneCnt(const struct FlmNthw *p, uint32_t *val, int get)
{
  assert(p->mpStatPrbDoneCnt);
  if (get) {
    *val = nthw_field_get_val32(p->mpStatPrbDoneCnt);
  }
}

void FlmNthw_StatPrbDoneUpdate(const struct FlmNthw *p)
{
  assert(p->mpStatPrbDone);
  nthw_register_update(p->mpStatPrbDone);
}

void FlmNthw_StatPrbIgnoreCnt(const struct FlmNthw *p, uint32_t *val, int get)
{
  assert(p->mpStatPrbIgnoreCnt);
  if (get) {
    *val = nthw_field_get_val32(p->mpStatPrbIgnoreCnt);
  }
}

void FlmNthw_StatPrbIgnoreUpdate(const struct FlmNthw *p)
{
  assert(p->mpStatPrbIgnore);
  nthw_register_update(p->mpStatPrbIgnore);
}

void FlmNthw_StatRelDoneCnt(const struct FlmNthw *p, uint32_t *val, int get)
{
  if (get) {
    *val = nthw_field_get_val32(p->mpStatRelDoneCnt);
  }
}

void FlmNthw_StatRelDoneUpdate(const struct FlmNthw *p)
{
  nthw_register_update(p->mpStatRelDone);
}

void FlmNthw_StatRelIgnoreCnt(const struct FlmNthw *p, uint32_t *val, int get)
{
  if (get) {
    *val = nthw_field_get_val32(p->mpStatRelIgnoreCnt);
  }
}

void FlmNthw_StatRelIgnoreUpdate(const struct FlmNthw *p)
{
  nthw_register_update(p->mpStatRelIgnore);
}

void FlmNthw_StatAulDoneCnt(const struct FlmNthw *p, uint32_t *val, int get)
{
  if (get) {
    *val = nthw_field_get_val32(p->mpStatAulDoneCnt);
  }
}

void FlmNthw_StatAulDoneUpdate(const struct FlmNthw *p)
{
  nthw_register_update(p->mpStatAulDone);
}

void FlmNthw_StatAulIgnoreCnt(const struct FlmNthw *p, uint32_t *val, int get)
{
  if (get) {
    *val = nthw_field_get_val32(p->mpStatAulIgnoreCnt);
  }
}

void FlmNthw_StatAulIgnoreUpdate(const struct FlmNthw *p)
{
  nthw_register_update(p->mpStatAulIgnore);
}

void FlmNthw_StatAulFailCnt(const struct FlmNthw *p, uint32_t *val, int get)
{
  if (get) {
    *val = nthw_field_get_val32(p->mpStatAulFailCnt);
  }
}

void FlmNthw_StatAulFailUpdate(const struct FlmNthw *p)
{
  nthw_register_update(p->mpStatAulFail);
}

void FlmNthw_StatTulDoneCnt(const struct FlmNthw *p, uint32_t *val, int get)
{
  if (get) {
    *val = nthw_field_get_val32(p->mpStatTulDoneCnt);
  }
}

void FlmNthw_StatTulDoneUpdate(const struct FlmNthw *p)
{
  nthw_register_update(p->mpStatTulDone);
}

void FlmNthw_StatFlowsCnt(const struct FlmNthw *p, uint32_t *val, int get)
{
  if (get) {
    *val = nthw_field_get_val32(p->mpStatFlowsCnt);
  }
}

void FlmNthw_StatFlowsUpdate(const struct FlmNthw *p)
{
  nthw_register_update(p->mpStatFlows);
}

void FlmNthw_StatStaDoneCnt(const struct FlmNthw *p, uint32_t *val, int get)
{
  assert(p->mpStatStaDoneCnt);
  if (get) {
    *val = nthw_field_get_val32(p->mpStatStaDoneCnt);
  }
}

void FlmNthw_StatStaDoneUpdate(const struct FlmNthw *p)
{
  assert(p->mpStatStaDone);
  nthw_register_update(p->mpStatStaDone);
}

void FlmNthw_StatInfDoneCnt(const struct FlmNthw *p, uint32_t *val, int get)
{
  assert(p->mpStatInfDoneCnt);
  if (get) {
    *val = nthw_field_get_val32(p->mpStatInfDoneCnt);
  }
}

void FlmNthw_StatInfDoneUpdate(const struct FlmNthw *p)
{
  assert(p->mpStatInfDone);
  nthw_register_update(p->mpStatInfDone);
}

void FlmNthw_StatInfSkipCnt(const struct FlmNthw *p, uint32_t *val, int get)
{
  assert(p->mpStatInfSkipCnt);
  if (get) {
    *val = nthw_field_get_val32(p->mpStatInfSkipCnt);
  }
}

void FlmNthw_StatInfSkipUpdate(const struct FlmNthw *p)
{
  assert(p->mpStatInfSkip);
  nthw_register_update(p->mpStatInfSkip);
}

void FlmNthw_StatPckHitCnt(const struct FlmNthw *p, uint32_t *val, int get)
{
  assert(p->mpStatPckHitCnt);
  if (get) {
    *val = nthw_field_get_val32(p->mpStatPckHitCnt);
  }
}

void FlmNthw_StatPckHitUpdate(const struct FlmNthw *p)
{
  assert(p->mpStatPckHit);
  nthw_register_update(p->mpStatPckHit);
}

void FlmNthw_StatPckMissCnt(const struct FlmNthw *p, uint32_t *val, int get)
{
  assert(p->mpStatPckMissCnt);
  if (get) {
    *val = nthw_field_get_val32(p->mpStatPckMissCnt);
  }
}

void FlmNthw_StatPckMissUpdate(const struct FlmNthw *p)
{
  assert(p->mpStatPckMiss);
  nthw_register_update(p->mpStatPckMiss);
}

void FlmNthw_StatPckUnhCnt(const struct FlmNthw *p, uint32_t *val, int get)
{
  assert(p->mpStatPckUnhCnt);
  if (get) {
    *val = nthw_field_get_val32(p->mpStatPckUnhCnt);
  }
}

void FlmNthw_StatPckUnhUpdate(const struct FlmNthw *p)
{
  assert(p->mpStatPckUnh);
  nthw_register_update(p->mpStatPckUnh);
}

void FlmNthw_StatPckDisCnt(const struct FlmNthw *p, uint32_t *val, int get)
{
  assert(p->mpStatPckDisCnt);
  if (get) {
    *val = nthw_field_get_val32(p->mpStatPckDisCnt);
  }
}

void FlmNthw_StatPckDisUpdate(const struct FlmNthw *p)
{
  assert(p->mpStatPckDis);
  nthw_register_update(p->mpStatPckDis);
}

void FlmNthw_StatCshHitCnt(const struct FlmNthw *p, uint32_t *val, int get)
{
  assert(p->mpStatCshHitCnt);
  if (get) {
    *val = nthw_field_get_val32(p->mpStatCshHitCnt);
  }
}

void FlmNthw_StatCshHitUpdate(const struct FlmNthw *p)
{
  assert(p->mpStatCshHit);
  nthw_register_update(p->mpStatCshHit);
}

void FlmNthw_StatCshMissCnt(const struct FlmNthw *p, uint32_t *val, int get)
{
  assert(p->mpStatCshMissCnt);
  if (get) {
    *val = nthw_field_get_val32(p->mpStatCshMissCnt);
  }
}

void FlmNthw_StatCshMissUpdate(const struct FlmNthw *p)
{
  assert(p->mpStatCshMiss);
  nthw_register_update(p->mpStatCshMiss);
}

void FlmNthw_StatCshUnhCnt(const struct FlmNthw *p, uint32_t *val, int get)
{
  assert(p->mpStatCshUnhCnt);
  if (get) {
    *val = nthw_field_get_val32(p->mpStatCshUnhCnt);
  }
}

void FlmNthw_StatCshUnhUpdate(const struct FlmNthw *p)
{
  assert(p->mpStatCshUnh);
  nthw_register_update(p->mpStatCshUnh);
}

void FlmNthw_StatCucStartCnt(const struct FlmNthw *p, uint32_t *val, int get)
{
  assert(p->mpStatCucStartCnt);
  if (get) {
    *val = nthw_field_get_val32(p->mpStatCucStartCnt);
  }
}

void FlmNthw_StatCucStartUpdate(const struct FlmNthw *p)
{
  assert(p->mpStatCucStart);
  nthw_register_update(p->mpStatCucStart);
}

void FlmNthw_StatCucMoveCnt(const struct FlmNthw *p, uint32_t *val, int get)
{
  assert(p->mpStatCucMoveCnt);
  if (get) {
    *val = nthw_field_get_val32(p->mpStatCucMoveCnt);
  }
}

void FlmNthw_StatCucMoveUpdate(const struct FlmNthw *p)
{
  assert(p->mpStatCucMove);
  nthw_register_update(p->mpStatCucMove);
}

void FlmNthw_ScrubSelect(const struct FlmNthw *p, uint32_t val)
{
  assert(p->mpScrubCtrlAdr);
  nthw_field_set_val32(p->mpScrubCtrlAdr, val);
}

void FlmNthw_ScrubCnt(const struct FlmNthw *p, uint32_t val)
{
  assert(p->mpScrubCtrlCnt);
  nthw_field_set_val32(p->mpScrubCtrlCnt, val);
}

void FlmNthw_ScrubT(const struct FlmNthw *p, uint32_t val)
{
  assert(p->mpScrubDataT);
  nthw_field_set_val32(p->mpScrubDataT, val);
}

void FlmNthw_ScrubFlush(const struct FlmNthw *p)
{
  assert(p->mpScrubCtrl);
  assert(p->mpScrubData);
  nthw_register_flush(p->mpScrubCtrl, 1);
  nthw_register_flush(p->mpScrubData, 1);
}

