/*
 * SPDX-License-Identifier: BSD-3-Clause
 * Copyright(c) 2023 Napatech A/S
 */

#include "ntlog.h"

#include "nthw_drv.h"
#include "nthw_register.h"

#include "flow_nthw_hsh.h"

#include <stdlib.h> /* malloc */
#include <string.h> /* memset */

void hshNthw_setDebugMode(struct HshNthw *p, unsigned int nDebugMode)
{
  nthw_module_set_debug_mode(p->mHsh, nDebugMode);
}

struct HshNthw *HshNthw_new(void)
{
  struct HshNthw *p = malloc(sizeof(struct HshNthw));
  if (p)
    (void)memset(p, 0, sizeof(*p));

  return p;
}

void HshNthw_delete(struct HshNthw *p)
{
  if (p) {
    (void)memset(p, 0, sizeof(*p));
    free(p);
  }
}

int HshNthw_init(struct HshNthw *p, nthw_fpga_t *pFpga, int nInstance)
{
  const char *const pAdapterIdStr = pFpga->p_fpga_info->mpAdapterIdStr;
  nthw_module_t* pMod = nthw_fpga_query_module(pFpga, MOD_HSH, nInstance);
  assert(nInstance >= 0 && nInstance < 256);

  if (p == NULL) {
    return pMod == NULL ? -1 : 0;
  }

  if (pMod == NULL) {
    NT_LOG(ERR, NTHW, "%s: Hsh %d: no such instance\n", pAdapterIdStr, nInstance);
    return -1;
  }

  p->mpFpga = pFpga;
  p->mPhysicalAdapterNo = (uint8_t)nInstance;
  p->mHsh = pMod;

#if defined(DEBUG) && (0)
  nthw_module_set_debug_mode(p->mHsh, 0xFF); /* DEBUG: */
#endif

  /* RCP */
  p->mpRcpCtrl = nthw_module_get_register(p->mHsh, HSH_RCP_CTRL);
  p->mpRcpAddr = nthw_register_get_field(p->mpRcpCtrl, HSH_RCP_CTRL_ADR);
  p->mpRcpCnt = nthw_register_get_field(p->mpRcpCtrl, HSH_RCP_CTRL_CNT);
  p->mpRcpData = nthw_module_get_register(p->mHsh, HSH_RCP_DATA);
  p->mpRcpDataLoadDistType =     nthw_register_get_field(p->mpRcpData, HSH_RCP_DATA_LOAD_DIST_TYPE);
  p->mpRcpDataMacPortMask =     nthw_register_get_field(p->mpRcpData, HSH_RCP_DATA_MAC_PORT_MASK);
  p->mpRcpDataSort = nthw_register_get_field(p->mpRcpData, HSH_RCP_DATA_SORT);
  p->mpRcpDataQw0Pe = nthw_register_get_field(p->mpRcpData, HSH_RCP_DATA_QW0_PE);
  p->mpRcpDataQw0Ofs = nthw_register_get_field(p->mpRcpData, HSH_RCP_DATA_QW0_OFS);
  p->mpRcpDataQw4Pe = nthw_register_get_field(p->mpRcpData, HSH_RCP_DATA_QW4_PE);
  p->mpRcpDataQw4Ofs = nthw_register_get_field(p->mpRcpData, HSH_RCP_DATA_QW4_OFS);
  p->mpRcpDataW8Pe = nthw_register_get_field(p->mpRcpData, HSH_RCP_DATA_W8_PE);
  p->mpRcpDataW8Ofs = nthw_register_get_field(p->mpRcpData, HSH_RCP_DATA_W8_OFS);
  p->mpRcpDataW8Sort = nthw_register_get_field(p->mpRcpData, HSH_RCP_DATA_W8_SORT);
  p->mpRcpDataW9Pe = nthw_register_get_field(p->mpRcpData, HSH_RCP_DATA_W9_PE);
  p->mpRcpDataW9Ofs = nthw_register_get_field(p->mpRcpData, HSH_RCP_DATA_W9_OFS);
  p->mpRcpDataW9Sort = nthw_register_get_field(p->mpRcpData, HSH_RCP_DATA_W9_SORT);
  p->mpRcpDataW9P = nthw_register_get_field(p->mpRcpData, HSH_RCP_DATA_W9_P);
  p->mpRcpDataPMask = nthw_register_get_field(p->mpRcpData, HSH_RCP_DATA_P_MASK);
  p->mpRcpDataWordMask =     nthw_register_get_field(p->mpRcpData, HSH_RCP_DATA_WORD_MASK);
  p->mpRcpDataSeed = nthw_register_get_field(p->mpRcpData, HSH_RCP_DATA_SEED);
  p->mpRcpDataTnlP = nthw_register_get_field(p->mpRcpData, HSH_RCP_DATA_TNL_P);
  p->mpRcpDataHshValid =     nthw_register_get_field(p->mpRcpData, HSH_RCP_DATA_HSH_VALID);
  p->mpRcpDataHshType = nthw_register_get_field(p->mpRcpData, HSH_RCP_DATA_HSH_TYPE);
  p->mpRcpDataAutoIpv4Mask = nthw_register_query_field(p->mpRcpData, HSH_RCP_DATA_AUTO_IPV4_MASK);

  /* Init */
  uint32_t val[10] = {0,0,0,0,0,0,0,0,0,0};
  nthw_field_set_val32(p->mpRcpAddr, 0);
  nthw_field_set_val32(p->mpRcpCnt, 1);

  nthw_field_set_val32(p->mpRcpDataLoadDistType, 0);
  nthw_field_set_val(p->mpRcpDataMacPortMask, val, p->mpRcpDataMacPortMask->mn_words);
  nthw_field_set_val32(p->mpRcpDataSort, 0);
  nthw_field_set_val32(p->mpRcpDataQw0Pe, 0);
  nthw_field_set_val32(p->mpRcpDataQw0Ofs, 0);
  nthw_field_set_val32(p->mpRcpDataQw4Pe, 0);
  nthw_field_set_val32(p->mpRcpDataQw4Ofs, 0);
  nthw_field_set_val32(p->mpRcpDataW8Pe, 0);
  nthw_field_set_val32(p->mpRcpDataW8Ofs, 0);
  nthw_field_set_val32(p->mpRcpDataW8Sort, 0);
  nthw_field_set_val32(p->mpRcpDataW9Pe, 0);
  nthw_field_set_val32(p->mpRcpDataW9Ofs, 0);
  nthw_field_set_val32(p->mpRcpDataW9Sort, 0);
  nthw_field_set_val32(p->mpRcpDataW9P, 0);
  nthw_field_set_val(p->mpRcpDataWordMask, val, 10);
  nthw_field_set_val32(p->mpRcpDataSeed, 0);
  nthw_field_set_val32(p->mpRcpDataTnlP, 0);
  nthw_field_set_val32(p->mpRcpDataHshValid, 0);
  nthw_field_set_val32(p->mpRcpDataHshType, 31);

  nthw_register_flush(p->mpRcpCtrl, 1);
  nthw_register_flush(p->mpRcpData, 1);

  return 0;
}

void HshNthw_RcpSelect(const struct HshNthw *p, uint32_t val)
{
  nthw_field_set_val32(p->mpRcpAddr, val);
}

void HshNthw_RcpCnt(const struct HshNthw *p, uint32_t val)
{
  nthw_field_set_val32(p->mpRcpCnt, val);
}

void HshNthw_RcpLoadDistType(const struct HshNthw *p, uint32_t val)
{
  nthw_field_set_val32(p->mpRcpDataLoadDistType, val);
}

void HshNthw_RcpMacPortMask(const struct HshNthw *p, uint32_t *val)
{
  nthw_field_set_val(p->mpRcpDataMacPortMask, val, p->mpRcpDataMacPortMask->mn_words);
}

void HshNthw_RcpSort(const struct HshNthw *p, uint32_t val)
{
  nthw_field_set_val32(p->mpRcpDataSort, val);
}

void HshNthw_RcpQw0Pe(const struct HshNthw *p, uint32_t val)
{
  nthw_field_set_val32(p->mpRcpDataQw0Pe, val);
}

void HshNthw_RcpQw0Ofs(const struct HshNthw *p, int32_t val)
{
  nthw_field_set_val32(p->mpRcpDataQw0Ofs, val);
}

void HshNthw_RcpQw4Pe(const struct HshNthw *p, uint32_t val)
{
  nthw_field_set_val32(p->mpRcpDataQw4Pe, val);
}

void HshNthw_RcpQw4Ofs(const struct HshNthw *p, int32_t val)
{
  nthw_field_set_val32(p->mpRcpDataQw4Ofs, val);
}

void HshNthw_RcpW8Pe(const struct HshNthw *p, uint32_t val)
{
  nthw_field_set_val32(p->mpRcpDataW8Pe, val);
}

void HshNthw_RcpW8Ofs(const struct HshNthw *p, int32_t val)
{
  nthw_field_set_val32(p->mpRcpDataW8Ofs, val);
}

void HshNthw_RcpW8Sort(const struct HshNthw *p, uint32_t val)
{
  nthw_field_set_val32(p->mpRcpDataW8Sort, val);
}

void HshNthw_RcpW9Pe(const struct HshNthw *p, uint32_t val)
{
  nthw_field_set_val32(p->mpRcpDataW9Pe, val);
}

void HshNthw_RcpW9Ofs(const struct HshNthw *p, int32_t val)
{
  nthw_field_set_val32(p->mpRcpDataW9Ofs, val);
}

void HshNthw_RcpW9Sort(const struct HshNthw *p, uint32_t val)
{
  nthw_field_set_val32(p->mpRcpDataW9Sort, val);
}

void HshNthw_RcpW9P(const struct HshNthw *p, uint32_t val)
{
  nthw_field_set_val32(p->mpRcpDataW9P, val);
}

void HshNthw_RcpPMask(const struct HshNthw *p, uint32_t val)
{
  nthw_field_set_val32(p->mpRcpDataPMask, val);
}

void HshNthw_RcpWordMask(const struct HshNthw *p, uint32_t *val)
{
  nthw_field_set_val(p->mpRcpDataWordMask, val, 10);
}

void HshNthw_RcpSeed(const struct HshNthw *p, uint32_t val)
{
  nthw_field_set_val32(p->mpRcpDataSeed, val);
}

void HshNthw_RcpTnlP(const struct HshNthw *p, uint32_t val)
{
  nthw_field_set_val32(p->mpRcpDataTnlP, val);
}

void HshNthw_RcpHshValid(const struct HshNthw *p, uint32_t val)
{
  nthw_field_set_val32(p->mpRcpDataHshValid, val);
}

void HshNthw_RcpHshType(const struct HshNthw *p, uint32_t val)
{
  nthw_field_set_val32(p->mpRcpDataHshType, val);
}

void HshNthw_RcpAutoIpv4Mask(const struct HshNthw *p, uint32_t val)
{
  if (p->mpRcpDataAutoIpv4Mask) {
    nthw_field_set_val32(p->mpRcpDataAutoIpv4Mask, val);
  }
}

void HshNthw_RcpFlush(const struct HshNthw *p)
{
  nthw_register_flush(p->mpRcpCtrl, 1);
  nthw_register_flush(p->mpRcpData, 1);
}
