/*
 * SPDX-License-Identifier: BSD-3-Clause
 * Copyright(c) 2023 Napatech A/S
 */

#include "ntlog.h"

#include "nthw_drv.h"
#include "nthw_register.h"

#include "flow_nthw_pdb.h"

#include <stdlib.h> /* malloc */
#include <string.h> /* memset */

void pdbNthw_setDebugMode(struct PdbNthw *p, unsigned int nDebugMode)
{
  nthw_module_set_debug_mode(p->mPdb, nDebugMode);
}

struct PdbNthw *PdbNthw_new(void)
{
  struct PdbNthw *p = malloc(sizeof(struct PdbNthw));
  if (p)
    (void)memset(p, 0, sizeof(*p));

  return p;
}

void PdbNthw_delete(struct PdbNthw *p)
{
  if (p) {
    (void)memset(p, 0, sizeof(*p));
    free(p);
  }
}

int PdbNthw_init(struct PdbNthw *p, nthw_fpga_t *pFpga, int nInstance)
{
  const char *const pAdapterIdStr = pFpga->p_fpga_info->mpAdapterIdStr;
  nthw_module_t* pMod = nthw_fpga_query_module(pFpga, MOD_PDB, nInstance);
  assert(nInstance >= 0 && nInstance < 256);

  if (p == NULL) {
    return pMod == NULL ? -1 : 0;
  }

  if (pMod == NULL) {
    NT_LOG(ERR, NTHW, "%s: Pdb %d: no such instance\n", pAdapterIdStr, nInstance);
    return -1;
  }

  p->mpFpga = pFpga;
  p->mPhysicalAdapterNo = (uint8_t)nInstance;
  p->mPdb = pMod;

#if defined(DEBUG) && (0)
  nthw_module_set_debug_mode(p->mPdb, 0xFF); /* DEBUG: */
#endif

  /* RCP */
  p->mpRcpCtrl = nthw_module_get_register(p->mPdb, PDB_RCP_CTRL);
  p->mpRcpAddr = nthw_register_get_field(p->mpRcpCtrl, PDB_RCP_CTRL_ADR);
  p->mpRcpCnt = nthw_register_get_field(p->mpRcpCtrl, PDB_RCP_CTRL_CNT);
  p->mpRcpData = nthw_module_get_register(p->mPdb, PDB_RCP_DATA);
  p->mpRcpDataDescriptor = nthw_register_get_field(p->mpRcpData, PDB_RCP_DATA_DESCRIPTOR);
  p->mpRcpDataDescLen = nthw_register_get_field(p->mpRcpData, PDB_RCP_DATA_DESC_LEN);
  p->mpRcpDataTxPort = nthw_register_get_field(p->mpRcpData, PDB_RCP_DATA_TX_PORT);
  p->mpRcpDataTxIgnore = nthw_register_get_field(p->mpRcpData, PDB_RCP_DATA_TX_IGNORE);
  p->mpRcpDataTxNow = nthw_register_get_field(p->mpRcpData, PDB_RCP_DATA_TX_NOW);
  p->mpRcpDataCrcOverwrite = nthw_register_get_field(p->mpRcpData, PDB_RCP_DATA_CRC_OVERWRITE);
  p->mpRcpDataAlign = nthw_register_get_field(p->mpRcpData, PDB_RCP_DATA_ALIGN);
  p->mpRcpDataOfs0Dyn = nthw_register_get_field(p->mpRcpData, PDB_RCP_DATA_OFS0_DYN);
  p->mpRcpDataOfs0Rel = nthw_register_get_field(p->mpRcpData, PDB_RCP_DATA_OFS0_REL);
  p->mpRcpDataOfs1Dyn = nthw_register_get_field(p->mpRcpData, PDB_RCP_DATA_OFS1_DYN);
  p->mpRcpDataOfs1Rel = nthw_register_get_field(p->mpRcpData, PDB_RCP_DATA_OFS1_REL);
  p->mpRcpDataOfs2Dyn = nthw_register_get_field(p->mpRcpData, PDB_RCP_DATA_OFS2_DYN);
  p->mpRcpDataOfs2Rel = nthw_register_get_field(p->mpRcpData, PDB_RCP_DATA_OFS2_REL);
  p->mpRcpDataIpProtTnl = nthw_register_get_field(p->mpRcpData, PDB_RCP_DATA_IP_PROT_TNL);
  p->mpRcpDataPpcHsh = nthw_register_get_field(p->mpRcpData, PDB_RCP_DATA_PPC_HSH);
  p->mpRcpDataDuplicateEn = nthw_register_get_field(p->mpRcpData, PDB_RCP_DATA_DUPLICATE_EN);
  p->mpRcpDataDuplicateBit = nthw_register_get_field(p->mpRcpData, PDB_RCP_DATA_DUPLICATE_BIT);
  p->mpRcpDataPcapKeepFcs = nthw_register_query_field(p->mpRcpData, PDB_RCP_DATA_PCAP_KEEP_FCS);
  /* CONFIG */
  p->mpConfig = nthw_module_get_register(p->mPdb, PDB_CONFIG);
  p->mpConfigTsFormat = nthw_register_get_field(p->mpConfig, PDB_CONFIG_TS_FORMAT);
  p->mpConfigPortOfs = nthw_register_get_field(p->mpConfig, PDB_CONFIG_PORT_OFS);

  return 0;
}

/* RCP */
void PdbNthw_RcpSelect(const struct PdbNthw *p, uint32_t val)
{
  nthw_field_set_val32(p->mpRcpAddr, val);
}

void PdbNthw_RcpCnt(const struct PdbNthw *p, uint32_t val)
{
  nthw_field_set_val32(p->mpRcpCnt, val);
}

void PdbNthw_RcpDescriptor(const struct PdbNthw *p, uint32_t val)
{
  nthw_field_set_val32(p->mpRcpDataDescriptor, val);
}

void PdbNthw_RcpDescLen(const struct PdbNthw *p, uint32_t val)
{
  nthw_field_set_val32(p->mpRcpDataDescLen, val);
}

void PdbNthw_RcpTxPort(const struct PdbNthw *p, uint32_t val)
{
  nthw_field_set_val32(p->mpRcpDataTxPort, val);
}

void PdbNthw_RcpTxIgnore(const struct PdbNthw *p, uint32_t val)
{
  nthw_field_set_val32(p->mpRcpDataTxIgnore, val);
}

void PdbNthw_RcpTxNow(const struct PdbNthw *p, uint32_t val)
{
  nthw_field_set_val32(p->mpRcpDataTxNow, val);
}

void PdbNthw_RcpCrcOverwrite(const struct PdbNthw *p, uint32_t val)
{
  nthw_field_set_val32(p->mpRcpDataCrcOverwrite, val);
}

void PdbNthw_RcpAlign(const struct PdbNthw *p, uint32_t val)
{
  nthw_field_set_val32(p->mpRcpDataAlign, val);
}

void PdbNthw_RcpOfs0Dyn(const struct PdbNthw *p, uint32_t val)
{
  nthw_field_set_val32(p->mpRcpDataOfs0Dyn, val);
}

void PdbNthw_RcpOfs0Rel(const struct PdbNthw *p, int32_t val)
{
  nthw_field_set_val32(p->mpRcpDataOfs0Rel, val);
}

void PdbNthw_RcpOfs1Dyn(const struct PdbNthw *p, uint32_t val)
{
  nthw_field_set_val32(p->mpRcpDataOfs1Dyn, val);
}

void PdbNthw_RcpOfs1Rel(const struct PdbNthw *p, int32_t val)
{
  nthw_field_set_val32(p->mpRcpDataOfs1Rel, val);
}

void PdbNthw_RcpOfs2Dyn(const struct PdbNthw *p, uint32_t val)
{
  nthw_field_set_val32(p->mpRcpDataOfs2Dyn, val);
}

void PdbNthw_RcpOfs2Rel(const struct PdbNthw *p, int32_t val)
{
  nthw_field_set_val32(p->mpRcpDataOfs2Rel, val);
}

void PdbNthw_RcpIpProtTnl(const struct PdbNthw *p, uint32_t val)
{
  nthw_field_set_val32(p->mpRcpDataIpProtTnl, val);
}

void PdbNthw_RcpPpcHsh(const struct PdbNthw *p, uint32_t val)
{
  nthw_field_set_val32(p->mpRcpDataPpcHsh, val);
}

void PdbNthw_RcpDuplicateEn(const struct PdbNthw *p, uint32_t val)
{
  nthw_field_set_val32(p->mpRcpDataDuplicateEn, val);
}

void PdbNthw_RcpDuplicateBit(const struct PdbNthw *p, uint32_t val)
{
  nthw_field_set_val32(p->mpRcpDataDuplicateBit, val);
}

void PdbNthw_RcpDataPcapKeepFcs(const struct PdbNthw *p, uint32_t val)
{
  if (p->mpRcpDataPcapKeepFcs) {
    nthw_field_set_val32(p->mpRcpDataPcapKeepFcs, val);
  }
}

void PdbNthw_RcpFlush(const struct PdbNthw *p)
{
  nthw_register_flush(p->mpRcpCtrl, 1);
  nthw_register_flush(p->mpRcpData, 1);
}

/* CONFIG */
void PdbNthw_ConfigTsFormat(const struct PdbNthw *p, uint32_t val)
{
  nthw_field_set_val32(p->mpConfigTsFormat, val);
}

void PdbNthw_ConfigPortOfs(const struct PdbNthw *p, uint32_t val)
{
  nthw_field_set_val32(p->mpConfigPortOfs, val);
}

void PdbNthw_ConfigFlush(const struct PdbNthw *p)
{
  nthw_register_flush(p->mpConfig, 1);
}
