/*
 * SPDX-License-Identifier: BSD-3-Clause
 * Copyright(c) 2023 Napatech A/S
 */

#include "ntlog.h"
#include "nthw_drv.h"
#include "nthw_register.h"

#include "flow_nthw_tx_rpl.h"

#include <stdlib.h>
#include <string.h>

void txRplNthw_setDebugMode(struct TxRplNthw *p, unsigned int nDebugMode)
{
  nthw_module_set_debug_mode(p->mTxRpl, nDebugMode);
}

struct TxRplNthw *TxRplNthw_new(void)
{
  struct TxRplNthw *p = malloc(sizeof(struct TxRplNthw));
  if (p) {
    (void)memset(p, 0, sizeof(*p));
  }
  return p;
}

void TxRplNthw_delete(struct TxRplNthw *p)
{
  if (p) {
    (void)memset(p, 0, sizeof(*p));
    free(p);
  }
}

int TxRplNthw_init(struct TxRplNthw *p, nthw_fpga_t *pFpga, int nInstance)
{
  const char *const pAdapterIdStr = pFpga->p_fpga_info->mpAdapterIdStr;
  nthw_module_t* pMod = nthw_fpga_query_module(pFpga, MOD_TX_RPL, nInstance);

  assert(nInstance >= 0 && nInstance < 256);

  if (p == NULL) {
    return pMod == NULL ? -1 : 0;
  }

  if (pMod == NULL) {
    NT_LOG(ERR, NTHW, "%s: TxRpl %d: no such instance\n", pAdapterIdStr, nInstance);
    return -1;
  }

  p->mpFpga = pFpga;
  p->mPhysicalAdapterNo = (uint8_t)nInstance;
  p->mTxRpl = nthw_fpga_query_module(pFpga, MOD_TX_RPL, nInstance);

  p->mpRcpCtrl = nthw_module_get_register(p->mTxRpl, RPL_RCP_CTRL);
  p->mpRcpCtrlAddr = nthw_register_get_field(p->mpRcpCtrl, RPL_RCP_CTRL_ADR);
  p->mpRcpCtrlCnt = nthw_register_get_field(p->mpRcpCtrl, RPL_RCP_CTRL_CNT);
  p->mpRcpData = nthw_module_get_register(p->mTxRpl, RPL_RCP_DATA);
  p->mpRcpDataDyn = nthw_register_get_field(p->mpRcpData, RPL_RCP_DATA_DYN);
  p->mpRcpDataOfs = nthw_register_get_field(p->mpRcpData, RPL_RCP_DATA_OFS);
  p->mpRcpDataLen = nthw_register_get_field(p->mpRcpData, RPL_RCP_DATA_LEN);
  p->mpRcpDataRplPtr = nthw_register_get_field(p->mpRcpData, RPL_RCP_DATA_RPL_PTR);
  p->mpRcpDataExtPrio = nthw_register_get_field(p->mpRcpData, RPL_RCP_DATA_EXT_PRIO);
  p->mpRcpDataEthTypeWr = nthw_register_query_field(p->mpRcpData, RPL_RCP_DATA_ETH_TYPE_WR);

  p->mpExtCtrl = nthw_module_get_register(p->mTxRpl, RPL_EXT_CTRL);
  p->mpExtCtrlAddr = nthw_register_get_field(p->mpExtCtrl, RPL_EXT_CTRL_ADR);
  p->mpExtCtrlCnt = nthw_register_get_field(p->mpExtCtrl, RPL_EXT_CTRL_CNT);
  p->mpExtData = nthw_module_get_register(p->mTxRpl, RPL_EXT_DATA);
  p->mpExtDataRplPtr = nthw_register_get_field(p->mpExtData, RPL_EXT_DATA_RPL_PTR);

  p->mpRplCtrl = nthw_module_get_register(p->mTxRpl, RPL_RPL_CTRL);
  p->mpRplCtrlAddr = nthw_register_get_field(p->mpRplCtrl, RPL_RPL_CTRL_ADR);
  p->mpRplCtrlCnt = nthw_register_get_field(p->mpRplCtrl, RPL_RPL_CTRL_CNT);
  p->mpRplData = nthw_module_get_register(p->mTxRpl, RPL_RPL_DATA);
  p->mpRplDataValue = nthw_register_get_field(p->mpRplData, RPL_RPL_DATA_VALUE);

  return 0;
}

void TxRplNthw_RcpSelect(const struct TxRplNthw *p, uint32_t val)
{
  nthw_field_set_val32(p->mpRcpCtrlAddr, val);
}

void TxRplNthw_RcpCnt(const struct TxRplNthw *p, uint32_t val)
{
  nthw_field_set_val32(p->mpRcpCtrlCnt, val);
}

void TxRplNthw_RcpDyn(const struct TxRplNthw *p, uint32_t val)
{
  nthw_field_set_val32(p->mpRcpDataDyn, val);
}

void TxRplNthw_RcpOfs(const struct TxRplNthw *p, uint32_t val)
{
  nthw_field_set_val32(p->mpRcpDataOfs, val);
}

void TxRplNthw_RcpLen(const struct TxRplNthw *p, uint32_t val)
{
  nthw_field_set_val32(p->mpRcpDataLen, val);
}

void TxRplNthw_RcpRplPtr(const struct TxRplNthw *p, uint32_t val)
{
  nthw_field_set_val32(p->mpRcpDataRplPtr, val);
}

void TxRplNthw_RcpExtPrio(const struct TxRplNthw *p, uint32_t val)
{
  nthw_field_set_val32(p->mpRcpDataExtPrio, val);
}

void TxRplNthw_RcpEthTypeWr(const struct TxRplNthw *p, uint32_t val)
{
  assert(p->mpRcpDataEthTypeWr);
  nthw_field_set_val32(p->mpRcpDataEthTypeWr, val);
}

void TxRplNthw_RcpFlush(const struct TxRplNthw *p)
{
  nthw_register_flush(p->mpRcpCtrl, 1);
  nthw_register_flush(p->mpRcpData, 1);
}

void TxRplNthw_ExtSelect(const struct TxRplNthw *p, uint32_t val)
{
  nthw_field_set_val32(p->mpExtCtrlAddr, val);
}

void TxRplNthw_ExtCnt(const struct TxRplNthw *p, uint32_t val)
{
  nthw_field_set_val32(p->mpExtCtrlCnt, val);
}

void TxRplNthw_ExtRplPtr(const struct TxRplNthw *p, uint32_t val)
{
  nthw_field_set_val32(p->mpExtDataRplPtr, val);
}

void TxRplNthw_ExtFlush(const struct TxRplNthw *p)
{
  nthw_register_flush(p->mpExtCtrl, 1);
  nthw_register_flush(p->mpExtData, 1);
}

void TxRplNthw_RplSelect(const struct TxRplNthw *p, uint32_t val)
{
  nthw_field_set_val32(p->mpRplCtrlAddr, val);
}

void TxRplNthw_RplCnt(const struct TxRplNthw *p, uint32_t val)
{
  nthw_field_set_val32(p->mpRplCtrlCnt, val);
}

void TxRplNthw_RplValue(const struct TxRplNthw *p, const uint32_t* val)
{
  nthw_field_set_val(p->mpRplDataValue, val, 4);
}

void TxRplNthw_RplFlush(const struct TxRplNthw *p)
{
  nthw_register_flush(p->mpRplCtrl, 1);
  nthw_register_flush(p->mpRplData, 1);
}
