/*
 * SPDX-License-Identifier: BSD-3-Clause
 * Copyright(c) 2023 Napatech A/S
 */

#include <errno.h>
#include "ntlog.h"

#include "nthw_drv.h"
#include "nthw_register.h"

#include "nthw_dbs.h"

#if (0)
#define DBS_PRINT_REGS
#else
#undef DBS_PRINT_REGS
#endif

static void setShadow_TxQosData(NthwDbs_t* p,
                                uint32_t index,
                                uint32_t enable,
                                uint32_t ir,
                                uint32_t bs);
static void flush_TxQosData(NthwDbs_t* p, uint32_t index);
static void setShadow_TxQpData(NthwDbs_t* p, uint32_t index, uint32_t virtualPort);
static void flush_TxQpData(NthwDbs_t* p, uint32_t index);
static void setShadow_TxDrData(NthwDbs_t* p,
                           uint32_t index,
                           uint64_t guestPhysicalAddress,
                           uint32_t hostId,
                           uint32_t queueSize,
                           uint32_t port,
                           uint32_t header,
                           uint32_t packed);
static void flush_TxDrData(NthwDbs_t* p, uint32_t index);
static void setShadow_RxDrData(NthwDbs_t* p,
                           uint32_t index,
                           uint64_t guestPhysicalAddress,
                           uint32_t hostId,
                           uint32_t queueSize,
                           uint32_t header,
                           uint32_t packed);
static void flush_RxDrData(NthwDbs_t* p, uint32_t index);
static void setShadow_TxUwData(NthwDbs_t* p,
                           uint32_t index,
                           uint64_t guestPhysicalAddress,
                           uint32_t hostId,
                           uint32_t queueSize,
                           uint32_t packed,
                           uint32_t int_enable,
                           uint32_t vec,
                           uint32_t istk,
                           uint32_t inOrder);
static void flush_TxUwData(NthwDbs_t* p, uint32_t index);
static void setShadow_RxUwData(NthwDbs_t* p,
                           uint32_t index,
                           uint64_t guestPhysicalAddress,
                           uint32_t hostId,
                           uint32_t queueSize,
                           uint32_t packed,
                           uint32_t int_enable,
                           uint32_t vec,
                           uint32_t istk);
static void flush_RxUwData(NthwDbs_t* p, uint32_t index);
static void setShadow_RxAmData(NthwDbs_t* p,
                               uint32_t index,
                               uint64_t guestPhysicalAddress,
                               uint32_t enable,
                               uint32_t hostId,
                               uint32_t packed,
                               uint32_t int_enable);
static void flush_RxAmData(NthwDbs_t* p, uint32_t index);
static void setShadow_TxAmData(NthwDbs_t* p,
                               uint32_t index,
                               uint64_t guestPhysicalAddress,
                               uint32_t enable,
                               uint32_t hostId,
                               uint32_t packed,
                               uint32_t int_enable);
static void flush_TxAmData(NthwDbs_t* p, uint32_t index);

NthwDbs_t* NthwDbs_new(void)
{
  NthwDbs_t* p = malloc(sizeof(NthwDbs_t));
  if (p) {
    memset(p, 0, sizeof(NthwDbs_t));
  }
  return p;
}

void NthwDbs_delete(NthwDbs_t* p)
{
  if (p) {
    memset(p, 0, sizeof(NthwDbs_t));
    free(p);
  }
}

int dbs_init(NthwDbs_t *p, nthw_fpga_t* pFpga, int nInstance)
{
  nthw_module_t* mod = nthw_fpga_query_module(pFpga, MOD_DBS, nInstance);

  if (p == NULL)
    return mod == NULL ? -1 : 0;

  if (mod == NULL) {
    NT_LOG(ERR, NTHW, "%s: DBS %d: no such instance\n", pFpga->p_fpga_info->mpAdapterIdStr,
        nInstance);
    return -1;
  }

  p->mpFpga = pFpga;
  p->mnInstance = nInstance;
  p->mpModDbs = mod;

#if defined(DEBUG) && (0)
  nthw_module_set_debug_mode(p->mpModDbs, 0x00); /* DEBUG: */
#endif

  p->mnParamDbsPresent = nthw_fpga_get_product_param(pFpga, NT_DBS_PRESENT, 0);
  if (p->mnParamDbsPresent == 0) {
    NT_LOG(WRN, NTHW, "%s: DBS %d: logical error: module found but not flagged at present\n", p->mpFpga->p_fpga_info->mpAdapterIdStr, p->mnInstance);
  }

  p->mpRegRxControl = nthw_module_get_register(p->mpModDbs, DBS_RX_CONTROL);
  p->mpFldRxControlLastQueue = nthw_register_get_field(p->mpRegRxControl, DBS_RX_CONTROL_LQ);
  p->mpFldRxControlAvailMonitorEnable = nthw_register_get_field(p->mpRegRxControl, DBS_RX_CONTROL_AME);
  p->mpFldRxControlAvailMonitorScanSpeed = nthw_register_get_field(p->mpRegRxControl, DBS_RX_CONTROL_AMS);
  p->mpFldRxControlUsedWriteEnable = nthw_register_get_field(p->mpRegRxControl, DBS_RX_CONTROL_UWE);
  p->mpFldRxControlUsedWriterUpdateSpeed = nthw_register_get_field(p->mpRegRxControl, DBS_RX_CONTROL_UWS);
  p->mpFldRxControlRxQueuesEnable = nthw_register_get_field(p->mpRegRxControl, DBS_RX_CONTROL_QE);

  p->mpRegTxControl = nthw_module_get_register(p->mpModDbs, DBS_TX_CONTROL);
  p->mpFldTxControlLastQueue = nthw_register_get_field(p->mpRegTxControl, DBS_TX_CONTROL_LQ);
  p->mpFldTxControlAvailMonitorEnable = nthw_register_get_field(p->mpRegTxControl, DBS_TX_CONTROL_AME);
  p->mpFldTxControlAvailMonitorScanSpeed = nthw_register_get_field(p->mpRegTxControl, DBS_TX_CONTROL_AMS);
  p->mpFldTxControlUsedWriteEnable = nthw_register_get_field(p->mpRegTxControl, DBS_TX_CONTROL_UWE);
  p->mpFldTxControlUsedWriterUpdateSpeed = nthw_register_get_field(p->mpRegTxControl, DBS_TX_CONTROL_UWS);
  p->mpFldTxControlTxQueuesEnable = nthw_register_get_field(p->mpRegTxControl, DBS_TX_CONTROL_QE);

  p->mpRegRxInit = nthw_module_get_register(p->mpModDbs, DBS_RX_INIT);
  p->mpFldRxInitInit = nthw_register_get_field(p->mpRegRxInit, DBS_RX_INIT_INIT);
  p->mpFldRxInitQueue = nthw_register_get_field(p->mpRegRxInit, DBS_RX_INIT_QUEUE);
  p->mpFldRxInitBusy = nthw_register_get_field(p->mpRegRxInit, DBS_RX_INIT_BUSY);

  p->mpRegRxInitVal = nthw_module_query_register(p->mpModDbs, DBS_RX_INIT_VAL);
  if (p->mpRegRxInitVal) {
    p->mpFldRxInitValIdx = nthw_register_query_field(p->mpRegRxInitVal, DBS_RX_INIT_VAL_IDX);
    p->mpFldRxInitValPtr = nthw_register_query_field(p->mpRegRxInitVal, DBS_RX_INIT_VAL_PTR);
  }

  p->mpRegRxPtr = nthw_module_query_register(p->mpModDbs, DBS_RX_PTR);
  if (p->mpRegRxPtr) {
    p->mpFldRxPtrPtr   = nthw_register_query_field(p->mpRegRxPtr, DBS_RX_PTR_PTR);
    p->mpFldRxPtrQueue = nthw_register_query_field(p->mpRegRxPtr, DBS_RX_PTR_QUEUE);
    p->mpFldRxPtrValid = nthw_register_query_field(p->mpRegRxPtr, DBS_RX_PTR_VALID);
  }

  p->mpRegTxInit = nthw_module_get_register(p->mpModDbs, DBS_TX_INIT);
  p->mpFldTxInitInit = nthw_register_get_field(p->mpRegTxInit, DBS_TX_INIT_INIT);
  p->mpFldTxInitQueue = nthw_register_get_field(p->mpRegTxInit, DBS_TX_INIT_QUEUE);
  p->mpFldTxInitBusy = nthw_register_get_field(p->mpRegTxInit, DBS_TX_INIT_BUSY);

  p->mpRegTxInitVal = nthw_module_query_register(p->mpModDbs, DBS_TX_INIT_VAL);
  if (p->mpRegTxInitVal) {
    p->mpFldTxInitValIdx = nthw_register_query_field(p->mpRegTxInitVal, DBS_TX_INIT_VAL_IDX);
    p->mpFldTxInitValPtr = nthw_register_query_field(p->mpRegTxInitVal, DBS_TX_INIT_VAL_PTR);
  }

  p->mpRegTxPtr = nthw_module_query_register(p->mpModDbs, DBS_TX_PTR);
  if (p->mpRegTxPtr) {
    p->mpFldTxPtrPtr   = nthw_register_query_field(p->mpRegTxPtr, DBS_TX_PTR_PTR);
    p->mpFldTxPtrQueue = nthw_register_query_field(p->mpRegTxPtr, DBS_TX_PTR_QUEUE);
    p->mpFldTxPtrValid = nthw_register_query_field(p->mpRegTxPtr, DBS_TX_PTR_VALID);
  }

  p->mpRegRxIdle = nthw_module_query_register(p->mpModDbs, DBS_RX_IDLE);
  if (p->mpRegRxIdle) {
    p->mpFldRxIdleIdle = nthw_register_query_field(p->mpRegRxIdle, DBS_RX_IDLE_IDLE);
    p->mpFldRxIdleQueue = nthw_register_query_field(p->mpRegRxIdle, DBS_RX_IDLE_QUEUE);
    p->mpFldRxIdleBusy = nthw_register_query_field(p->mpRegRxIdle, DBS_RX_IDLE_BUSY);
  }

  p->mpRegTxIdle = nthw_module_query_register(p->mpModDbs, DBS_TX_IDLE);
  if (p->mpRegTxIdle) {
    p->mpFldTxIdleIdle = nthw_register_query_field(p->mpRegTxIdle, DBS_TX_IDLE_IDLE);
    p->mpFldTxIdleQueue = nthw_register_query_field(p->mpRegTxIdle, DBS_TX_IDLE_QUEUE);
    p->mpFldTxIdleBusy = nthw_register_query_field(p->mpRegTxIdle, DBS_TX_IDLE_BUSY);
  }

  p->mpRegRxAvailMonitorControl = nthw_module_get_register(p->mpModDbs, DBS_RX_AM_CTRL);
  p->mpFldRxAvailMonitorControlAdr = nthw_register_get_field(p->mpRegRxAvailMonitorControl, DBS_RX_AM_CTRL_ADR);
  p->mpFldRxAvailMonitorControlCnt = nthw_register_get_field(p->mpRegRxAvailMonitorControl, DBS_RX_AM_CTRL_CNT);

  p->mpRegRxAvailMonitorData = nthw_module_get_register(p->mpModDbs, DBS_RX_AM_DATA);
  p->mpFldRxAvailMonitorDataGuestPhysicalAddress = nthw_register_get_field(p->mpRegRxAvailMonitorData, DBS_RX_AM_DATA_GPA);
  p->mpFldRxAvailMonitorDataEnable = nthw_register_get_field(p->mpRegRxAvailMonitorData, DBS_RX_AM_DATA_ENABLE);
  p->mpFldRxAvailMonitorDataHostID = nthw_register_get_field(p->mpRegRxAvailMonitorData, DBS_RX_AM_DATA_HID);
  p->mpFldRxAvailMonitorDataPacked = nthw_register_query_field(p->mpRegRxAvailMonitorData, DBS_RX_AM_DATA_PCKED);
  p->mpFldRxAvailMonitorDataInt = nthw_register_query_field(p->mpRegRxAvailMonitorData, DBS_RX_AM_DATA_INT);

  p->mpRegTxAvailMonitorControl = nthw_module_get_register(p->mpModDbs, DBS_TX_AM_CTRL);
  p->mpFldTxAvailMonitorControlAdr = nthw_register_get_field(p->mpRegTxAvailMonitorControl, DBS_TX_AM_CTRL_ADR);
  p->mpFldTxAvailMonitorControlCnt = nthw_register_get_field(p->mpRegTxAvailMonitorControl, DBS_TX_AM_CTRL_CNT);

  p->mpRegTxAvailMonitorData = nthw_module_get_register(p->mpModDbs, DBS_TX_AM_DATA);
  p->mpFldTxAvailMonitorDataGuestPhysicalAddress = nthw_register_get_field(p->mpRegTxAvailMonitorData, DBS_TX_AM_DATA_GPA);
  p->mpFldTxAvailMonitorDataEnable = nthw_register_get_field(p->mpRegTxAvailMonitorData, DBS_TX_AM_DATA_ENABLE);
  p->mpFldTxAvailMonitorDataHostID = nthw_register_get_field(p->mpRegTxAvailMonitorData, DBS_TX_AM_DATA_HID);
  p->mpFldTxAvailMonitorDataPacked = nthw_register_query_field(p->mpRegTxAvailMonitorData, DBS_TX_AM_DATA_PCKED);
  p->mpFldTxAvailMonitorDataInt = nthw_register_query_field(p->mpRegTxAvailMonitorData, DBS_TX_AM_DATA_INT);

  p->mpRegRxUsedWriterControl = nthw_module_get_register(p->mpModDbs, DBS_RX_UW_CTRL);
  p->mpFldRxUsedWriterControlAdr = nthw_register_get_field(p->mpRegRxUsedWriterControl, DBS_RX_UW_CTRL_ADR);
  p->mpFldRxUsedWriterControlCnt = nthw_register_get_field(p->mpRegRxUsedWriterControl, DBS_RX_UW_CTRL_CNT);

  p->mpRegRxUsedWriterData = nthw_module_get_register(p->mpModDbs, DBS_RX_UW_DATA);
  p->mpFldRxUsedWriterDataGuestPhysicalAddress = nthw_register_get_field(p->mpRegRxUsedWriterData, DBS_RX_UW_DATA_GPA);
  p->mpFldRxUsedWriterDataHostID = nthw_register_get_field(p->mpRegRxUsedWriterData, DBS_RX_UW_DATA_HID);
  p->mpFldRxUsedWriterDataQueueSize = nthw_register_get_field(p->mpRegRxUsedWriterData, DBS_RX_UW_DATA_QS);
  p->mpFldRxUsedWriterDataPacked = nthw_register_query_field(p->mpRegRxUsedWriterData, DBS_RX_UW_DATA_PCKED);
  p->mpFldRxUsedWriterDataInt = nthw_register_query_field(p->mpRegRxUsedWriterData, DBS_RX_UW_DATA_INT);
  p->mpFldRxUsedWriterDataVec = nthw_register_query_field(p->mpRegRxUsedWriterData, DBS_RX_UW_DATA_VEC);
  p->mpFldRxUsedWriterDataIstk = nthw_register_query_field(p->mpRegRxUsedWriterData, DBS_RX_UW_DATA_ISTK);

  p->mpRegTxUsedWriterControl = nthw_module_get_register(p->mpModDbs, DBS_TX_UW_CTRL);
  p->mpFldTxUsedWriterControlAdr = nthw_register_get_field(p->mpRegTxUsedWriterControl, DBS_TX_UW_CTRL_ADR);
  p->mpFldTxUsedWriterControlCnt = nthw_register_get_field(p->mpRegTxUsedWriterControl, DBS_TX_UW_CTRL_CNT);

  p->mpRegTxUsedWriterData = nthw_module_get_register(p->mpModDbs, DBS_TX_UW_DATA);
  p->mpFldTxUsedWriterDataGuestPhysicalAddress = nthw_register_get_field(p->mpRegTxUsedWriterData, DBS_TX_UW_DATA_GPA);
  p->mpFldTxUsedWriterDataHostID = nthw_register_get_field(p->mpRegTxUsedWriterData, DBS_TX_UW_DATA_HID);
  p->mpFldTxUsedWriterDataQueueSize = nthw_register_get_field(p->mpRegTxUsedWriterData, DBS_TX_UW_DATA_QS);
  p->mpFldTxUsedWriterDataPacked = nthw_register_query_field(p->mpRegTxUsedWriterData, DBS_TX_UW_DATA_PCKED);
  p->mpFldTxUsedWriterDataInt = nthw_register_query_field(p->mpRegTxUsedWriterData, DBS_TX_UW_DATA_INT);
  p->mpFldTxUsedWriterDataVec = nthw_register_query_field(p->mpRegTxUsedWriterData, DBS_TX_UW_DATA_VEC);
  p->mpFldTxUsedWriterDataIstk = nthw_register_query_field(p->mpRegTxUsedWriterData, DBS_TX_UW_DATA_ISTK);
  p->mpFldTxUsedWriterDataInOrder = nthw_register_query_field(p->mpRegTxUsedWriterData, DBS_TX_UW_DATA_INO);

  p->mpRegRxDescriptorReaderControl = nthw_module_get_register(p->mpModDbs, DBS_RX_DR_CTRL);
  p->mpFldRxDescriptorReaderControlAdr = nthw_register_get_field(p->mpRegRxDescriptorReaderControl, DBS_RX_DR_CTRL_ADR);
  p->mpFldRxDescriptorReaderControlCnt = nthw_register_get_field(p->mpRegRxDescriptorReaderControl, DBS_RX_DR_CTRL_CNT);

  p->mpRegRxDescriptorReaderData = nthw_module_get_register(p->mpModDbs, DBS_RX_DR_DATA);
  p->mpFldRxDescriptorReaderDataGuestPhysicalAddress = nthw_register_get_field(p->mpRegRxDescriptorReaderData, DBS_RX_DR_DATA_GPA);
  p->mpFldRxDescriptorReaderDataHostID = nthw_register_get_field(p->mpRegRxDescriptorReaderData, DBS_RX_DR_DATA_HID);
  p->mpFldRxDescriptorReaderDataQueueSize = nthw_register_get_field(p->mpRegRxDescriptorReaderData, DBS_RX_DR_DATA_QS);
  p->mpFldRxDescriptorReaderDataHeader = nthw_register_get_field(p->mpRegRxDescriptorReaderData, DBS_RX_DR_DATA_HDR);
  p->mpFldRxDescriptorReaderDataPacked = nthw_register_query_field(p->mpRegRxDescriptorReaderData, DBS_RX_DR_DATA_PCKED);

  p->mpRegTxDescriptorReaderControl = nthw_module_get_register(p->mpModDbs, DBS_TX_DR_CTRL);
  p->mpFldTxDescriptorReaderControlAdr = nthw_register_get_field(p->mpRegTxDescriptorReaderControl, DBS_TX_DR_CTRL_ADR);
  p->mpFldTxDescriptorReaderControlCnt = nthw_register_get_field(p->mpRegTxDescriptorReaderControl, DBS_TX_DR_CTRL_CNT);

  p->mpRegTxDescriptorReaderData = nthw_module_get_register(p->mpModDbs, DBS_TX_DR_DATA);
  p->mpFldTxDescriptorReaderDataGuestPhysicalAddress = nthw_register_get_field(p->mpRegTxDescriptorReaderData, DBS_TX_DR_DATA_GPA);
  p->mpFldTxDescriptorReaderDataHostID = nthw_register_get_field(p->mpRegTxDescriptorReaderData, DBS_TX_DR_DATA_HID);
  p->mpFldTxDescriptorReaderDataQueueSize = nthw_register_get_field(p->mpRegTxDescriptorReaderData, DBS_TX_DR_DATA_QS);
  p->mpFldTxDescriptorReaderDataHeader = nthw_register_get_field(p->mpRegTxDescriptorReaderData, DBS_TX_DR_DATA_HDR);
  p->mpFldTxDescriptorReaderDataPort = nthw_register_get_field(p->mpRegTxDescriptorReaderData, DBS_TX_DR_DATA_PORT);
  p->mpFldTxDescriptorReaderDataPacked = nthw_register_query_field(p->mpRegTxDescriptorReaderData, DBS_TX_DR_DATA_PCKED);

  p->mpRegTxQueuePropertyControl = nthw_module_get_register(p->mpModDbs, DBS_TX_QP_CTRL);
  p->mpFldTxQueuePropertyControlAdr = nthw_register_get_field(p->mpRegTxQueuePropertyControl, DBS_TX_QP_CTRL_ADR);
  p->mpFldTxQueuePropertyControlCnt = nthw_register_get_field(p->mpRegTxQueuePropertyControl, DBS_TX_QP_CTRL_CNT);

  p->mpRegTxQueuePropertyData = nthw_module_get_register(p->mpModDbs, DBS_TX_QP_DATA);
  p->mpFldTxQueuePropertyDataVPort = nthw_register_get_field(p->mpRegTxQueuePropertyData, DBS_TX_QP_DATA_VPORT);

    /* HW QoS Tx rate limiting policing RFC2697/RFC4111 */
  p->mpRegTxQueueQosControl = nthw_module_query_register(p->mpModDbs, DBS_TX_QOS_CTRL);
  p->mpRegTxQueueQosData = nthw_module_query_register(p->mpModDbs, DBS_TX_QOS_DATA);
  if (p->mpRegTxQueueQosControl) {
    p->mpRegTxQueueQosControlAdr = nthw_register_query_field(p->mpRegTxQueueQosControl, DBS_TX_QOS_CTRL_ADR);
    p->mpRegTxQueueQosControlCnt = nthw_register_query_field(p->mpRegTxQueueQosControl, DBS_TX_QOS_CTRL_CNT);

    if (p->mpRegTxQueueQosData) {
      p->mpRegTxQueueQosDataEn = nthw_register_query_field(p->mpRegTxQueueQosData, DBS_TX_QOS_DATA_EN);
      p->mpRegTxQueueQosDataIR = nthw_register_query_field(p->mpRegTxQueueQosData, DBS_TX_QOS_DATA_IR);
      p->mpRegTxQueueQosDataBS = nthw_register_query_field(p->mpRegTxQueueQosData, DBS_TX_QOS_DATA_BS);
    }
  }

  p->mpRegTxQueueQosRate = nthw_module_query_register(p->mpModDbs, DBS_TX_QOS_RATE);
  if (p->mpRegTxQueueQosRate) {
    p->mpRegTxQueueQosRateMul = nthw_register_query_field(p->mpRegTxQueueQosRate, DBS_TX_QOS_RATE_MUL);
    p->mpRegTxQueueQosRateDiv = nthw_register_query_field(p->mpRegTxQueueQosRate, DBS_TX_QOS_RATE_DIV);
  }

  return 0;
}

int dbs_reset_rx_control(NthwDbs_t* p)
{
  nthw_field_set_val32(p->mpFldRxControlLastQueue, 0);
  nthw_field_set_val32(p->mpFldRxControlAvailMonitorEnable, 0);
  nthw_field_set_val32(p->mpFldRxControlAvailMonitorScanSpeed, 8);
  nthw_field_set_val32(p->mpFldRxControlUsedWriteEnable, 0);
  nthw_field_set_val32(p->mpFldRxControlUsedWriterUpdateSpeed, 5);
  nthw_field_set_val32(p->mpFldRxControlRxQueuesEnable, 0);
  nthw_register_flush(p->mpRegRxControl, 1);
  return 0;
}

int dbs_reset_tx_control(NthwDbs_t* p)
{
  nthw_field_set_val32(p->mpFldTxControlLastQueue, 0);
  nthw_field_set_val32(p->mpFldTxControlAvailMonitorEnable, 0);
  nthw_field_set_val32(p->mpFldTxControlAvailMonitorScanSpeed, 5);
  nthw_field_set_val32(p->mpFldTxControlUsedWriteEnable, 0);
  nthw_field_set_val32(p->mpFldTxControlUsedWriterUpdateSpeed, 8);
  nthw_field_set_val32(p->mpFldTxControlTxQueuesEnable, 0);
  nthw_register_flush(p->mpRegTxControl, 1);
  return 0;
}

void dbs_reset(NthwDbs_t* p)
{
  uint32_t i;

  NT_LOG(DBG, NTHW, "NthwDbs::%s: resetting DBS", __func__);

  dbs_reset_rx_control(p);
  dbs_reset_tx_control(p);

  /* Reset RX memory banks and shado */
  for (i = 0; i < NT_DBS_RX_QUEUES_MAX; ++i) {

    setShadow_RxAmData(p, i, 0, 0, 0, 0, 0);
    flush_RxAmData(p, i);

    setShadow_RxUwData(p, i, 0, 0, 0, 0, 0, 0, 0);
    flush_RxUwData(p, i);

    setShadow_RxDrData(p, i, 0, 0, 0, 0, 0);
    flush_RxDrData(p, i);
  }

  /* Reset TX memory banks and shado */
  for (i = 0; i < NT_DBS_TX_QUEUES_MAX; ++i) {
    setShadow_TxAmData(p, i, 0, 0, 0, 0, 0);
    flush_TxAmData(p, i);

    setShadow_TxUwData(p, i, 0, 0, 0, 0, 0, 0, 0, 0);
    flush_TxUwData(p, i);

    setShadow_TxDrData(p, i, 0, 0, 0, 0, 0, 0);
    flush_TxDrData(p, i);

    setShadow_TxQpData(p, i, 0);
    flush_TxQpData(p, i);

    setShadow_TxQosData(p, i, 0, 0, 0);
    flush_TxQosData(p, i);
  }
}

int set_rx_control(NthwDbs_t* p,
                         uint32_t lastQueue,
                         uint32_t availMonitorEnable,
                         uint32_t availMonitorSpeed,
                         uint32_t usedWriteEnable,
                         uint32_t usedWriteSpeed,
                         uint32_t rxQueueEnable)
{
#ifdef DBS_PRINT_REGS
  printf("lastQueue %u\n", lastQueue);
  printf("availMonitorEnable %u\n", availMonitorEnable);
  printf("availMonitorSpeed %u\n", availMonitorSpeed);
  printf("usedWriteEnable %u\n", usedWriteEnable);
  printf("usedWriteSpeed %u\n", usedWriteSpeed);
  printf("rxQueueEnable %u\n", rxQueueEnable);
#endif

  nthw_field_set_val32(p->mpFldRxControlLastQueue, lastQueue);
  nthw_field_set_val32(p->mpFldRxControlAvailMonitorEnable, availMonitorEnable);
  nthw_field_set_val32(p->mpFldRxControlAvailMonitorScanSpeed, availMonitorSpeed);
  nthw_field_set_val32(p->mpFldRxControlUsedWriteEnable, usedWriteEnable);
  nthw_field_set_val32(p->mpFldRxControlUsedWriterUpdateSpeed, usedWriteSpeed);
  nthw_field_set_val32(p->mpFldRxControlRxQueuesEnable, rxQueueEnable);
  nthw_register_flush(p->mpRegRxControl, 1);
  return 0;
}

int NthwDbs_getRxControl(NthwDbs_t* p,
                         uint32_t *lastQueue,
                         uint32_t *availMonitorEnable,
                         uint32_t *availMonitorSpeed,
                         uint32_t *usedWriteEnable,
                         uint32_t *usedWriteSpeed,
                         uint32_t *rxQueueEnable)
{
  *lastQueue = nthw_field_get_val32(p->mpFldRxControlLastQueue);
  *availMonitorEnable = nthw_field_get_val32(p->mpFldRxControlAvailMonitorEnable);
  *availMonitorSpeed = nthw_field_get_val32(p->mpFldRxControlAvailMonitorScanSpeed);
  *usedWriteEnable = nthw_field_get_val32(p->mpFldRxControlUsedWriteEnable);
  *usedWriteSpeed = nthw_field_get_val32(p->mpFldRxControlUsedWriterUpdateSpeed);
  *rxQueueEnable = nthw_field_get_val32(p->mpFldRxControlRxQueuesEnable);
  return 0;
}

int set_tx_control(NthwDbs_t* p,
                         uint32_t lastQueue,
                         uint32_t availMonitorEnable,
                         uint32_t availMonitorSpeed,
                         uint32_t usedWriteEnable,
                         uint32_t usedWriteSpeed,
                         uint32_t txQueueEnable)
{
#ifdef DBS_PRINT_REGS
  printf("lastQueue %u\n", lastQueue);
  printf("availMonitorEnable %u\n", availMonitorEnable);
  printf("availMonitorSpeed %u\n", availMonitorSpeed);
  printf("usedWriteEnable %u\n", usedWriteEnable);
  printf("usedWriteSpeed %u\n", usedWriteSpeed);
#endif

  nthw_field_set_val32(p->mpFldTxControlLastQueue, lastQueue);
  nthw_field_set_val32(p->mpFldTxControlAvailMonitorEnable, availMonitorEnable);
  nthw_field_set_val32(p->mpFldTxControlAvailMonitorScanSpeed, availMonitorSpeed);
  nthw_field_set_val32(p->mpFldTxControlUsedWriteEnable, usedWriteEnable);
  nthw_field_set_val32(p->mpFldTxControlUsedWriterUpdateSpeed, usedWriteSpeed);
  nthw_field_set_val32(p->mpFldTxControlTxQueuesEnable, txQueueEnable);
  nthw_register_flush(p->mpRegTxControl, 1);
  return 0;
}

int NthwDbs_getTxControl(NthwDbs_t* p,
                         uint32_t *lastQueue,
                         uint32_t *availMonitorEnable,
                         uint32_t *availMonitorSpeed,
                         uint32_t *usedWriteEnable,
                         uint32_t *usedWriteSpeed,
                         uint32_t *txQueueEnable)
{
  *lastQueue = nthw_field_get_val32(p->mpFldTxControlLastQueue);
  *availMonitorEnable = nthw_field_get_val32(p->mpFldTxControlAvailMonitorEnable);
  *availMonitorSpeed = nthw_field_get_val32(p->mpFldTxControlAvailMonitorScanSpeed);
  *usedWriteEnable = nthw_field_get_val32(p->mpFldTxControlUsedWriteEnable);
  *usedWriteSpeed = nthw_field_get_val32(p->mpFldTxControlUsedWriterUpdateSpeed);
  *txQueueEnable = nthw_field_get_val32(p->mpFldTxControlTxQueuesEnable);
  return 0;
}


int set_rx_init(NthwDbs_t* p,
                      uint32_t startIdx,
                      uint32_t startPtr,
                      uint32_t init,
                      uint32_t queue)
{
  if (p->mpRegRxInitVal) {
    nthw_field_set_val32(p->mpFldRxInitValIdx, startIdx);
    nthw_field_set_val32(p->mpFldRxInitValPtr, startPtr);
#if 0
    printf("Set RxVq init [%i]: idx,ptr %08x,%08x\n", queue, startIdx, startPtr);
#endif
    nthw_register_flush(p->mpRegRxInitVal, 1);
  }
  nthw_field_set_val32(p->mpFldRxInitInit, init);
  nthw_field_set_val32(p->mpFldRxInitQueue, queue);
  nthw_register_flush(p->mpRegRxInit, 1);
  return 0;
}

int get_rx_init(NthwDbs_t* p,
                      uint32_t *init,
                      uint32_t *queue,
                      uint32_t *busy)
{
  *init = nthw_field_get_val32(p->mpFldRxInitInit);
  *queue = nthw_field_get_val32(p->mpFldRxInitQueue);
  *busy = nthw_field_get_val32(p->mpFldRxInitBusy);
  return 0;
}

int set_tx_init(NthwDbs_t* p,
                      uint32_t startIdx,
                      uint32_t startPtr,
                      uint32_t init,
                      uint32_t queue)
{
  if (p->mpRegTxInitVal) {
    nthw_field_set_val32(p->mpFldTxInitValIdx, startIdx);
    nthw_field_set_val32(p->mpFldTxInitValPtr, startPtr);
#if 0
    printf("Set TxVq init [%i]: idx,ptr %08x,%08x\n",queue, startIdx, startPtr);
#endif
    nthw_register_flush(p->mpRegTxInitVal, 1);
  }
  nthw_field_set_val32(p->mpFldTxInitInit, init);
  nthw_field_set_val32(p->mpFldTxInitQueue, queue);
  nthw_register_flush(p->mpRegTxInit, 1);
  return 0;
}

int get_tx_init(NthwDbs_t* p,
                      uint32_t *init,
                      uint32_t *queue,
                      uint32_t *busy)
{
  *init = nthw_field_get_val32(p->mpFldTxInitInit);
  *queue = nthw_field_get_val32(p->mpFldTxInitQueue);
  *busy = nthw_field_get_val32(p->mpFldTxInitBusy);
  return 0;
}

int set_rx_idle(NthwDbs_t* p,
                      uint32_t idle,
                      uint32_t queue)

{
  if (!p->mpRegRxIdle)
    return -ENOTSUP;

  nthw_field_set_val32(p->mpFldRxIdleIdle, idle);
  nthw_field_set_val32(p->mpFldRxIdleQueue, queue);
  nthw_register_flush(p->mpRegRxIdle, 1);
  return 0;
}

int get_rx_idle(NthwDbs_t* p,
                      uint32_t *idle,
                      uint32_t *queue,
                      uint32_t *busy)
{
  if (!p->mpRegRxIdle)
    return -ENOTSUP;

  *idle =  nthw_field_get_updated(p->mpFldRxIdleIdle);
  *queue =  0;
  *busy = nthw_field_get_updated(p->mpFldRxIdleBusy);
  return 0;
}

int set_tx_idle(NthwDbs_t* p,
                      uint32_t idle,
                      uint32_t queue)

{
  if (!p->mpRegTxIdle)
    return -ENOTSUP;

#if 0
  printf("set idle Tx, queue %i\n", queue);
#endif
  nthw_field_set_val32(p->mpFldTxIdleIdle, idle);
  nthw_field_set_val32(p->mpFldTxIdleQueue, queue);
  nthw_register_flush(p->mpRegTxIdle, 1);
  return 0;
}

int get_tx_idle(NthwDbs_t* p,
                      uint32_t *idle,
                      uint32_t *queue,
                      uint32_t *busy)
{
  if (!p->mpRegTxIdle)
    return -ENOTSUP;

  *idle =  nthw_field_get_updated(p->mpFldTxIdleIdle);
  *queue =  0;
  *busy = nthw_field_get_updated(p->mpFldTxIdleBusy);
#if 0
  printf("get idle Tx queue %i\n", nthw_field_get_updated(p->mpFldTxIdleQueue));
#endif
  return 0;
}

int set_rx_ptr_queue(NthwDbs_t* p, uint32_t queue)
{
  if (!p->mpRegRxPtr)
    return -ENOTSUP;

  nthw_field_set_val32(p->mpFldRxPtrQueue, queue);
  nthw_register_flush(p->mpRegRxPtr, 1);
  return 0;
}

int get_rx_ptr(NthwDbs_t* p,
                     uint32_t *ptr,
                     uint32_t *queue,
                     uint32_t *valid)
{
  if (!p->mpRegRxPtr)
    return -ENOTSUP;

  *ptr = nthw_field_get_updated(p->mpFldRxPtrPtr);
  *queue =  0;
  *valid = nthw_field_get_updated(p->mpFldRxPtrValid);
  return 0;
}

int set_tx_ptr_queue(NthwDbs_t* p, uint32_t queue)
{
  if (!p->mpRegTxPtr)
    return -ENOTSUP;

  nthw_field_set_val32(p->mpFldTxPtrQueue, queue);
  nthw_register_flush(p->mpRegTxPtr, 1);
  return 0;
}

int get_tx_ptr(NthwDbs_t* p,
                     uint32_t *ptr,
                     uint32_t *queue,
                     uint32_t *valid)
{
  if (!p->mpRegTxPtr)
    return -ENOTSUP;

  *ptr =  nthw_field_get_updated(p->mpFldTxPtrPtr);
  *queue =  0;
  *valid = nthw_field_get_updated(p->mpFldTxPtrValid);
  return 0;
}

static void setRxAmDataIndex(NthwDbs_t* p, uint32_t index)
{
  nthw_field_set_val32(p->mpFldRxAvailMonitorControlAdr, index);
  nthw_field_set_val32(p->mpFldRxAvailMonitorControlCnt, 1);
  nthw_register_flush(p->mpRegRxAvailMonitorControl, 1);
}

static void setShadow_RxAmDataGuestPhysicalAddress(NthwDbs_t* p, uint32_t index, uint64_t guestPhysicalAddress)
{
  p->mRxAmShadow[index].guestPhysicalAddress = guestPhysicalAddress;
}

static void NthwDbs_setShadow_RxAmDataEnable(NthwDbs_t* p, uint32_t index, uint32_t enable)
{
  p->mRxAmShadow[index].enable = enable;
}

static void setShadow_RxAmDataHostId(NthwDbs_t* p, uint32_t index, uint32_t hostId)
{
  p->mRxAmShadow[index].hostId = hostId;
}

static void setShadow_RxAmDataPacked(NthwDbs_t* p, uint32_t index, uint32_t packed)
{
  p->mRxAmShadow[index].packed = packed;
}

static void setShadow_RxAmDataIntEnable(NthwDbs_t* p, uint32_t index, uint32_t int_enable)
{
  p->mRxAmShadow[index].int_enable = int_enable;
}

static void setShadow_RxAmData(NthwDbs_t* p,
                               uint32_t index,
                               uint64_t guestPhysicalAddress,
                               uint32_t enable,
                               uint32_t hostId,
                               uint32_t packed,
                               uint32_t int_enable)
{
  setShadow_RxAmDataGuestPhysicalAddress(p, index, guestPhysicalAddress);
  NthwDbs_setShadow_RxAmDataEnable(p, index, enable);
  setShadow_RxAmDataHostId(p, index, hostId);
  setShadow_RxAmDataPacked(p, index, packed);
  setShadow_RxAmDataIntEnable(p, index, int_enable);
}

static void flush_RxAmData(NthwDbs_t* p, uint32_t index)
{
  nthw_field_set_val(p->mpFldRxAvailMonitorDataGuestPhysicalAddress, (uint32_t*)&p->mRxAmShadow[index].guestPhysicalAddress, 2);
  nthw_field_set_val32(p->mpFldRxAvailMonitorDataEnable, p->mRxAmShadow[index].enable);
  nthw_field_set_val32(p->mpFldRxAvailMonitorDataHostID, p->mRxAmShadow[index].hostId);
  if (p->mpFldRxAvailMonitorDataPacked) {
    nthw_field_set_val32(p->mpFldRxAvailMonitorDataPacked, p->mRxAmShadow[index].packed);
  }
  if (p->mpFldRxAvailMonitorDataInt) {
    nthw_field_set_val32(p->mpFldRxAvailMonitorDataInt, p->mRxAmShadow[index].int_enable);
  }

  setRxAmDataIndex(p, index);
  nthw_register_flush(p->mpRegRxAvailMonitorData, 1);
}

int set_rx_am_data(NthwDbs_t* p,
                        uint32_t index,
                        uint64_t guestPhysicalAddress,
                        uint32_t enable,
                        uint32_t hostId,
                        uint32_t packed,
                        uint32_t int_enable)
{
  if (!p->mpRegRxAvailMonitorData)
    return -ENOTSUP;

  setShadow_RxAmData(p, index, guestPhysicalAddress, enable, hostId, packed, int_enable);
  flush_RxAmData(p, index);
  return 0;
}

static void setTxAmDataIndex(NthwDbs_t* p, uint32_t index)
{
  nthw_field_set_val32(p->mpFldTxAvailMonitorControlAdr, index);
  nthw_field_set_val32(p->mpFldTxAvailMonitorControlCnt, 1);
  nthw_register_flush(p->mpRegTxAvailMonitorControl, 1);
}

static void setShadow_TxAmData(NthwDbs_t* p,
                               uint32_t index,
                               uint64_t guestPhysicalAddress,
                               uint32_t enable,
                               uint32_t hostId,
                               uint32_t packed,
                               uint32_t int_enable)
{
  p->mTxAmShadow[index].guestPhysicalAddress = guestPhysicalAddress;
  p->mTxAmShadow[index].enable = enable;
  p->mTxAmShadow[index].hostId = hostId;
  p->mTxAmShadow[index].packed = packed;
  p->mTxAmShadow[index].int_enable = int_enable;
}

static void flush_TxAmData(NthwDbs_t* p, uint32_t index)
{
  nthw_field_set_val(p->mpFldTxAvailMonitorDataGuestPhysicalAddress, (uint32_t*)&p->mTxAmShadow[index].guestPhysicalAddress, 2);
  nthw_field_set_val32(p->mpFldTxAvailMonitorDataEnable, p->mTxAmShadow[index].enable);
  nthw_field_set_val32(p->mpFldTxAvailMonitorDataHostID, p->mTxAmShadow[index].hostId);
  if (p->mpFldTxAvailMonitorDataPacked) {
    nthw_field_set_val32(p->mpFldTxAvailMonitorDataPacked, p->mTxAmShadow[index].packed);
  }
  if (p->mpFldTxAvailMonitorDataInt) {
    nthw_field_set_val32(p->mpFldTxAvailMonitorDataInt, p->mTxAmShadow[index].int_enable);
  }

  setTxAmDataIndex(p, index);
  nthw_register_flush(p->mpRegTxAvailMonitorData, 1);
}

int set_tx_am_data(NthwDbs_t* p,
                        uint32_t index,
                        uint64_t guestPhysicalAddress,
                        uint32_t enable,
                        uint32_t hostId,
                        uint32_t packed,
                        uint32_t int_enable)
{
  if (!p->mpRegTxAvailMonitorData)
    return -ENOTSUP;

  setShadow_TxAmData(p, index, guestPhysicalAddress, enable, hostId, packed, int_enable);
  flush_TxAmData(p, index);
  return 0;
}

static void setRxUwDataIndex(NthwDbs_t* p, uint32_t index)
{
  nthw_field_set_val32(p->mpFldRxUsedWriterControlAdr, index);
  nthw_field_set_val32(p->mpFldRxUsedWriterControlCnt, 1);
  nthw_register_flush(p->mpRegRxUsedWriterControl, 1);
}

static void setShadow_RxUwDataGuestPhysicalAddress(NthwDbs_t* p, uint32_t index, uint64_t guestPhysicalAddress)
{
  p->mRxUwShadow[index].guestPhysicalAddress = guestPhysicalAddress;
}

static void setShadow_RxUwDataHostId(NthwDbs_t* p, uint32_t index, uint32_t hostId)
{
  p->mRxUwShadow[index].hostId = hostId;
}

static void setShadow_RxUwDataQueueSize(NthwDbs_t* p, uint32_t index, uint32_t queueSize)
{
  p->mRxUwShadow[index].queueSize = queueSize;
}

static void setShadow_RxUwDataPacked(NthwDbs_t* p, uint32_t index, uint32_t packed)
{
  p->mRxUwShadow[index].packed = packed;
}

static void setShadow_RxUwDataIntEnable(NthwDbs_t* p, uint32_t index, uint32_t int_enable)
{
  p->mRxUwShadow[index].int_enable = int_enable;
}

static void setShadow_RxUwDataVec(NthwDbs_t* p, uint32_t index, uint32_t vec)
{
  p->mRxUwShadow[index].vec = vec;
}

static void setShadow_RxUwDataIstk(NthwDbs_t* p, uint32_t index, uint32_t istk)
{
  p->mRxUwShadow[index].istk = istk;
}

static void setShadow_RxUwData(NthwDbs_t* p,
                           uint32_t index,
                           uint64_t guestPhysicalAddress,
                           uint32_t hostId,
                           uint32_t queueSize,
                           uint32_t packed,
                           uint32_t int_enable,
                           uint32_t vec,
                           uint32_t istk)
{
  setShadow_RxUwDataGuestPhysicalAddress(p, index, guestPhysicalAddress);
  setShadow_RxUwDataHostId(p, index, hostId);
  setShadow_RxUwDataQueueSize(p, index, queueSize);
  setShadow_RxUwDataPacked(p, index, packed);
  setShadow_RxUwDataIntEnable(p, index, int_enable);
  setShadow_RxUwDataVec(p, index, vec);
  setShadow_RxUwDataIstk(p, index, istk);
}

static void flush_RxUwData(NthwDbs_t* p, uint32_t index)
{
  nthw_field_set_val(p->mpFldRxUsedWriterDataGuestPhysicalAddress, (uint32_t*)&p->mRxUwShadow[index].guestPhysicalAddress, 2);
  nthw_field_set_val32(p->mpFldRxUsedWriterDataHostID, p->mRxUwShadow[index].hostId);
  if (nthw_module_is_version_newer(p->mpModDbs, 0, 8)) {
    nthw_field_set_val32(p->mpFldRxUsedWriterDataQueueSize, (1U << p->mRxUwShadow[index].queueSize) - 1U);
  } else {
    nthw_field_set_val32(p->mpFldRxUsedWriterDataQueueSize, p->mRxUwShadow[index].queueSize);
  }
  if (p->mpFldRxUsedWriterDataPacked) {
    nthw_field_set_val32(p->mpFldRxUsedWriterDataPacked, p->mRxUwShadow[index].packed);
  }
  if (p->mpFldRxUsedWriterDataInt) {
    nthw_field_set_val32(p->mpFldRxUsedWriterDataInt, p->mRxUwShadow[index].int_enable);
    nthw_field_set_val32(p->mpFldRxUsedWriterDataVec, p->mRxUwShadow[index].vec);
    nthw_field_set_val32(p->mpFldRxUsedWriterDataIstk, p->mRxUwShadow[index].istk);
  }

  setRxUwDataIndex(p, index);
  nthw_register_flush(p->mpRegRxUsedWriterData, 1);
}

int set_rx_uw_data(NthwDbs_t* p,
                        uint32_t index,
                        uint64_t guestPhysicalAddress,
                        uint32_t hostId,
                        uint32_t queueSize,
                        uint32_t packed,
                        uint32_t int_enable,
                        uint32_t vec,
                        uint32_t istk)
{
  if (!p->mpRegRxUsedWriterData)
    return -ENOTSUP;

  setShadow_RxUwData(p, index, guestPhysicalAddress, hostId, queueSize, packed, int_enable, vec, istk);
  flush_RxUwData(p, index);
  return 0;
}

static void setTxUwDataIndex(NthwDbs_t* p, uint32_t index)
{
  nthw_field_set_val32(p->mpFldTxUsedWriterControlAdr, index);
  nthw_field_set_val32(p->mpFldTxUsedWriterControlCnt, 1);
  nthw_register_flush(p->mpRegTxUsedWriterControl, 1);
}

static void setShadow_TxUwDataGuestPhysicalAddress(NthwDbs_t* p, uint32_t index, uint64_t guestPhysicalAddress)
{
  p->mTxUwShadow[index].guestPhysicalAddress = guestPhysicalAddress;
}

static void setShadow_TxUwDataHostId(NthwDbs_t* p, uint32_t index, uint32_t hostId)
{
  p->mTxUwShadow[index].hostId = hostId;
}

static void setShadow_TxUwDataQueueSize(NthwDbs_t* p, uint32_t index, uint32_t queueSize)
{
  p->mTxUwShadow[index].queueSize = queueSize;
}

static void setShadow_TxUwDataPacked(NthwDbs_t* p, uint32_t index, uint32_t packed)
{
  p->mTxUwShadow[index].packed = packed;
}

static void setShadow_TxUwDataIntEnable(NthwDbs_t* p, uint32_t index, uint32_t int_enable)
{
  p->mTxUwShadow[index].int_enable = int_enable;
}

static void setShadow_TxUwDataVec(NthwDbs_t* p, uint32_t index, uint32_t vec)
{
  p->mTxUwShadow[index].vec = vec;
}

static void setShadow_TxUwDataIstk(NthwDbs_t* p, uint32_t index, uint32_t istk)
{
  p->mTxUwShadow[index].istk = istk;
}

static void setShadow_TxUwDataInOrder(NthwDbs_t* p, uint32_t index, uint32_t inOrder)
{
  p->mTxUwShadow[index].inOrder = inOrder;
}

static void setShadow_TxUwData(NthwDbs_t* p,
                           uint32_t index,
                           uint64_t guestPhysicalAddress,
                           uint32_t hostId,
                           uint32_t queueSize,
                           uint32_t packed,
                           uint32_t int_enable,
                           uint32_t vec,
                           uint32_t istk,
                           uint32_t inOrder)
{
  setShadow_TxUwDataGuestPhysicalAddress(p, index, guestPhysicalAddress);
  setShadow_TxUwDataHostId(p, index, hostId);
  setShadow_TxUwDataQueueSize(p, index, queueSize);
  setShadow_TxUwDataPacked(p, index, packed);
  setShadow_TxUwDataIntEnable(p, index, int_enable);
  setShadow_TxUwDataVec(p, index, vec);
  setShadow_TxUwDataIstk(p, index, istk);
  setShadow_TxUwDataInOrder(p, index, inOrder);
}

static void flush_TxUwData(NthwDbs_t* p, uint32_t index)
{
  nthw_field_set_val(p->mpFldTxUsedWriterDataGuestPhysicalAddress, (uint32_t*)&p->mTxUwShadow[index].guestPhysicalAddress, 2);
  nthw_field_set_val32(p->mpFldTxUsedWriterDataHostID, p->mTxUwShadow[index].hostId);
  if (nthw_module_is_version_newer(p->mpModDbs, 0,8)) {
    nthw_field_set_val32(p->mpFldTxUsedWriterDataQueueSize, (1U << p->mTxUwShadow[index].queueSize) - 1U);
  } else {
    nthw_field_set_val32(p->mpFldTxUsedWriterDataQueueSize, p->mTxUwShadow[index].queueSize);
  }
  if (p->mpFldTxUsedWriterDataPacked) {
    nthw_field_set_val32(p->mpFldTxUsedWriterDataPacked, p->mTxUwShadow[index].packed);
  }
  if (p->mpFldTxUsedWriterDataInt) {
    nthw_field_set_val32(p->mpFldTxUsedWriterDataInt, p->mTxUwShadow[index].int_enable);
    nthw_field_set_val32(p->mpFldTxUsedWriterDataVec, p->mTxUwShadow[index].vec);
    nthw_field_set_val32(p->mpFldTxUsedWriterDataIstk, p->mTxUwShadow[index].istk);
  }
  if (p->mpFldTxUsedWriterDataInOrder) {
    nthw_field_set_val32(p->mpFldTxUsedWriterDataInOrder, p->mTxUwShadow[index].inOrder);
  }

  setTxUwDataIndex(p, index);
  nthw_register_flush(p->mpRegTxUsedWriterData, 1);
}

int set_tx_uw_data(NthwDbs_t* p,
                        uint32_t index,
                        uint64_t guestPhysicalAddress,
                        uint32_t hostId,
                        uint32_t queueSize,
                        uint32_t packed,
                        uint32_t int_enable,
                        uint32_t vec,
                        uint32_t istk,
                        uint32_t inOrder)
{
  if (!p->mpRegTxUsedWriterData)
    return -ENOTSUP;

  setShadow_TxUwData(p, index, guestPhysicalAddress, hostId, queueSize, packed, int_enable, vec, istk, inOrder);
  flush_TxUwData(p, index);
  return 0;
}

static void setRxDrDataIndex(NthwDbs_t* p, uint32_t index)
{
  nthw_field_set_val32(p->mpFldRxDescriptorReaderControlAdr, index);
  nthw_field_set_val32(p->mpFldRxDescriptorReaderControlCnt, 1);
  nthw_register_flush(p->mpRegRxDescriptorReaderControl, 1);
}

static void setShadow_RxDrDataGuestPhysicalAddress(NthwDbs_t* p, uint32_t index, uint64_t guestPhysicalAddress)
{
  p->mRxDrShadow[index].guestPhysicalAddress = guestPhysicalAddress;
}

static void setShadow_RxDrDataHostId(NthwDbs_t* p, uint32_t index, uint32_t hostId)
{
  p->mRxDrShadow[index].hostId = hostId;
}

static void setShadow_RxDrDataQueueSize(NthwDbs_t* p, uint32_t index, uint32_t queueSize)
{
  p->mRxDrShadow[index].queueSize = queueSize;
}

static void setShadow_RxDrDataHeader(NthwDbs_t* p, uint32_t index, uint32_t header)
{
  p->mRxDrShadow[index].header = header;
}

static void setShadow_RxDrDataPacked(NthwDbs_t* p, uint32_t index, uint32_t packed)
{
  p->mRxDrShadow[index].packed = packed;
}

static void setShadow_RxDrData(NthwDbs_t* p,
                           uint32_t index,
                           uint64_t guestPhysicalAddress,
                           uint32_t hostId,
                           uint32_t queueSize,
                           uint32_t header,
                           uint32_t packed)
{
  setShadow_RxDrDataGuestPhysicalAddress(p, index, guestPhysicalAddress);
  setShadow_RxDrDataHostId(p, index, hostId);
  setShadow_RxDrDataQueueSize(p, index, queueSize);
  setShadow_RxDrDataHeader(p, index, header);
  setShadow_RxDrDataPacked(p, index, packed);
}

static void flush_RxDrData(NthwDbs_t* p, uint32_t index)
{
  nthw_field_set_val(p->mpFldRxDescriptorReaderDataGuestPhysicalAddress, (uint32_t*)&p->mRxDrShadow[index].guestPhysicalAddress, 2);
  nthw_field_set_val32(p->mpFldRxDescriptorReaderDataHostID, p->mRxDrShadow[index].hostId);
  if (nthw_module_is_version_newer(p->mpModDbs, 0,8)) {
    nthw_field_set_val32(p->mpFldRxDescriptorReaderDataQueueSize, (1U << p->mRxDrShadow[index].queueSize) - 1U);
  } else {
    nthw_field_set_val32(p->mpFldRxDescriptorReaderDataQueueSize, p->mRxDrShadow[index].queueSize);
  }
  nthw_field_set_val32(p->mpFldRxDescriptorReaderDataHeader, p->mRxDrShadow[index].header);
  if (p->mpFldRxDescriptorReaderDataPacked) {
    nthw_field_set_val32(p->mpFldRxDescriptorReaderDataPacked, p->mRxDrShadow[index].packed);
  }

  setRxDrDataIndex(p, index);
  nthw_register_flush(p->mpRegRxDescriptorReaderData, 1);
}

int set_rx_dr_data(NthwDbs_t* p,
                        uint32_t index,
                        uint64_t guestPhysicalAddress,
                        uint32_t hostId,
                        uint32_t queueSize,
                        uint32_t header,
                        uint32_t packed)
{
  if (!p->mpRegRxDescriptorReaderData)
    return -ENOTSUP;

  setShadow_RxDrData(p, index, guestPhysicalAddress, hostId, queueSize, header, packed);
  flush_RxDrData(p, index);
  return 0;
}

static void setTxDrDataIndex(NthwDbs_t* p, uint32_t index)
{
  nthw_field_set_val32(p->mpFldTxDescriptorReaderControlAdr, index);
  nthw_field_set_val32(p->mpFldTxDescriptorReaderControlCnt, 1);
  nthw_register_flush(p->mpRegTxDescriptorReaderControl, 1);
}

static void setShadow_TxDrDataGuestPhysicalAddress(NthwDbs_t* p, uint32_t index, uint64_t guestPhysicalAddress)
{
  p->mTxDrShadow[index].guestPhysicalAddress = guestPhysicalAddress;
}

static void setShadow_TxDrDataHostId(NthwDbs_t* p, uint32_t index, uint32_t hostId)
{
  p->mTxDrShadow[index].hostId = hostId;
}

static void setShadow_TxDrDataQueueSize(NthwDbs_t* p, uint32_t index, uint32_t queueSize)
{
  p->mTxDrShadow[index].queueSize = queueSize;
}

static void setShadow_TxDrDataHeader(NthwDbs_t* p, uint32_t index, uint32_t header)
{
  p->mTxDrShadow[index].header = header;
}

static void setShadow_TxDrDataPort(NthwDbs_t* p, uint32_t index, uint32_t port)
{
  p->mTxDrShadow[index].port = port;
}

static void setShadow_TxDrDataPacked(NthwDbs_t* p, uint32_t index, uint32_t packed)
{
  p->mTxDrShadow[index].packed = packed;
}

static void setShadow_TxDrData(NthwDbs_t* p,
                           uint32_t index,
                           uint64_t guestPhysicalAddress,
                           uint32_t hostId,
                           uint32_t queueSize,
                           uint32_t port,
                           uint32_t header,
                           uint32_t packed)
{
  setShadow_TxDrDataGuestPhysicalAddress(p, index, guestPhysicalAddress);
  setShadow_TxDrDataHostId(p, index, hostId);
  setShadow_TxDrDataQueueSize(p, index, queueSize);
  setShadow_TxDrDataHeader(p, index, header);
  setShadow_TxDrDataPort(p, index, port);
  setShadow_TxDrDataPacked(p, index, packed);
}

static void flush_TxDrData(NthwDbs_t* p, uint32_t index)
{
  nthw_field_set_val(p->mpFldTxDescriptorReaderDataGuestPhysicalAddress, (uint32_t*)&p->mTxDrShadow[index].guestPhysicalAddress, 2);
  nthw_field_set_val32(p->mpFldTxDescriptorReaderDataHostID, p->mTxDrShadow[index].hostId);
  if (nthw_module_is_version_newer(p->mpModDbs, 0,8)) {
    nthw_field_set_val32(p->mpFldTxDescriptorReaderDataQueueSize, (1U << p->mTxDrShadow[index].queueSize) - 1U);
  } else {
    nthw_field_set_val32(p->mpFldTxDescriptorReaderDataQueueSize, p->mTxDrShadow[index].queueSize);
  }
  nthw_field_set_val32(p->mpFldTxDescriptorReaderDataHeader, p->mTxDrShadow[index].header);
  nthw_field_set_val32(p->mpFldTxDescriptorReaderDataPort, p->mTxDrShadow[index].port);
  if (p->mpFldTxDescriptorReaderDataPacked) {
    nthw_field_set_val32(p->mpFldTxDescriptorReaderDataPacked, p->mTxDrShadow[index].packed);
  }

  setTxDrDataIndex(p, index);
  nthw_register_flush(p->mpRegTxDescriptorReaderData, 1);
}

int set_tx_dr_data(NthwDbs_t* p,
                        uint32_t index,
                        uint64_t guestPhysicalAddress,
                        uint32_t hostId,
                        uint32_t queueSize,
                        uint32_t port,
                        uint32_t header,
                        uint32_t packed)
{
  if (!p->mpRegTxDescriptorReaderData)
    return -ENOTSUP;

  setShadow_TxDrData(p, index, guestPhysicalAddress, hostId, queueSize, port, header, packed);
  flush_TxDrData(p, index);
  return 0;
}

static void setTxQpDataIndex(NthwDbs_t* p, uint32_t index)
{
  nthw_field_set_val32(p->mpFldTxQueuePropertyControlAdr, index);
  nthw_field_set_val32(p->mpFldTxQueuePropertyControlCnt, 1);
  nthw_register_flush(p->mpRegTxQueuePropertyControl, 1);
}

static void setShadow_TxQpDataVirtualPort(NthwDbs_t* p, uint32_t index, uint32_t virtualPort)
{
  p->mTxQpShadow[index].virtualPort = virtualPort;
}

static void setShadow_TxQpData(NthwDbs_t* p, uint32_t index, uint32_t virtualPort)
{
  setShadow_TxQpDataVirtualPort(p, index, virtualPort);
}

static void flush_TxQpData(NthwDbs_t* p, uint32_t index)
{
  nthw_field_set_val32(p->mpFldTxQueuePropertyDataVPort, p->mTxQpShadow[index].virtualPort);

  setTxQpDataIndex(p, index);
  nthw_register_flush(p->mpRegTxQueuePropertyData, 1);
}

int NthwDbs_setTxQpData(NthwDbs_t* p,
                        uint32_t index,
                        uint32_t virtualPort)
{
  if (!p->mpRegTxQueuePropertyData)
    return -ENOTSUP;

  setShadow_TxQpData(p, index, virtualPort);
  flush_TxQpData(p, index);
  return 0;
}

static void setTxQosDataIndex(NthwDbs_t* p, uint32_t index)
{
  nthw_field_set_val32(p->mpRegTxQueueQosControlAdr, index);
  nthw_field_set_val32(p->mpRegTxQueueQosControlCnt, 1);
  nthw_register_flush(p->mpRegTxQueueQosControl, 1);
}

static void setShadow_TxQosDataEnable(NthwDbs_t* p, uint32_t index, uint32_t enable)
{
  p->mTxQosShadow[index].enable = enable;
}

static void setShadow_TxQosDataIr(NthwDbs_t* p, uint32_t index, uint32_t ir)
{
  p->mTxQosShadow[index].ir = ir;
}

static void setShadow_TxQosDataBs(NthwDbs_t* p, uint32_t index, uint32_t bs)
{
  p->mTxQosShadow[index].bs = bs;
}

static void setShadow_TxQosData(NthwDbs_t* p,
                                uint32_t index,
                                uint32_t enable,
                                uint32_t ir,
                                uint32_t bs)
{
  setShadow_TxQosDataEnable(p, index, enable);
  setShadow_TxQosDataIr(p, index, ir);
  setShadow_TxQosDataBs(p, index, bs);
}

static void flush_TxQosData(NthwDbs_t* p, uint32_t index)
{
  nthw_field_set_val32(p->mpRegTxQueueQosDataEn, p->mTxQosShadow[index].enable);
  nthw_field_set_val32(p->mpRegTxQueueQosDataIR, p->mTxQosShadow[index].ir);
  nthw_field_set_val32(p->mpRegTxQueueQosDataBS, p->mTxQosShadow[index].bs);

  setTxQosDataIndex(p, index);
  nthw_register_flush(p->mpRegTxQueueQosData, 1);
}

int set_tx_qos_data(NthwDbs_t* p,
                         uint32_t index,
                         uint32_t enable,
                         uint32_t ir,
                         uint32_t bs)
{
  if (!p->mpRegTxQueueQosData)
    return -ENOTSUP;

  setShadow_TxQosData(p, index, enable, ir, bs);
  flush_TxQosData(p, index);
  return 0;
}

int set_tx_qos_rate(NthwDbs_t* p,
                         uint32_t mul,
                         uint32_t div)
{
  if (!p->mpRegTxQueueQosRate)
    return -ENOTSUP;

  nthw_field_set_val32(p->mpRegTxQueueQosRateMul, mul);
  nthw_field_set_val32(p->mpRegTxQueueQosRateDiv, div);
  nthw_register_flush(p->mpRegTxQueueQosRate, 1);
  return 0;
}
