/*
 * SPDX-License-Identifier: BSD-3-Clause
 * Copyright(c) 2023 Napatech A/S
 */

#include <rte_byteorder.h>

#include "nt_util.h"

#include "sensors.h"
#include "ntlog.h"


static const char* translate_type(NtSensorType_t Type)
{
  switch (Type)
  {
  case NT_SENSOR_TYPE_UNKNOWN      : return "Unknown";
  case NT_SENSOR_TYPE_TEMPERATURE  : return "Temp.";
  case NT_SENSOR_TYPE_VOLTAGE      : return "Voltage";
  case NT_SENSOR_TYPE_CURRENT      : return "Current";
  case NT_SENSOR_TYPE_POWER        : return "Power";
  case NT_SENSOR_TYPE_FAN          : return "Speed";
  case NT_SENSOR_TYPE_HIGH_POWER   : return "Power";
  default:
    assert(false);
    return "Unhandled";
  }
}

static float get_multiplier(enum NtSensorType_e type)
{
  switch (type)
  {
  case NT_SENSOR_TYPE_TEMPERATURE  : return (float)0.1;   /* Unit: 0.1 degree Celsius */
  case NT_SENSOR_TYPE_VOLTAGE      : return (float)1e-3;  /* Unit: 1 mV */
  case NT_SENSOR_TYPE_CURRENT      : return (float)1e-6;  /* Unit: 1 uA */
  case NT_SENSOR_TYPE_POWER        : return (float)1e-7;  /* Unit: 0.1 uW */
  case NT_SENSOR_TYPE_FAN          : return (float)1;     /* Unit: 1 RPM (Revolutions Per Minute) */
  case NT_SENSOR_TYPE_HIGH_POWER   : return (float)1e-3;  /* Unit: 1 mW */
  default:
    assert(false);
    return 0.0;
  }
}

static const char* get_unit(enum NtSensorType_e type)
{
  switch (type)
  {
  case NT_SENSOR_TYPE_UNKNOWN      : return "?";
  case NT_SENSOR_TYPE_TEMPERATURE  : return "C";
  case NT_SENSOR_TYPE_VOLTAGE      : return "V";
  case NT_SENSOR_TYPE_CURRENT      : return "A";
  case NT_SENSOR_TYPE_POWER        : return "W";
  case NT_SENSOR_TYPE_FAN          : return "RPM";
  case NT_SENSOR_TYPE_HIGH_POWER   : return "W";
  default:
    assert(false);
    return "Unhandled";
  }
}

static char* translate_value(char*          Str,
                                     size_t         Size,
                                     int            Value,
                                     NtSensorType_t Type)
{
  char  UnitStr[32];
  float FloatVal;
  float PrefixMultiplier;

  if (Value == (int)NT_SENSOR_NAN)
  {
    snprintf(Str, Size, "%s", "");
    return Str;
  }

  switch (Type)
  {
  case NT_SENSOR_TYPE_HIGH_POWER:
  case NT_SENSOR_TYPE_VOLTAGE:
  case NT_SENSOR_TYPE_CURRENT:
    PrefixMultiplier = 1e3;
    snprintf(UnitStr, sizeof(UnitStr), "m");
    break;
  case NT_SENSOR_TYPE_POWER:
    PrefixMultiplier = 1e6;
    snprintf(UnitStr, sizeof(UnitStr), "u");
    break;
  default:
    PrefixMultiplier = 1;
    snprintf(UnitStr, sizeof(UnitStr), "%s", "");
    break;
  }

  /*
   * Multiply with sensor multiplier in order to get a value that is in
   * correspondence with the sensor unit
   */
  FloatVal = (float)Value * get_multiplier(Type);

  strncat(UnitStr, get_unit(Type), sizeof(UnitStr) - strlen(UnitStr) - 1);
  FloatVal *= PrefixMultiplier;

  switch (Type)
  {
  case NT_SENSOR_TYPE_TEMPERATURE: /* Unit: 0.1 degree Celsius */
  case NT_SENSOR_TYPE_HIGH_POWER:  /* Unit: 1 mW */
    snprintf(Str, Size, "%.1f%s", FloatVal, UnitStr);
    break;
  case NT_SENSOR_TYPE_VOLTAGE:     /* Unit: 1 mV */
  case NT_SENSOR_TYPE_CURRENT:     /* Unit: 1 uA */
    snprintf(Str, Size, "%.2f%s", FloatVal, UnitStr);
    break;
  case NT_SENSOR_TYPE_POWER:       /* Unit: 0.1 uW */
  case NT_SENSOR_TYPE_FAN:         /* Unit: 1 RPM (Revolutions Per Minute) */
    snprintf(Str, Size, "%.0f%s", FloatVal, UnitStr);
    break;
  default:
    snprintf(Str, Size, "%.0f", FloatVal);
    break;
  }
  return Str;
}

static void dump_sensor(struct nt_adapter_sensor* sensor)
{
  char buffer1[20];
  char buffer2[20];
  char buffer3[20];

  if (sensor->info.source == NT_SENSOR_SOURCE_ADAPTER ||
      sensor->info.source == NT_SENSOR_SOURCE_LEVEL1_ADAPTER ||
      sensor->info.source == NT_SENSOR_SOURCE_LEVEL2_ADAPTER) {
    NT_LOG(DBG, SENSOR, "Adapter - \"%s\" %s [%s..%s]. Value %s",
                        sensor->info.name,
                        translate_type(sensor->info.type),
                        translate_value(buffer1, 20, sensor->info.valueLowest, sensor->info.type),
                        translate_value(buffer2, 20, sensor->info.valueHighest, sensor->info.type),
                        translate_value(buffer3, 20, sensor->info.value, sensor->info.type));
  } else {
    NT_LOG(DBG, SENSOR, "Port %u - \"%s\" %s [%s..%s]. Value %s",
                        sensor->mIntfNo,
                        sensor->info.name,
                        translate_type(sensor->info.type),
                        translate_value(buffer1, 20, sensor->info.valueLowest, sensor->info.type),
                        translate_value(buffer2, 20, sensor->info.valueHighest, sensor->info.type),
                        translate_value(buffer3, 20, sensor->info.value, sensor->info.type));
  }
}

void sensor_deinit(struct nt_sensor_group* sg) {
  if(sg) {
    if(sg->sensor){free(sg->sensor);}
    if(sg->monitor){free(sg->monitor);}
    free(sg);
  }
}

struct nt_adapter_sensor* allocate_sensor(
  uint8_t adapterOrPortIndex,
  const char* pName,
  enum NtSensorSource_e ssrc,
  enum NtSensorType_e type,
  unsigned int index,
  enum NtSensorEventAlarm_e eventAlarm,
  enum sensor_mon_sign si
) {
  struct nt_adapter_sensor* sensor = (struct nt_adapter_sensor*)malloc(sizeof(struct nt_adapter_sensor));
  if(sensor == NULL) {
    NT_LOG(ERR, ETHDEV, "%s: sensor is NULL", __func__);
    return NULL;
  }

  sensor->alarm = eventAlarm;
  sensor->mEnableAlarm = true;
  sensor->mIntfNo = 0xFF;
  sensor->mAdapterNo = 0xFF;
  sensor->si = si;

  sensor->info.source = ssrc;
  sensor->info.sourceIndex = adapterOrPortIndex;
  sensor->info.sensorIndex = index;
  sensor->info.type = type;
  sensor->info.subType = NT_SENSOR_SUBTYPE_NA;
  sensor->info.state = NT_SENSOR_STATE_INITIALIZING;
  sensor->info.value = NT_SENSOR_NAN;
  sensor->info.valueLowest = NT_SENSOR_NAN;
  sensor->info.valueHighest = NT_SENSOR_NAN;
  memset(sensor->info.name, 0, NT_INFO_SENSOR_NAME);
  memcpy(sensor->info.name, pName, (strlen(pName) > NT_INFO_SENSOR_NAME) ? NT_INFO_SENSOR_NAME : strlen(pName));
  sensor->info.name[NT_INFO_SENSOR_NAME] = '\0';

  return sensor;
}

void update_sensor_value(struct nt_adapter_sensor* sensor, int32_t value)
{
  if(sensor == NULL){return;}
  sensor->info.value = value;
  if (sensor->info.valueHighest < value || sensor->info.valueHighest == (int)NT_SENSOR_NAN) {
    sensor->info.valueHighest = value;
  }
  if (sensor->info.valueLowest > value || sensor->info.valueLowest == (int)NT_SENSOR_NAN || sensor->info.valueLowest == 0) {
    sensor->info.valueLowest = value;
  }
  dump_sensor(sensor);
}

struct nt_adapter_sensor* allocate_sensor_by_description(uint8_t adapterOrPortIndex, enum NtSensorSource_e ssrc, struct nt_adapter_sensor_description* descr)
{
  struct nt_adapter_sensor* sensor = (struct nt_adapter_sensor*)malloc(sizeof(struct nt_adapter_sensor));
  if(sensor == NULL) {
    NT_LOG(ERR, ETHDEV, "%s: sensor is NULL", __func__);
    return NULL;
  }

  sensor->alarm = descr->eventAlarm;
  sensor->mEnableAlarm = true;
  sensor->mIntfNo = adapterOrPortIndex;
  sensor->mAdapterNo = 0xFF;
  sensor->si = SENSOR_MON_UNSIGNED;

  sensor->info.sourceIndex = adapterOrPortIndex;
  sensor->info.source = ssrc;
  sensor->info.type = descr->type;
  sensor->info.sensorIndex = descr->index;
  memset(sensor->info.name, 0, NT_INFO_SENSOR_NAME);
  memcpy(sensor->info.name, descr->name, (strlen(descr->name) > NT_INFO_SENSOR_NAME) ? NT_INFO_SENSOR_NAME : strlen(descr->name));
  sensor->info.name[NT_INFO_SENSOR_NAME] = '\0';

  return sensor;
}

void init_sensor_group(struct nt_sensor_group* sg)
{
  /* Set all pointers to NULL */
  sg->sensor = NULL;
  sg->monitor = NULL;
  sg->next = NULL;
  sg->read = NULL;
  sg->conv_func = NULL;
}

/* Getters */
int32_t get_value(struct nt_sensor_group* sg) {return sg->sensor->info.value;};
int32_t get_lowest(struct nt_sensor_group* sg) {return sg->sensor->info.valueLowest;};
int32_t get_highest(struct nt_sensor_group* sg) {return sg->sensor->info.valueHighest;};
char* get_name(struct nt_sensor_group* sg) {return sg->sensor->info.name;};

/* Conversion functions */
int null_signed(uint32_t pSensorResult) {
  return (int16_t)pSensorResult;
}

int null_unsigned(uint32_t pSensorResult) {
  return (uint16_t)pSensorResult;
}


/*
 * For EXAR7724: Convert a read Vch value to Napatech internal representation
 * Doc: Vout = ReadVal * 0.015 (PRESCALE is accounted for)
 */
int Exar7724_Vch(uint32_t pSensorResult)
{
  return pSensorResult * 15; /* NT unit: 1mV */
}

/*
 * For EXAR7724: Convert a read Vin value to Napatech internal representation
 * Doc: Vout = ReadVal * 0.0125
 */
int Exar7724_Vin(uint32_t pSensorResult)
{
  return (pSensorResult * 25) / 2; /* NT unit: 1mV */
}

/*
 * For EXAR7724: Convert a read Tj value to Napatech internal representation
 * Doc: Temp (in Kelvin) = (((ReadVal * 10mV) - 600mV) / (2mV/K)) + 300K =
 *                      = ReadVal * 5K
 */
int Exar7724_Tj(uint32_t pSensorResult)
{
  /*
   * A value of 2730 is used instead of 2732 which is more correct but since
   * the temperature step is 5 degrees it is more natural to show these steps
   */
  return pSensorResult * 50 - 2730; /* NT unit: 0.1C */
}

/*
 * Conversion function for Linear Tecnology Linear_5s_11s format.
 * The functions returns Y * 2**N, where N = b[15:11] is a 5-bit two's complement
 * integer and Y = b[10:0] is an 11-bit two's complement integer.
 * The multiplier value is used for scaling to Napatech units.
 */
static int conv5s_11s(uint16_t value, int multiplier)
{
  int  N, Y;

  Y = value & 0x07FF;

  if (value & 0x0400) {
    Y -= 0x0800; /* The MSBit is a sign bit */
  }

  N = (value >> 11) & 0x1F;

  if (N & 0x10) {
    N -= 0x20;   /* The MSBit is a sign bit */
  }

  Y *= multiplier;

  if (N > 0) {
    Y *= (1 << N);
  }
  else if (N < 0) {
    Y /= (1 << (-N));
  }

  return Y;
}

/*
 * Temperature conversion from Linear_5s_11s format.
 */
int Ltm4676_Tj(uint32_t pSensorResult)
{
  return (uint16_t)conv5s_11s(pSensorResult, 10);  /* NT unit: 0.1C */
}

/*
 * For MP2886a: Convert a read Tj value to Napatech internal representation
 */
int Mp2886a_Tj(uint32_t pSensorResult)
{
  /*
   * MPS-2886p: READ_TEMPERATURE (register 0x8Dh)
   * READ_TEMPERATURE is a 2-byte, unsigned integer.
   */
  return (uint16_t)pSensorResult; /* NT unit: 0.1C */
}

/*
 * For MAX6642: Convert a read temperature value to Napatech internal representation
 */
int Max6642_T(uint32_t pSensorResult)
{
  if ((pSensorResult >> 8) == 0xFF) {
    return NT_SENSOR_NAN;
  }

  /* The six lower bits are not used */
  return (int)(((pSensorResult >> 6) * 5) / 2); /* NT unit: 0.25 deg, Native unit: 0.1C */
}

/*
 * For DS1775: Convert a read temperature value to Napatech internal representation
 */
int Ds1775_T(uint32_t pSensorResult)
{
  return (pSensorResult * 10) / 256; /* NT unit: 0.1 deg, Native unit: 1/256 C */
}

/*
 * For FAN: Convert a tick count to RPM
 * NT unit: RPM, Native unit: 2 ticks/revolution
 */
int Fan(uint32_t pSensorResult)
{
  return (pSensorResult * 60U / 4);
}

int Fan_nt400(uint32_t sensorResult)
{
  sensorResult = RTE_BE32(sensorResult);
  sensor_mon_result32_t *pSensorResult = (sensor_mon_result32_t *)&sensorResult;
  return (int)(pSensorResult->uint16 * 60U / 4);
}

int NullSigned(uint32_t sensorResult)
{
  sensorResult = RTE_BE32(sensorResult);
  sensor_mon_result32_t *pSensorResult = (sensor_mon_result32_t *)&sensorResult;
  return pSensorResult->int16;
}

/*
 * For TMP464 : Convert a read temperature value to Napatech internal representation
 */

int Tmp464p_T(uint32_t sensorResult)
{

  /*
   * Explainer for 1/128 : The measurement is delivered as a 13 bit value,
   * alligned with the MSB and therefore has to be shifted 3 bits equal to dividing by 8 or 2^3
   * The 13 bit temperature has a resolution of 0.0625 C/bit (1/0.0625 = 16 = 2^4).
   * Therefore, to reach the final result we need to divide by a total of (2^3 + 2^4 = 2^7) 128.
   */

  sensorResult = RTE_BE32(sensorResult);
  sensor_mon_result32_t *pSensorResult = (sensor_mon_result32_t *)&sensorResult;

  uint16_t bin_uint;
  int16_t temp;

  bin_uint = pSensorResult->uint16; /* msb, lsb */
  /*
   * Test: bin_uint = -25.000  = -3200 = 1111001110000000 = 0xF380
   *       bin_uint = -12.125  = -1552 = 1111100111110000 = 0xf9f0
   *       bin_uint =  12.125  =  1552 = 0000011000010000 = 0x0610
   *       bin_uint =  10.0625 =  1288 = 0000010100001000 = 0x0508
   * bin_uint = 0xF380;
   */
  if (bin_uint & 0x8000) {
    bin_uint = (uint16_t)(~(bin_uint - 1)); /* convert from two's compliment */
    temp = (int16_t)(-bin_uint*10/128);
  } else {
    temp = (int16_t)(bin_uint*10/128);
  }

  return temp;
}

/*
 * For MP8645p: Convert a read Tj value to Napatech internal representation
 */
int Mp8645p_Tj(uint32_t sensorResult)
{
  /*
   * MPS-8645p: READ_TEMPERATURE_1 (register 0x8Dh)
   * READ_TEMPERATURE is a 2-byte, twos complement integer.
   * The bit9 is the sign bit. The default value is 0032h.
   * The corresponding value is 25 degree Celsius.
   * TBD: Sign-bit handling missing !
   */

  sensorResult = RTE_BE32(sensorResult);
  sensor_mon_result32_t *pSensorResult = (sensor_mon_result32_t *)&sensorResult;
  return (pSensorResult->uint16*10); /* NT unit: 0.1C */
}

/*
 * For MP2978 : Convert a read temperature value to Napatech internal representation
 */
int Mp2978_T(uint32_t sensorResult)
{
  sensorResult = RTE_BE32(sensorResult);
  sensor_mon_result32_t *pSensorResult = (sensor_mon_result32_t *)&sensorResult;
  return (pSensorResult->int16 * 10); /* NT unit: 0.1 deg, Native unit: 1 C/LSB */
}
