/*
 * SPDX-License-Identifier: BSD-3-Clause
 * Copyright(c) 2023 Napatech A/S
 */

#include <getopt.h>
#include <pthread.h> /* pthread_create */
#include <signal.h> /* signal, SIGINT */
#include <stdatomic.h> /* atomic_fetch_add, atomic_load, atomic_store */
#include <stdint.h> /* uint8_t, ... */
#include <stdio.h> /* printf */
#include <stdlib.h> /* free */
#include <string.h> /* memcpy, memset */
#include <unistd.h> /* sleep, usleep */

#include <rte_eal.h>    /* rte_eal_init */
#include <rte_errno.h>  /* rte_errno */
#include <rte_ethdev.h> /* rte_eth_dev_count_avail */
#include <rte_flow.h>   /* rte_flow_create */
#include <rte_lcore.h>  /* rte_lcore_<xx> */
#include <rte_mbuf.h>   /* rte_pktmbuf_pool_create */
#include <rte_version.h> /* rte_version */

#define wait_usec(x) rte_delay_us_sleep(x)

/* Driver parameters */
static const uint16_t RX_RINGS = 1U;
static const uint16_t TX_RINGS = 1U;
static const uint16_t RX_RING_SIZE = 128U;
static const uint16_t TX_RING_SIZE = 128U;
static const unsigned int NUM_MBUFS = 8192U;
static const unsigned int MBUF_CACHE_SIZE = 256U;
static const uint16_t MAX_PKT_SIZE = 1500U;
static const size_t COUNTER_OFFSET = 100U;

/* Command line flags */
static unsigned int NUM_PKTS = 10000U;
static unsigned int DURATION;
static unsigned int RX_PORT = 1U;
static unsigned int TX_PORT;
static const char *progname;

/* Variables shared by RX and TX threads */
static _Atomic(uint8_t) quit_flag;
static _Atomic(uint64_t) total_packets;
static _Atomic(uint64_t) total_bytes;
static _Atomic(uint64_t) total_pkt_errors;

static void int_handler(int sig_num)
{
  /* set quit flag for rx thread to exit */
  atomic_store(&quit_flag, true);
  printf("Exiting on signal %d\n", sig_num);
  fflush(stdout);
}

static int port_init(uint8_t port_id, struct rte_mempool **mbuf_pool)
{
  struct rte_eth_conf port_conf;
  int retval;

  uint16_t nb_ports = rte_eth_dev_count_avail();

  memset(&port_conf, 0x0, sizeof(struct rte_eth_conf));

  if (*mbuf_pool == NULL) {
    /* Creates a new mempool in memory to hold the mbufs. */
    *mbuf_pool = rte_pktmbuf_pool_create("MBUF_POOL", NUM_MBUFS * nb_ports,
      MBUF_CACHE_SIZE, 0, MAX_PKT_SIZE + RTE_PKTMBUF_HEADROOM, rte_socket_id());

    if (*mbuf_pool == NULL) {
      rte_exit(EXIT_FAILURE, "Cannot create mbuf pool. rte_errno %d\n", rte_errno);
    }
  }

  /* Get the number of ports which are usable for the application */
  if (port_id >= rte_eth_dev_count_avail()) {
    return -1;
  }

  /* Configure the Ethernet device. */
  retval = rte_eth_dev_configure(port_id, RX_RINGS, TX_RINGS, &port_conf);

  if (retval != 0) {
    return retval;
  }

  /* Allocate and set up "RX_RINGS" RX queue per Ethernet port. */
  for (uint16_t q = 0; q < RX_RINGS; q++) {
    retval = rte_eth_rx_queue_setup(port_id, q, RX_RING_SIZE,
      rte_eth_dev_socket_id(port_id), NULL, *mbuf_pool);

    if (retval < 0) {
      return retval;
    }
  }

  /* Allocate and set up "TX_RINGS" TX queue per Ethernet port. */
  for (uint16_t q = 0; q < TX_RINGS; q++) {
    retval = rte_eth_tx_queue_setup(port_id, q, TX_RING_SIZE,
      rte_eth_dev_socket_id(port_id), NULL);

    if (retval < 0) {
      return retval;
    }
  }

  /* Start the Ethernet port. */
  retval = rte_eth_dev_start(port_id);
  if (retval < 0) {
    return retval;
  }

  return 0;
}

static int port_close(uint8_t port_id)
{
  rte_eth_dev_stop(port_id);
  rte_eth_dev_close(port_id);
  return 0;
}

static struct rte_flow* setup_pass_all_filter(uint8_t port_id)
{
  struct rte_flow_attr attr;
  struct rte_flow_item pattern[10];
  struct rte_flow_action actions[10];
  struct rte_flow_error error;

  uint32_t action_count = 0;
  uint32_t pattern_count = 0;

  memset(&error,  0x0, sizeof(error));
  memset(&attr,   0x0, sizeof(attr));
  memset(pattern, 0x0, sizeof(pattern));
  memset(actions, 0x0, sizeof(actions));

  attr.group = 0;
  attr.priority = 1;

  /* Pass all ethernet filter */
  struct rte_flow_item_eth eth;
  memset(&eth, 0x0, sizeof(eth));

  pattern[pattern_count].type = RTE_FLOW_ITEM_TYPE_ETH;
  pattern[pattern_count].spec = &eth;
  pattern[pattern_count].mask = &eth;
  pattern[pattern_count].last = NULL;
  pattern_count += 1;

  /* Pass all IPv4 filter */
  struct rte_flow_item_ipv4 ipv4;
  memset(&ipv4, 0x0, sizeof(ipv4));

  pattern[pattern_count].type = RTE_FLOW_ITEM_TYPE_IPV4;
  pattern[pattern_count].spec = &ipv4;
  pattern[pattern_count].mask = &ipv4;
  pattern[pattern_count].last = NULL;
  pattern_count += 1;

  /* Pass all UDP filter */
  struct rte_flow_item_udp udp;
  memset(&udp, 0x0, sizeof(udp));

  pattern[pattern_count].type = RTE_FLOW_ITEM_TYPE_UDP;
  pattern[pattern_count].spec = &udp;
  pattern[pattern_count].mask = &udp;
  pattern[pattern_count].last = NULL;
  pattern_count += 1;

  /* Add pattern end */
  pattern[pattern_count++].type = RTE_FLOW_ITEM_TYPE_END;

  /* Send all packets to queue 0 */
  struct rte_flow_action_queue queue;
  memset(&queue, 0x0, sizeof(queue));
  queue.index = 0;

  actions[action_count].type = RTE_FLOW_ACTION_TYPE_QUEUE;
  actions[action_count].conf = &queue;
  action_count += 1;

  /* Add action end */
  actions[action_count++].type = RTE_FLOW_ACTION_TYPE_END;

  struct rte_flow *flow = rte_flow_create(port_id, &attr, pattern, actions, &error);

  if (flow == NULL) {
    printf("rte_flow_create error %d: %s\n", (int)error.type, error.message);
  }

  return flow;
}

/*
 * Threads setup
 */

static const uint8_t ipv4_packet[246] = {
  /* Ethernet Header, 14 bytes */
  0x08, 0x00, 0x34, 0x00, 0x02, 0x31, /* DST MAC address */
  0xd2, 0x94, 0x66, 0x24, 0x87, 0xed, /* SRC MAC address */
  0x08, 0x00,                         /* Ether type, 0x0800 = IPv4 datagram */
  /* IPv4 Header, 20 bytes */
  0x45,                   /* IPv4, IHL = 5 */
  0x00,                   /* DSCP = 0, ECN = 0 */
  0x00, 0xe8,             /* Total Length = 232 */
  0x00, 0x01,             /* Identification = 1 */
  0x00, 0x00,             /* Flags = 0, Fragment Offset = 0 */
  0x40,                   /* TTL = 64 */
  0x11,                   /* Protocol = 17 (UDP) */
  0x17, 0xd9,             /* Header Checksum */
  0x67, 0xec, 0xf0, 0x3e, /* Source IP Address */
  0x00, 0x00, 0x0a, 0x01, /* Destination IP Address */
  /* UDP header, 8 bytes */
  0xad, 0xa5, /* Source Port */
  0xb1, 0xbd, /* Destination Port */
  0x00, 0xd4, /* Length = 212 (header: 8, udp data: 204) */
  0x00, 0x00, /* Checksum (optional in IPv4, zero means ignore) */
  /* UDP payload, 212 bytes */
  0x42, 0x44, 0x46, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42,
  0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42,
  0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42,
  0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42,
  0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42,
  0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42,
  0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42,
  0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42,
  0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42,
  0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42,
  0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42,
  0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42,
  0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42,
  0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42,
  0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42,
  0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0xfe, 0xff};

struct worker_info {
  int port;
  int queue;
};

static void lcore_main(struct rte_mempool *mbuf_pool)
{
#if (0)
  unsigned int wait_time = DURATION;
#endif

  /* TX packet from port 1 */
  struct rte_mbuf *mbuf = rte_pktmbuf_alloc(mbuf_pool);
  uint8_t *const payload = (uint8_t *)mbuf->buf_addr + RTE_PKTMBUF_HEADROOM;
  uint32_t *const p = (uint32_t *)&payload[COUNTER_OFFSET];
  uint32_t value = 0U;
  const uint16_t port_id = (uint16_t)TX_PORT;
  const uint16_t queue_id = 0U;

  mbuf->data_len = sizeof(ipv4_packet);
  mbuf->pkt_len = sizeof(ipv4_packet);
  memcpy(payload, ipv4_packet, sizeof(ipv4_packet));

  printf("Transmitting up to %u packets for up to %u seconds\n", NUM_PKTS,
         DURATION);
  fflush(stdout);

  size_t pkts_sent = 0U;
  while (atomic_load(&quit_flag) == false) {
    bool stop_pkts = NUM_PKTS > 0U && pkts_sent >= NUM_PKTS;
    if (stop_pkts)
      break;
    /* Send one packet */
    *p = value;
    wait_usec(10);
    const uint16_t num_sent = rte_eth_tx_burst(port_id, queue_id, &mbuf, 1);
    if (num_sent == 1) {
      pkts_sent++;
      value++;
    }
  }

  printf("Transmission complete.\n");
  fflush(stdout);

  rte_pktmbuf_free(mbuf);
}

static int lcore_worker(void* p)
{
  const uint16_t BURST_SIZE = 32U;

  struct worker_info *const info = (struct worker_info *)p;

  uint32_t expected_counter = 0U;
  while (!atomic_load(&quit_flag)) {
    /* Get burst of RX packets */
    struct rte_mbuf *bufs[BURST_SIZE];
    const uint16_t nb_rx = rte_eth_rx_burst(info->port, info->queue, bufs, BURST_SIZE);

    if (nb_rx > 0U) {
      for (uint16_t buf = 0; buf < nb_rx; buf++) {
        struct rte_mbuf *mbuf = bufs[buf];
        uint32_t pkt_len = mbuf->pkt_len;
#if (0)
        uint32_t data_len = mbuf->data_len;
#endif
        uint8_t *payload = (uint8_t *)mbuf->buf_addr + RTE_PKTMBUF_HEADROOM;
        const uint32_t counter = *((uint32_t *)&payload[COUNTER_OFFSET]);

#if (0)
        printf("payload = %p, pkt_len = %u, data_len = %u\n", payload, pkt_len,
               data_len);
        for (unsigned int i = 0; i < pkt_len; i++) {
          printf("%03u:%02x ", i, payload[i]);
        }
        printf("\n");
        fflush(stdout);
#endif

        atomic_fetch_add(&total_bytes, pkt_len + 4U /* checksum ?? */);
        atomic_fetch_add(&total_packets, 1);

        if (expected_counter != counter) {
          atomic_fetch_add(&total_pkt_errors, 1U);
          fprintf(stderr, "packet# %" PRIu64 ": counter = %u but expected %u\n",
                  total_packets, counter, expected_counter);
          fflush(stdout);
        }
        expected_counter = counter + 1U;

        rte_pktmbuf_free(bufs[buf]);
      }
    }
  }

  return 0;
}

static const char *version(void)
{
  static char str[128];

  snprintf(str, sizeof(str), "%s 1.0 (%s)\n", progname, rte_version());
  return str;
}

/* Parse numeric argument from command line */
static unsigned long get_uint(const char *arg, const char *name,
                              unsigned int limit)
{
  char *endp;

  unsigned long u = strtoul(arg, &endp, 0);
  if (*arg == '\0' || *endp != '\0')
    rte_exit(EXIT_FAILURE, "Specified %s \"%s\" is not a valid number\n", name,
             arg);
  if (limit && u > limit)
    rte_exit(EXIT_FAILURE,
             "Specified %s \"%s\" is too large (greater than %u)\n", name, arg,
             limit);
  return u;
}

static void usage(void)
{
  printf("Usage: %s [options] ...\n", progname);
  printf("\n"
         "Port selections:\n"
         "  -r <port>               receive on this port (default: 1)\n"
         "  -t <port>               transmit on this port (default: 0)\n"
         "Stop conditions:\n"
         "  -c <packet count>       stop after n packets (def: 10,000)\n"
         "  -d <duration_in_secs>   stop after duration (seconds) (def: 0)\n"
         "\n"
         "Miscellaneous:\n"
         "  -v, --version           print version information and exit\n"
         "  -h, --help              display this help and exit\n"
         "\n"
         "Use Ctrl-C to stop at any time.\n");
}


/*
 * Parse command line options.
 */
static void parse_opts(int argc, char **argv)
{
  static const struct option long_options[] = {
    {"count", required_argument, NULL, 'c'}, /* Packet count */
    {"duration", required_argument, NULL, 'd'}, /* Duration in seconds */
    {"help", required_argument, NULL, 'h'}, /* Display help (usage) */
    {"rx", required_argument, NULL, 'r'}, /* RX port number */
    {"tx", required_argument, NULL, 't'}, /* TX port number */
    {"version", no_argument, NULL, 'v'},
    {NULL},
  };
  int option_index, c;

  for (;;) {
    c = getopt_long(argc, argv, "c:d:hr:t:v", long_options, &option_index);

    if (c == -1)
      break;

    switch (c) {
    case 'c':
      NUM_PKTS = get_uint(optarg, "count", 2000000000U);
      break;
    case 'd':
      DURATION = get_uint(optarg, "duration", 60U);
      break;
    case 'h':
      printf("%s\n", version());
      usage();
      exit(0);
    case 'r':
      RX_PORT = get_uint(optarg, "rx", rte_eth_dev_count_avail() - 1U);
      break;
    case 't':
      TX_PORT = get_uint(optarg, "tx", rte_eth_dev_count_avail() - 1U);
      break;
    case 'v':
      printf("%s\n", version());
      exit(0);
    default:
      fprintf(stderr, "Invalid option: %s\n", argv[optind - 1]);
      usage();
      exit(1);
    }
  }
}

static void *monitor_thread(void *arg)
{
  (void)arg;

  printf("HELLO FROM THE MONITOR THREAD! HOW ARE YOU?\n");

  const time_t start_secs = time(NULL);

  /*
   * Run until:
   * User presses ctrl-c
   * total_packets >= NUM_PKTS
   * DURATION seconds has elapsed
   */
  bool do_stop = false;
  do {
    wait_usec(1 * 1000 * 1000);
    bool ctrl_c = atomic_load(&quit_flag) ? true : false;
    const uint64_t packets = atomic_load(&total_packets);
    const time_t now_secs = time(NULL);
    const time_t duration = now_secs >= start_secs ? now_secs - start_secs : 0U;
    if (NUM_PKTS > 0U && packets >= NUM_PKTS)
      do_stop = true;
    if (DURATION > 0U && duration >= DURATION)
      do_stop = true;
    if (ctrl_c)
      do_stop = true;
    if (!do_stop) {
      const uint64_t bytes = atomic_load(&total_bytes);
      printf("seconds = %u, ", (unsigned int)duration);
      printf("total_packets: %" PRIu64 ", ", packets);
      printf("total_bytes: %" PRIu64 "\r", bytes);
      fflush(stdout);
    }
  } while (!do_stop);
  printf("\n");
  fflush(stdout);

  atomic_store(&quit_flag, true); /* Signal stop */

  return NULL;
}

int main(int argc, char *argv[])
{
  struct worker_info *info = NULL;
  struct rte_mempool *mbuf_pool = NULL;

  progname = argv[0];

  /* catch ctrl-c so we can print on exit */
  signal(SIGINT, int_handler);

  /* Initialize the Environment Abstraction Layer (EAL). */
  int ret = rte_eal_init(argc, argv);
  if (ret < 0)
    rte_exit(EXIT_FAILURE, "Error with EAL initialization\n");

  argc -= ret;
  argv += ret;

  /* Get the number of ports available to the application */
  const uint16_t nb_ports = rte_eth_dev_count_avail();
  if (nb_ports < 1)
    rte_exit(EXIT_FAILURE, "No ports found\n");

  parse_opts(argc, argv);

  /* Initialize all ports. */
  for (uint8_t port_id = 0; port_id < nb_ports; port_id++) {
    if (port_init(port_id, &mbuf_pool) != 0)
      rte_exit(EXIT_FAILURE, "Cannot init port %"PRIu8 "\n", port_id);
  }

  struct rte_flow *flow = setup_pass_all_filter(RX_PORT);

  /* Check required resources */
  if (rte_lcore_count() < nb_ports) {
    rte_exit(EXIT_FAILURE, "Need at least %d cores specified", nb_ports);
  }

  /* Launch worker threads. */
  atomic_store(&total_packets, 0);
  atomic_store(&total_bytes, 0);

  /* Wait for link state machine to make link-up */
  wait_usec(5 * 1000 * 1000);

  pthread_t mon_thread;
  int rv = pthread_create(&mon_thread, NULL, monitor_thread, NULL);
  if (rv != 0) {
    perror("pthread_create");
    return -2;
  }

  info = calloc(nb_ports, sizeof(struct worker_info));

  printf("Transmit on port %u, receive on port %u\n", TX_PORT, RX_PORT);
  unsigned int start_core = (unsigned int)-1;
  const int skip_main_flag = true;
  const int wrap_flag = false;
  unsigned int lcore_id =
    rte_get_next_lcore(start_core, skip_main_flag, wrap_flag);
  for (unsigned int i = 0; i < nb_ports; i++) {
    info[i].port = i;
    info[i].queue = 0;
#if (0)
    printf("Launching rx worker thread %u on lcore_id = %u\n", i, lcore_id);
#endif
    rte_eal_remote_launch((lcore_function_t *)lcore_worker, &info[i], lcore_id);
    lcore_id = rte_get_next_lcore(lcore_id, skip_main_flag, wrap_flag);
  }

  /* Call lcore_main on the master core only. */
  lcore_main(mbuf_pool);

  /* Wait for worker threads to stop. */
  lcore_id = rte_get_next_lcore(start_core, skip_main_flag, wrap_flag);
  for (unsigned int i = 0; i < nb_ports; i++) {
    if (rte_eal_wait_lcore(lcore_id) < 0) {
    free(info);
      return -1;
  }
    lcore_id = rte_get_next_lcore(lcore_id, skip_main_flag, wrap_flag);
  }

  /* Wait for monitor thread */
  (void)pthread_join(mon_thread, NULL);

  /* Destroy flow created by setup_pass_all_filter(0); */
  struct rte_flow_error error;
  memset(&error, 0x0, sizeof(error));
  rte_flow_destroy(0, flow, &error);

  free(info);

  /* Stop and close down ports. */
  for (uint8_t port_id = 0; port_id < nb_ports; port_id++) {
    port_close(port_id);
  }

  printf("total_packets: %" PRIu64 "\n", total_packets);
  printf("total_bytes: %" PRIu64 "\n", total_bytes);
  printf("total_errors: %" PRIu64 "\n", total_pkt_errors);

  bool success = total_packets > 0U && total_pkt_errors == 0U ? true : false;

  printf("Test %s.\n", (success ? "passed" : "failed"));

  return success ? 0 : 1;
}
