/*
 * BSD LICENSE
 *
 * Copyright(c) 2010-2015 Intel Corporation. All rights reserved.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in
 *     the documentation and/or other materials provided with the
 *     distribution.
 *   * Neither the name of Intel Corporation nor the names of its
 *     contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <rte_flow.h>
#include "meter_setup.h"

int meter_handle_create(struct meter_context_s *context,
  uint8_t port_id, uint32_t id, uint64_t rate, uint64_t size)
{
  struct rte_mtr_error error;

  context->id = id;
  context->port_id = port_id;

  /* 2. The application creates the required meter profiles */
  context->profile.alg = RTE_MTR_TRTCM_RFC2698;
  context->profile.trtcm_rfc2698.cir = rate;          /* Bucket 2 rate */
  context->profile.trtcm_rfc2698.pir = rate;          /* Bucket 1 rate */
  context->profile.trtcm_rfc2698.cbs = size;          /* Bucket 2 size */
  context->profile.trtcm_rfc2698.pbs = size;          /* Bucket 1 size */
  context->profile.packet_mode = 0;                   /* 0 -> byte mode; !0 -> packet mode */

  if (rte_mtr_meter_profile_add(port_id, id, &context->profile, &error)) {
    printf("Error: Cannot add meter profile: %s\n", error.message);
    return -1;
  }

  /* 3. The application creates the required meter policies */
  struct rte_flow_action actions_green[2];
  struct rte_flow_action actions_yellow[2];
  struct rte_flow_action actions_red[2];

  actions_green[0].type = RTE_FLOW_ACTION_TYPE_VOID;
  actions_green[1].type = RTE_FLOW_ACTION_TYPE_END;

  actions_yellow[0].type = RTE_FLOW_ACTION_TYPE_DROP;
  actions_yellow[1].type = RTE_FLOW_ACTION_TYPE_END;

  actions_red[0].type = RTE_FLOW_ACTION_TYPE_DROP;
  actions_red[1].type = RTE_FLOW_ACTION_TYPE_END;

  context->policy.actions[RTE_COLOR_GREEN] = actions_green;
  context->policy.actions[RTE_COLOR_YELLOW] = actions_yellow;
  context->policy.actions[RTE_COLOR_RED] = actions_red;

  if (rte_mtr_meter_policy_add(port_id, id, &context->policy, &error)) {
    printf("Error: Cannot add meter policy : %s\n", error.message);
    return -1;
  }

  return 0;
}


int meter_handle_destroy(struct meter_context_s *context)
{
  struct rte_mtr_error error;

  if (rte_mtr_meter_profile_delete(context->port_id, context->id, &error)) {
    printf("Error: Cannot delete meter profile : %s\n", error.message);
    return -1;
  }

  if (rte_mtr_meter_policy_delete(context->port_id, context->id, &error)) {
    printf("Error: Cannot delete meter policy %s\n", error.message);
    return -1;
  }

  return 0;
}

int meter_learn(struct meter_context_s *context, uint32_t mtr_id, uint64_t stats_mask)
{
  struct rte_mtr_error error;

  /* 4. The application creates a meter object */
  struct rte_mtr_params params;
  int shared = 1;
  int res;

  params.meter_profile_id = context->id;
  params.use_prev_mtr_color = 0;
  params.dscp_table = NULL;
  params.vlan_table = NULL;
  params.meter_enable = 1;
  params.stats_mask = stats_mask;
  params.meter_policy_id = context->id;

  res = rte_mtr_create(context->port_id, mtr_id, &params, shared, &error);

  if (res != 0) {
    printf("Error: Cannot create meter[%d]: %s\n", mtr_id, error.message);
  }

  return res;
}

int meter_unlearn(struct meter_context_s *context, uint32_t mtr_id)
{
  struct rte_mtr_error error;
  int res;

  res = rte_mtr_destroy(context->port_id, mtr_id, &error);

  if (res != 0) {
    printf("Error: Cannot destroy meter[%d]: %s\n", mtr_id, error.message);
  }

  return res;
}
