/*
 * SPDX-License-Identifier: BSD-3-Clause
 * Copyright(c) 2023 Napatech A/S
 */

#include "unordered_multimap.h"

#include <stdlib.h>

struct unordered_multimap_bucket_t {
  uint64_t key;
  void*    value;

  struct unordered_multimap_bucket_t* next;
};

static uint64_t round_power_of_2(uint64_t value)
{
  value--;
  value |= value >> 1;
  value |= value >> 2;
  value |= value >> 4;
  value |= value >> 8;
  value |= value >> 16;
  value |= value >> 32;
  value++;
  return value;
}

static int unordered_multimap_rehash(struct unordered_multimap_t* umm,
  uint64_t num_buckets)
{
  const uint64_t new_mask = num_buckets - 1;
  struct unordered_multimap_bucket_t** new_buckets = calloc(num_buckets,
    sizeof(struct unordered_multimap_bucket_t*));

  for (uint64_t i = 0; i < umm->num_buckets; ++i) {
    struct unordered_multimap_bucket_t* next_bucket = umm->buckets[i];
    while (next_bucket != NULL) {
      struct unordered_multimap_bucket_t* bucket = next_bucket;
      next_bucket = bucket->next;

      bucket->next = new_buckets[bucket->key & new_mask];
      new_buckets[bucket->key & new_mask] = bucket;
    }
  }

  free(umm->buckets);
  umm->buckets = new_buckets;
  umm->num_buckets = num_buckets;

  return 0;
}

int unordered_multimap_init(struct unordered_multimap_t** umm,
  uint64_t (*hash_func)(void*), bool (*compare)(void*, void*))
{
  *umm = malloc(sizeof(struct unordered_multimap_t));

  (*umm)->num_buckets = 1024;
  (*umm)->num_elements = 0;

  (*umm)->max_load_factor = 1.0f;

  (*umm)->hash_func = hash_func;
  (*umm)->compare = compare;

  (*umm)->buckets = calloc((*umm)->num_buckets,
    sizeof(struct unordered_multimap_bucket_t*));

  return 0;
}

void unordered_multimap_destroy(struct unordered_multimap_t* umm,
  bool free_elements)
{
  if (umm == NULL) {
    return;
  }

  for (uint64_t i = 0; i < umm->num_buckets; ++i) {
    struct unordered_multimap_bucket_t* bucket_next = umm->buckets[i];
    while (bucket_next != NULL) {
      struct unordered_multimap_bucket_t* bucket = bucket_next;
      bucket_next = bucket->next;
      if (free_elements) {
        free(bucket->value);
      }
      free(bucket);
    }
  }

  free(umm->buckets);
  free(umm);
}

int unordered_multimap_reserve(struct unordered_multimap_t* umm,
  uint64_t count)
{
  count = round_power_of_2(count);
  if (count == umm->num_buckets) {
    return  0;
  }

  float new_load_factor = (float)umm->num_elements / (float)count;
  if (new_load_factor > umm->max_load_factor) {
    return -1;
  }

  return unordered_multimap_rehash(umm, count);
}

int unordered_multimap_insert(struct unordered_multimap_t* umm,
  void* element)
{
  struct unordered_multimap_bucket_t* new_bucket = calloc(1,
    sizeof(struct unordered_multimap_bucket_t));

  new_bucket->key = umm->hash_func(element);
  new_bucket->value = element;

  const uint64_t index = new_bucket->key & (umm->num_buckets - 1);

  new_bucket->next = umm->buckets[index];
  umm->buckets[index] = new_bucket;
  umm->num_elements += 1;

  float load_factor = (float)umm->num_elements / (float)umm->num_buckets;
  return (load_factor > umm->max_load_factor)
    ? unordered_multimap_rehash(umm, umm->num_buckets << 1)
    : 0;
}

int unordered_multimap_insert_unique(struct unordered_multimap_t* umm,
  void* element)
{
  const uint64_t key = umm->hash_func(element);
  const uint64_t index = key & (umm->num_buckets - 1);

  /* Return if element already exists in container */
  struct unordered_multimap_bucket_t* bucket = umm->buckets[index];
  while (bucket != NULL) {
    if (bucket->key == key && umm->compare(element, bucket->value)) {
      return 0;
    }
    bucket = bucket->next;
  }

  /* Add element */
  struct unordered_multimap_bucket_t* new_bucket = calloc(1,
    sizeof(struct unordered_multimap_bucket_t));

  new_bucket->key = key;
  new_bucket->value = element;
  new_bucket->next = umm->buckets[index];

  umm->buckets[index] = new_bucket;
  umm->num_elements += 1;

  /* Rehash if needed */
  float load_factor = (float)umm->num_elements / (float)umm->num_buckets;
  return (load_factor > umm->max_load_factor)
    ? unordered_multimap_rehash(umm, umm->num_buckets << 1)
    : 0;
}

void unordered_multimap_erase(struct unordered_multimap_t* umm,
  void* element)
{
  const uint64_t key = umm->hash_func(element);
  const uint64_t index = key & (umm->num_buckets - 1);

  struct unordered_multimap_bucket_t** bucket_prev = &umm->buckets[index];
  struct unordered_multimap_bucket_t* bucket = *bucket_prev;
  while (bucket != NULL) {
    if (bucket->key == key && umm->compare(element, bucket->value)) {
      *bucket_prev = bucket->next;
      free(bucket);
      break;
    }
    bucket_prev = &bucket->next;
    bucket = *bucket_prev;
  }
}

bool unordered_multimap_contains(struct unordered_multimap_t* umm,
  void* element)
{
  const uint64_t key = umm->hash_func(element);
  const uint64_t index = key & (umm->num_buckets - 1);

  struct unordered_multimap_bucket_t* bucket = umm->buckets[index];
  while (bucket != NULL) {
    if (bucket->key == key && umm->compare(element, bucket->value)) {
      return true;
    }
    bucket = bucket->next;
  }

  return false;
}
