/*
 * SPDX-License-Identifier: BSD-3-Clause
 * Copyright(c) 2023 Napatech A/S
 */

#ifndef NT_UNORDERED_MULTIMAP_H
#define NT_UNORDERED_MULTIMAP_H

#include <stdbool.h>
#include <stdint.h>

struct unordered_multimap_t {
  uint64_t num_buckets;
  uint64_t num_elements;

  float max_load_factor;

  uint64_t (*hash_func)(void*);
  bool (*compare)(void*, void*);

  struct unordered_multimap_bucket_t** buckets;
};

/*
 * Initialize data-structure with needed function pointers.
 * Returns 0 on success.
 */
int unordered_multimap_init(struct unordered_multimap_t** umm,
  uint64_t (*hash_func)(void*), bool (*compare)(void*, void*));

/*
 * Destroy data-structure.
 */
void unordered_multimap_destroy(struct unordered_multimap_t* umm,
  bool free_elements);

/*
 * Reserve memory for a number of elements. Causes rehashing.
 * Return 0 on success.
 */
int unordered_multimap_reserve(struct unordered_multimap_t* umm,
  uint64_t count);

/*
 * Insert element into data-structure.
 * Causes rehashing if max_load_factor > 1.0.
 * Return 0 on success.
 */
int unordered_multimap_insert(struct unordered_multimap_t* umm,
  void* element);

/*
 * Insert element into data-structure if the element is not already contained.
 * Causes rehashing if max_load_factor > 1.0.
 * Return 0 on success.
 */
int unordered_multimap_insert_unique(struct unordered_multimap_t* umm,
  void* element);

/*
 * Delete element from data-structure.
 * If the element is contained multiple times,
 * only the first instance is erased.
 */
void unordered_multimap_erase(struct unordered_multimap_t* umm,
  void* element);

/*
 * Check if an element exists in the data-structure.
 * Returns true if the element exists.
 */
bool unordered_multimap_contains(struct unordered_multimap_t* umm,
  void* element);

#endif  /* NT_UNORDERED_MULTIMAP_H */
