/*
 * BSD LICENSE
 *
 * Copyright(c) 2010-2015 Intel Corporation. All rights reserved.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in
 *     the documentation and/or other materials provided with the
 *     distribution.
 *   * Neither the name of Intel Corporation nor the names of its
 *     contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdint.h>
#include <inttypes.h>
#include <signal.h>
#include <rte_eal.h>
#include <rte_ethdev.h>
#include <rte_lcore.h>
#include <rte_mbuf.h>
#include <rte_flow.h>
#include <rte_net.h>
#include <rte_version.h>
#include <unistd.h>
#include <getopt.h>
#include <arpa/inet.h>
#include <rte_build_config.h>

#define wait_usec(x) rte_delay_us_sleep(x)

#define RX_RING_SIZE 128
#define TX_RING_SIZE 512

#define NUM_MBUFS 8192
#define MBUF_CACHE_SIZE 250
#define BURST_SIZE 32

static const struct rte_eth_conf port_conf_default = {
#if (RTE_VERSION_NUM(21, 11, 0, 0) <= RTE_VERSION)
  .rxmode = { .max_lro_pkt_size = RTE_ETHER_MAX_LEN }
#else
  .rxmode = { .max_rx_pkt_len = RTE_ETHER_MAX_LEN }
#endif
};

volatile uint8_t quit_signal;
volatile unsigned int nb_ports = 0;
volatile uint64_t packets[2] = { 0, 0};

volatile int  opt_queues = 1;

char version[256];

static void int_handler(int sig_num)
{
  printf("Exiting on signal %d\n", sig_num);
  /* set quit flag for rx thread to exit */
  quit_signal = 1;
}

/*
 * Initializes a given port using global settings and with the RX buffers
 * coming from the mbuf_pool passed as a parameter.
 */
static inline int port_init(uint8_t port, struct rte_mempool *mbuf_pool, uint8_t queues)
{
  struct rte_eth_conf port_conf = port_conf_default;
  struct rte_eth_dev_info dev_info;
  int retval;
  uint16_t q;

  if (port >= rte_eth_dev_count_avail())
    return -1;

  retval = rte_eth_dev_info_get(port, &dev_info);
  if (retval != 0)
    return retval;

  /* Configure the Ethernet device. */
  retval = rte_eth_dev_configure(port, queues, queues, &port_conf);
  if (retval != 0)
    return retval;

  /* Allocate and set up 1 TX queue per Ethernet port. */
  for (q = 0; q < queues; q++) {
    retval = rte_eth_tx_queue_setup(port, q, TX_RING_SIZE, rte_eth_dev_socket_id(port), NULL);
    if (retval < 0)
      return retval;
  }

  /* Allocate and set up 1 RX queue per Ethernet port. */
  for (q = 0; q < queues; q++) {
    retval = rte_eth_rx_queue_setup(port, q, RX_RING_SIZE, rte_eth_dev_socket_id(port), NULL, mbuf_pool);
    if (retval < 0)
      return retval;
  }

  /* Start the Ethernet port. */
  retval = rte_eth_dev_start(port);
  if (retval < 0)
    return retval;

  /* Disable RX in promiscuous mode for the Ethernet device. */
  rte_eth_promiscuous_disable(port);

  return 0;
}

static int lcore_main(void)
{
  unsigned int port;

  /* Run until the application is quit or killed. */
  printf("NT Main App started\n");
  for (;;) {
    if (quit_signal) break;
    for (port = 0; port < nb_ports; port++) {
      /* Get burst of RX packets */
      struct rte_mbuf *bufs[BURST_SIZE];
      const uint16_t nb_rx = rte_eth_rx_burst(port, 0, bufs, BURST_SIZE);
      if (likely(nb_rx)) {
        packets[port] += nb_rx;
        rte_pktmbuf_free_bulk(bufs, nb_rx);
      }
    }
  }
  return 0;
}

static int lcore_stat(__rte_unused void *p)
{
  uint8_t port;

  return 0;
  printf("\n\n");
  while (quit_signal == 0) {
    wait_usec(1 * 1000 * 1000);
    for (port = 0; port < 2; port++) {
      printf("P%u: %16llu   ", port, (unsigned long long)packets[port]);
  }
    printf("\r"); fflush(stdout);
  }
  printf("\n\n");
  return 0;
}

static const char short_options[] = "h";

static struct option long_options[] = {
            {"queues",    required_argument, 0,  0 },
            {0,           0,                 0,  0 }
        };
int option_index = 0;

/* display usage */
static void forward_usage(const char *prgname)
{
  printf("\n%s [-h][-v]\n"
         "  -h                 : Display help\n"
         "  --queues <number>  : Number of queues to use for VFs\n"
         "\n\n", prgname);
}

/* Parse the argument given in the command line of the application */
static int parse_args(int argc, char **argv)
{
  int opt, ret;
  char **argvopt;
  char *prgname = argv[0];

  argvopt = argv;
  int option_index = 0;


 while ((opt = getopt_long_only(argc, argvopt, short_options, long_options, &option_index)) != -1) {
   printf("opt: %u, option_index %u\n", opt, option_index);
   switch (opt) {
   case 'h':
     forward_usage(prgname);
     exit(0);
   case 0:
     switch (option_index) {
      case 0:
        opt_queues = strtoul(optarg, NULL, 10);
        break;
      }
    }
  }

  if (optind >= 0)
    argv[optind-1] = prgname;

  ret = optind-1;
  optind = 1; /* reset getopt lib */
  return ret;
}

int main(int argc, char *argv[])
{
  struct rte_mempool *mbuf_pool;
  unsigned int portid;
  unsigned int lcore_id;
  int q;
  int ret;

  quit_signal = 0;

  /* catch ctrl-c so we can print on exit */
  signal(SIGINT, int_handler);

  /* Initialize the Environment Abstraction Layer (EAL). */
  ret = rte_eal_init(argc, argv);
  if (ret < 0) {
    rte_exit(EXIT_FAILURE, "Error with EAL initialization\n");
  }

  argc -= ret;
  argv += ret;

  parse_args(argc, argv);

  /* Check that there is any ports to send/receive on. */
  nb_ports = rte_eth_dev_count_avail();
  if (nb_ports < 2) {
    rte_exit(EXIT_FAILURE, "Need 2 ports\n");
  }

  if (nb_ports > RTE_MAX_ETHPORTS) {
    rte_exit(EXIT_FAILURE, "Too many ports. Increase the RTE_MAX_ETHPORTS define\n");
  }

  /* Get the firmware version */
  ret = rte_eth_dev_fw_version_get(0, version, 256);
  if (ret > 0) {
      printf("Firmware version: %s\n", version);
  }

  /* Creates a new mempool in memory to hold the mbufs. */
  mbuf_pool = rte_pktmbuf_pool_create("MBUF_POOL", NUM_MBUFS * nb_ports, MBUF_CACHE_SIZE, 0, RTE_MBUF_DEFAULT_BUF_SIZE, rte_socket_id());

  if (mbuf_pool == NULL) {
    rte_exit(EXIT_FAILURE, "Cannot create mbuf pool\n");
  }

  /* Initialize the main ports. */
  for (portid = 0; portid < 2; portid++) {
    if (port_init(portid, mbuf_pool, 1) != 0) {
      rte_exit(EXIT_FAILURE, "Cannot init port %"PRIu8 "\n", portid);
    }
  }

  /* Initialize the virtual ports for the VFs. */
  for (; portid < nb_ports; portid++) {
    if (port_init(portid, mbuf_pool, opt_queues) != 0) {
      rte_exit(EXIT_FAILURE, "Cannot init port %"PRIu8 "\n", portid);
    }
  }

  if (rte_lcore_count() < 1) {
    rte_exit(EXIT_FAILURE, "Need at least 1 cores specified");
  }

  /* Start queues for the main ports */
  for (portid = 0; portid < 2; portid++) {
    if (rte_eth_dev_rx_queue_start(portid, 0) < 0)
      rte_exit(EXIT_FAILURE, "Error starting queue");

    if (rte_eth_dev_tx_queue_start(portid, 0) < 0)
      rte_exit(EXIT_FAILURE, "Error starting queue");
  }

  /* Start queues for the virtual VF ports */
  for (; portid < nb_ports; portid++) {
    for (q = 0; q < opt_queues; q++) {
      if (rte_eth_dev_rx_queue_start(portid, q) < 0)
        rte_exit(EXIT_FAILURE, "Error starting queue");
    }
    for (q = 0; q < opt_queues; q++) {
      if (rte_eth_dev_tx_queue_start(portid, q) < 0)
        rte_exit(EXIT_FAILURE, "Error starting queue");
    }
  }

  /* launch worker thread for statistics. */
  lcore_id = rte_get_next_lcore(-1, 1, 0);
  if (rte_eal_remote_launch((lcore_function_t *)lcore_stat, NULL, lcore_id) < 0) {
    return -1;
  }

  lcore_main();

  if (rte_eal_wait_lcore(lcore_id) < 0) {
    return -1;
  }

  /* Stop queues for the main ports */
  for (portid = 0; portid < 2; portid++) {
    if (rte_eth_dev_rx_queue_stop(portid, 0) < 0)
      rte_exit(EXIT_FAILURE, "Error stopping queue");

    if (rte_eth_dev_tx_queue_stop(portid, 0) < 0)
      rte_exit(EXIT_FAILURE, "Error stopping queue");
  }

  /* Stop queues for the virtual VF ports */
  for (; portid < nb_ports; portid++) {
    for (q = 0; q < opt_queues; q++) {
      if (rte_eth_dev_rx_queue_stop(portid, q) < 0)
        rte_exit(EXIT_FAILURE, "Error stopping queue");
    }
    for (q = 0; q < opt_queues; q++) {
      if (rte_eth_dev_tx_queue_stop(portid, q) < 0)
        rte_exit(EXIT_FAILURE, "Error stopping queue");
    }
  }

  /* Stop and close down ports */
  for (portid = 0; portid < nb_ports; portid++) {
    rte_eth_dev_stop(portid);
  }

  for (portid = 0; portid < nb_ports; portid++) {
    rte_eth_dev_close(portid);
  }

  /* Do final cleanup */
  ret = rte_eal_cleanup();
  if (ret != 0) {
     rte_exit(EXIT_FAILURE, "EAL cleanup failed\n");
  }

  return 0;
}
