# Napatech Device Plugin (NTDevPlugin) for Kubernetes

In order to use Napatech Smart NIC capabilities inside container managed by Kubernetes, a Napatech Device
Plugin is required. It will provide a requested capabilities to selected Kubernetes PODs, so containerized
applications can benefit from hardware acceleration features of Napatech Smart NIC adapters.
Following resources can be allocated to Kubernetes PODs:
* PF / Physical Function - it can be provided up to one container. It is expected to be used by a main
  DPDK application with NT NIC PMD. In case that Napatech main-app is executed automatically as part
  of NTDevPlugin POD, then PF resource manager is disabled.
* VFs / Virtual Functions - they can be provided to multiple containers and each container can be provided
  with multiple VFs. These VFs shall be used by DPDK application, which processes networking traffic.
  This application shall use Virtio NIC PMD.

NOTE: Napatech Device Plugin doesn't provide any networking capabilities on top of allocated NT SmartNIC
resources.

## Directory structure

```
ntdevplugin/
|-- Dockerfile.ntdevplugin              # dockerfile for NTDevPlugin container image
|-- Dockerfile.ntmain                   # dockerfile for NT Main application container image
|-- README.md                           # this documentation
|-- ntdevplugin.go                      # NTDevPlugin implementation
|-- ntdevplugin.yaml                    # manifest file for deployment and configuration of NTDevPlugin POD
|-- start-init.sh                       # start script of NT SmartNIC configuration container
`-- start-ntmain.sh                     # start script of NT Main application container
```

## NTDevPlugin k8s POD architecture

The Kubernetes POD of NTDevPlugin consists of three containers and its configuration is defined at k8s ConfigMap.

Schema:
```
 ......................    ....................................
 : ConfigMap          :    : Napatech Device Plugin POD       :
 : - NicPci           :    :         +-------------+          :
 : - NicAutoInit      :    :         |    init     |          :
 : - MainApp          :    :         |  container  |          :
 : - vfCount          :    :         +-------------+          :
 : - vfMacStart       :    : +-------------+  +-------------+ :
 : - vfVlanStart      :    : | NTDevPlugin |  | NT main-app | :
 : - vfNumberOfQueues :    : |  container  |  |  container  | :
 :....................:    : +-------------+  +-------------+ :
                           :..................................:
```

### Configuration

The NTDevPlugin ConfigMap `napatech-device-plugin-config` can be found at manifest file `ntdevplugin.yaml`.
Its configuration parameters affect behaviour of NTDevPlugin containers.

Following parameters are supported:

| Parameter        | Values                   | Description 
|------------------|--------------------------|--------------|
| NicPci           | autodetect / PCI address | specifies NT SmartNIC PCI address or `autodetect` for automatic detection
| NicAutoInit      | enabled / disabled       | enables NT SmartNIC automatic configuration; The value `disabled` causes POD failure if NIC configuration is incorrect.
| MainApp          | enabled / disabled       | enables NT main-app and turns off `nt_pf` resource availability; The value `disabled` turns off NT main-app and enables `nt_pf` resource.
| vfCount          | uint                     | specifies a number of VFs to be configured at NIC and provided by NTDevPlugin
| vfMacStart       | MAC                      | specifies a MAC address of the 1st VF
| vfVlanStart      | uint                     | specifies a VLAN id of the 1st VF
| vfNumberOfQueues | uint                     | specifies a number of queues to be configured per VF

Configuration example:

```yaml
apiVersion: v1
kind: ConfigMap
metadata:
  name: napatech-device-plugin-config
  namespace: kube-system
data:
  NicPci: "autodetect"
  NicAutoInit: "enabled"
  MainApp: "enabled"
  vfCount: "2"
  vfMacStart: "02:11:22:33:44:00"
  vfVlanStart: "44"
  vfNumberOfQueues: "1"
immutable: true
```

### Init container

The init container is started before NTDevPlugin and NT main-app containers. It verifies that NT SmartNIC specified
by `NicPci` is properly bound to DPDK driver and requested `vfCount` number of VFs is available. In case that
option `NicAutoInit` is set to `enabled`, then NIC will be automatically bound to correct driver and required
number of VFs will be configured.
In case that NT SmartNIC is not available, its configuration is incorrect or its auto configuration fails, then
error will be reported into container log file. In that case NTDevPlugin and NT main-app containers won't be started.

### NTDevPlugin container

The NTDevPlugin container runs Napatech Device Plugin for Kubernetes. Based on the value of `MainApp` parameter,
it provides to Kubernetes NT SmartNIC resources `nt_vf` and `nt_pf` (`MainApp` is `disabled`) or only `nt_vf` resources
(`MainApp` is `enabled`).

The NTDevPlugin preferred allocation of VFs (i.e. `nt_vf` resources) is *in order*, i.e. it tries to allocate available
VFs with the lowest number (e.g. for request `nt_vf: 2` to allocate VFs 4 and 5). However, the final decision and order
of allocated VFs is driven by Kubernetes.

### Napatech main-app container

The Napatech main-app container runs `ntmain` application if `MainApp` parameter is set to `enabled`. This application will
attach to NT SmartNIC PF and configure flows to dispatch traffic among VFs based on VLAN ids. The VLAN id of the 1st VF
is specified by `vfVlanStart` parameter and it is increased by 1 for each other VF until flows for all VFs specified by
`vfCount` are created.

## Prerequisites

Following configuration is expected and out of scope of this document:
* Kubernetes cluster with compatible Napatech SmartNIC (e.g. NT200A02) installed at worker node
* Napatech SmartNIC flashed with Inline FPGA FW version 9563-55-16 or newer
* 1GB hugepages configured and mounted

## Build of container images

Provided docker files can be used to generate a local container images with NTDevPlugin and ntmain app. It is possible
to use `docker`, `podman` or `nerdctl` to build images. Please select a build method, which suits
your Kubernetes configuration, especially chosen container runtime.

Example of images build from package root directory:
```bash
docker build -t k8s.io/ntdevplugin:1.5 -f kubernetes_v22.11.1/ntdevplugin/Dockerfile.ntdevplugin kubernetes_v22.11.1/ntdevplugin/.
docker build -t k8s.io/ntmain:1.1 -f kubernetes_v22.11.1/ntdevplugin/Dockerfile.ntmain .
```
NOTE: Both *NT main-app* and *Init* containers use the same image *k8s.io/ntmain*.

## Plugin configuration and deployment

The NTDevPlugin POD has to be configured to properly detect and configure NT SmartNIC and to successfully start all three
containers. Please refer to *naptech-device-plugin-config* ConfigMap and its parameters described above.

NOTE: Provided manifest file `ntdevplugin.yaml` enables NT SmartNIC autodetection feature, so manual modification
is required only in case that autodetection fails. Please note, that only a single NT SmartNIC can be installed at host.

Provided `ntdevplugin.yaml` can be used for NTDevPlugin deployment as follows:

```bash
kubectl apply -f kubernetes_v22.11.1/ntdevplugin/ntdevplugin.yaml
```

Please check NTDevPlugin POD log messages to verify, that NT SmartNIC was properly detected and both NTDevPlugin and NT
main-app containers have started correctly, e.g.:

```bash
kubectl logs -n kube-system -l app=napatech-device-plugin --all-containers --tail=-1
```

### Plugin deployment at selected k8s nodes

The NTDevPlugin POD shall be deployed only at k8s nodes, where compatible NT SmartNIC with Inline FPGA FW is available.
Thus it is recommended to label such nodes with NT specific label and to add appropriate node selector into NTDevPlugin
POD yaml file.

Following `nodeSelector` section shall be added into `ntdevplugin.yaml` manifest file before NTDevPlugin POD deployment:
```bash
...
spec:
  selector:
    matchLabels:
      app: napatech-device-plugin
  template:
    metadata:
      labels:
        app: napatech-device-plugin
    spec:
      nodeSelector:
        napatech.com/ntdevplugin: enabled
...
```

After `nodeSelector` is properly configured, then k8s nodes with NT SmartNIC shall be labeled, e.g. following command
will enable NTDevPlugin at worker nodes 'worker03` and 'worker04':
```bash
kubectl label nodes worker03 worker04 napatech.com/ntdevplugin=enabled
```

## Plugin deletion

In case that NTDevPlugin has been deployed by `ntdevplugin.yaml` manifest described above, then it can
be deleted from Kubernetes cluster as follows:

```bash
kubectl delete -f kubernetes_v22.11.1/ntdevplugin/ntdevplugin.yaml
```

## NTDevPlugin usage

After proper installation and configuration of Napatech Device Plugin, it is necessary to properly configure
your application manifest and startup script to properly assign and use NT Smart NIC resources.

### Resource allocation

Requested NT SmartNIC resource type shall be set at manifest file of your application. Following resources
are supported:
* `napatech.com/nt_pf: 1` - NT SmartNIC Physical Function allocation; It can be allocated to only one container.
  Container will get information about PF0 device PCI address and details about all configured virtual functions.
* `napatech.com/nt_vf: x` - NT SmartNIC Virtual Function allocation; It can be allocated to multiple containers,
  where 'x' specifies a number of requested VFs. In case that requested amount of VFs is not available, then
  container will remain at *Pending* state.

Example of manifest yaml file for container with NT *forward* demo application built from Vanilla DPDK:
```yaml
apiVersion: apps/v1
kind: Deployment
metadata:
  name: forward
  namespace: default
spec:
  replicas: 1
  selector:
    matchLabels:
      app: forward
  template:
    metadata:
      labels:
        app: forward
    spec:
      containers:
      - name: forward
        image: k8s.io/ntforward:1.1
        imagePullPolicy: IfNotPresent
        command: ["/opt/kubernetes/start.sh"]
        resources:
          limits:
            napatech.com/nt_vf: 2
            hugepages-1Gi: 2Gi
            memory: 512Mi
        volumeMounts:
        - name: hugepage
          mountPath: /hugepages
      volumes:
      - name: hugepage
        emptyDir:
          medium: HugePages
```

### Containerized Application configuration

After successful allocation of resources by NTDevPlugin, a set of environment variables is set with
details about allocated resources.

#### Physical Function related environment variables

Following shell variables will be set by NTDevPlugin after allocation of physical function:
* `NT_PF_NUM` - a number of allocated PF; It will be always "0"
* `NT_PF_DEV` - a PCI address of Napatech SmartNIC, e.g. "0000:65:00.0"
* `NT_DPDK_PREFIX` - a unique DPDK prefix to avoid clash with other DPDK based applications, e.g. "nt-container0"
* `NT_VF_NUM` - a list of numbers of all configured virtual functions, e.g. "5 6 7 8"
* `NT_VF_QUEUES` - a number of queues to be configured per VF, e.g. "3"
* `NT_VF_DEV` - a list of PCI addresses of all configured VFs, e.g. "0000:65:00.4 0000:65:00.5 0000:65:00.6 0000:65:00.7"
* `NT_VF_MAC` - a list of MAC addresses of all configured VFs, e.g. "02:11:22:33:44:01 02:11:22:33:44:02 02:11:22:33:44:03 02:11:22:33:44:00"
* `NT_VF_VLAN` - a list of VLAN IDs of all configured VFs, e.g. "44 45 46 47"
* `NT_VF_SOCKET` - a list of sockets of all configured VFs, e.g. "/usr/local/var/run/stdvio4/stdvio4 /usr/local/var/run/stdvio5/stdvio5 /usr/local/var/run/stdvio6/stdvio6 /usr/local/var/run/stdvio7"

#### Virtual Function related environment variables

Following shell variables will be set by NTDevPlugin after allocation of virtual function:
* `NT_VF_NUM` - a list of numbers of allocated virtual functions, e.g. "4 5"
* `NT_VF_QUEUES` - a number of queues to be configured per VF, e.g. "3"
* `NT_VF_DEV` - a list of PCI addresses of allocated VFs, e.g. "0000:65:00.4 0000:65:00.5"
* `NT_VF_MAC` - a list of MAC addresses of allocated VFs, e.g. "02:11:22:33:44:00 02:11:22:33:44:01"
* `NT_VF_VLAN` - a list of VLAN IDs of allocated VFs, e.g. "44 45"
* `NT_VF_SOCKET` - a list of sockets of allocated VFs, e.g. "/usr/local/var/run/stdvio4/stdvio4 /usr/local/var/run/stdvio5/stdvio5"
* `NT_DPDK_PREFIX` - a unique DPDK prefix to avoid clash with other DPDK based applications, e.g. "nt-container4"

### Example of startup script

Exmaple below can be used to start NT forward app with `napatech.com/nt_vf: 2` resource request.

```bash
#!/usr/bin/env bash
NT_VF_MAC_ARRAY=($NT_VF_MAC)
NT_VF_SOCKET_ARRAY=($NT_VF_SOCKET)

# prepare dpdk vdev configuration for allocated VFs
APP_VDEV=""
for i in ${!NT_VF_MAC_ARRAY[@]}; do
        APP_VDEV="$APP_VDEV --vdev=virtio_user${i},path=${NT_VF_SOCKET_ARRAY[$i]},server=1,mac=${NT_VF_MAC_ARRAY[$i]},packed_vq=1,queues=${NT_VF_QUEUES}"
        # remove old socket
        rm -f ${NT_VF_SOCKET_ARRAY[$i]}
done

# Execute app
forward -c 0xfff0 -n 4 --no-pci ${APP_VDEV} --file-prefix=${NT_DPDK_PREFIX} --socket-mem=1024 -- -i 0 -e 1 -q ${NT_VF_QUEUES}
```

## NTDevPlugin constraints

* NTDevPlugin init and NT main-app containers must be run as privileged containers.
* NTDevPlugin assigns to container any of available resource of given type. It is not possible to assign particular
  resources to given container. It means, that application has to assure, that allocated resources (e.g. VFs) are
  properly configured to receive required traffic.
* Containers requesting `napatech.com/nt_pf` resource, i.e. DPDK applications using NTNIC PMD, shall be configured with `IPC_LOCK` securityContext capability.
  It is required for proper allocation of memory for DMA.
