package main

import (
        "errors"
        "flag"
        "fmt"
        "net"
        "os"
        "path"
        "sort"
        "strconv"
        "strings"
        "sync"
        "time"

        "github.com/golang/glog"
        "github.com/jaypipes/ghw"
        pciaddress "github.com/jaypipes/ghw/pkg/pci/address"
        "golang.org/x/net/context"
        "google.golang.org/grpc"
        pluginapi "k8s.io/kubelet/pkg/apis/deviceplugin/v1beta1"
)

//////////////////////////////////////////////////////////////////////////////
///////////////////////////// Constants //////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
const ntPluginVersion = "1.5"
const ntNT200AFirstVFID = 4
const ntNT400DFirstVFID = 1
const ntVendorID = "18f4"
const ntVFProductID = "051a"
const ntNT200AProductID = "01c5"
const ntNT400DProductID = "0215"
const ntDpdkPrefix = "nt-container%d"
const ntNICSocket = "/usr/local/var/run/stdvio%d/stdvio%d"
const ntConnectSocketDir = "/var/run/ntconnect"
const ntResourceNamePF = "napatech.com/nt_pf"
const ntResourceNameVF = "napatech.com/nt_vf"
const ntSocketNamePF = "ntPhysFunc"
const ntSocketNameVF = "ntVirtFunc"
const ntPluginNameLong = "Napatech Device Plugin"
const ntPluginNameShort = "NTDevPlugin"
const ntNICAutodetect = "autodetect"
const ntWait = 5 * time.Second
var ntEnvVars = []string{"NicPci", "vfMacStart", "vfVlanStart", "vfNumberOfQueues", "MainApp"}

//////////////////////////////////////////////////////////////////////////////
///////////////////// Global variables & types ///////////////////////////////
//////////////////////////////////////////////////////////////////////////////
type ntInfoType struct {
        pf_address []string
        vf_address []string
        vf_first_vfid uint64
        vf_id map[string]uint64
        vf_mac uint64
        vf_vlan uint64
        vf_queues uint64
}
func NewNTInfo() *ntInfoType {
        return &ntInfoType{
                pf_address: make([]string, 0),
                vf_address: make([]string, 0),
                vf_first_vfid: ntNT200AFirstVFID,
                vf_id: make(map[string]uint64),
                vf_mac: 0,
                vf_vlan: 0,
                vf_queues: 0,
        }
}

// Device manager interface
type ntDeviceManager interface {
        // plugin interface required by k8s
        Allocate(ctx context.Context, rqt *pluginapi.AllocateRequest) (*pluginapi.AllocateResponse, error)
        GetDevicePluginOptions(ctx context.Context, empty *pluginapi.Empty) (*pluginapi.DevicePluginOptions, error)
        GetPreferredAllocation(ctx context.Context, rqt *pluginapi.PreferredAllocationRequest) (*pluginapi.PreferredAllocationResponse, error)
        PreStartContainer(ctx context.Context, rqt *pluginapi.PreStartContainerRequest) (*pluginapi.PreStartContainerResponse, error)
        ListAndWatch(emtpy *pluginapi.Empty, stream pluginapi.DevicePlugin_ListAndWatchServer) error
        // NT specific methods
        Init()
        GetSocket() string
        GetEndpoint() string
        GetResourceName() string
        GetServer() *grpc.Server
        SetServer(server *grpc.Server)
        CheckNTDriver() []string
        GetMounts(deviceIDs []string) []*pluginapi.Mount
}

// ntFuncManager generic struct
type ntFuncManager struct {
        devices map[string]*pluginapi.Device
        dev_addresses *[]string
        server  *grpc.Server
        socket_name string
        resource_name string
        endpoint string
}

// ntPhysFuncManager manages Napatech PF devices
type ntPhysFuncManager struct {
        ntFuncManager           // inherit generic ntFuncManager
}

// ntVirtFuncManager manages Napatech VF devices
type ntVirtFuncManager struct {
        ntFuncManager           // inherit generic ntFuncManager
}

var wg sync.WaitGroup
var ntInfo *ntInfoType
var ntPhysFunc *ntPhysFuncManager
var ntVirtFunc *ntVirtFuncManager

//////////////////////////////////////////////////////////////////////////////
////////////////// Generic methods for Phys&Virt Functions ///////////////////
//////////////////////////////////////////////////////////////////////////////
func (ntFunc *ntFuncManager) GetSocket() string {
        return ntFunc.socket_name
}

func (ntFunc *ntFuncManager) SetServer(server *grpc.Server) {
        ntFunc.server = server
}

func (ntFunc *ntFuncManager) GetServer() *grpc.Server {
        return ntFunc.server
}

func (ntFunc *ntFuncManager) GetResourceName() string {
        return ntFunc.resource_name
}

func (ntFunc *ntFuncManager) GetEndpoint() string {
        return ntFunc.endpoint
}

func (ntFunc *ntFuncManager) PreStartContainer(ctx context.Context, rqt *pluginapi.PreStartContainerRequest) (*pluginapi.PreStartContainerResponse, error) {
        return nil, errors.New("PreStartContainer() should not be called as this device plugin doesn't implement it")
}

func (ntFunc *ntFuncManager) CheckNTDriver() []string {
        var unheathlyDrivers []string
        pci, err := ghw.PCI()
        if err != nil {
                glog.Errorf("CheckNTDriver(): error getting PCI info: %v", err)
                return *ntFunc.dev_addresses
        }
        devices := pci.ListDevices()
        if len(devices) == 0 {
                glog.Warningf("CheckNTDriver(): no PCI network device found")
                return *ntFunc.dev_addresses
        }
        // check availability of all detected devices at plugin start
        for _, dev := range ntFunc.devices {
                found := false
                for _, PCIDev := range devices {
                        if PCIDev.Address == dev.ID && PCIDev.Driver == "vfio-pci" {
                                found = true
                                break
                        }
                }
                if found {
                    dev.Health = pluginapi.Healthy
                } else {
                    dev.Health = pluginapi.Unhealthy
                    unheathlyDrivers = append(unheathlyDrivers, dev.ID)
                }
        }
        return unheathlyDrivers
}

func (ntFunc *ntFuncManager) GetMounts(deviceIDs []string) []*pluginapi.Mount {
        // mount all socket dirs
        mnt := GetSocketMounts(deviceIDs)
        // Export directory where ntconnect.sock is created, but avoid access to /var/run at host
        mnt = append(mnt, &pluginapi.Mount{HostPath: ntConnectSocketDir,
                ContainerPath: ntConnectSocketDir,
                ReadOnly:      false})

        return mnt
}

//////////////////////////////////////////////////////////////////////////////
///////////////////////////// Support functions //////////////////////////////
//////////////////////////////////////////////////////////////////////////////

// In case of configuration error, print usage and exit the plugin
func Usage() {
        glog.Errorf("Napatech Device Plugin requires following environment variables to be set:\n")
        glog.Errorf("    NicPci           - Napatech SmartNIC PCI address, e.g. 0000:65:00.0 or %s\n", ntNICAutodetect)
        glog.Errorf("    vfMacStart       - starting MAC address of VFs, e.g. 11:22:33:44:55:66\n")
        glog.Errorf("    vfVlanStart      - starting VLAN ID of VFs, e.g. 44\n")
        glog.Errorf("    vfNumberOfQueues - specifies a number of queues to be configured per VF\n")
        glog.Errorf("    MainApp          - controls automatic execution of NT Main application, values: enabled or disabled\n")
        os.Exit(1)
}

// Get vfio device name for given VF pci address
func GetVfioPath(pciAddress string) string {
        var iommuGroupLnk = fmt.Sprintf("/sys/bus/pci/devices/%s/iommu_group", pciAddress)
        iommuGroup, err := os.Readlink(iommuGroupLnk)
        if err != nil {
                glog.Errorf("Cannot find iommu_group for device %s - file %s not found.", pciAddress, iommuGroupLnk)
                os.Exit(1)
        }
        return fmt.Sprintf("/dev/vfio/%s", path.Base(iommuGroup))
}

// Convert MAC string (e.g. aa:bb:cc:dd:ee:ff) to uint64
func mac2umac(mac string) (uint64, error) {
        mac = strings.ReplaceAll(mac, ":", "")
        umac, err := strconv.ParseUint(mac, 16, 64)
        return umac, err
}

// Convert MAC uint64 form to []bytes
func umac2bmac(umac uint64) []byte {
        return []byte {
                byte(umac >> 40 & 0xff),
                byte(umac >> 32 & 0xff),
                byte(umac >> 24 & 0xff),
                byte(umac >> 16 & 0xff),
                byte(umac >> 8 & 0xff),
                byte(umac & 0xff)}
}

// Convert MAC uint64 form to MAC string
func umac2mac(umac uint64) string {
        return bmac2mac(umac2bmac(umac))
}

// Convert MAC []byte form to MAC string
func bmac2mac(bmac []byte) string {
        return fmt.Sprintf("%02x:%02x:%02x:%02x:%02x:%02x", bmac[0], bmac[1], bmac[2], bmac[3], bmac[4], bmac[5])
}

// Register given NTDevPlugin (pf/vf)resource into kubelet
func Register(kubeletEndpoint string, pluginEndpoint string, resourceName string) error {
        conn, err := grpc.Dial(kubeletEndpoint, grpc.WithInsecure(), grpc.WithDialer(func(addr string, timeout time.Duration) (net.Conn, error) {
                return net.DialTimeout("unix", addr, timeout)
        }))
        defer conn.Close()
        if err != nil {
                return fmt.Errorf("%s: cannot connect to kubelet service: %v", ntPluginNameShort, err)
        }
        client := pluginapi.NewRegistrationClient(conn)
        reqt := &pluginapi.RegisterRequest{Version: pluginapi.Version, Endpoint: pluginEndpoint, ResourceName: resourceName}

        _, err = client.Register(context.Background(), reqt)
        if err != nil {
                return fmt.Errorf("%s: cannot register to kubelet service: %v", ntPluginNameShort, err)
        }
        return nil
}

//Detect NT SmartNICs automatically and return list of their PCI addresses
func DetectNTSmartNIC() []string {
        var ntSmartNIC []string
        pci, err := ghw.PCI()
        if err != nil {
                glog.Errorf("DetectNTSmartNIC(): error getting PCI info: %v", err)
                return ntSmartNIC
        }

        devices := pci.ListDevices()
        for _, dev := range devices {
                if dev.Vendor.ID == ntVendorID && dev.Product.ID != ntVFProductID {
                        ntSmartNIC = append(ntSmartNIC, dev.Address)
                }
        }
        return ntSmartNIC
}

// Get socket mount points for given device IDs
func GetSocketMounts(deviceIDs []string) []*pluginapi.Mount {
        mnt := make([]*pluginapi.Mount, 0)
        for _, id := range deviceIDs {
                deviceNum := ntInfo.vf_id[id]
                socketDir := path.Dir(fmt.Sprintf(ntNICSocket, deviceNum, deviceNum))
                mnt = append(mnt, &pluginapi.Mount{HostPath: socketDir,
                        ContainerPath: socketDir,
                        ReadOnly:      false})
        }
        return mnt
}

//////////////////////////////////////////////////////////////////////////////
////////////////// Functions for the Virtual Function ///////////////////////
//////////////////////////////////////////////////////////////////////////////
func NewNTVirtFuncManager() (*ntVirtFuncManager, error) {
        return &ntVirtFuncManager{
            ntFuncManager: ntFuncManager{
                devices: make(map[string]*pluginapi.Device),
                socket_name: ntSocketNameVF,
                resource_name: ntResourceNameVF,
                endpoint: fmt.Sprintf("%s-%d.sock", ntSocketNameVF, time.Now().Unix()),
            },
        }, nil
}

func (ntVirtFunc *ntVirtFuncManager) Init() {
        ntVirtFunc.dev_addresses = &ntInfo.vf_address
        for address, _ := range ntInfo.vf_id {
                NTDev := pluginapi.Device{ID: address, Health: pluginapi.Healthy}
                ntVirtFunc.devices[address] = &NTDev
                glog.Infof("Creating VF device: %s\n", address)
        }
}

func (ntVirtFunc *ntVirtFuncManager) CheckNTDriver() []string {
        var unheathlyDrivers []string
        pci, err := ghw.PCI()
        if err != nil {
                glog.Errorf("CheckNTDriver(): error getting PCI info: %v", err)
                return ntInfo.vf_address
        }
        devices := pci.ListDevices()
        if len(devices) == 0 {
                glog.Warningf("CheckNTDriver(): no PCI network device found")
                return ntInfo.vf_address
        }
        // check availability of all VFs detected at plugin start
        for _, dev := range ntVirtFunc.devices {
                found := false
                for _, PCIDev := range devices {
                        if PCIDev.Address == dev.ID && PCIDev.Driver == "vfio-pci" {
                                found = true
                                break
                        }
                }
                if found {
                    dev.Health = pluginapi.Healthy
                } else {
                    dev.Health = pluginapi.Unhealthy
                    unheathlyDrivers = append(unheathlyDrivers, dev.ID)
                }
        }
        return unheathlyDrivers
}

func (m *ntVirtFuncManager) GetDevicePluginOptions(ctx context.Context, empty *pluginapi.Empty) (*pluginapi.DevicePluginOptions, error) {
        options := new(pluginapi.DevicePluginOptions)
        options.PreStartRequired = false
        options.GetPreferredAllocationAvailable = true
        glog.Infof("ntVirtFuncManager GetOptions: %v\n", options)
        return options, nil
}

func (m *ntVirtFuncManager) GetPreferredAllocation(ctx context.Context, rqt *pluginapi.PreferredAllocationRequest) (*pluginapi.PreferredAllocationResponse, error) {
        rsp := new(pluginapi.PreferredAllocationResponse)
        for _, creq := range rqt.ContainerRequests {
            cres := new(pluginapi.ContainerPreferredAllocationResponse)
            sort.Strings(creq.AvailableDeviceIDs)
            cres.DeviceIDs = append(cres.DeviceIDs, creq.MustIncludeDeviceIDs...)
            cres.DeviceIDs = append(cres.DeviceIDs, creq.AvailableDeviceIDs...)
            if creq.AllocationSize > (int32)(len(cres.DeviceIDs)) {
                return nil, fmt.Errorf("ntVirtFuncManager GetPreferredAllocation: cannot allocate %d from available: %v", creq.AllocationSize, cres.DeviceIDs)
            }
            cres.DeviceIDs = cres.DeviceIDs[0:creq.AllocationSize]
            rsp.ContainerResponses = append(rsp.ContainerResponses, cres)
        }
        glog.Infof("ntVirtFuncManager GetPreferredAllocation response: %v\n", rsp)
        return rsp, nil
}

// Implements DevicePlugin service functions
func (ntVirtFunc *ntVirtFuncManager) ListAndWatch(emtpy *pluginapi.Empty, stream pluginapi.DevicePlugin_ListAndWatchServer) error {
        ok := true
        for {
                unhealthyDevices := ntVirtFunc.CheckNTDriver()
                sort.Strings(unhealthyDevices)
                if len(unhealthyDevices) != 0 {
                        glog.Errorf("ntVirtFuncManager driver error - %d of %d VFs are unhealthy: %v", len(unhealthyDevices), len(ntVirtFunc.devices), unhealthyDevices)
                        ok = false
                } else {
                    if ! ok {
                        glog.Infof("ntVirtFuncManager all devices are healthy")
                        ok = true
                    }
                }
                resp := new(pluginapi.ListAndWatchResponse)
                for _, dev := range ntVirtFunc.devices {
                        resp.Devices = append(resp.Devices, dev)
                }
                if err := stream.Send(resp); err != nil {
                        glog.Errorf("ntVirtFuncManager failed to send response to kubelet: %v\n", err)
                }
                time.Sleep(ntWait)
        }
}

func (ntVirtFunc *ntVirtFuncManager) GetEnvs(deviceIDs []string) map[string]string {
        var appSockets, appMacs, appVlans, appVFs []string
        glog.Infof("Creating info for VF deviceIDs: %v\n", deviceIDs)
        envs := make(map[string]string)
        for _, id := range deviceIDs {
                if _, ok := ntVirtFunc.devices[id]; ok {
                        deviceNum := ntInfo.vf_id[id]
                        fullSocketName := fmt.Sprintf(ntNICSocket, deviceNum, deviceNum)
                        appSockets = append(appSockets, fullSocketName)
                        appMacs = append(appMacs, umac2mac(ntInfo.vf_mac + deviceNum - ntInfo.vf_first_vfid))
                        appVlans = append(appVlans, fmt.Sprintf("%d", ntInfo.vf_vlan + deviceNum - ntInfo.vf_first_vfid))
                        appVFs = append(appVFs, fmt.Sprintf("%d", deviceNum))
                } else {
                        glog.Warningf("Ignoring request for unknown VF ID: %v\n", id)
                }
        }

        envs["NT_VF_NUM"] = strings.Join(appVFs, " ")
        envs["NT_VF_QUEUES"] = fmt.Sprintf("%d", ntInfo.vf_queues)
        envs["NT_VF_DEV"] = strings.Join(deviceIDs, " ")
        envs["NT_VF_SOCKET"] = strings.Join(appSockets, " ")
        envs["NT_VF_MAC"] = strings.Join(appMacs, " ")
        envs["NT_VF_VLAN"] = strings.Join(appVlans, " ")
        envs["NT_DPDK_PREFIX"] = fmt.Sprintf(ntDpdkPrefix, ntInfo.vf_id[deviceIDs[0]])
        return envs
}

func (ntVirtFunc *ntVirtFuncManager) GetDeviceSpecs(deviceIDs []string) []*pluginapi.DeviceSpec {
        devSpecs := make([]*pluginapi.DeviceSpec, 0)
        devSpecs = append(devSpecs, &pluginapi.DeviceSpec{HostPath: "/dev/vfio/vfio",
                ContainerPath: "/dev/vfio/vfio",
                Permissions:   "mrw"})
        for _, id := range deviceIDs {
                if _, ok := ntVirtFunc.devices[id]; ok {
                        devSpecs = append(devSpecs, &pluginapi.DeviceSpec{HostPath: GetVfioPath(id),
                                ContainerPath: GetVfioPath(id),
                                Permissions:   "mrw"})
                } else {
                        glog.Warningf("Ignoring request for unknown device: %v\n", id)
                }
        }
        return devSpecs
}

func (ntVirtFunc *ntVirtFuncManager) Allocate(ctx context.Context, rqt *pluginapi.AllocateRequest) (*pluginapi.AllocateResponse, error) {
        resp := new(pluginapi.AllocateResponse)
        for _, creq := range rqt.ContainerRequests {
                cres := new(pluginapi.ContainerAllocateResponse)
                // kubelet can reorder DeviceIDs hinted by GetPreferredAllocation, so sort it again
                sort.Strings(creq.DevicesIDs)
                cres.Envs = ntVirtFunc.GetEnvs(creq.DevicesIDs)
                cres.Mounts = ntVirtFunc.GetMounts(creq.DevicesIDs)
                cres.Devices = ntVirtFunc.GetDeviceSpecs(creq.DevicesIDs)
                glog.Infof("ntVirtFuncManager allocated interfaces: %v, %v\n", creq.DevicesIDs, cres)

                resp.ContainerResponses = append(resp.ContainerResponses, cres)
        }
        return resp, nil
}

//////////////////////////////////////////////////////////////////////////////
////////////////// Functions for the Physical Function ///////////////////////
//////////////////////////////////////////////////////////////////////////////

func NewNTPhysFuncManager() (*ntPhysFuncManager, error) {
        return &ntPhysFuncManager{
            ntFuncManager: ntFuncManager{
                devices: make(map[string]*pluginapi.Device),
                socket_name: ntSocketNamePF,
                resource_name: ntResourceNamePF,
                endpoint: fmt.Sprintf("%s-%d.sock", ntSocketNamePF, time.Now().Unix()),
            },
        }, nil
}

func (ntPhysFunc *ntPhysFuncManager) Init() {
        ntPhysFunc.dev_addresses = &ntInfo.pf_address
        NTDev := pluginapi.Device{ID: ntInfo.pf_address[0], Health: pluginapi.Healthy}
        ntPhysFunc.devices[ntInfo.pf_address[0]] = &NTDev
        glog.Infof("Creating PF device: %s\n", ntInfo.pf_address[0])
}

func (m *ntPhysFuncManager) GetDevicePluginOptions(ctx context.Context, empty *pluginapi.Empty) (*pluginapi.DevicePluginOptions, error) {
        options := new(pluginapi.DevicePluginOptions)
        options.PreStartRequired = false
        options.GetPreferredAllocationAvailable = false
        glog.Infof("ntPhysFuncManager GetOptions: %v\n", options)
        return options, nil
}

func (m *ntPhysFuncManager) GetPreferredAllocation(ctx context.Context, rqt *pluginapi.PreferredAllocationRequest) (*pluginapi.PreferredAllocationResponse, error) {
        glog.Infof("ntPhysFuncManager GetPreferredAllocation rqt: %v\n", rqt)
        return nil, errors.New("GetPreferredAllocation should not be called as this device plugin doesn't implement it")
}

// Implements DevicePlugin service functions
func (ntPhysFunc *ntPhysFuncManager) ListAndWatch(emtpy *pluginapi.Empty, stream pluginapi.DevicePlugin_ListAndWatchServer) error {
        ok := true
        for {
                unhealthyDevices := ntPhysFunc.CheckNTDriver()
                sort.Strings(unhealthyDevices)
                if len(unhealthyDevices) != 0 {
                        glog.Errorf("ntPhysFuncManager driver error - %d of %d PFs are unhealthy: %v", len(unhealthyDevices), len(ntPhysFunc.devices), unhealthyDevices)
                        ok = false
                } else {
                    if ! ok {
                        glog.Infof("ntPhysFuncManager all devices are healthy")
                        ok = true
                    }
                }
                resp := new(pluginapi.ListAndWatchResponse)
                for _, dev := range ntPhysFunc.devices {
                        resp.Devices = append(resp.Devices, dev)
                }
                if err := stream.Send(resp); err != nil {
                        glog.Errorf("ntPhysFuncManager failed to send response to kubelet: %v\n", err)
                }
                time.Sleep(ntWait)
        }
}

func (ntPhysFunc *ntPhysFuncManager) GetEnvs() map[string]string {
        glog.Infof("Creating info for physical device: %s\n", ntInfo.pf_address[0])
        envs := make(map[string]string)
        var appDevs, appVFs, appMacs, appVlans, appSockets []string
        for i := 0; i < len(ntInfo.vf_address); i++ {
                appVFs = append(appVFs, fmt.Sprintf("%d", uint64(i) + ntInfo.vf_first_vfid))
                appDevs = append(appDevs, ntInfo.vf_address[i])
                appMacs = append(appMacs, umac2mac(ntInfo.vf_mac + uint64(i)))
                appVlans = append(appVlans, fmt.Sprintf("%d", ntInfo.vf_vlan + uint64(i)))
                appSockets = append(appSockets, fmt.Sprintf(ntNICSocket, uint64(i) + ntInfo.vf_first_vfid, uint64(i) + ntInfo.vf_first_vfid))
        }
        envs["NT_DPDK_PREFIX"] = fmt.Sprintf(ntDpdkPrefix, 0)
        envs["NT_PF_NUM"] = "0"
        envs["NT_PF_DEV"] = ntInfo.pf_address[0]
        envs["NT_VF_NUM"] = strings.Join(appVFs, " ")
        envs["NT_VF_QUEUES"] = fmt.Sprintf("%d", ntInfo.vf_queues)
        envs["NT_VF_DEV"] = strings.Join(appDevs, " ")
        envs["NT_VF_MAC"] = strings.Join(appMacs, " ")
        envs["NT_VF_VLAN"] = strings.Join(appVlans, " ")
        envs["NT_VF_SOCKET"] = strings.Join(appSockets, " ")
        envs["NT_VF_FIRST_ID"] = fmt.Sprintf("%d", ntInfo.vf_first_vfid)
        return envs
}

func (ntPhysFunc *ntPhysFuncManager) GetDeviceSpecs() []*pluginapi.DeviceSpec {
        devSpecs := make([]*pluginapi.DeviceSpec, 0)
        devSpecs = append(devSpecs, &pluginapi.DeviceSpec{HostPath: "/dev/vfio/vfio",
                ContainerPath: "/dev/vfio/vfio",
                Permissions:   "mrw"})
        devSpecs = append(devSpecs, &pluginapi.DeviceSpec{HostPath: GetVfioPath(ntInfo.pf_address[0]),
                ContainerPath: GetVfioPath(ntInfo.pf_address[0]),
                Permissions:   "mrw"})

        numberOfDevices := len(ntInfo.vf_address)
        for i := 0; i < numberOfDevices; i++ {
                devSpecs = append(devSpecs, &pluginapi.DeviceSpec{HostPath: GetVfioPath(ntInfo.vf_address[i]),
                        ContainerPath: GetVfioPath(ntInfo.vf_address[i]),
                        Permissions:   "mrw"})
        }
        return devSpecs
}

func (ntPhysFunc *ntPhysFuncManager) Allocate(ctx context.Context, rqt *pluginapi.AllocateRequest) (*pluginapi.AllocateResponse, error) {
        resp := new(pluginapi.AllocateResponse)
        for _, creq := range rqt.ContainerRequests {
                cres := new(pluginapi.ContainerAllocateResponse)
                for _, id := range creq.DevicesIDs {
                        glog.Infof("Request IDs: %v\n", id)
                        if _, ok := ntPhysFunc.devices[id]; ok {
                                // Make environment variables
                                cres.Envs = ntPhysFunc.GetEnvs()
                                cres.Mounts = ntPhysFunc.GetMounts(ntInfo.vf_address)
                                cres.Devices = ntPhysFunc.GetDeviceSpecs()
                                glog.Infof("ntPhysFuncManager allocated interface: %v, %v\n", id, cres)
                        }
                }
                resp.ContainerResponses = append(resp.ContainerResponses, cres)
        }
        return resp, nil
}

//////////////////////////////////////////////////////////////////////////////
/////////////////////////// Device Plugin Main ///////////////////////////////
//////////////////////////////////////////////////////////////////////////////

func DiscoverNTDevices() bool {
        pci, err := ghw.PCI()
        if err != nil {
                glog.Errorf("DiscoverNTDevices(): error getting PCI info: %v", err)
                return false
        }

        devices := pci.ListDevices()
        if len(devices) == 0 {
                glog.Warningf("DiscoverNTDevices(): no PCI network device found")
                return false
        }

        found := false
        for _, dev := range devices {
                if ntInfo.pf_address[0] == dev.Address {
                        found = true
                        if dev.Driver != "vfio-pci" {
                                glog.Errorf("ERROR vfio-pci driver is not loaded for: %s\n", dev.Address)
                                return false
                        }
                        switch strings.ToLower(dev.Product.ID) {
                        case ntNT200AProductID:
                                ntInfo.vf_first_vfid = ntNT200AFirstVFID
                        case ntNT400DProductID:
                                ntInfo.vf_first_vfid = ntNT400DFirstVFID
                        default:
                                glog.Warningf("Unknown NT SmartiNIC product id %v\n")
                                // let's assume that 1st VF is 4th device
                                ntInfo.vf_first_vfid = ntNT200AFirstVFID
                        }
                }
        }

        if ! found {
                glog.Errorf("ERROR: PF address %v cannot be found\n", ntInfo.pf_address[0])
                return false
        }

        PFPCIAdress := pciaddress.FromString(ntInfo.pf_address[0])
        for _, dev := range devices {
                PCIAdress := pciaddress.FromString(dev.Address)
                if PCIAdress.Domain == PFPCIAdress.Domain && PCIAdress.Bus == PFPCIAdress.Bus {
                        device, _ := strconv.ParseUint(PCIAdress.Device, 16, 32)
                        function, _ := strconv.ParseUint(PCIAdress.Function, 16, 32)
                        if (device*8 + function) >= ntInfo.vf_first_vfid {
                                if dev.Driver == "vfio-pci" {
                                        ntInfo.vf_address = append(ntInfo.vf_address, dev.Address)
                                        ntInfo.vf_id[dev.Address] = (device*8 + function)
                                } else {
                                        glog.Errorf("ERROR vfio-pci driver is not loaded for: %s\n", dev.Address)
                                }
                        }
                }
        }

        if len(ntInfo.vf_address) == 0 {
                glog.Errorf("ERROR: NO VFs are found\n")
                return false
        }

        glog.Infof("Found: PF:  %v\n", ntInfo.pf_address)
        glog.Infof("       VFs: %v\n", ntInfo.vf_address)
        return true
}

func PluginInit(manager ntDeviceManager) {
        manager.Init()
        manager.CheckNTDriver()

        // start the servers
        wg.Add(1)
        // Starts device plugin service.
        go func() {
                defer wg.Done()

                glog.Infof("%s: DevicePluginPath %s, pluginEndpoint %s\n", manager.GetSocket(), pluginapi.DevicePluginPath, manager.GetEndpoint())
                glog.Infof("            %s start server at: %s\n", ntPluginNameShort, path.Join(pluginapi.DevicePluginPath, manager.GetEndpoint()))

                lis, err := net.Listen("unix", path.Join(pluginapi.DevicePluginPath, manager.GetEndpoint()))
                if err != nil {
                        glog.Fatal(err)
                        return
                }

                manager.SetServer(grpc.NewServer())
                pluginapi.RegisterDevicePluginServer(manager.GetServer(), manager)
                manager.GetServer().Serve(lis)

        }()

        time.Sleep(ntWait)

        // Registers plugin with Kubelet.
        err := Register(pluginapi.KubeletSocket, manager.GetEndpoint(), manager.GetResourceName())
        if err != nil {
                glog.Fatal(err)
        }
}

func main() {
        flag.Parse()
        var err error

        ntInfo = NewNTInfo()

        flag.Lookup("logtostderr").Value.Set("true")

        glog.Infof("%s version %s\n", ntPluginNameLong, ntPluginVersion)

        glog.Infof("Configuration:")
        for _, key := range ntEnvVars {
                glog.Infof("    %-17s: %s\n", key, os.Getenv(key))
        }

        ntInfo.pf_address = append(ntInfo.pf_address, os.Getenv("NicPci"))

        if strings.ToLower(ntInfo.pf_address[0]) == ntNICAutodetect {
                ntSmartNIC := DetectNTSmartNIC()
                if len(ntSmartNIC) == 0 {
                        glog.Error("ERROR: Auto detection of NT SmartNIC failed. Please configure its PCI address manually.\n")
                        Usage()
                }
                if len(ntSmartNIC) > 1 {
                        glog.Error("ERROR: Multiple NT SmartNICs were detected. Please configure PCI address of preferred card manually.\n")
                        os.Exit(1)
                }
                ntInfo.pf_address = ntSmartNIC
                glog.Infof("NT SmartNIC detected at PCI address: %s\n", ntInfo.pf_address[0])
        } else {
                // normalize PCI address and fail if it is not valid
                tmp_address := pciaddress.FromString(ntInfo.pf_address[0])
                if tmp_address == nil {
                        glog.Error("ERROR: NicPci is not a valid PCI address.\n")
                        Usage()
                }
                ntInfo.pf_address[0] = tmp_address.String()
        }

        ntInfo.vf_mac, err = mac2umac(os.Getenv("vfMacStart"))
        if err != nil {
                glog.Error("ERROR: Unsupported value of vfMacStart.")
                Usage()
        }

        ntInfo.vf_vlan, err = strconv.ParseUint(os.Getenv("vfVlanStart"), 10, 64)
        if err != nil {
                glog.Error("ERROR: Unsupported value of vfVlanStart.\n")
                Usage()
        }

        ntInfo.vf_queues, err = strconv.ParseUint(os.Getenv("vfNumberOfQueues"), 10, 64)
        if err != nil {
                glog.Error("ERROR: Unsupported value of vfNumberOfQueues.\n")
                Usage()
        }

        // Find devices for Phys and Virt
        found := DiscoverNTDevices()
        if !found {
                glog.Errorf("ERROR: No Napatech adapters are present or the Napatech driver is not loaded\n")
                os.Exit(1)
        }

        // Create managers for Phys and Virt
        switch strings.ToLower(os.Getenv("MainApp")) {
        case "disabled":
                ntPhysFunc, err = NewNTPhysFuncManager()
                if err != nil {
                        glog.Fatal(err)
                        os.Exit(1)
                }
               PluginInit(ntPhysFunc)
        case "enabled":
                glog.Info("ntPhysFuncManager is disabled by configuration")
        default:
                glog.Error("ERROR: Unsupported value of MainApp.\n")
                Usage()
        }

        ntVirtFunc, err = NewNTVirtFuncManager()
        if err != nil {
                glog.Fatal(err)
                os.Exit(1)
        }
        PluginInit(ntVirtFunc)

        // Start watcher to monitor kubelet restart after which NTDevPlugin has to register itself again
        go func() {
                defer wg.Done()
                glog.Infof("Socket health check started\n")
                for {
                        time.Sleep(ntWait)
                        // kubelet removes ALL registered sockets during (re)start
                        _, err := os.Lstat(path.Join(pluginapi.DevicePluginPath, ntVirtFunc.endpoint))
                        if err != nil {
                            glog.Infof("%s socket is not available probably due to kubelet restart - plugin will restart now\n", ntVirtFunc.endpoint)
                            return
                        }
                }
        }()

        glog.Infof("%s registered\n", ntPluginNameLong)
        wg.Wait()
}
