/*
 * SPDX-License-Identifier: BSD-3-Clause
 * Copyright(c) 2023 Napatech A/S
 */

#include <menu.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include <ncurses.h>
#include <inttypes.h>
#include "monitoring.h"

#define countof(_arr) sizeof(_arr) / sizeof(*_arr)

#define PORTS_HEADER      "  Port  Type   PCI   Intf       NIM    Link    Speed    MTU    Rx Mbps    Tx Mbps  Rx Mpps  Tx Mpps   "
#define QUEUE_HEADER      "  HWQ  P:Q  Owner  Dest           Packets              Octets      VQ Drop pkts      VQ Drop octets   "
#define COLOR_STAT_HEADER "  Color              Packets               Bytes    tcp flags                                         "
#define FLM_STAT_HEADER "                                                                                                        "

#define PORTS_FORMAT " %4s  %02x.%x  %5s  %10s  %4s  %7s  %5d  %9s  %9s %8s %8s"
#define QUEUE_FORMAT "%17s  #%015lu  #%017lu  #%015lu  #%017lu"
#define COLOR_STAT_FORMAT "    #%015lu  #%017lu        %02x"
#define FLM_STAT_FORMAT " "

#define STR_ID_SIZE 5

#define THROUGHPUT_STRLEN 32
#define PPS_STRLEN        32

#define MAX_QUEUE_NUM   128

#define MAX_WIDTH  106
#define MAX_HEIGHT  100

ITEM **ports_items;
ITEM **queue_items;
ITEM **color_stat_items;
MENU *top_menu;

static int curr_port_index;

#define MAX_PORTS 136

static char ports_nums[MAX_PORTS][STR_ID_SIZE + 1];
static char ports_data[MAX_PORTS][sizeof(PORTS_HEADER) + 1];

static char queue_nums[MAX_QUEUE_NUM][STR_ID_SIZE + 1];
static char queue_data[MAX_QUEUE_NUM][sizeof(QUEUE_HEADER) + 1];
static struct mon_port_queue_s port_queue_mapping[MAX_QUEUE_NUM];

static char (*color_stat_nums)[STR_ID_SIZE + 1];
static char (*color_stat_data)[sizeof(COLOR_STAT_HEADER)];
static int is_inited_color_stat_items;
static int is_inited_queue_items;
static int color_counters_number;
static int queue_counters_number;
static int maxy, maxx, menu_size = 3;
static int last_failed, bottom_height = 5;

#define max_xwin ((maxx > MAX_WIDTH) ? MAX_WIDTH : maxx)
#define max_ywin ((maxy > MAX_HEIGHT) ? MAX_HEIGHT : maxy)


enum presentation_state_e {
  PRESENT_STAT_NONE,
  PRESENT_PORT_RX_STATS,
  PRESENT_PORT_TX_STATS,
  PRESENT_QUEUE_STATS,
  PRESENT_COLOR_STATS,
  PRESENT_FLM_STATS
};

typedef struct _window_handler_s {
  enum presentation_state_e state;
  void (*update_top_window_data)(void);
  void (*update_bottom_window)(void);
} window_handler_t;

static window_handler_t win_h;

static WINDOW *main_window, *top_window, *bottom_window;
static int width, height;

static struct mon_data_s *mon;

/*
 * Reset base
 */
struct res_base_s {
  uint64_t *clr;
  uint64_t *que;
  uint64_t *rxc;
  uint64_t *txc;
  uint64_t *flm;
} res_base;

const char *data_buttons[] = {
  "Reset",
  "Port",
  "TX/Rx",
  "Queue",
  "Color",
  "Flowstat"
};


static void _wprint(WINDOW *win, int y, int x, const char *fmt, ...)
{
#define SIZE 1024

  if (win == bottom_window && y >= (bottom_height - 1)) return;

  char buf[SIZE + 1];
  va_list ap;
  va_start(ap, fmt);
  vsnprintf(buf, SIZE, fmt, ap);
  va_end(ap);

  const int len = strlen(buf);
  if (len > maxx - x - 3) {
    buf[maxx - x - 3] = '\0';
  }
  mvwprintw(win, y, x, "%s", buf);
}


static void update_data_buttons(void)
{
  for (int i = 0; (unsigned int)i < countof(data_buttons); i++) {
    wattron(main_window, last_failed ? A_DIM : A_STANDOUT);
    _wprint(main_window, height - 3, 1 + (i * 7), "%c", data_buttons[i][0]);
    wattroff(main_window, A_STANDOUT);
    _wprint(main_window, height - 3, 2 + (i * 7), "%s", data_buttons[i] + 1);
  }
  wrefresh(main_window);
}


/*
 * returns port stat index for the selected logical port index (dpdk_port_id)
 */
static int get_stat_port_index(int logical_port_index)
{
  return mon->ifs[mon->sel].intf[logical_port_index].type != PORT_TYPE_VIRT ? logical_port_index :
        mon->ifs[mon->sel].intf[logical_port_index].pci_id.devid * 8 + mon->ifs[mon->sel].intf[logical_port_index].pci_id.function;
}


static void update_port_data(void)
{
  int line_cnt = 1;

  if (win_h.state != PRESENT_PORT_RX_STATS && win_h.state != PRESENT_PORT_TX_STATS) {
    return;
  }

  const char *pmode = win_h.state == PRESENT_PORT_RX_STATS ? "Rx counters" : "Tx counters";

  werase(bottom_window);
  box(bottom_window, 0, 0);

  wattron(bottom_window, last_failed ? A_DIM : A_BOLD);
  _wprint(bottom_window, line_cnt++, max_xwin / 2 - 7, "%s", pmode);
  wattroff(bottom_window, A_BOLD);

  int port_index = get_stat_port_index(curr_port_index);

  int err = 0;
  if (mon->stat_layout_version == 6) {
    if (mon->rxc->is_virt) {
      /* Virtualization counters */
      struct rtx_type_fields_virt_s *rtxc;
      struct rtx_type_fields_virt_s *base;
      if (win_h.state == PRESENT_PORT_RX_STATS) {
        rtxc = (struct rtx_type_fields_virt_s *)mon->rxc->data;
        base = (struct rtx_type_fields_virt_s *)res_base.rxc;
        if (port_index > (int)mon->rxc->nb_counters) {
          err = 1;
        }
      } else if (win_h.state == PRESENT_PORT_TX_STATS) {
        rtxc = (struct rtx_type_fields_virt_s *)mon->txc->data;
        base = (struct rtx_type_fields_virt_s *)res_base.txc;
        if (port_index > (int)mon->txc->nb_counters) {
          err = 1;
        }
      } else {
        return;
      }

      if (err) {
        _wprint(bottom_window, 2, 2, "No %s available", pmode);
      } else {
        rtxc += port_index;
        if (base) {
          base += port_index;
        }

        _wprint(bottom_window, line_cnt++, 2, "%-16s: %015llu  %-16s:  %017llu",
            "Packets", rtxc->pkts - (base ? base->pkts : 0),
            "Octets", rtxc->octets - (base ? base->octets : 0));
        _wprint(bottom_window, line_cnt++, 2, "%-16s: %015llu  %-16s:  %017llu",
            "Qos drop packets", rtxc->qos_drop_pkts - (base ? base->qos_drop_pkts : 0),
            "Qos drop octets", rtxc->qos_drop_octets - (base ? base->qos_drop_octets : 0));
        _wprint(bottom_window, line_cnt++, 2, "%-16s: %015llu",
            "Other drops", rtxc->drop_events - (base ? base->drop_events : 0));
      }
    } else {
      /* Capture counters */
      struct stat_rmon_s *rmon;
      struct stat_rmon_s *base;

      if (win_h.state == PRESENT_PORT_RX_STATS) {
        if (port_index > (int)mon->rxc->nb_counters) {
          return;
        }
        rmon = &((struct rx_type_fields_cap_s *)mon->rxc->data + port_index)->rmon;
        base = res_base.rxc ? &((struct rx_type_fields_cap_s *)res_base.rxc + port_index)->rmon : NULL;
      } else if (win_h.state == PRESENT_PORT_TX_STATS) {
        if (port_index > (int)mon->txc->nb_counters) {
          return;
        }
        rmon = &((struct tx_type_fields_cap_s *)mon->txc->data + port_index)->rmon;
        base = res_base.txc ? &((struct tx_type_fields_cap_s *)res_base.txc + port_index)->rmon : NULL;
      } else {
        return;
      }

      /*
       * Present RMON counters
       */
      _wprint(bottom_window, line_cnt++, 2, "%-16s: %017llu  %-16s:  %017llu",
          "Packets", rmon->pkts - (base ? base->pkts : 0), "Octets", rmon->octets - (base ? base->octets : 0));
      _wprint(bottom_window, line_cnt++, 2, "%-16s: %017llu  %-16s:  %017llu",
          "Broadcast", rmon->broadcast_pkts - (base ? base->broadcast_pkts : 0), "Multicast", rmon->multicast_pkts - (base ? base->multicast_pkts : 0));
      _wprint(bottom_window, line_cnt++, 2, "%-16s: %017llu  %-16s:  %017llu",
          "Unicast", rmon->unicast_pkts - (base ? base->unicast_pkts : 0), "Crc errors", rmon->pkts_crc - (base ? base->pkts_crc : 0));
      _wprint(bottom_window, line_cnt++, 2, "%-16s: %017llu  %-16s:  %017llu",
          "64 octets", rmon->pkts_64_octets - (base ? base->pkts_64_octets : 0), "65-127 octets", rmon->pkts_65_to_127_octets - (base ? base->pkts_65_to_127_octets : 0));
      _wprint(bottom_window, line_cnt++, 2, "%-16s: %017llu  %-16s:  %017llu",
          "128-255 octets", rmon->pkts_128_to_255_octets - (base ? base->pkts_128_to_255_octets : 0), "256-511 octets", rmon->pkts_256_to_511_octets - (base ? base->pkts_256_to_511_octets : 0));
      _wprint(bottom_window, line_cnt++, 2, "%-16s: %017llu  %-16s:  %017llu",
          "512-1023 octets", rmon->pkts_512_to_1023_octets - (base ? base->pkts_512_to_1023_octets : 0), "1024-1518 octets", rmon->pkts_1024_to_1518_octets - (base ? base->pkts_1024_to_1518_octets : 0));
      _wprint(bottom_window, line_cnt++, 2, "%-16s: %017llu  %-16s:  %017llu",
          "1519-2047 octets", rmon->pkts_1519_to_2047_octets - (base ? base->pkts_1519_to_2047_octets : 0), "2048-4095 octets", rmon->pkts_2048_to_4095_octets - (base ? base->pkts_2048_to_4095_octets : 0));
      _wprint(bottom_window, line_cnt++, 2, "%-16s: %017llu  %-16s:  %017llu",
          "4096-8191 octets", rmon->pkts_4096_to_8191_octets - (base ? base->pkts_4096_to_8191_octets : 0), "8192-max octets", rmon->pkts_8192_to_max_octets - (base ? base->pkts_8192_to_max_octets : 0));
      _wprint(bottom_window, line_cnt++, 2, "%-16s: %017llu  %-16s:  %017llu",
          "Undersize", rmon->undersize_pkts - (base ? base->undersize_pkts : 0), "Oversize", rmon->oversize_pkts - (base ? base->oversize_pkts : 0));
      _wprint(bottom_window, line_cnt++, 2, "%-16s: %017llu  %-16s:  %017llu",
          "Fragments", rmon->fragments - (base ? base->fragments : 0), "Drop events", rmon->drop_events - (base ? base->drop_events : 0));

      if (win_h.state == PRESENT_PORT_RX_STATS) {
#if (0)
        /* only RX additional possibilities */
        uint64_t pkts_alignment;
        uint64_t pkts_code_violation;
        uint64_t jabbers_not_truncated;
        uint64_t jabbers_truncated;
        /* Rx only port counters */
        uint64_t pkts_ip_chksum_error;
        uint64_t pkts_udp_chksum_error;
        uint64_t pkts_tcp_chksum_error;
        uint64_t pkts_giant_undersize;
        uint64_t pkts_baby_giant;
        uint64_t pkts_not_isl_vlan_mpls;
        uint64_t pkts_isl;
        uint64_t pkts_vlan;
        uint64_t pkts_isl_vlan;
        uint64_t pkts_mpls;
        uint64_t pkts_isl_mpls;
        uint64_t pkts_vlan_mpls;
        uint64_t pkts_isl_vlan_mpls;
        uint64_t pkts_no_filter;
        uint64_t pkts_dedup;
        uint64_t pkts_overflow;
        uint64_t octets_no_filter;
        uint64_t octets_dedup;
        uint64_t octets_overflow;
        /* Rx: IPF added with STA module 0.2 */
        uint64_t ipft_first_hit;
        uint64_t ipft_first_not_hit;
        uint64_t ipft_mid_hit;
        uint64_t ipft_mid_not_hit;
        uint64_t ipft_last_hit;
        uint64_t ipft_last_not_hit;
#endif
      }

    }
  } else {
    _wprint(bottom_window, 1, 2, "Unknown statistics layout version. Cannot interpret statistics information.");
    _wprint(bottom_window, 2, 2, "Version expected %i, version reported %i", 6, mon->stat_layout_version);
  }
  wrefresh(bottom_window);
}



static const char *get_speed_str(enum port_speed port_speed)
{
  switch (port_speed) {
  case PORT_LINK_SPEED_10M:
    return "10M";
  case PORT_LINK_SPEED_100M:
    return "100M";
  case PORT_LINK_SPEED_1G:
    return "1G";
  case PORT_LINK_SPEED_10G:
    return "10G";
  case PORT_LINK_SPEED_25G:
    return "25G";
  case PORT_LINK_SPEED_40G:
    return "40G";
  case PORT_LINK_SPEED_50G:
    return "50G";
  case PORT_LINK_SPEED_100G:
    return "100G";
  case PORT_LINK_SPEED_NONE_REPORTED:
    return "N/A";
  case PORT_LINK_SPEED_UNKNOWN: /* intentional fall-through */
  default:
    return "unknown";
  }
}


static void update_status_bar(void)
{
  int curr_item_index = item_index(current_item(top_menu));
  int pointer_location = ((menu_size - 1) * curr_item_index) /  (item_count(top_menu) - 1);

  for (int i = 0; i < menu_size; i++) {
    if (i == pointer_location) {
      mvwaddch(top_window, 2 + i, width - 4, ACS_DIAMOND);
    } else {
      mvwaddch(top_window, 2 + i, width - 4, ACS_CKBOARD);
    }
  }
}


static void init_port_items(void)
{
  ports_items = (ITEM **)calloc(mon->ifs[mon->sel].nb_ports + 1, sizeof(ITEM *));
  for (uint8_t i = 0; i < mon->ifs[mon->sel].nb_ports; ++i) {
    sprintf(ports_data[i], PORTS_FORMAT,
        "---", 0,0, "---", "---", "---", "---", 0, "---", "---", "---", "---");
    sprintf(ports_nums[i], "%4d", i);
    ports_items[i] = new_item(ports_nums[i], ports_data[i]);
  }
  ports_items[mon->ifs[mon->sel].nb_ports] = (ITEM *)NULL;
}


static void deinit_port_items(void)
{
  for (unsigned int i = 0; i < mon->ifs[mon->sel].nb_ports; ++i) {
    free_item(ports_items[i]);
  }
  free(ports_items);
}


static uint64_t _calc_throughput(uint64_t pkts, uint64_t last_pkts, uint64_t octets, uint64_t last_octets, uint64_t tdiff, uint64_t *pps)
{
  if (last_pkts > pkts || last_octets > octets) {
    *pps = 0;
    return 0;
  }
  /* Interframe gap to be added */
  uint64_t ifg = (pkts - last_pkts) * 20;
  /* calculate PPS */
  *pps = (pkts - last_pkts) * 100000 / tdiff;
  /* calculate and convert into Mbps * 100 for fraction */
  return (octets - last_octets + ifg) * 100000 * 8 / tdiff;
}


static uint64_t get_throughput(int logical_port_index, uint64_t *pps, int rx)
{
  int stat_index, num_counters;
  uint64_t tdiff;
  enum port_type port_type = mon->ifs[mon->sel].intf[logical_port_index].type;

  *pps = 0;
  if ((rx && !mon->last_rxc) || (!rx && !mon->last_txc)) {
    return 0;
  }


  if (rx) {
    tdiff = mon->rxc->timestamp - mon->last_rxc->timestamp;
  } else {
    tdiff = mon->txc->timestamp - mon->last_txc->timestamp;
  }

  num_counters = rx ? mon->rxc->nb_counters : mon->txc->nb_counters;

  stat_index = get_stat_port_index(logical_port_index);

  if (rx || port_type != PORT_TYPE_VIRT) {
    if (stat_index >= num_counters) {
      return 0;
    }
  }

  if ((rx && mon->rxc->is_virt) || (!rx && mon->txc->is_virt)) {
    /*
     * Virtualization NIC
     */
    struct rtx_type_fields_virt_s *rtxc;
    struct rtx_type_fields_virt_s *last_rtxc;
    struct queue_type_fields_s *que;
    struct queue_type_fields_s *last_que;

    if (rx) {
      rtxc = (struct rtx_type_fields_virt_s *)mon->rxc->data;
      last_rtxc = (struct rtx_type_fields_virt_s *)mon->last_rxc->data;
    } else {
      if (port_type != PORT_TYPE_VIRT) {
        rtxc = (struct rtx_type_fields_virt_s *)mon->txc->data;
        last_rtxc = (struct rtx_type_fields_virt_s *)mon->last_txc->data;
      } else {
        /*
         * we do not have Tx stat counters on virtual ports in FPGA
         * therefore we need to collect that from vport attached queues (outgoing queues)
         */
        que = (struct queue_type_fields_s *)mon->que->data;
        last_que = (struct queue_type_fields_s *)mon->last_que->data;

        uint64_t trp = 0;
        uint64_t pkts = 0;

        /* only calculate queues with output direction */
        for (int i = 0; i < mon->ifs[mon->sel].intf[logical_port_index].num_queues; i++) {

          if (mon->ifs[mon->sel].intf[logical_port_index].queue[i].dir == QUEUE_OUTPUT) {
            int idx = mon->ifs[mon->sel].intf[logical_port_index].queue[i].idx;
            if (idx < (int)mon->que->nb_counters) {
              /* since DBS drop counters are not atomically in sync with pkt/octet counters, we need to ensure non negative values */
              uint64_t subtract = (last_que + idx)->fwd_pkts + ((que + idx)->dbs_drop_pkts - (last_que + idx)->dbs_drop_pkts);

              if ((que + idx)->fwd_pkts > subtract)
                pkts += (que + idx)->fwd_pkts - subtract;

              subtract = (last_que + idx)->fwd_octets + ((que + idx)->dbs_drop_octets - (last_que + idx)->dbs_drop_octets);
              if ((que + idx)->fwd_octets > subtract)
              trp += (que + idx)->fwd_octets - subtract;
            }
          }
        }

        return _calc_throughput(pkts, 0, trp, 0, tdiff, pps);
      }
    }

    rtxc += stat_index;
    last_rtxc += stat_index;
    return _calc_throughput(rtxc->pkts, last_rtxc->pkts, rtxc->octets, last_rtxc->octets, tdiff, pps);

  } else {
    /*
     * Capture/inline
     */
    if (rx) {
      struct rx_type_fields_cap_s *rx = (struct rx_type_fields_cap_s *)mon->rxc->data;
      struct rx_type_fields_cap_s *last_rx = (struct rx_type_fields_cap_s *)mon->last_rxc->data;

      rx += stat_index;
      last_rx += stat_index;
      return _calc_throughput(rx->rmon.pkts, last_rx->rmon.pkts, rx->rmon.octets, last_rx->rmon.octets, tdiff, pps);

    } else {
      struct tx_type_fields_cap_s *tx = (struct tx_type_fields_cap_s *)mon->txc->data;
      struct tx_type_fields_cap_s *last_tx = (struct tx_type_fields_cap_s *)mon->last_txc->data;

      tx += stat_index;
      last_tx += stat_index;
      return _calc_throughput(tx->rmon.pkts, last_tx->rmon.pkts, tx->rmon.octets, last_tx->rmon.octets, tdiff, pps);
    }

  }

  return 0;
}




static void set_reset_base(void)
{
  int size_clr = mon->clr->nb_counters * sizeof(struct color_type_fields_s);
  int size_que = mon->que->nb_counters * sizeof(struct queue_type_fields_s);
  int size_rxc = mon->rxc->nb_counters * (mon->rxc->is_virt ? sizeof(struct rtx_type_fields_virt_s) : sizeof(struct rx_type_fields_cap_s));
  int size_txc = mon->txc->nb_counters * (mon->rxc->is_virt ? sizeof(struct rtx_type_fields_virt_s) : sizeof(struct tx_type_fields_cap_s));
  int size_flm = mon->flm->nb_counters * sizeof(struct flowmatcher_type_fields_s);

  if (res_base.clr) {
    free(res_base.clr);
  }

  uint8_t *ptr = malloc(size_clr + size_que + size_rxc + size_txc + size_flm);
  if (!ptr) {
    res_base.clr = NULL;
    res_base.que = NULL;
    res_base.rxc = NULL;
    res_base.txc = NULL;
    res_base.flm = NULL;
    return;
  }
  res_base.clr = (uint64_t *)ptr;
  ptr += size_clr;
  res_base.que = (uint64_t *)ptr;
  ptr += size_que;
  res_base.rxc = (uint64_t *)ptr;
  ptr += size_rxc;
  res_base.txc = (uint64_t *)ptr;
  ptr += size_txc;
  res_base.flm = (uint64_t *)ptr;

  memcpy(res_base.clr, mon->clr->data, size_clr);
  memcpy(res_base.que, mon->que->data, size_que);
  memcpy(res_base.rxc, mon->rxc->data, size_rxc);
  memcpy(res_base.txc, mon->txc->data, size_txc);
  memcpy(res_base.flm, mon->flm->data, size_flm);
}



static const char *get_port_state_str(enum port_states state)
{
  switch (state) {
  case PORT_STATE_DISABLED:
    return "disabled";
  case PORT_STATE_NIM_PRESENT:
    return "present";
  case PORT_STATE_NIM_ABSENT:
    return "absent";
  case PORT_STATE_VIRTUAL_UNATTACHED:
    return "unattached";
  case PORT_STATE_VIRTUAL_SPLIT:
    return "split";
  case PORT_STATE_VIRTUAL_PACKED:
    return "packed";
  case PORT_STATE_VIRTUAL_RELAY:
    return "relay";
  }

  return "-err-";
}


static void update_port_list(void)
{
  char rx_throughput[THROUGHPUT_STRLEN + 1];
  char tx_throughput[THROUGHPUT_STRLEN + 1];
  char rxpps[PPS_STRLEN + 1];
  char txpps[PPS_STRLEN + 1];

  curr_port_index = item_index(current_item(top_menu));

  for (uint8_t i = 0; i < mon->ifs[mon->sel].nb_ports; i++) {
    uint64_t pps_val = 0;
    uint64_t throughput = 0;
    if (mon->ifs[mon->sel].intf[i].type != PORT_TYPE_VIRT) {
      throughput = get_throughput(i, &pps_val, 1);
    }
    snprintf(rx_throughput, THROUGHPUT_STRLEN, "%5lu.%02lu", throughput / 100, throughput % 100);
    snprintf(rxpps, PPS_STRLEN, "%5lu.%02lu", pps_val / 100, pps_val % 100);

    if (mon->ifs[mon->sel].intf[i].type != PORT_TYPE_VIRT) {
      throughput = get_throughput(i, &pps_val, 0);
    }
    snprintf(tx_throughput, THROUGHPUT_STRLEN, "%5lu.%02lu", throughput / 100, throughput % 100);
    snprintf(txpps, PPS_STRLEN, "%5lu.%02lu", pps_val / 100, pps_val % 100);


    char func[32];
    if (mon->ifs[mon->sel].intf[i].type != PORT_TYPE_VIRT) {
      snprintf(func, 31, "phy%i", i);
    } else {
      int num = mon->ifs[mon->sel].intf[i].pci_id.devid * 8 + mon->ifs[mon->sel].intf[i].pci_id.function;
      snprintf(func, 31, "VF%i", num);
    }

    char port_type[PORT_TYPE_LAST][5] = {
                   "phys",  /* Normal phy port */
                   "lagA",  /* Active/active LAG phy port */
                   "lagP",  /* Primary LAG phy port */
                   "lagB",  /* Backup LAG phy port */
                   "virt"   /* Virtuel port */
                   };

    sprintf(ports_data[i], PORTS_FORMAT,
         port_type[mon->ifs[mon->sel].intf[i].type],
         mon->ifs[mon->sel].intf[i].pci_id.devid, mon->ifs[mon->sel].intf[i].pci_id.function,
         func,
        get_port_state_str(mon->ifs[mon->sel].intf[i].port_state),
        mon->ifs[mon->sel].intf[i].link == PORT_LINK_UP ? "up" : "down",
        get_speed_str(mon->ifs[mon->sel].intf[i].port_speed), mon->ifs[mon->sel].intf[i].mtu, rx_throughput, tx_throughput, rxpps, txpps);
  }

  if (mon->ifs[mon->sel].intf[curr_port_index].type != PORT_TYPE_VIRT) {
  _wprint(main_window, menu_size + 3, 1, "MAC: %02x:%02x:%02x:%02x:%02x:%02x ",
                mon->ifs[mon->sel].intf[curr_port_index].mac.addr_b[0], mon->ifs[mon->sel].intf[curr_port_index].mac.addr_b[1],
                mon->ifs[mon->sel].intf[curr_port_index].mac.addr_b[2], mon->ifs[mon->sel].intf[curr_port_index].mac.addr_b[3],
                mon->ifs[mon->sel].intf[curr_port_index].mac.addr_b[4], mon->ifs[mon->sel].intf[curr_port_index].mac.addr_b[5]);
  } else {
    _wprint(main_window, menu_size + 3, 1, "RSS queues: %i                       ", mon->ifs[mon->sel].intf[curr_port_index].num_queues - 1);
  }
  wrefresh(main_window);
}




static void init_queue_items(void)
{
  queue_items = (ITEM **)calloc(queue_counters_number + 1, sizeof(ITEM *));

  for (int i = 0; i < queue_counters_number; ++i) {
    sprintf(queue_data[i], QUEUE_FORMAT, "", (uint64_t)0, (uint64_t)0, (uint64_t)0, (uint64_t)0);
    sprintf(queue_nums[i], "%3d", i);
    queue_items[i] = new_item(queue_nums[i], queue_data[i]);
    port_queue_mapping[i].port_id = -1;
    port_queue_mapping[i].queue_id = -1;
  }
  queue_items[queue_counters_number] = (ITEM *)NULL;

  is_inited_queue_items = 1;
}

static void deinit_queue_items(void)
{
  for (int i = 0; i < queue_counters_number; ++i) {
    free_item(queue_items[i]);
  }
  is_inited_queue_items = 0;
  free(queue_items);
}


static const char *find_pci_func(char *buf, int queue_idx)
{
  for (int i = 0; i < mon->ifs[mon->sel].nb_ports; i++) {
    int ii = 0;
    while (ii < mon->ifs[mon->sel].intf[i].num_queues) {
      if (mon->ifs[mon->sel].intf[i].queue[ii].idx == queue_idx) {
        int port_id = mon->ifs[mon->sel].intf[i].port_id;
        int queue_id = mon->ifs[mon->sel].intf[i].queue[ii].queue_id;
        port_queue_mapping[queue_idx].port_id = port_id;
        port_queue_mapping[queue_idx].queue_id = queue_id;
        if (mon->ifs[mon->sel].intf[i].type != PORT_TYPE_VIRT) {
          if (mon->ifs[mon->sel].intf[i].type == PORT_TYPE_PHY_NORMAL) {
            snprintf(buf, 31,   "%2d:%2d phy%i ctl/in", port_id, queue_id, i);
          } else {
            snprintf(buf, 31,   "%2d:%2d lag ctl/in", port_id, queue_id);  /* Queue representing link aggregated phy ports */
          }
        } else {
          int num = mon->ifs[mon->sel].intf[i].pci_id.devid * 8 + mon->ifs[mon->sel].intf[i].pci_id.function;
          if (mon->ifs[mon->sel].intf[i].queue[ii].dir == QUEUE_INPUT) {
            snprintf(buf, 31, "%2d:%2d VF%i  ctl/in", port_id, queue_id, num);
          } else {
            snprintf(buf, 31, "%2d:%2d VF%i     out", port_id, queue_id, num);
          }
        }
        return buf;
      }
      ii++;
    }
  }
  return "-unused-  ";
}


static void update_queue_list(void)
{
  if (is_inited_queue_items == 0) {
    init_queue_items();
  }
  if (mon->stat_layout_version == 6) {
    char buf[32];
    struct queue_type_fields_s *que = (struct queue_type_fields_s *)mon->que->data;
    struct queue_type_fields_s *base = (struct queue_type_fields_s *)res_base.que;
    for (int i = 0; i < queue_counters_number; ++i) {
      sprintf(queue_data[i], QUEUE_FORMAT, find_pci_func(buf, i),
          que->fwd_pkts - (base ? base->fwd_pkts : 0),
          que->fwd_octets - (base ? base->fwd_octets : 0),
          que->dbs_drop_pkts - (base ? base->dbs_drop_pkts : 0),
          que->dbs_drop_octets - (base ? base->dbs_drop_octets : 0));
      que++;
      if (base) {
        base++;
      }
    }
  }
  _wprint(main_window, menu_size + 3, 1, "                                                                                             ");
  wrefresh(main_window);
}




static void update_queue_info(void)
{
  int line_cnt = 1;
  int curr_qid = item_index(current_item(top_menu));

  werase(bottom_window);
  box(bottom_window, 0, 0);

  _wprint(bottom_window, line_cnt++, 2, "%-12s%d", "HW QueueId:", curr_qid);
  _wprint(bottom_window, line_cnt++, 2, "%-6s%d", "Port:", port_queue_mapping[curr_qid].port_id);
  _wprint(bottom_window, line_cnt++, 2, "%-7s%d", "Queue:", port_queue_mapping[curr_qid].queue_id);

  struct queue_type_fields_s *que = (struct queue_type_fields_s *)mon->que->data;
  struct queue_type_fields_s *base = (struct queue_type_fields_s *)res_base.que;
  que += curr_qid;
  if (base) {
    base += curr_qid;
  }

  _wprint(bottom_window, line_cnt++, 2, "%-16s: %015llu  %-16s: %017llu",
      "flush packets", que->flush_pkts - (base ? base->flush_pkts : 0),
      "flush octets", que->flush_octets - (base ? base->flush_octets : 0));
  _wprint(bottom_window, line_cnt++, 2, "%-16s: %015llu  %-16s: %017llu",
      "drop packets", que->drop_pkts - (base ? base->drop_pkts : 0),
      "drop octets", que->drop_octets - (base ? base->drop_octets : 0));

  wrefresh(bottom_window);
}



static void init_color_stat_items(void)
{
  color_stat_nums = calloc(color_counters_number, STR_ID_SIZE + 1);
  color_stat_data = calloc(color_counters_number, sizeof(COLOR_STAT_HEADER));
  color_stat_items = (ITEM **)calloc(color_counters_number + 1, sizeof(ITEM *));

  if (!color_stat_nums || !color_stat_data || !color_stat_items) {
    exit(0);
  }

  for (int i = 0; i < color_counters_number; ++i) {
    sprintf(color_stat_nums[i], "%5d", i);
    sprintf(color_stat_data[i], COLOR_STAT_FORMAT, (uint64_t)0, (uint64_t)0, 0);
    color_stat_items[i] = new_item(color_stat_nums[i], color_stat_data[i]);
  }
  color_stat_items[color_counters_number] = (ITEM *)NULL;

  is_inited_color_stat_items = 1;
}

static void deinit_color_stat_items(void)
{
  if (is_inited_color_stat_items != 0) {
    for (int i = 0; i < color_counters_number; ++i) {
      free_item(color_stat_items[i]);
    }
  }

  if (color_stat_nums) {
    free(color_stat_nums);
  }

  if (color_stat_data) {
    free(color_stat_data);
  }

  if (color_stat_items) {
    free(color_stat_items);
  }

  is_inited_color_stat_items = 0;
}

static void update_color_stats_list(void)
{
  if (curr_port_index < 0) {
    return;
  }

  if (is_inited_color_stat_items == 0) {
    init_color_stat_items();
  }

  struct color_type_fields_s *clr = (struct color_type_fields_s *)mon->clr->data;
  struct color_type_fields_s *base = (struct color_type_fields_s *)res_base.clr;

  for (int i = 0; i < color_counters_number; ++i) {
    sprintf(color_stat_data[i], COLOR_STAT_FORMAT,
            clr->pkts - (base ? base->pkts : 0),
            clr->octets - (base ? base->octets : 0),
            (uint8_t)clr->tcp_flgs);  /* TCP flags cannot be subtracted from base, so cannot be reset */
    clr++;
    if (base) {
      base++;
    }
  }

  _wprint(main_window, menu_size + 3, 1, "                                                                                             ");
  wrefresh(main_window);
}

static void update_color_stats_info(void)
{
  werase(bottom_window);
  box(bottom_window, 0, 0);

  /* what to add here? - move tcp flags down here maybe? */

  wrefresh(bottom_window);
}

static void update_flm_list(void)
{
  _wprint(main_window, menu_size + 3, 1, "FLM statistics                                                                               ");
  wrefresh(main_window);
}

#define FLM_TEXT_WIDTH -21
#define FLM_TEXT_WIDTH_2 -10
#define FLM_COUNTER_WIDTH 17

#define PRINT_FLM_COUNTER(param)                   \
  _wprint(bottom_window,                           \
          line_cnt++,                              \
          2,                                       \
          "%*s: %*" PRIu64,                        \
          FLM_TEXT_WIDTH,                          \
          #param,                                  \
          FLM_COUNTER_WIDTH,                       \
          flm->param - (base ? base->param : 0))

#define PRINT_FLM_COUNTER_2(param1, param2)             \
  _wprint(bottom_window,                                \
          line_cnt++,                                   \
          2,                                            \
          "%*s: %*" PRIu64 "       %*s: %*" PRIu64 ,    \
          FLM_TEXT_WIDTH,                               \
          #param1,                                      \
          FLM_COUNTER_WIDTH,                            \
          flm->param1 - (base ? base->param1 : 0),      \
          FLM_TEXT_WIDTH_2,                             \
          #param2,                                      \
          FLM_COUNTER_WIDTH,                            \
          flm->param2 - (base ? base->param2 : 0))

static void update_flm_info(void)
{
  int line_cnt = 1;
  int curr_flmid = 0;  /* Currently only 1 FLM supported */

  werase(bottom_window);
  box(bottom_window, 0, 0);

  _wprint(bottom_window, line_cnt++, 2, "%-15s%d", "FlowMatcherId:", curr_flmid);

  struct flowmatcher_type_fields_s *flm = (struct flowmatcher_type_fields_s *)mon->flm->data;
  struct flowmatcher_type_fields_s *base = (struct flowmatcher_type_fields_s *)res_base.flm;
  flm += curr_flmid;
  if (base) {
    base += curr_flmid;
    /* FLM counter `current` must be always actual even after reset */
    base->current = 0;
  }

  if (mon->stat_flm_layout_version < 2) {
    PRINT_FLM_COUNTER(current);
    PRINT_FLM_COUNTER(learn_done);
    PRINT_FLM_COUNTER(learn_ignore);
    PRINT_FLM_COUNTER(learn_fail);
    PRINT_FLM_COUNTER(unlearn_done);
    PRINT_FLM_COUNTER(unlearn_ignore);
    PRINT_FLM_COUNTER(auto_unlearn_done);
    PRINT_FLM_COUNTER(auto_unlearn_ignore);
    PRINT_FLM_COUNTER(auto_unlearn_fail);
    PRINT_FLM_COUNTER(timeout_unlearn_done);
    PRINT_FLM_COUNTER(rel_done);
    PRINT_FLM_COUNTER(rel_ignore);
    PRINT_FLM_COUNTER(prb_done);
    PRINT_FLM_COUNTER(prb_ignore);
  } else if (mon->stat_flm_layout_version == 2) {
    PRINT_FLM_COUNTER_2(current, sta_done);
    PRINT_FLM_COUNTER_2(learn_done, inf_done);
    PRINT_FLM_COUNTER_2(learn_ignore, inf_skip);
    PRINT_FLM_COUNTER_2(learn_fail, pck_hit);
    PRINT_FLM_COUNTER_2(unlearn_done, pck_miss);
    PRINT_FLM_COUNTER_2(unlearn_ignore, pck_unh);
    PRINT_FLM_COUNTER_2(auto_unlearn_done, pck_dis);
    PRINT_FLM_COUNTER_2(auto_unlearn_ignore, csh_hit);
    PRINT_FLM_COUNTER_2(auto_unlearn_fail, csh_miss);
    PRINT_FLM_COUNTER_2(timeout_unlearn_done, csh_unh);
    PRINT_FLM_COUNTER_2(rel_done, cuc_start);
    PRINT_FLM_COUNTER_2(rel_ignore, cuc_move);
    PRINT_FLM_COUNTER(prb_done);
    PRINT_FLM_COUNTER(prb_ignore);
  }

  wrefresh(bottom_window);
}

static void update_menu_state(void) {
  static ITEM *highlighted_port_item;
  static ITEM *highlighted_queue_item;
  static ITEM *highlighted_color_stat_item;
  const char *pb;

  ITEM **current_items = menu_items(top_menu);

  if (win_h.state == PRESENT_QUEUE_STATS) {
    if (current_items == queue_items) {
      highlighted_queue_item = current_item(top_menu);
    }
    unpost_menu(top_menu);
    set_menu_items(top_menu, queue_items);
    if (highlighted_queue_item) {
      set_current_item(top_menu, highlighted_queue_item);
    }
    post_menu(top_menu);

    pb = QUEUE_HEADER;
  } else if (win_h.state == PRESENT_COLOR_STATS) {
    if (current_items == color_stat_items) {
      highlighted_color_stat_item = current_item(top_menu);
    }
    unpost_menu(top_menu);
    set_menu_items(top_menu, color_stat_items);
    if (highlighted_color_stat_item) {
      set_current_item(top_menu, highlighted_color_stat_item);
    }
    post_menu(top_menu);

    pb = COLOR_STAT_HEADER;
  } else if (win_h.state == PRESENT_FLM_STATS) {
    unpost_menu(top_menu);
    set_menu_items(top_menu, NULL);
    post_menu(top_menu);

    pb = FLM_STAT_HEADER;
  } else {
    if (current_items == ports_items) {
      highlighted_port_item = current_item(top_menu);
    }
    unpost_menu(top_menu);
    set_menu_items(top_menu, ports_items);
    post_menu(top_menu);
    if (highlighted_port_item) {
      set_current_item(top_menu, highlighted_port_item);
    }
    pb = PORTS_HEADER;
  }

  int len = strlen(pb);
  if (width - 5 < len) {
    char buf[width];
    memcpy(buf, pb, width - 5);
    buf[width - 5] = '\0';
    _wprint(top_window, 1, 1, buf);
  } else {
    _wprint(top_window, 1, 1, pb);
  }

}


static void write_top_window_border(void)
{
  box(top_window, 0, 0);
  _wprint(main_window, 0, (width / 2 > 25) ? width / 2 - 25 : 0, " monitoring pci:%s fpga:%s ", mon->nics->nic_pci_id[mon->sel],mon->info->fw_version);
}


static void init_port_menu(void) {
  init_port_items();

  top_menu = new_menu((ITEM **)ports_items);

  set_menu_win(top_menu, top_window);
  set_menu_sub(top_menu, derwin(top_window, menu_size, width - 4, 2, 1));
  set_menu_format(top_menu, menu_size, 1);
  set_menu_mark(top_menu, " *");
  post_menu(top_menu);
}

static int menu_scale = 2, force_recalc;


static int recalc_windows(void)
{
  getmaxyx(stdscr, maxy, maxx);
  int _height, _width;

  _height = max_ywin;
  _width = max_xwin;

  if (force_recalc || _height != height || _width != width) {
    height = _height;
    width = _width;
    /* we need to resize */

    force_recalc = 0;

    if (!main_window) {
      main_window = newwin(height, width, 1, 1);
    }

    wresize(main_window, height, width);

    _width = ((width - 2) < 2) ? 2 : width - 2;

    menu_size = menu_scale * _height / 10;
    if (menu_size < 3) {
      menu_size = 3;
    }

    free_menu(top_menu);
    if (top_window) {
      delwin(top_window);
    }
    top_window = subwin(main_window, menu_size + 3, _width, 1, 1);
    scrollok(top_window, TRUE);
    write_top_window_border();

    init_port_menu();

    if (bottom_window) {
      delwin(bottom_window);
    }

    bottom_height = (height > menu_size + 8) ? height - menu_size - 7 : 2;
    bottom_window = subwin(main_window, bottom_height, _width, menu_size + 5, 1);

    return 1;
  }

  return 0;
}


static void _show_screen(void)
{
  update_menu_state();
  win_h.update_top_window_data();
  update_status_bar();
  wrefresh(top_window);
  win_h.update_bottom_window();
  wrefresh(bottom_window);
  wrefresh(main_window);
}



static void _update_and_calc_window_data(void)
{
  recalc_windows();
  update_port_data();
  update_data_buttons();
  init_port_menu();

  color_counters_number = mon->clr->nb_counters;
  queue_counters_number = mon->que->nb_counters;

  update_queue_list();
  update_color_stats_list();
  update_flm_list();

  update_port_list();
}



void display(int failed)
{
  if (failed) {
    wattron(main_window, A_DIM);
    write_top_window_border();
    wattroff(main_window, A_DIM);
    wattron(main_window, A_BOLD);
    _wprint(main_window, 0, 2, "  CONNECTION LOST  ");
    wattroff(main_window, A_BOLD);
    wattron(main_window, A_DIM);

    wattron(top_window, A_DIM);
    wattron(bottom_window, A_DIM);
    last_failed = 1;

  } else
  if (last_failed) {
    if (res_base.clr) {
      free(res_base.clr);
      memset(&res_base, 0, sizeof(res_base));
    }
    _update_and_calc_window_data();

    wattroff(main_window, A_DIM);
    wattroff(top_window, A_DIM);
    wattroff(bottom_window, A_DIM);
    write_top_window_border();
    last_failed = 0;
  }


  flushinp();
  int ch = getch();
  switch (ch) {
    case 'R':
    case 'r':
      set_reset_base();
      break;

    case 'P':
    case 'p':
      win_h.state = PRESENT_STAT_NONE; /* let it select Rx and fall through */
      __attribute__ ((fallthrough));
    case 'T':
    case 't':
      win_h.state = win_h.state == PRESENT_PORT_RX_STATS ? PRESENT_PORT_TX_STATS : PRESENT_PORT_RX_STATS;
      win_h.update_top_window_data = update_port_list;
      win_h.update_bottom_window = update_port_data;
      write_top_window_border();
      break;

    case 'Q':
    case 'q':
      win_h.state = PRESENT_QUEUE_STATS;
      win_h.update_top_window_data = update_queue_list;
      win_h.update_bottom_window = update_queue_info;
      write_top_window_border();

      if (queue_counters_number != (int)mon->que->nb_counters) {
        deinit_queue_items();
        queue_counters_number = mon->que->nb_counters;
        if (queue_counters_number > MAX_QUEUE_NUM)
          queue_counters_number = MAX_QUEUE_NUM;
      }

      break;

    case 'F':
    case 'f':
      win_h.state = PRESENT_FLM_STATS;
      win_h.update_top_window_data = update_flm_list;
      win_h.update_bottom_window = update_flm_info;
      write_top_window_border();
      break;

    case 'C':
    case 'c':
      win_h.state = PRESENT_COLOR_STATS;
      win_h.update_top_window_data = update_color_stats_list;
      win_h.update_bottom_window = update_color_stats_info;
      write_top_window_border();

      if (color_counters_number != (int)mon->clr->nb_counters) {
        deinit_color_stat_items();
        color_counters_number = mon->clr->nb_counters;
      }

      break;
    case KEY_UP:
      menu_driver(top_menu, REQ_UP_ITEM);
      break;
    case KEY_DOWN:
      menu_driver(top_menu, REQ_NEXT_ITEM);
      break;
    case KEY_NPAGE:
      menu_driver(top_menu, REQ_SCR_DPAGE);
      break;
    case KEY_PPAGE:
      menu_driver(top_menu, REQ_SCR_UPAGE);
      break;

    case 'W':
    case 'w':
      menu_scale += (ch == 'W') ? -1 : 1;

      if (menu_scale > 8)
        menu_scale = 1;
      if (menu_scale < 1)
        menu_scale = 8;

      force_recalc = 1;
      __attribute__ ((fallthrough));

    case KEY_RESIZE:
      if (recalc_windows()) {
        werase(main_window);
        wrefresh(top_window);
        wrefresh(bottom_window);
        box(top_window, 0, 0);
        box(bottom_window, 0, 0);
        write_top_window_border();
        update_port_list();
        update_queue_list();
        update_color_stats_list();
        update_flm_list();
        update_data_buttons();
      }
      return;
    default:
      break;
  }

  _show_screen();
}


void init_display(struct mon_data_s *mon_data)
{
  initscr();
  cbreak();
  noecho();
  halfdelay(5);
  curs_set(0);
  keypad(stdscr, TRUE);

  refresh();
  mon = mon_data;

  _update_and_calc_window_data();

  update_port_list();
  win_h.state = PRESENT_PORT_RX_STATS;
  win_h.update_top_window_data = update_port_list;
  win_h.update_top_window_data();

  win_h.update_bottom_window = update_port_data;
  win_h.update_bottom_window();

  _show_screen();
}


void deinit_display(void)
{
  if (res_base.clr) {
    free(res_base.clr);
    memset(&res_base, 0, sizeof(res_base));
  }

  wborder(main_window, ' ', ' ', ' ',' ',' ',' ',' ',' ');
  wrefresh(main_window);
  delwin(top_window);
  delwin(bottom_window);
  delwin(main_window);
  unpost_menu(top_menu);
  free_menu(top_menu);
  deinit_port_items();
  deinit_queue_items();
  deinit_color_stat_items();

  endwin();
}

/*
 * EOF
 */
