/*
 * SPDX-License-Identifier: BSD-3-Clause
 * Copyright(c) 2023 Napatech A/S
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "ntconn_client.h"
#include "utils.h"

static int send_request(int fd, const char *pci_id, const char *module, const char *func, void *blob, uint32_t blob_len)
{
  struct ntconn_header_s hdr;
  char msg[MESSAGE_BUFFER];

  char request[MESSAGE_BUFFER - NTCMOD_HDR_LEN - 1];
  snprintf(request, sizeof(request), "%s;%s;%s", pci_id, module, func);

  size_t req_len = strlen(request) + 1;
  hdr.tag = NTCONN_TAG_REQUEST;
  hdr.len = NTCMOD_HDR_LEN + req_len; /* including 0 -termination */
  hdr.blob_len = blob_len;

  memcpy(msg, &hdr, NTCMOD_HDR_LEN);
  memcpy(&msg[NTCMOD_HDR_LEN], &request, req_len);

  if (blob_len > 0 && blob_len < (uint32_t)(MESSAGE_BUFFER - hdr.len)) {
    memcpy(&msg[hdr.len], blob, hdr.blob_len);
    int res = write_all(fd, (uint8_t *)msg, hdr.len + hdr.blob_len);
    if (res) {
      return res;
    }
  }
  else {
    int res = write_all(fd, (uint8_t *)msg, hdr.len);
    if (res) {
      return res;
    }
    if (blob_len > 0) {
      res = write_all(fd, (uint8_t *)blob, blob_len);
      if (res) {
        return res;
      }
    }
  }

  return 0;
}

static int get_reply(int fd, struct ntconn_header_s *hdr, uint8_t **rdata, size_t *len)
{
  hdr->tag = 0;
  hdr->len = 0;
  *len = 0;

  uint8_t buffer[MESSAGE_BUFFER];
  size_t recv_len = 0;
  int ret = read_data(fd, MESSAGE_BUFFER, buffer, &recv_len, -1);
  if (ret) {
    return -1;
  }

  memcpy(hdr, buffer, NTCMOD_HDR_LEN);
  size_t length = hdr->len + hdr->blob_len - NTCMOD_HDR_LEN;
  uint8_t *data = malloc(length * sizeof(uint8_t));
  if (!data) {
    fprintf(stderr, "memory allocation failed\n");
    return -1;
  }

  memcpy(data, &buffer[NTCMOD_HDR_LEN], recv_len - NTCMOD_HDR_LEN);
  if (length + NTCMOD_HDR_LEN > MESSAGE_BUFFER) {
    uint8_t *new_data = realloc(data, length * sizeof(uint8_t));
    if (!new_data) {
      fprintf(stderr, "memory reallocation failed\n");
      free(data);
      return -1;
    }
    data = new_data;
    ret = read_all(fd, &data[recv_len - NTCMOD_HDR_LEN], length + NTCMOD_HDR_LEN - recv_len);
    if (ret) {
      free(data);
      return -1;
    }
  }

  *len = length;
  *rdata = data;

  return 0;
}

static int is_reply_ok(struct ntconn_header_s *hdr, uint8_t *data, size_t len, struct reply_err_s *err)
{
  if (hdr->tag == NTCONN_TAG_REPLY) {
    if (err) {
      err->err_code = NTCONN_ERR_CODE_NONE;
      strncpy(err->msg, "Success", MAX_ERR_MESSAGE_LENGTH);
    }
    return 1;
  }

  if (hdr->tag == NTCONN_TAG_ERROR) {
    /* text error message returned */
    const uint32_t err_code = *(const uint32_t *)data;
    const char *message = (len > 0) ? (const char *)data + sizeof(uint32_t) : "<no error text replied>";

    if (err) {
      err->err_code = err_code;
      strncpy(err->msg, message, MAX_ERR_MESSAGE_LENGTH);
    } else {
      fprintf(stderr, "ERROR Reply: %s (code %u)\n", message, err_code);
    }
  } else {
    char err_msg[64];
    snprintf(err_msg, 64, "Cannot interpret reply: Got tag %i", hdr->tag);
    if (err) {
      err->err_code = NTCONN_ERR_CODE_INTERNAL_ERROR;
      strncpy(err->msg, err_msg, MAX_ERR_MESSAGE_LENGTH);
    } else {
      fprintf(stderr, "%s\n", err_msg);
    }
  }
  if (len) {
    free(data);
  }

  return 0;
}

/*
 * ntconn client API
 */

int ntconn_client_connect(void)
{
  return unix_connect(NTCONNECT_SOCKET);
}

uint64_t ntconn_client_get_version(int fd, const char *pci_id, const char *module)
{
  struct ntconn_header_s hdr;
  uint8_t *data = NULL;
  size_t len = 0;

  if (send_request(fd, pci_id, module, "version", NULL, 0)) {
    return 0;
  }

  if (get_reply(fd, &hdr, &data, &len)) {
    return 0;
  }

  if (is_reply_ok(&hdr, data, len, NULL)) {
    if (len == sizeof(uint64_t)) {
      uint64_t val = *(uint64_t *)data;
      free(data);
      return val;
    }
    if (len) {
      free(data);
    }
  }

  return 0;
}

int ntconn_client_request(int fd, const char *pci_id, const char *module, const char *func, uint8_t **data, struct reply_err_s *err)
{
  struct ntconn_header_s hdr;
  size_t len = 0;
  if (send_request(fd, pci_id, module, func, NULL, 0)) {
    return -1;
  }

  if (get_reply(fd, &hdr, data, &len)) {
    return -1;
  }

  if (is_reply_ok(&hdr, *data, len, err)) {
    return (int)len;
  }

  return 0;
}

int ntconn_client_request_with_data(int fd,
                                    const char *pci_id,
                                    const char *module,
                                    const char *func,
                                    uint8_t **data,
                                    struct reply_err_s *err,
                                    void *blob,
                                    uint32_t blob_len)
{
  struct ntconn_header_s hdr;
  size_t len = 0;
  if (send_request(fd, pci_id, module, func, blob, blob_len)) {
    return -1;
  }

  if (get_reply(fd, &hdr, data, &len)) {
    return -1;
  }

  if (is_reply_ok(&hdr, *data, len, err)) {
    return (int)len;
  }

  return 0;
}
