/*
 * SPDX-License-Identifier: BSD-3-Clause
 * Copyright(c) 2023 Napatech A/S
 */

#ifndef _NTCONN_CLIENT_H_
#define _NTCONN_CLIENT_H_

#include <sys/un.h>

#include "ntconnect_api.h"

/*
 * return a file descriptor on success connected with ntconnect unix server in PMD
 */
int ntconn_client_connect(void);

/**
 * fd       is the file descriptor got from ntconn_connect()
 * pci_id   specifies the physical adapter to target (ie. 0000:17:00.0)
 * module   name of the ntconnect implemented backend module (ie. adapter or stat)
 *
 * NOTE: this gets the version of the implemented ntconnect module. Any other versions
 * regarding FPGA modules or other SW structure version must be negotiated otherwise through a request
 */
uint64_t ntconn_client_get_version(int fd, const char *pci_id, const char *module);

/**
 * Make a request or send a command/configuration to a specific module on a specific NIC (pci_id)
 *
 * fd       is the file descriptor got from ntconn_connect()
 * pci_id   specifies the physical adapter to target (ie. 0000:17:00.0)
 * module   name of the ntconnect implemented backend module (ie. adapter or stat)
 * func     the requested instruction string (ie. "stat,get,rx_counters" - getting rx counter statistics or
 *          "stat,snapshot" make a stat snapshot to get all stat counters synchronized (with same timestamp)
 *            by using a following "stat,get,snapshot,..." or
 *          "adapter,set,interface,port0,speed=10G" to set the port speed of an interface - is not completely implemented yet)
 * data     address of a data pointer to point to the delivered, allocated, resulting information. NOTE when length > 0 this will
 *          always be memory allocated thus must be freed with free() after use by application
 * err      if errors happened in the protocol layers (excluding errors from backend-specific protocol), the error message may be
 *          returned here if needed. Set to NULL if not needed and the error message will got to stderr.
 *          each backend that has the need for specific error codes must establish its own error codes in the reply
 * return   length of data allocated and put into *data. In the situation of a command or config setting function this can
 *          return 0 and will then indicate success (and no allocated reply in *data)
 */
int ntconn_client_request(int fd, const char *pci_id, const char *module, const char *func, uint8_t **data, struct reply_err_s *err);

/**
 * Make a request or send a command/configuration with a blob of data to a specific module on a specific NIC (pci_id)
 *
 * fd       is the file descriptor got from ntconn_connect()
 * pci_id   specifies the physical adapter to target (ie. 0000:17:00.0)
 * module   name of the ntconnect implemented backend module (ie. adapter or stat)
 * func     the requested instruction string (ie. "stat,get,rx_counters" - getting rx counter statistics or
 *          "stat,snapshot" make a stat snapshot to get all stat counters synchronized (with same timestamp)
 *          by using a following "stat,get,snapshot,..." or
 *          "adapter,set,interface,port0,speed=10G" to set the port speed of an interface - is not completely implemented yet)
 * data     address of a data pointer to point to the delivered, allocated, resulting information. NOTE when length > 0 this will
 *          always be memory allocated thus must be freed with free() after use by application
 * err      if errors happened in the protocol layers (excluding errors from backend-specific protocol), the error message may be
 *          returned here if needed. Set to NULL if not needed and the error message will got to stderr.
 *          each backend that has the need for specific error codes must establish its own error codes in the reply
 * blob     pointer to data blob to send. The data blob must be identified by the instruction string func (ie. "filter,set,settings")
 * blob_len length of data blob to send
 * return   length of data allocated and put into *data. In the situation of a command or config setting function this can
 *          return 0 and will then indicate success (and no allocated reply in *data)
 */
int ntconn_client_request_with_data(int fd,
                                    const char *pci_id,
                                    const char *module,
                                    const char *func,
                                    uint8_t **data,
                                    struct reply_err_s *err,
                                    void *blob,
                                    uint32_t blob_len);

#endif /* _NTCONN_CLIENT_H_ */
